#line 1 "./prim"
\ Gforth primitives

\ Copyright (C) 1995,1996,1997,1998,2000 Free Software Foundation, Inc.

\ This file is part of Gforth.

\ Gforth is free software; you can redistribute it and/or
\ modify it under the terms of the GNU General Public License
\ as published by the Free Software Foundation; either version 2
\ of the License, or (at your option) any later version.

\ This program is distributed in the hope that it will be useful,
\ but WITHOUT ANY WARRANTY; without even the implied warranty of
\ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
\ GNU General Public License for more details.

\ You should have received a copy of the GNU General Public License
\ along with this program; if not, write to the Free Software
\ Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.


\ WARNING: This file is processed by m4. Make sure your identifiers
\ don't collide with m4's (e.g. by undefining them).
\ 
\ 
\ 
\ This file contains primitive specifications in the following :
\ 
\ forth name	( stack effect )	category	[pronunciation]
\ [""glossary entry""]
\ C code
\ [:
\ Forth code]
\ 
\ Note: Fields in brackets are optional.  Word specifications have to
\ be separated by at least one empty line
\
\ Both pronounciation and stack items (in the stack effect) must
\ conform to the C identifier syntax or the C compiler will complain.
\ If you don't have a pronounciation field, the Forth name is used,
\ and has to conform to the C identifier syntax.
\ 
\ These specifications are automatically translated into C-code for the
\ interpreter and into some other files. I hope that your C compiler has
\ decent optimization, otherwise the automatically generated code will
\ be somewhat slow. The Forth version of the code is included for manual
\ compilers, so they will need to compile only the important words.
\ 
\ Note that stack pointer adjustment is performed according to stack
\ effect by automatically generated code and NEXT is automatically
\ appended to the C code. Also, you can use the names in the stack
\ effect in the C code. Stack access is automatic. One exception: if
\ your code does not fall through, the results are not stored into the
\ stack. Use different names on both sides of the '--', if you change a
\ value (some stores to the stack are optimized away).
\
\ For superinstructions the syntax is:
\
\ forth-name [/ c-name] = forth-name forth-name ...
\
\ 
\ The stack variables have the following types:
\ 
\ name matches	type
\ f.*		Bool
\ c.*		Char
\ [nw].*	Cell
\ u.*		UCell
\ d.*		DCell
\ ud.*		UDCell
\ r.*		Float
\ a_.*		Cell *
\ c_.*		Char *
\ f_.*		Float *
\ df_.*		DFloat *
\ sf_.*		SFloat *
\ xt.*		XT
\ f83name.*	F83Name *

\E stack data-stack   sp Cell
\E stack fp-stack     fp Float
\E stack return-stack rp Cell
\E
\E get-current prefixes set-current
\E 
\E s" Bool"		single data-stack type-prefix f
\E s" Char"		single data-stack type-prefix c
\E s" Cell"		single data-stack type-prefix n
\E s" Cell"		single data-stack type-prefix w
\E s" UCell"		single data-stack type-prefix u
\E s" DCell"		double data-stack type-prefix d
\E s" UDCell"		double data-stack type-prefix ud
\E s" Float"		single fp-stack   type-prefix r
\E s" Cell *"		single data-stack type-prefix a_
\E s" Char *"		single data-stack type-prefix c_
\E s" Float *"		single data-stack type-prefix f_
\E s" DFloat *"		single data-stack type-prefix df_
\E s" SFloat *"		single data-stack type-prefix sf_
\E s" Xt"		single data-stack type-prefix xt
\E s" struct F83Name *"	single data-stack type-prefix f83name
\E s" struct Longname *" single data-stack type-prefix longname
\E 
\E return-stack stack-prefix R:
\E inst-stream  stack-prefix #
\E 
\E set-current

\ 
\ 
\ 
\ In addition the following names can be used:
\ ip	the instruction pointer
\ sp	the data stack pointer
\ rp	the parameter stack pointer
\ lp	the locals stack pointer
\ NEXT	executes NEXT
\ cfa	
\ NEXT1	executes NEXT1
\ FLAG(x)	makes a Forth flag from a C flag
\ 
\ 
\ 
\ Percentages in comments are from Koopmans book: average/maximum use
\ (taken from four, not very representative benchmarks)
\ 
\ 
\ 
\ To do:
\ 
\ throw execute, cfa and NEXT1 out?
\ macroize *ip, ip++, *ip++ (pipelining)?

\ these m4 macros would collide with identifiers




\g control

noop	( -- )		gforth
:
 ;

lit	( #w -- w )		gforth
:
 r> dup @ swap cell+ >r ;

execute	( xt -- )		core
""Perform the semantics represented by the execution token, @i{xt}.""
ip=IP;
IF_spTOS(spTOS = sp[0]);
SUPER_END;
EXEC(xt);

perform	( a_addr -- )	gforth
""@code{@@ execute}.""
/* and pfe */
ip=IP;
IF_spTOS(spTOS = sp[0]);
SUPER_END;
EXEC(*(Xt *)a_addr);
:
 @ execute ;

\fhas? skipbranchprims 0= [IF]
\+glocals

branch-lp+!#	( #ndisp #nlocals -- )	gforth	branch_lp_plus_store_number
/* this will probably not be used */
lp += nlocals;
SET_IP((Xt *)(((Cell)(IP-2))+ndisp));

\+

branch	( #ndisp -- )		gforth
SET_IP((Xt *)(((Cell)(IP-1))+ndisp));
:
 r> dup @ + >r ;

\ condbranch(forthname,stackeffect,restline,code,forthcode)
\ this is non-syntactical: code must open a brace that is closed by the macro
#line 200


#line 209
?branch ( #ndisp f -- ) f83	question_branch
#line 209
if (f==0) {
#line 209
	SET_IP((Xt *)(((Cell)(IP-1))+ndisp));
#line 209
TAIL;
#line 209
}
#line 209
SUPER_CONTINUE;
#line 209
:
#line 209
 0= dup     \ !f !f
#line 209
 r> dup @   \ !f !f IP branchoffset
#line 209
 rot and +  \ !f IP|IP+branchoffset
#line 209
 swap 0= cell and + \ IP''
#line 209
 >r ;
#line 209

#line 209
\+glocals
#line 209

#line 209
?branch-lp+!# ( #ndisp #nlocals f -- ) f83	question_branch_lp_plus_store_number
#line 209
if (f==0) {
#line 209
    lp += nlocals;
#line 209
SET_IP((Xt *)(((Cell)(IP-2))+ndisp));
#line 209
TAIL;
#line 209
}
#line 209
SUPER_CONTINUE;
#line 209

#line 209
\+
#line 209


\ we don't need an lp_plus_store version of the ?dup-stuff, because it
\ is only used in if's (yet)

\+xconds

?dup-?branch	( #ndisp f -- f )	new	question_dupe_question_branch
""The run-time procedure compiled by @code{?DUP-IF}.""
if (f==0) {
  sp++;
  IF_spTOS(spTOS = sp[0]);
  SET_IP((Xt *)(((Cell)(IP-1))+ndisp));
  TAIL;
}
SUPER_CONTINUE;

?dup-0=-?branch	( #ndisp f -- )	new	question_dupe_zero_equals_question_branch
""The run-time procedure compiled by @code{?DUP-0=-IF}.""
/* the approach taken here of declaring the word as having the stack
effect ( f -- ) and correcting for it in the branch-taken case costs a
few cycles in that case, but is easy to convert to a CONDBRANCH
invocation */
if (f!=0) {
  sp--;
  SET_IP((Xt *)(((Cell)(IP-1))+ndisp));
  NEXT;
}
SUPER_CONTINUE;

\+
\f[THEN]
\fhas? skiploopprims 0= [IF]

#line 248
(next) ( #ndisp R:n1 -- R:n2 ) cmFORTH	paren_next
#line 248
n2=n1-1;
#line 248
if (n1) {
#line 248
	SET_IP((Xt *)(((Cell)(IP-1))+ndisp));
#line 248
TAIL;
#line 248
}
#line 248
SUPER_CONTINUE;
#line 248
:
#line 248
 r> r> dup 1- >r
#line 248
 IF dup @ + >r ELSE cell+ >r THEN ;
#line 248

#line 248
\+glocals
#line 248

#line 248
(next)-lp+!# ( #ndisp #nlocals R:n1 -- R:n2 ) cmFORTH	paren_next_lp_plus_store_number
#line 248
n2=n1-1;
#line 248
if (n1) {
#line 248
    lp += nlocals;
#line 248
SET_IP((Xt *)(((Cell)(IP-2))+ndisp));
#line 248
TAIL;
#line 248
}
#line 248
SUPER_CONTINUE;
#line 248

#line 248
\+
#line 248


#line 256
(loop) ( #ndisp R:nlimit R:n1 -- R:nlimit R:n2 ) gforth	paren_loop
#line 256
n2=n1+1;
#line 256
if (n2 != nlimit) {
#line 256
	SET_IP((Xt *)(((Cell)(IP-1))+ndisp));
#line 256
TAIL;
#line 256
}
#line 256
SUPER_CONTINUE;
#line 256
:
#line 256
 r> r> 1+ r> 2dup =
#line 256
 IF >r 1- >r cell+ >r
#line 256
 ELSE >r >r dup @ + >r THEN ;
#line 256

#line 256
\+glocals
#line 256

#line 256
(loop)-lp+!# ( #ndisp #nlocals R:nlimit R:n1 -- R:nlimit R:n2 ) gforth	paren_loop_lp_plus_store_number
#line 256
n2=n1+1;
#line 256
if (n2 != nlimit) {
#line 256
    lp += nlocals;
#line 256
SET_IP((Xt *)(((Cell)(IP-2))+ndisp));
#line 256
TAIL;
#line 256
}
#line 256
SUPER_CONTINUE;
#line 256

#line 256
\+
#line 256


#line 272
(+loop) ( #ndisp n R:nlimit R:n1 -- R:nlimit R:n2 ) gforth paren_plus_loop
#line 272
/* !! check this thoroughly */
#line 272
/* sign bit manipulation and test: (x^y)<0 is equivalent to (x<0) != (y<0) */
#line 272
/* dependent upon two's complement arithmetic */
#line 272
Cell olddiff = n1-nlimit;
#line 272
n2=n1+n;	
#line 272
if ((olddiff^(olddiff+n))>=0   /* the limit is not crossed */
#line 272
    || (olddiff^n)>=0          /* it is a wrap-around effect */) {
#line 272
	SET_IP((Xt *)(((Cell)(IP-1))+ndisp));
#line 272
TAIL;
#line 272
}
#line 272
SUPER_CONTINUE;
#line 272
:
#line 272
 r> swap
#line 272
 r> r> 2dup - >r
#line 272
 2 pick r@ + r@ xor 0< 0=
#line 272
 3 pick r> xor 0< 0= or
#line 272
 IF    >r + >r dup @ + >r
#line 272
 ELSE  >r >r drop cell+ >r THEN ;
#line 272

#line 272
\+glocals
#line 272

#line 272
(+loop)-lp+!# ( #ndisp #nlocals n R:nlimit R:n1 -- R:nlimit R:n2 ) gforth paren_plus_loop_lp_plus_store_number
#line 272
/* !! check this thoroughly */
#line 272
/* sign bit manipulation and test: (x^y)<0 is equivalent to (x<0) != (y<0) */
#line 272
/* dependent upon two's complement arithmetic */
#line 272
Cell olddiff = n1-nlimit;
#line 272
n2=n1+n;	
#line 272
if ((olddiff^(olddiff+n))>=0   /* the limit is not crossed */
#line 272
    || (olddiff^n)>=0          /* it is a wrap-around effect */) {
#line 272
    lp += nlocals;
#line 272
SET_IP((Xt *)(((Cell)(IP-2))+ndisp));
#line 272
TAIL;
#line 272
}
#line 272
SUPER_CONTINUE;
#line 272

#line 272
\+
#line 272


\+xconds

#line 280
(-loop) ( #ndisp u R:nlimit R:n1 -- R:nlimit R:n2 ) gforth paren_minus_loop
#line 280
UCell olddiff = n1-nlimit;
#line 280
n2=n1-u;
#line 280
if (olddiff>u) {
#line 280
	SET_IP((Xt *)(((Cell)(IP-1))+ndisp));
#line 280
TAIL;
#line 280
}
#line 280
SUPER_CONTINUE;
#line 280

#line 280

#line 280
\+glocals
#line 280

#line 280
(-loop)-lp+!# ( #ndisp #nlocals u R:nlimit R:n1 -- R:nlimit R:n2 ) gforth paren_minus_loop_lp_plus_store_number
#line 280
UCell olddiff = n1-nlimit;
#line 280
n2=n1-u;
#line 280
if (olddiff>u) {
#line 280
    lp += nlocals;
#line 280
SET_IP((Xt *)(((Cell)(IP-2))+ndisp));
#line 280
TAIL;
#line 280
}
#line 280
SUPER_CONTINUE;
#line 280

#line 280
\+
#line 280


#line 295
(s+loop) ( #ndisp n R:nlimit R:n1 -- R:nlimit R:n2 ) gforth	paren_symmetric_plus_loop
#line 295
""The run-time procedure compiled by S+LOOP. It loops until the index
#line 295
crosses the boundary between limit and limit-sign(n). I.e. a symmetric
#line 295
version of (+LOOP).""
#line 295
/* !! check this thoroughly */
#line 295
Cell diff = n1-nlimit;
#line 295
Cell newdiff = diff+n;
#line 295
if (n<0) {
#line 295
    diff = -diff;
#line 295
    newdiff = -newdiff;
#line 295
}
#line 295
n2=n1+n;
#line 295
if (diff>=0 || newdiff<0) {
#line 295
	SET_IP((Xt *)(((Cell)(IP-1))+ndisp));
#line 295
TAIL;
#line 295
}
#line 295
SUPER_CONTINUE;
#line 295

#line 295

#line 295
\+glocals
#line 295

#line 295
(s+loop)-lp+!# ( #ndisp #nlocals n R:nlimit R:n1 -- R:nlimit R:n2 ) gforth	paren_symmetric_plus_loop_lp_plus_store_number
#line 295
""The run-time procedure compiled by S+LOOP. It loops until the index
#line 295
crosses the boundary between limit and limit-sign(n). I.e. a symmetric
#line 295
version of (+LOOP).""
#line 295
/* !! check this thoroughly */
#line 295
Cell diff = n1-nlimit;
#line 295
Cell newdiff = diff+n;
#line 295
if (n<0) {
#line 295
    diff = -diff;
#line 295
    newdiff = -newdiff;
#line 295
}
#line 295
n2=n1+n;
#line 295
if (diff>=0 || newdiff<0) {
#line 295
    lp += nlocals;
#line 295
SET_IP((Xt *)(((Cell)(IP-2))+ndisp));
#line 295
TAIL;
#line 295
}
#line 295
SUPER_CONTINUE;
#line 295

#line 295
\+
#line 295


\+

unloop	( R:w1 R:w2 -- )	core
/* !! alias for 2rdrop */
:
 r> rdrop rdrop >r ;

(for)	( ncount -- R:nlimit R:ncount )		cmFORTH		paren_for
/* or (for) = >r -- collides with unloop! */
nlimit=0;
:
 r> swap 0 >r >r >r ;

(do)	( nlimit nstart -- R:nlimit R:nstart )	gforth		paren_do
:
 r> swap rot >r >r >r ;

(?do)	( #ndisp nlimit nstart -- R:nlimit R:nstart )	gforth	paren_question_do
if (nstart == nlimit) {
    SET_IP((Xt *)(((Cell)(IP-1))+ndisp));
    TAIL;
}
SUPER_CONTINUE;
:
  2dup =
  IF   r> swap rot >r >r
       dup @ + >r
  ELSE r> swap rot >r >r
       cell+ >r
  THEN ;				\ --> CORE-EXT

\+xconds

(+do)	( #ndisp nlimit nstart -- R:nlimit R:nstart )	gforth	paren_plus_do
if (nstart >= nlimit) {
    SET_IP((Xt *)(((Cell)(IP-1))+ndisp));
    TAIL;
}
SUPER_CONTINUE;
:
 swap 2dup
 r> swap >r swap >r
 >=
 IF
     dup @ +
 ELSE
     cell+
 THEN  >r ;

(u+do)	( #ndisp ulimit ustart -- R:ulimit R:ustart )	gforth	paren_u_plus_do
if (ustart >= ulimit) {
    SET_IP((Xt *)(((Cell)(IP-1))+ndisp));
    TAIL;
}
SUPER_CONTINUE;
:
 swap 2dup
 r> swap >r swap >r
 u>=
 IF
     dup @ +
 ELSE
     cell+
 THEN  >r ;

(-do)	( #ndisp nlimit nstart -- R:nlimit R:nstart )	gforth	paren_minus_do
if (nstart <= nlimit) {
    SET_IP((Xt *)(((Cell)(IP-1))+ndisp));
    TAIL;
}
SUPER_CONTINUE;
:
 swap 2dup
 r> swap >r swap >r
 <=
 IF
     dup @ +
 ELSE
     cell+
 THEN  >r ;

(u-do)	( #ndisp ulimit ustart -- R:ulimit R:ustart )	gforth	paren_u_minus_do
if (ustart <= ulimit) {
    SET_IP((Xt *)(((Cell)(IP-1))+ndisp));
    TAIL;
}
SUPER_CONTINUE;
:
 swap 2dup
 r> swap >r swap >r
 u<=
 IF
     dup @ +
 ELSE
     cell+
 THEN  >r ;

\+

\ don't make any assumptions where the return stack is!!
\ implement this in machine code if it should run quickly!

i	( R:n -- R:n n )		core
:
\ rp@ cell+ @ ;
  r> r> tuck >r >r ;

i'	( R:w R:w2 -- R:w R:w2 w )		gforth		i_tick
:
\ rp@ cell+ cell+ @ ;
  r> r> r> dup itmp ! >r >r >r itmp @ ;
variable itmp

j	( R:n R:d1 -- n R:n R:d1 )		core
:
\ rp@ cell+ cell+ cell+ @ ;
  r> r> r> r> dup itmp ! >r >r >r >r itmp @ ;
[IFUNDEF] itmp variable itmp [THEN]

k	( R:n R:d1 R:d2 -- n R:n R:d1 R:d2 )		gforth
:
\ rp@ [ 5 cells ] Literal + @ ;
  r> r> r> r> r> r> dup itmp ! >r >r >r >r >r >r itmp @ ;
[IFUNDEF] itmp variable itmp [THEN]

\f[THEN]

\ digit is high-level: 0/0%

\g strings

move	( c_from c_to ucount -- )		core
""Copy the contents of @i{ucount} aus at @i{c-from} to
@i{c-to}. @code{move} works correctly even if the two areas overlap.""
/* !! note that the standard specifies addr, not c-addr */
memmove(c_to,c_from,ucount);
/* make an Ifdef for bsd and others? */
:
 >r 2dup u< IF r> cmove> ELSE r> cmove THEN ;

cmove	( c_from c_to u -- )	string	c_move
""Copy the contents of @i{ucount} characters from data space at
@i{c-from} to @i{c-to}. The copy proceeds @code{char}-by-@code{char}
from low address to high address; i.e., for overlapping areas it is
safe if @i{c-to}=<@i{c-from}.""
while (u-- > 0)
  *c_to++ = *c_from++;
:
 bounds ?DO  dup c@ I c! 1+  LOOP  drop ;

cmove>	( c_from c_to u -- )	string	c_move_up
""Copy the contents of @i{ucount} characters from data space at
@i{c-from} to @i{c-to}. The copy proceeds @code{char}-by-@code{char}
from high address to low address; i.e., for overlapping areas it is
safe if @i{c-to}>=@i{c-from}.""
while (u-- > 0)
  c_to[u] = c_from[u];
:
 dup 0= IF  drop 2drop exit  THEN
 rot over + -rot bounds swap 1-
 DO  1- dup c@ I c!  -1 +LOOP  drop ;

fill	( c_addr u c -- )	core
""Store @i{c} in @i{u} chars starting at @i{c-addr}.""
memset(c_addr,c,u);
:
 -rot bounds
 ?DO  dup I c!  LOOP  drop ;

compare	( c_addr1 u1 c_addr2 u2 -- n )	string
""Compare two strings lexicographically. If they are equal, @i{n} is 0; if
the first string is smaller, @i{n} is -1; if the first string is larger, @i{n}
is 1. Currently this is based on the machine's character
comparison. In the future, this may change to consider the current
locale and its collation order.""
/* close ' to keep fontify happy */ 
n = memcmp(c_addr1, c_addr2, u1<u2 ? u1 : u2);
if (n==0)
  n = u1-u2;
if (n<0)
  n = -1;
else if (n>0)
  n = 1;
:
 rot 2dup swap - >r min swap -text dup
 IF  rdrop  ELSE  drop r> sgn  THEN ;
: sgn ( n -- -1/0/1 )
 dup 0= IF EXIT THEN  0< 2* 1+ ;

-text	( c_addr1 u c_addr2 -- n )	new	dash_text
n = memcmp(c_addr1, c_addr2, u);
if (n<0)
  n = -1;
else if (n>0)
  n = 1;
:
 swap bounds
 ?DO  dup c@ I c@ = WHILE  1+  LOOP  drop 0
 ELSE  c@ I c@ - unloop  THEN  sgn ;
: sgn ( n -- -1/0/1 )
 dup 0= IF EXIT THEN  0< 2* 1+ ;

toupper	( c1 -- c2 )	gforth
""If @i{c1} is a lower-case character (in the current locale), @i{c2}
is the equivalent upper-case character. All other characters are unchanged.""
c2 = toupper(c1);
:
 dup [char] a - [ char z char a - 1 + ] Literal u<  bl and - ;

capscomp	( c_addr1 u c_addr2 -- n )	new
n = memcasecmp(c_addr1, c_addr2, u); /* !! use something that works in all locales */
if (n<0)
  n = -1;
else if (n>0)
  n = 1;
:
 swap bounds
 ?DO  dup c@ I c@ <>
     IF  dup c@ toupper I c@ toupper =
     ELSE  true  THEN  WHILE  1+  LOOP  drop 0
 ELSE  c@ toupper I c@ toupper - unloop  THEN  sgn ;

-trailing	( c_addr u1 -- c_addr u2 )		string	dash_trailing
""Adjust the string specified by @i{c-addr, u1} to remove all trailing
spaces. @i{u2} is the length of the modified string.""
u2 = u1;
while (u2>0 && c_addr[u2-1] == ' ')
  u2--;
:
 BEGIN  1- 2dup + c@ bl =  WHILE
        dup  0= UNTIL  ELSE  1+  THEN ;

/string	( c_addr1 u1 n -- c_addr2 u2 )	string	slash_string
""Adjust the string specified by @i{c-addr1, u1} to remove @i{n}
characters from the start of the string.""
c_addr2 = c_addr1+n;
u2 = u1-n;
:
 tuck - >r + r> dup 0< IF  - 0  THEN ;

\g arith

+	( n1 n2 -- n )		core	plus
n = n1+n2;

\ PFE-0.9.14 has it differently, but the next release will have it as follows
under+	( n1 n2 n3 -- n n2 )	gforth	under_plus
""add @i{n3} to @i{n1} (giving @i{n})""
n = n1+n3;
:
 rot + swap ;

-	( n1 n2 -- n )		core	minus
n = n1-n2;
:
 negate + ;

negate	( n1 -- n2 )		core
/* use minus as alias */
n2 = -n1;
:
 invert 1+ ;

1+	( n1 -- n2 )		core		one_plus
n2 = n1+1;
:
 1 + ;

1-	( n1 -- n2 )		core		one_minus
n2 = n1-1;
:
 1 - ;

max	( n1 n2 -- n )	core
if (n1<n2)
  n = n2;
else
  n = n1;
:
 2dup < IF swap THEN drop ;

min	( n1 n2 -- n )	core
if (n1<n2)
  n = n1;
else
  n = n2;
:
 2dup > IF swap THEN drop ;

abs	( n -- u )	core
if (n<0)
  u = -n;
else
  u = n;
:
 dup 0< IF negate THEN ;

*	( n1 n2 -- n )		core	star
n = n1*n2;
:
 um* drop ;

/	( n1 n2 -- n )		core	slash
n = n1/n2;
:
 /mod nip ;

mod	( n1 n2 -- n )		core
n = n1%n2;
:
 /mod drop ;

/mod	( n1 n2 -- n3 n4 )		core		slash_mod
n4 = n1/n2;
n3 = n1%n2; /* !! is this correct? look into C standard! */
:
 >r s>d r> fm/mod ;

2*	( n1 -- n2 )		core		two_star
""Shift left by 1; also works on unsigned numbers""
n2 = 2*n1;
:
 dup + ;

2/	( n1 -- n2 )		core		two_slash
""Arithmetic shift right by 1.  For signed numbers this is a floored
division by 2 (note that @code{/} not necessarily floors).""
n2 = n1>>1;
:
 dup MINI and IF 1 ELSE 0 THEN
 [ bits/byte cell * 1- ] literal 
 0 DO 2* swap dup 2* >r MINI and 
     IF 1 ELSE 0 THEN or r> swap
 LOOP nip ;

fm/mod	( d1 n1 -- n2 n3 )		core		f_m_slash_mod
""Floored division: @i{d1} = @i{n3}*@i{n1}+@i{n2}, @i{n1}>@i{n2}>=0 or 0>=@i{n2}>@i{n1}.""
#ifdef BUGGY_LONG_LONG
DCell r = fmdiv(d1,n1);
n2=r.hi;
n3=r.lo;
#else
/* assumes that the processor uses either floored or symmetric division */
n3 = d1/n1;
n2 = d1%n1;
/* note that this 1%-3>0 is optimized by the compiler */
if (1%-3>0 && (d1<0) != (n1<0) && n2!=0) {
  n3--;
  n2+=n1;
}
#endif
:
 dup >r dup 0< IF  negate >r dnegate r>  THEN
 over       0< IF  tuck + swap  THEN
 um/mod
 r> 0< IF  swap negate swap  THEN ;

sm/rem	( d1 n1 -- n2 n3 )		core		s_m_slash_rem
""Symmetric division: @i{d1} = @i{n3}*@i{n1}+@i{n2}, sign(@i{n2})=sign(@i{d1}) or 0.""
#ifdef BUGGY_LONG_LONG
DCell r = smdiv(d1,n1);
n2=r.hi;
n3=r.lo;
#else
/* assumes that the processor uses either floored or symmetric division */
n3 = d1/n1;
n2 = d1%n1;
/* note that this 1%-3<0 is optimized by the compiler */
if (1%-3<0 && (d1<0) != (n1<0) && n2!=0) {
  n3++;
  n2-=n1;
}
#endif
:
 over >r dup >r abs -rot
 dabs rot um/mod
 r> r@ xor 0< IF       negate       THEN
 r>        0< IF  swap negate swap  THEN ;

m*	( n1 n2 -- d )		core	m_star
#ifdef BUGGY_LONG_LONG
d = mmul(n1,n2);
#else
d = (DCell)n1 * (DCell)n2;
#endif
:
 2dup      0< and >r
 2dup swap 0< and >r
 um* r> - r> - ;

um*	( u1 u2 -- ud )		core	u_m_star
/* use u* as alias */
#ifdef BUGGY_LONG_LONG
ud = ummul(u1,u2);
#else
ud = (UDCell)u1 * (UDCell)u2;
#endif
:
   >r >r 0 0 r> r> [ 8 cells ] literal 0
   DO
       over >r dup >r 0< and d2*+ drop
       r> 2* r> swap
   LOOP 2drop ;
: d2*+ ( ud n -- ud+n c )
   over MINI
   and >r >r 2dup d+ swap r> + swap r> ;

um/mod	( ud u1 -- u2 u3 )		core	u_m_slash_mod
""ud=u3*u1+u2, u1>u2>=0""
#ifdef BUGGY_LONG_LONG
UDCell r = umdiv(ud,u1);
u2=r.hi;
u3=r.lo;
#else
u3 = ud/u1;
u2 = ud%u1;
#endif
:
   0 swap [ 8 cells 1 + ] literal 0
   ?DO /modstep
   LOOP drop swap 1 rshift or swap ;
: /modstep ( ud c R: u -- ud-?u c R: u )
   >r over r@ u< 0= or IF r@ - 1 ELSE 0 THEN  d2*+ r> ;
: d2*+ ( ud n -- ud+n c )
   over MINI
   and >r >r 2dup d+ swap r> + swap r> ;

m+	( d1 n -- d2 )		double		m_plus
#ifdef BUGGY_LONG_LONG
d2.lo = d1.lo+n;
d2.hi = d1.hi - (n<0) + (d2.lo<d1.lo);
#else
d2 = d1+n;
#endif
:
 s>d d+ ;

d+	( d1 d2 -- d )		double	d_plus
#ifdef BUGGY_LONG_LONG
d.lo = d1.lo+d2.lo;
d.hi = d1.hi + d2.hi + (d.lo<d1.lo);
#else
d = d1+d2;
#endif
:
 rot + >r tuck + swap over u> r> swap - ;

d-	( d1 d2 -- d )		double		d_minus
#ifdef BUGGY_LONG_LONG
d.lo = d1.lo - d2.lo;
d.hi = d1.hi-d2.hi-(d1.lo<d2.lo);
#else
d = d1-d2;
#endif
:
 dnegate d+ ;

dnegate	( d1 -- d2 )		double	d_negate
/* use dminus as alias */
#ifdef BUGGY_LONG_LONG
d2 = dnegate(d1);
#else
d2 = -d1;
#endif
:
 invert swap negate tuck 0= - ;

d2*	( d1 -- d2 )		double		d_two_star
""Shift left by 1; also works on unsigned numbers""
#ifdef BUGGY_LONG_LONG
d2.lo = d1.lo<<1;
d2.hi = (d1.hi<<1) | (d1.lo>>(CELL_BITS-1));
#else
d2 = 2*d1;
#endif
:
 2dup d+ ;

d2/	( d1 -- d2 )		double		d_two_slash
""Arithmetic shift right by 1.  For signed numbers this is a floored
division by 2.""
#ifdef BUGGY_LONG_LONG
d2.hi = d1.hi>>1;
d2.lo= (d1.lo>>1) | (d1.hi<<(CELL_BITS-1));
#else
d2 = d1>>1;
#endif
:
 dup 1 and >r 2/ swap 2/ [ 1 8 cells 1- lshift 1- ] Literal and
 r> IF  [ 1 8 cells 1- lshift ] Literal + THEN  swap ;

and	( w1 w2 -- w )		core
w = w1&w2;

or	( w1 w2 -- w )		core
w = w1|w2;
:
 invert swap invert and invert ;

xor	( w1 w2 -- w )		core	x_or
w = w1^w2;

invert	( w1 -- w2 )		core
w2 = ~w1;
:
 MAXU xor ;

rshift	( u1 n -- u2 )		core	r_shift
""Logical shift right by @i{n} bits.""
  u2 = u1>>n;
:
    0 ?DO 2/ MAXI and LOOP ;

lshift	( u1 n -- u2 )		core	l_shift
  u2 = u1<<n;
:
    0 ?DO 2* LOOP ;

\ comparisons(prefix, args, prefix, arg1, arg2, wordsets...)
#line 864


0=	( n -- f )		core	zero_equals
#line 866
f = FLAG(n==0);
#line 866
:
#line 866
    [ char 0x char 0 = [IF]
#line 866
	] IF false ELSE true THEN [
#line 866
    [ELSE]
#line 866
	] xor 0= [
#line 866
    [THEN] ] ;
#line 866

#line 866
0<>	( n -- f )		core-ext	zero_not_equals
#line 866
f = FLAG(n!=0);
#line 866
:
#line 866
    [ char 0x char 0 = [IF]
#line 866
	] IF true ELSE false THEN [
#line 866
    [ELSE]
#line 866
	] xor 0<> [
#line 866
    [THEN] ] ;
#line 866

#line 866
0<	( n -- f )		core	zero_less_than
#line 866
f = FLAG(n<0);
#line 866
:
#line 866
    [ char 0x char 0 = [IF]
#line 866
	] MINI and 0<> [
#line 866
    [ELSE] char 0x char u = [IF]
#line 866
	]   2dup xor 0<  IF nip ELSE - THEN 0<  [
#line 866
	[ELSE]
#line 866
	    ] MINI xor >r MINI xor r> u< [
#line 866
	[THEN]
#line 866
    [THEN] ] ;
#line 866

#line 866
0>	( n -- f )		core-ext	zero_greater_than
#line 866
f = FLAG(n>0);
#line 866
:
#line 866
    [ char 0x char 0 = [IF] ] negate [ [ELSE] ] swap [ [THEN] ]
#line 866
    0< ;
#line 866

#line 866
0<=	( n -- f )		gforth	zero_less_or_equal
#line 866
f = FLAG(n<=0);
#line 866
:
#line 866
    0> 0= ;
#line 866

#line 866
0>=	( n -- f )		gforth	zero_greater_or_equal
#line 866
f = FLAG(n>=0);
#line 866
:
#line 866
    [ char 0x char 0 = [IF] ] negate [ [ELSE] ] swap [ [THEN] ]
#line 866
    0<= ;
#line 866

#line 866

=	( n1 n2 -- f )		core	equals
#line 867
f = FLAG(n1==n2);
#line 867
:
#line 867
    [ char x char 0 = [IF]
#line 867
	] IF false ELSE true THEN [
#line 867
    [ELSE]
#line 867
	] xor 0= [
#line 867
    [THEN] ] ;
#line 867

#line 867
<>	( n1 n2 -- f )		core-ext	not_equals
#line 867
f = FLAG(n1!=n2);
#line 867
:
#line 867
    [ char x char 0 = [IF]
#line 867
	] IF true ELSE false THEN [
#line 867
    [ELSE]
#line 867
	] xor 0<> [
#line 867
    [THEN] ] ;
#line 867

#line 867
<	( n1 n2 -- f )		core	less_than
#line 867
f = FLAG(n1<n2);
#line 867
:
#line 867
    [ char x char 0 = [IF]
#line 867
	] MINI and 0<> [
#line 867
    [ELSE] char x char u = [IF]
#line 867
	]   2dup xor 0<  IF nip ELSE - THEN 0<  [
#line 867
	[ELSE]
#line 867
	    ] MINI xor >r MINI xor r> u< [
#line 867
	[THEN]
#line 867
    [THEN] ] ;
#line 867

#line 867
>	( n1 n2 -- f )		core	greater_than
#line 867
f = FLAG(n1>n2);
#line 867
:
#line 867
    [ char x char 0 = [IF] ] negate [ [ELSE] ] swap [ [THEN] ]
#line 867
    < ;
#line 867

#line 867
<=	( n1 n2 -- f )		gforth	less_or_equal
#line 867
f = FLAG(n1<=n2);
#line 867
:
#line 867
    > 0= ;
#line 867

#line 867
>=	( n1 n2 -- f )		gforth	greater_or_equal
#line 867
f = FLAG(n1>=n2);
#line 867
:
#line 867
    [ char x char 0 = [IF] ] negate [ [ELSE] ] swap [ [THEN] ]
#line 867
    <= ;
#line 867

#line 867

u=	( u1 u2 -- f )		gforth	u_equals
#line 868
f = FLAG(u1==u2);
#line 868
:
#line 868
    [ char ux char 0 = [IF]
#line 868
	] IF false ELSE true THEN [
#line 868
    [ELSE]
#line 868
	] xor 0= [
#line 868
    [THEN] ] ;
#line 868

#line 868
u<>	( u1 u2 -- f )		gforth	u_not_equals
#line 868
f = FLAG(u1!=u2);
#line 868
:
#line 868
    [ char ux char 0 = [IF]
#line 868
	] IF true ELSE false THEN [
#line 868
    [ELSE]
#line 868
	] xor 0<> [
#line 868
    [THEN] ] ;
#line 868

#line 868
u<	( u1 u2 -- f )		core	u_less_than
#line 868
f = FLAG(u1<u2);
#line 868
:
#line 868
    [ char ux char 0 = [IF]
#line 868
	] MINI and 0<> [
#line 868
    [ELSE] char ux char u = [IF]
#line 868
	]   2dup xor 0<  IF nip ELSE - THEN 0<  [
#line 868
	[ELSE]
#line 868
	    ] MINI xor >r MINI xor r> u< [
#line 868
	[THEN]
#line 868
    [THEN] ] ;
#line 868

#line 868
u>	( u1 u2 -- f )		core-ext	u_greater_than
#line 868
f = FLAG(u1>u2);
#line 868
:
#line 868
    [ char ux char 0 = [IF] ] negate [ [ELSE] ] swap [ [THEN] ]
#line 868
    u< ;
#line 868

#line 868
u<=	( u1 u2 -- f )		gforth	u_less_or_equal
#line 868
f = FLAG(u1<=u2);
#line 868
:
#line 868
    u> 0= ;
#line 868

#line 868
u>=	( u1 u2 -- f )		gforth	u_greater_or_equal
#line 868
f = FLAG(u1>=u2);
#line 868
:
#line 868
    [ char ux char 0 = [IF] ] negate [ [ELSE] ] swap [ [THEN] ]
#line 868
    u<= ;
#line 868

#line 868


\ dcomparisons(prefix, args, prefix, arg1, arg2, wordsets...)
#line 914


\+dcomps

d=	( d1 d2 -- f )		double	d_equals
#line 918
#ifdef BUGGY_LONG_LONG
#line 918
f = FLAG(d1.lo==d2.lo && d1.hi==d2.hi);
#line 918
#else
#line 918
f = FLAG(d1==d2);
#line 918
#endif
#line 918

#line 918
d<>	( d1 d2 -- f )		gforth	d_not_equals
#line 918
#ifdef BUGGY_LONG_LONG
#line 918
f = FLAG(d1.lo!=d2.lo || d1.hi!=d2.hi);
#line 918
#else
#line 918
f = FLAG(d1!=d2);
#line 918
#endif
#line 918

#line 918
d<	( d1 d2 -- f )		double	d_less_than
#line 918
#ifdef BUGGY_LONG_LONG
#line 918
f = FLAG(d1.hi==d2.hi ? d1.lo<d2.lo : d1.hi<d2.hi);
#line 918
#else
#line 918
f = FLAG(d1<d2);
#line 918
#endif
#line 918

#line 918
d>	( d1 d2 -- f )		gforth	d_greater_than
#line 918
#ifdef BUGGY_LONG_LONG
#line 918
f = FLAG(d1.hi==d2.hi ? d1.lo>d2.lo : d1.hi>d2.hi);
#line 918
#else
#line 918
f = FLAG(d1>d2);
#line 918
#endif
#line 918

#line 918
d<=	( d1 d2 -- f )		gforth	d_less_or_equal
#line 918
#ifdef BUGGY_LONG_LONG
#line 918
f = FLAG(d1.hi==d2.hi ? d1.lo<=d2.lo : d1.hi<=d2.hi);
#line 918
#else
#line 918
f = FLAG(d1<=d2);
#line 918
#endif
#line 918

#line 918
d>=	( d1 d2 -- f )		gforth	d_greater_or_equal
#line 918
#ifdef BUGGY_LONG_LONG
#line 918
f = FLAG(d1.hi==d2.hi ? d1.lo>=d2.lo : d1.hi>=d2.hi);
#line 918
#else
#line 918
f = FLAG(d1>=d2);
#line 918
#endif
#line 918

#line 918

d0=	( d -- f )		double	d_zero_equals
#line 919
#ifdef BUGGY_LONG_LONG
#line 919
f = FLAG(d.lo==DZERO.lo && d.hi==DZERO.hi);
#line 919
#else
#line 919
f = FLAG(d==DZERO);
#line 919
#endif
#line 919

#line 919
d0<>	( d -- f )		gforth	d_zero_not_equals
#line 919
#ifdef BUGGY_LONG_LONG
#line 919
f = FLAG(d.lo!=DZERO.lo || d.hi!=DZERO.hi);
#line 919
#else
#line 919
f = FLAG(d!=DZERO);
#line 919
#endif
#line 919

#line 919
d0<	( d -- f )		double	d_zero_less_than
#line 919
#ifdef BUGGY_LONG_LONG
#line 919
f = FLAG(d.hi==DZERO.hi ? d.lo<DZERO.lo : d.hi<DZERO.hi);
#line 919
#else
#line 919
f = FLAG(d<DZERO);
#line 919
#endif
#line 919

#line 919
d0>	( d -- f )		gforth	d_zero_greater_than
#line 919
#ifdef BUGGY_LONG_LONG
#line 919
f = FLAG(d.hi==DZERO.hi ? d.lo>DZERO.lo : d.hi>DZERO.hi);
#line 919
#else
#line 919
f = FLAG(d>DZERO);
#line 919
#endif
#line 919

#line 919
d0<=	( d -- f )		gforth	d_zero_less_or_equal
#line 919
#ifdef BUGGY_LONG_LONG
#line 919
f = FLAG(d.hi==DZERO.hi ? d.lo<=DZERO.lo : d.hi<=DZERO.hi);
#line 919
#else
#line 919
f = FLAG(d<=DZERO);
#line 919
#endif
#line 919

#line 919
d0>=	( d -- f )		gforth	d_zero_greater_or_equal
#line 919
#ifdef BUGGY_LONG_LONG
#line 919
f = FLAG(d.hi==DZERO.hi ? d.lo>=DZERO.lo : d.hi>=DZERO.hi);
#line 919
#else
#line 919
f = FLAG(d>=DZERO);
#line 919
#endif
#line 919

#line 919

du=	( ud1 ud2 -- f )		gforth	d_u_equals
#line 920
#ifdef BUGGY_LONG_LONG
#line 920
f = FLAG(ud1.lo==ud2.lo && ud1.hi==ud2.hi);
#line 920
#else
#line 920
f = FLAG(ud1==ud2);
#line 920
#endif
#line 920

#line 920
du<>	( ud1 ud2 -- f )		gforth	d_u_not_equals
#line 920
#ifdef BUGGY_LONG_LONG
#line 920
f = FLAG(ud1.lo!=ud2.lo || ud1.hi!=ud2.hi);
#line 920
#else
#line 920
f = FLAG(ud1!=ud2);
#line 920
#endif
#line 920

#line 920
du<	( ud1 ud2 -- f )		double-ext	d_u_less_than
#line 920
#ifdef BUGGY_LONG_LONG
#line 920
f = FLAG(ud1.hi==ud2.hi ? ud1.lo<ud2.lo : ud1.hi<ud2.hi);
#line 920
#else
#line 920
f = FLAG(ud1<ud2);
#line 920
#endif
#line 920

#line 920
du>	( ud1 ud2 -- f )		gforth	d_u_greater_than
#line 920
#ifdef BUGGY_LONG_LONG
#line 920
f = FLAG(ud1.hi==ud2.hi ? ud1.lo>ud2.lo : ud1.hi>ud2.hi);
#line 920
#else
#line 920
f = FLAG(ud1>ud2);
#line 920
#endif
#line 920

#line 920
du<=	( ud1 ud2 -- f )		gforth	d_u_less_or_equal
#line 920
#ifdef BUGGY_LONG_LONG
#line 920
f = FLAG(ud1.hi==ud2.hi ? ud1.lo<=ud2.lo : ud1.hi<=ud2.hi);
#line 920
#else
#line 920
f = FLAG(ud1<=ud2);
#line 920
#endif
#line 920

#line 920
du>=	( ud1 ud2 -- f )		gforth	d_u_greater_or_equal
#line 920
#ifdef BUGGY_LONG_LONG
#line 920
f = FLAG(ud1.hi==ud2.hi ? ud1.lo>=ud2.lo : ud1.hi>=ud2.hi);
#line 920
#else
#line 920
f = FLAG(ud1>=ud2);
#line 920
#endif
#line 920

#line 920


\+

within	( u1 u2 u3 -- f )		core-ext
""u2=<u1<u3 or: u3=<u2 and u1 is not in [u3,u2).  This works for
unsigned and signed numbers (but not a mixture).  Another way to think
about this word is to consider the numbers as a circle (wrapping
around from @code{max-u} to 0 for unsigned, and from @code{max-n} to
min-n for signed numbers); now consider the range from u2 towards
increasing numbers up to and excluding u3 (giving an empty range if
u2=u3); if u1 is in this range, @code{within} returns true.""
f = FLAG(u1-u2 < u3-u2);
:
 over - >r - r> u< ;

\g internal

sp@	( -- a_addr )		gforth		sp_fetch
a_addr = sp+1;

sp!	( a_addr -- )		gforth		sp_store
sp = a_addr;
/* works with and without spTOS caching */

rp@	( -- a_addr )		gforth		rp_fetch
a_addr = rp;

rp!	( a_addr -- )		gforth		rp_store
rp = a_addr;

\+floating

fp@	( -- f_addr )	gforth	fp_fetch
f_addr = fp;

fp!	( f_addr -- )	gforth	fp_store
fp = f_addr;

\+

;s	( R:w -- )		gforth	semis
""The primitive compiled by @code{EXIT}.""
SET_IP((Xt *)w);

\g stack

>r	( w -- R:w )		core	to_r
:
 (>r) ;
: (>r)  rp@ cell+ @ rp@ ! rp@ cell+ ! ;

r>	( R:w -- w )		core	r_from
:
 rp@ cell+ @ rp@ @ rp@ cell+ ! (rdrop) rp@ ! ;
Create (rdrop) ' ;s A,

rdrop	( R:w -- )		gforth
:
 r> r> drop >r ;

2>r	( w1 w2 -- R:w1 R:w2 )	core-ext	two_to_r
:
 swap r> swap >r swap >r >r ;

2r>	( R:w1 R:w2 -- w1 w2 )	core-ext	two_r_from
:
 r> r> swap r> swap >r swap ;

2r@	( R:w1 R:w2 -- R:w1 R:w2 w1 w2 )	core-ext	two_r_fetch
:
 i' j ;

2rdrop	(  R:w1 R:w2 -- )		gforth	two_r_drop
:
 r> r> drop r> drop >r ;

over	( w1 w2 -- w1 w2 w1 )		core
:
 sp@ cell+ @ ;

drop	( w -- )		core
:
 IF THEN ;

swap	( w1 w2 -- w2 w1 )		core
:
 >r (swap) ! r> (swap) @ ;
Variable (swap)

dup	( w -- w w )		core	dupe
:
 sp@ @ ;

rot	( w1 w2 w3 -- w2 w3 w1 )	core	rote
:
[ defined? (swap) [IF] ]
    (swap) ! (rot) ! >r (rot) @ (swap) @ r> ;
Variable (rot)
[ELSE] ]
    >r swap r> swap ;
[THEN]

-rot	( w1 w2 w3 -- w3 w1 w2 )	gforth	not_rote
:
 rot rot ;

nip	( w1 w2 -- w2 )		core-ext
:
 swap drop ;

tuck	( w1 w2 -- w2 w1 w2 )	core-ext
:
 swap over ;

?dup	( w -- w )			core	question_dupe
""Actually the stack effect is: @code{( w -- 0 | w w )}.  It performs a
@code{dup} if w is nonzero.""
if (w!=0) {
  IF_spTOS(*sp-- = w;)
#ifndef USE_TOS
  *--sp = w;
#endif
}
:
 dup IF dup THEN ;

pick	( u -- w )			core-ext
""Actually the stack effect is @code{ x0 ... xu u -- x0 ... xu x0 }.""
w = sp[u+1];
:
 1+ cells sp@ + @ ;

2drop	( w1 w2 -- )		core	two_drop
:
 drop drop ;

2dup	( w1 w2 -- w1 w2 w1 w2 )	core	two_dupe
:
 over over ;

2over	( w1 w2 w3 w4 -- w1 w2 w3 w4 w1 w2 )	core	two_over
:
 3 pick 3 pick ;

2swap	( w1 w2 w3 w4 -- w3 w4 w1 w2 )	core	two_swap
:
 rot >r rot r> ;

2rot	( w1 w2 w3 w4 w5 w6 -- w3 w4 w5 w6 w1 w2 )	double-ext	two_rote
:
 >r >r 2swap r> r> 2swap ;

2nip	( w1 w2 w3 w4 -- w3 w4 )	gforth	two_nip
:
 2swap 2drop ;

2tuck	( w1 w2 w3 w4 -- w3 w4 w1 w2 w3 w4 )	gforth	two_tuck
:
 2swap 2over ;

\ toggle is high-level: 0.11/0.42%

@	( a_addr -- w )		core	fetch
""@i{w} is the cell stored at @i{a_addr}.""
w = *a_addr;

!	( w a_addr -- )		core	store
""Store @i{w} into the cell at @i{a-addr}.""
*a_addr = w;

+!	( n a_addr -- )		core	plus_store
""Add @i{n} to the cell at @i{a-addr}.""
*a_addr += n;
:
 tuck @ + swap ! ;

c@	( c_addr -- c )		core	c_fetch
""@i{c} is the char stored at @i{c_addr}.""
c = *c_addr;
:
[ bigendian [IF] ]
    [ cell>bit 4 = [IF] ]
	dup [ 0 cell - ] Literal and @ swap 1 and
	IF  $FF and  ELSE  8>>  THEN  ;
    [ [ELSE] ]
	dup [ cell 1- ] literal and
	tuck - @ swap [ cell 1- ] literal xor
 	0 ?DO 8>> LOOP $FF and
    [ [THEN] ]
[ [ELSE] ]
    [ cell>bit 4 = [IF] ]
	dup [ 0 cell - ] Literal and @ swap 1 and
	IF  8>>  ELSE  $FF and  THEN
    [ [ELSE] ]
	dup [ cell  1- ] literal and 
	tuck - @ swap
	0 ?DO 8>> LOOP 255 and
    [ [THEN] ]
[ [THEN] ]
;
: 8>> 2/ 2/ 2/ 2/  2/ 2/ 2/ 2/ ;

c!	( c c_addr -- )		core	c_store
""Store @i{c} into the char at @i{c-addr}.""
*c_addr = c;
:
[ bigendian [IF] ]
    [ cell>bit 4 = [IF] ]
	tuck 1 and IF  $FF and  ELSE  8<<  THEN >r
	dup -2 and @ over 1 and cells masks + @ and
	r> or swap -2 and ! ;
	Create masks $00FF , $FF00 ,
    [ELSE] ]
	dup [ cell 1- ] literal and dup 
	[ cell 1- ] literal xor >r
	- dup @ $FF r@ 0 ?DO 8<< LOOP invert and
	rot $FF and r> 0 ?DO 8<< LOOP or swap ! ;
    [THEN]
[ELSE] ]
    [ cell>bit 4 = [IF] ]
	tuck 1 and IF  8<<  ELSE  $FF and  THEN >r
	dup -2 and @ over 1 and cells masks + @ and
	r> or swap -2 and ! ;
	Create masks $FF00 , $00FF ,
    [ELSE] ]
	dup [ cell 1- ] literal and dup >r
	- dup @ $FF r@ 0 ?DO 8<< LOOP invert and
	rot $FF and r> 0 ?DO 8<< LOOP or swap ! ;
    [THEN]
[THEN]
: 8<< 2* 2* 2* 2*  2* 2* 2* 2* ;

2!	( w1 w2 a_addr -- )		core	two_store
""Store @i{w2} into the cell at @i{c-addr} and @i{w1} into the next cell.""
a_addr[0] = w2;
a_addr[1] = w1;
:
 tuck ! cell+ ! ;

2@	( a_addr -- w1 w2 )		core	two_fetch
""@i{w2} is the content of the cell stored at @i{a-addr}, @i{w1} is
the content of the next cell.""
w2 = a_addr[0];
w1 = a_addr[1];
:
 dup cell+ @ swap @ ;

cell+	( a_addr1 -- a_addr2 )	core	cell_plus
""@code{1 cells +}""
a_addr2 = a_addr1+1;
:
 cell + ;

cells	( n1 -- n2 )		core
"" @i{n2} is the number of address units of @i{n1} cells.""
n2 = n1 * sizeof(Cell);
:
 [ cell
 2/ dup [IF] ] 2* [ [THEN]
 2/ dup [IF] ] 2* [ [THEN]
 2/ dup [IF] ] 2* [ [THEN]
 2/ dup [IF] ] 2* [ [THEN]
 drop ] ;

char+	( c_addr1 -- c_addr2 )	core	char_plus
""@code{1 chars +}.""
c_addr2 = c_addr1 + 1;
:
 1+ ;

(chars)	( n1 -- n2 )	gforth	paren_chars
n2 = n1 * sizeof(Char);
:
 ;

count	( c_addr1 -- c_addr2 u )	core
""@i{c-addr2} is the first character and @i{u} the length of the
counted string at @i{c-addr1}.""
u = *c_addr1;
c_addr2 = c_addr1+1;
:
 dup 1+ swap c@ ;

(f83find)	( c_addr u f83name1 -- f83name2 )	new	paren_f83find
for (; f83name1 != NULL; f83name1 = (struct F83Name *)(f83name1->next))
  if ((UCell)F83NAME_COUNT(f83name1)==u &&
      memcasecmp(c_addr, f83name1->name, u)== 0 /* or inline? */)
    break;
f83name2=f83name1;
:
    BEGIN  dup WHILE  (find-samelen)  dup  WHILE
	>r 2dup r@ cell+ char+ capscomp  0=
	IF  2drop r>  EXIT  THEN
	r> @
    REPEAT  THEN  nip nip ;
: (find-samelen) ( u f83name1 -- u f83name2/0 )
    BEGIN  2dup cell+ c@ $1F and <> WHILE  @  dup 0= UNTIL THEN ;

\+hash

(hashfind)	( c_addr u a_addr -- f83name2 )	new	paren_hashfind
struct F83Name *f83name1;
f83name2=NULL;
while(a_addr != NULL)
{
   f83name1=(struct F83Name *)(a_addr[1]);
   a_addr=(Cell *)(a_addr[0]);
   if ((UCell)F83NAME_COUNT(f83name1)==u &&
       memcasecmp(c_addr, f83name1->name, u)== 0 /* or inline? */)
     {
	f83name2=f83name1;
	break;
     }
}
:
 BEGIN  dup  WHILE
        2@ >r >r dup r@ cell+ c@ $1F and =
        IF  2dup r@ cell+ char+ capscomp 0=
	    IF  2drop r> rdrop  EXIT  THEN  THEN
	rdrop r>
 REPEAT nip nip ;

(tablefind)	( c_addr u a_addr -- f83name2 )	new	paren_tablefind
""A case-sensitive variant of @code{(hashfind)}""
struct F83Name *f83name1;
f83name2=NULL;
while(a_addr != NULL)
{
   f83name1=(struct F83Name *)(a_addr[1]);
   a_addr=(Cell *)(a_addr[0]);
   if ((UCell)F83NAME_COUNT(f83name1)==u &&
       memcmp(c_addr, f83name1->name, u)== 0 /* or inline? */)
     {
	f83name2=f83name1;
	break;
     }
}
:
 BEGIN  dup  WHILE
        2@ >r >r dup r@ cell+ c@ $1F and =
        IF  2dup r@ cell+ char+ -text 0=
	    IF  2drop r> rdrop  EXIT  THEN  THEN
	rdrop r>
 REPEAT nip nip ;

(hashkey)	( c_addr u1 -- u2 )		gforth	paren_hashkey
u2=0;
while(u1--)
   u2+=(Cell)toupper(*c_addr++);
:
 0 -rot bounds ?DO  I c@ toupper +  LOOP ;

(hashkey1)	( c_addr u ubits -- ukey )		gforth	paren_hashkey1
""ukey is the hash key for the string c_addr u fitting in ubits bits""
/* this hash function rotates the key at every step by rot bits within
   ubits bits and xors it with the character. This function does ok in
   the chi-sqare-test.  Rot should be <=7 (preferably <=5) for
   ASCII strings (larger if ubits is large), and should share no
   divisors with ubits.
*/
unsigned rot = ((char []){5,0,1,2,3,4,5,5,5,5,3,5,5,5,5,7,5,5,5,5,7,5,5,5,5,6,5,5,5,5,7,5,5})[ubits];
Char *cp = c_addr;
for (ukey=0; cp<c_addr+u; cp++)
    ukey = ((((ukey<<rot) | (ukey>>(ubits-rot))) 
	     ^ toupper(*cp))
	    & ((1<<ubits)-1));
:
 dup rot-values + c@ over 1 swap lshift 1- >r
 tuck - 2swap r> 0 2swap bounds
 ?DO  dup 4 pick lshift swap 3 pick rshift or
      I c@ toupper xor
      over and  LOOP
 nip nip nip ;
Create rot-values
  5 c, 0 c, 1 c, 2 c, 3 c,  4 c, 5 c, 5 c, 5 c, 5 c,
  3 c, 5 c, 5 c, 5 c, 5 c,  7 c, 5 c, 5 c, 5 c, 5 c,
  7 c, 5 c, 5 c, 5 c, 5 c,  6 c, 5 c, 5 c, 5 c, 5 c,
  7 c, 5 c, 5 c,

\+

(parse-white)	( c_addr1 u1 -- c_addr2 u2 )	gforth	paren_parse_white
/* use !isgraph instead of isspace? */
Char *endp = c_addr1+u1;
while (c_addr1<endp && isspace(*c_addr1))
  c_addr1++;
if (c_addr1<endp) {
  for (c_addr2 = c_addr1; c_addr1<endp && !isspace(*c_addr1); c_addr1++)
    ;
  u2 = c_addr1-c_addr2;
}
else {
  c_addr2 = c_addr1;
  u2 = 0;
}
:
 BEGIN  dup  WHILE  over c@ bl <=  WHILE  1 /string
 REPEAT  THEN  2dup
 BEGIN  dup  WHILE  over c@ bl >   WHILE  1 /string
 REPEAT  THEN  nip - ;

aligned	( c_addr -- a_addr )	core
"" @i{a-addr} is the first aligned address greater than or equal to @i{c-addr}.""
a_addr = (Cell *)((((Cell)c_addr)+(sizeof(Cell)-1))&(-sizeof(Cell)));
:
 [ cell 1- ] Literal + [ -1 cells ] Literal and ;

faligned	( c_addr -- f_addr )	float	f_aligned
"" @i{f-addr} is the first float-aligned address greater than or equal to @i{c-addr}.""
f_addr = (Float *)((((Cell)c_addr)+(sizeof(Float)-1))&(-sizeof(Float)));
:
 [ 1 floats 1- ] Literal + [ -1 floats ] Literal and ;

>body	( xt -- a_addr )	core	to_body
"" Get the address of the body of the word represented by @i{xt} (the address
of the word's data field).""
a_addr = PFA(xt);
:
    2 cells + ;

\ threading stuff is currently only interesting if we have a compiler
\fhas? standardthreading has? compiler and [IF]

>code-address	( xt -- c_addr )		gforth	to_code_address
""@i{c-addr} is the code address of the word @i{xt}.""
/* !! This behaves installation-dependently for DOES-words */
c_addr = (Address)CODE_ADDRESS(xt);
:
    @ ;

>does-code	( xt -- a_addr )		gforth	to_does_code
""If @i{xt} is the execution token of a child of a @code{DOES>} word,
@i{a-addr} is the start of the Forth code after the @code{DOES>};
Otherwise @i{a-addr} is 0.""
a_addr = (Cell *)DOES_CODE(xt);
:
    cell+ @ ;

code-address!	( c_addr xt -- )		gforth	code_address_store
""Create a code field with code address @i{c-addr} at @i{xt}.""
MAKE_CF(xt, c_addr);
:
    ! ;

does-code!	( a_addr xt -- )		gforth	does_code_store
""Create a code field at @i{xt} for a child of a @code{DOES>}-word;
@i{a-addr} is the start of the Forth code after @code{DOES>}.""
MAKE_DOES_CF(xt, a_addr);
:
    dodoes: over ! cell+ ! ;

does-handler!	( a_addr -- )	gforth	does_handler_store
""Create a @code{DOES>}-handler at address @i{a-addr}. Normally,
@i{a-addr} points just behind a @code{DOES>}.""
MAKE_DOES_HANDLER(a_addr);
:
    drop ;

/does-handler	( -- n )	gforth	slash_does_handler
""The size of a @code{DOES>}-handler (includes possible padding).""
/* !! a constant or environmental query might be better */
n = DOES_HANDLER_SIZE;
:
    2 cells ;

threading-method	( -- n )	gforth	threading_method
""0 if the engine is direct threaded. Note that this may change during
the lifetime of an image.""
#if defined(DOUBLY_INDIRECT)
n=2;
#else
# if defined(DIRECT_THREADED)
n=0;
# else
n=1;
# endif
#endif
:
 1 ;

\f[THEN]

\g hostos

key-file	( wfileid -- n )		gforth	paren_key_file
#ifdef HAS_FILE
fflush(stdout);
n = key((FILE*)wfileid);
#else
n = key(stdin);
#endif

key?-file	( wfileid -- n )		facility	key_q_file
#ifdef HAS_FILE
fflush(stdout);
n = key_query((FILE*)wfileid);
#else
n = key_query(stdin);
#endif

\+os

stdin	( -- wfileid )	gforth
wfileid = (Cell)stdin;

stdout	( -- wfileid )	gforth
wfileid = (Cell)stdout;

stderr	( -- wfileid )	gforth
wfileid = (Cell)stderr;

form	( -- urows ucols )	gforth
""The number of lines and columns in the terminal. These numbers may change
with the window size.""
/* we could block SIGWINCH here to get a consistent size, but I don't
 think this is necessary or always beneficial */
urows=rows;
ucols=cols;

flush-icache	( c_addr u -- )	gforth	flush_icache
""Make sure that the instruction cache of the processor (if there is
one) does not contain stale data at @i{c-addr} and @i{u} bytes
afterwards. @code{END-CODE} performs a @code{flush-icache}
automatically. Caveat: @code{flush-icache} might not work on your
installation; this is usually the case if direct threading is not
supported on your machine (take a look at your @file{machine.h}) and
your machine has a separate instruction cache. In such cases,
@code{flush-icache} does nothing instead of flushing the instruction
cache.""
FLUSH_ICACHE(c_addr,u);

(bye)	( n -- )	gforth	paren_bye
SUPER_END;
return (Label *)n;

(system)	( c_addr u -- wretval wior )	gforth	peren_system
#ifndef MSDOS
int old_tp=terminal_prepped;
deprep_terminal();
#endif
wretval=system(cstr(c_addr,u,1)); /* ~ expansion on first part of string? */
wior = IOR(wretval==-1 || (wretval==127 && errno != 0));
#ifndef MSDOS
if (old_tp)
  prep_terminal();
#endif

getenv	( c_addr1 u1 -- c_addr2 u2 )	gforth
""The string @i{c-addr1 u1} specifies an environment variable. The string @i{c-addr2 u2}
is the host operating system's expansion of that environment variable. If the
environment variable does not exist, @i{c-addr2 u2} specifies a string 0 characters
in length.""
/* close ' to keep fontify happy */
c_addr2 = getenv(cstr(c_addr1,u1,1));
u2 = (c_addr2 == NULL ? 0 : strlen(c_addr2));

open-pipe	( c_addr u wfam -- wfileid wior )	gforth	open_pipe
wfileid=(Cell)popen(cstr(c_addr,u,1),pfileattr[wfam]); /* ~ expansion of 1st arg? */
wior = IOR(wfileid==0); /* !! the man page says that errno is not set reliably */

close-pipe	( wfileid -- wretval wior )		gforth	close_pipe
wretval = pclose((FILE *)wfileid);
wior = IOR(wretval==-1);

time&date	( -- nsec nmin nhour nday nmonth nyear )	facility-ext	time_and_date
""Report the current time of day. Seconds, minutes and hours are numbered from 0.
Months are numbered from 1.""
struct timeval time1;
struct timezone zone1;
struct tm *ltime;
gettimeofday(&time1,&zone1);
/* !! Single Unix specification: 
   If tzp is not a null pointer, the behaviour is unspecified. */
ltime=localtime((time_t *)&time1.tv_sec);
nyear =ltime->tm_year+1900;
nmonth=ltime->tm_mon+1;
nday  =ltime->tm_mday;
nhour =ltime->tm_hour;
nmin  =ltime->tm_min;
nsec  =ltime->tm_sec;

ms	( n -- )	facility-ext
""Wait at least @i{n} milli-second.""
struct timeval timeout;
timeout.tv_sec=n/1000;
timeout.tv_usec=1000*(n%1000);
(void)select(0,0,0,0,&timeout);

allocate	( u -- a_addr wior )	memory
""Allocate @i{u} address units of contiguous data space. The initial
contents of the data space is undefined. If the allocation is successful,
@i{a-addr} is the start address of the allocated region and @i{wior}
is 0. If the allocation fails, @i{a-addr} is undefined and @i{wior}
is a non-zero I/O result code.""
a_addr = (Cell *)malloc(u?u:1);
wior = IOR(a_addr==NULL);

free	( a_addr -- wior )		memory
""Return the region of data space starting at @i{a-addr} to the system.
The region must originally have been obtained using @code{allocate} or
@code{resize}. If the operational is successful, @i{wior} is 0.
If the operation fails, @i{wior} is a non-zero I/O result code.""
free(a_addr);
wior = 0;

resize	( a_addr1 u -- a_addr2 wior )	memory
""Change the size of the allocated area at @i{a-addr1} to @i{u}
address units, possibly moving the contents to a different
area. @i{a-addr2} is the address of the resulting area.
If the operation is successful, @i{wior} is 0.
If the operation fails, @i{wior} is a non-zero
I/O result code. If @i{a-addr1} is 0, Gforth's (but not the Standard)
@code{resize} @code{allocate}s @i{u} address units.""
/* the following check is not necessary on most OSs, but it is needed
   on SunOS 4.1.2. */
/* close ' to keep fontify happy */
if (a_addr1==NULL)
  a_addr2 = (Cell *)malloc(u);
else
  a_addr2 = (Cell *)realloc(a_addr1, u);
wior = IOR(a_addr2==NULL);	/* !! Define a return code */

strerror	( n -- c_addr u )	gforth
c_addr = strerror(n);
u = strlen(c_addr);

strsignal	( n -- c_addr u )	gforth
c_addr = strsignal(n);
u = strlen(c_addr);

call-c	( w -- )	gforth	call_c
""Call the C function pointed to by @i{w}. The C function has to
access the stack itself. The stack pointers are exported in the global
variables @code{SP} and @code{FP}.""
/* This is a first attempt at support for calls to C. This may change in
   the future */
IF_fpTOS(fp[0]=fpTOS);
FP=fp;
SP=sp;
((void (*)())w)();
sp=SP;
fp=FP;
IF_spTOS(spTOS=sp[0]);
IF_fpTOS(fpTOS=fp[0]);

\+
\+file

close-file	( wfileid -- wior )		file	close_file
wior = IOR(fclose((FILE *)wfileid)==EOF);

open-file	( c_addr u wfam -- wfileid wior )	file	open_file
wfileid = (Cell)fopen(tilde_cstr(c_addr, u, 1), fileattr[wfam]);
wior =  IOR(wfileid == 0);

create-file	( c_addr u wfam -- wfileid wior )	file	create_file
Cell	fd;
fd = open(tilde_cstr(c_addr, u, 1), O_CREAT|O_TRUNC|ufileattr[wfam], 0666);
if (fd != -1) {
  wfileid = (Cell)fdopen(fd, fileattr[wfam]);
  wior = IOR(wfileid == 0);
} else {
  wfileid = 0;
  wior = IOR(1);
}

delete-file	( c_addr u -- wior )		file	delete_file
wior = IOR(unlink(tilde_cstr(c_addr, u, 1))==-1);

rename-file	( c_addr1 u1 c_addr2 u2 -- wior )	file-ext	rename_file
""Rename file @i{c_addr1 u1} to new name @i{c_addr2 u2}""
char *s1=tilde_cstr(c_addr2, u2, 1);
wior = IOR(rename(tilde_cstr(c_addr1, u1, 0), s1)==-1);

file-position	( wfileid -- ud wior )	file	file_position
/* !! use tell and lseek? */
ud = LONG2UD(ftell((FILE *)wfileid));
wior = IOR(UD2LONG(ud)==-1);

reposition-file	( ud wfileid -- wior )	file	reposition_file
wior = IOR(fseek((FILE *)wfileid, UD2LONG(ud), SEEK_SET)==-1);

file-size	( wfileid -- ud wior )	file	file_size
struct stat buf;
wior = IOR(fstat(fileno((FILE *)wfileid), &buf)==-1);
ud = LONG2UD(buf.st_size);

resize-file	( ud wfileid -- wior )	file	resize_file
wior = IOR(ftruncate(fileno((FILE *)wfileid), UD2LONG(ud))==-1);

read-file	( c_addr u1 wfileid -- u2 wior )	file	read_file
/* !! fread does not guarantee enough */
u2 = fread(c_addr, sizeof(Char), u1, (FILE *)wfileid);
wior = FILEIO(u2<u1 && ferror((FILE *)wfileid));
/* !! is the value of ferror errno-compatible? */
if (wior)
  clearerr((FILE *)wfileid);

read-line	( c_addr u1 wfileid -- u2 flag wior )	file	read_line
/* this may one day be replaced with : read-line (read-line) nip ; */
Cell c;
flag=-1;
for(u2=0; u2<u1; u2++)
{
   c = getc((FILE *)wfileid);
   if (c=='\n') break;
   if (c=='\r') {
     if ((c = getc((FILE *)wfileid))!='\n')
       ungetc(c,(FILE *)wfileid);
     break;
   }
   if (c==EOF) {
	flag=FLAG(u2!=0);
	break;
     }
   c_addr[u2] = (Char)c;
}
wior=FILEIO(ferror((FILE *)wfileid));

\+

write-file	( c_addr u1 wfileid -- wior )	file	write_file
/* !! fwrite does not guarantee enough */
#ifdef HAS_FILE
{
  UCell u2 = fwrite(c_addr, sizeof(Char), u1, (FILE *)wfileid);
  wior = FILEIO(u2<u1 && ferror((FILE *)wfileid));
  if (wior)
    clearerr((FILE *)wfileid);
}
#else
TYPE(c_addr, u1);
#endif

emit-file	( c wfileid -- wior )	gforth	emit_file
#ifdef HAS_FILE
wior = FILEIO(putc(c, (FILE *)wfileid)==EOF);
if (wior)
  clearerr((FILE *)wfileid);
#else
PUTC(c);
#endif

\+file

flush-file	( wfileid -- wior )		file-ext	flush_file
wior = IOR(fflush((FILE *) wfileid)==EOF);

file-status	( c_addr u -- wfam wior )	file-ext	file_status
char *filename=tilde_cstr(c_addr, u, 1);
if (access (filename, F_OK) != 0) {
  wfam=0;
  wior=IOR(1);
}
else if (access (filename, R_OK | W_OK) == 0) {
  wfam=2; /* r/w */
  wior=0;
}
else if (access (filename, R_OK) == 0) {
  wfam=0; /* r/o */
  wior=0;
}
else if (access (filename, W_OK) == 0) {
  wfam=4; /* w/o */
  wior=0;
}
else {
  wfam=1; /* well, we cannot access the file, but better deliver a legal
	    access mode (r/o bin), so we get a decent error later upon open. */
  wior=0;
}

\+
\+floating

\g floating

f=	( r1 r2 -- f )		gforth	f_equals
#line 1698
f = FLAG(r1==r2);
#line 1698
:
#line 1698
    [ char fx char 0 = [IF]
#line 1698
	] IF false ELSE true THEN [
#line 1698
    [ELSE]
#line 1698
	] xor 0= [
#line 1698
    [THEN] ] ;
#line 1698

#line 1698
f<>	( r1 r2 -- f )		gforth	f_not_equals
#line 1698
f = FLAG(r1!=r2);
#line 1698
:
#line 1698
    [ char fx char 0 = [IF]
#line 1698
	] IF true ELSE false THEN [
#line 1698
    [ELSE]
#line 1698
	] xor 0<> [
#line 1698
    [THEN] ] ;
#line 1698

#line 1698
f<	( r1 r2 -- f )		float	f_less_than
#line 1698
f = FLAG(r1<r2);
#line 1698
:
#line 1698
    [ char fx char 0 = [IF]
#line 1698
	] MINI and 0<> [
#line 1698
    [ELSE] char fx char u = [IF]
#line 1698
	]   2dup xor 0<  IF nip ELSE - THEN 0<  [
#line 1698
	[ELSE]
#line 1698
	    ] MINI xor >r MINI xor r> u< [
#line 1698
	[THEN]
#line 1698
    [THEN] ] ;
#line 1698

#line 1698
f>	( r1 r2 -- f )		gforth	f_greater_than
#line 1698
f = FLAG(r1>r2);
#line 1698
:
#line 1698
    [ char fx char 0 = [IF] ] negate [ [ELSE] ] swap [ [THEN] ]
#line 1698
    f< ;
#line 1698

#line 1698
f<=	( r1 r2 -- f )		gforth	f_less_or_equal
#line 1698
f = FLAG(r1<=r2);
#line 1698
:
#line 1698
    f> 0= ;
#line 1698

#line 1698
f>=	( r1 r2 -- f )		gforth	f_greater_or_equal
#line 1698
f = FLAG(r1>=r2);
#line 1698
:
#line 1698
    [ char fx char 0 = [IF] ] negate [ [ELSE] ] swap [ [THEN] ]
#line 1698
    f<= ;
#line 1698

#line 1698

f0=	( r -- f )		float	f_zero_equals
#line 1699
f = FLAG(r==0.);
#line 1699
:
#line 1699
    [ char f0x char 0 = [IF]
#line 1699
	] IF false ELSE true THEN [
#line 1699
    [ELSE]
#line 1699
	] xor 0= [
#line 1699
    [THEN] ] ;
#line 1699

#line 1699
f0<>	( r -- f )		gforth	f_zero_not_equals
#line 1699
f = FLAG(r!=0.);
#line 1699
:
#line 1699
    [ char f0x char 0 = [IF]
#line 1699
	] IF true ELSE false THEN [
#line 1699
    [ELSE]
#line 1699
	] xor 0<> [
#line 1699
    [THEN] ] ;
#line 1699

#line 1699
f0<	( r -- f )		float	f_zero_less_than
#line 1699
f = FLAG(r<0.);
#line 1699
:
#line 1699
    [ char f0x char 0 = [IF]
#line 1699
	] MINI and 0<> [
#line 1699
    [ELSE] char f0x char u = [IF]
#line 1699
	]   2dup xor 0<  IF nip ELSE - THEN 0<  [
#line 1699
	[ELSE]
#line 1699
	    ] MINI xor >r MINI xor r> u< [
#line 1699
	[THEN]
#line 1699
    [THEN] ] ;
#line 1699

#line 1699
f0>	( r -- f )		gforth	f_zero_greater_than
#line 1699
f = FLAG(r>0.);
#line 1699
:
#line 1699
    [ char f0x char 0 = [IF] ] negate [ [ELSE] ] swap [ [THEN] ]
#line 1699
    f0< ;
#line 1699

#line 1699
f0<=	( r -- f )		gforth	f_zero_less_or_equal
#line 1699
f = FLAG(r<=0.);
#line 1699
:
#line 1699
    f0> 0= ;
#line 1699

#line 1699
f0>=	( r -- f )		gforth	f_zero_greater_or_equal
#line 1699
f = FLAG(r>=0.);
#line 1699
:
#line 1699
    [ char f0x char 0 = [IF] ] negate [ [ELSE] ] swap [ [THEN] ]
#line 1699
    f0<= ;
#line 1699

#line 1699


d>f	( d -- r )		float	d_to_f
#ifdef BUGGY_LONG_LONG
extern double ldexp(double x, int exp);
r = ldexp((Float)d.hi,CELL_BITS) + (Float)d.lo;
#else
r = d;
#endif

f>d	( r -- d )		float	f_to_d
#ifdef BUGGY_LONG_LONG
d.hi = ldexp(r,-(int)(CELL_BITS)) - (r<0);
d.lo = r-ldexp((Float)d.hi,CELL_BITS);
#else
d = r;
#endif

f!	( r f_addr -- )	float	f_store
""Store @i{r} into the float at address @i{f-addr}.""
*f_addr = r;

f@	( f_addr -- r )	float	f_fetch
""@i{r} is the float at address @i{f-addr}.""
r = *f_addr;

df@	( df_addr -- r )	float-ext	d_f_fetch
""Fetch the double-precision IEEE floating-point value @i{r} from the address @i{df-addr}.""
#ifdef IEEE_FP
r = *df_addr;
#else
!! df@
#endif

df!	( r df_addr -- )	float-ext	d_f_store
""Store @i{r} as double-precision IEEE floating-point value to the
address @i{df-addr}.""
#ifdef IEEE_FP
*df_addr = r;
#else
!! df!
#endif

sf@	( sf_addr -- r )	float-ext	s_f_fetch
""Fetch the single-precision IEEE floating-point value @i{r} from the address @i{sf-addr}.""
#ifdef IEEE_FP
r = *sf_addr;
#else
!! sf@
#endif

sf!	( r sf_addr -- )	float-ext	s_f_store
""Store @i{r} as single-precision IEEE floating-point value to the
address @i{sf-addr}.""
#ifdef IEEE_FP
*sf_addr = r;
#else
!! sf!
#endif

f+	( r1 r2 -- r3 )	float	f_plus
r3 = r1+r2;

f-	( r1 r2 -- r3 )	float	f_minus
r3 = r1-r2;

f*	( r1 r2 -- r3 )	float	f_star
r3 = r1*r2;

f/	( r1 r2 -- r3 )	float	f_slash
r3 = r1/r2;

f**	( r1 r2 -- r3 )	float-ext	f_star_star
""@i{r3} is @i{r1} raised to the @i{r2}th power.""
r3 = pow(r1,r2);

fnegate	( r1 -- r2 )	float	f_negate
r2 = - r1;

fdrop	( r -- )		float	f_drop

fdup	( r -- r r )	float	f_dupe

fswap	( r1 r2 -- r2 r1 )	float	f_swap

fover	( r1 r2 -- r1 r2 r1 )	float	f_over

frot	( r1 r2 r3 -- r2 r3 r1 )	float	f_rote

fnip	( r1 r2 -- r2 )	gforth	f_nip

ftuck	( r1 r2 -- r2 r1 r2 )	gforth	f_tuck

float+	( f_addr1 -- f_addr2 )	float	float_plus
""@code{1 floats +}.""
f_addr2 = f_addr1+1;

floats	( n1 -- n2 )	float
""@i{n2} is the number of address units of @i{n1} floats.""
n2 = n1*sizeof(Float);

floor	( r1 -- r2 )	float
""Round towards the next smaller integral value, i.e., round toward negative infinity.""
/* !! unclear wording */
r2 = floor(r1);

fround	( r1 -- r2 )	float	f_round
""Round to the nearest integral value.""
/* !! unclear wording */
#ifdef HAVE_RINT
r2 = rint(r1);
#else
r2 = floor(r1+0.5);
/* !! This is not quite true to the rounding rules given in the standard */
#endif

fmax	( r1 r2 -- r3 )	float	f_max
if (r1<r2)
  r3 = r2;
else
  r3 = r1;

fmin	( r1 r2 -- r3 )	float	f_min
if (r1<r2)
  r3 = r1;
else
  r3 = r2;

represent	( r c_addr u -- n f1 f2 )	float
char *sig;
int flag;
int decpt;
sig=ecvt(r, u, &decpt, &flag);
n=(r==0 ? 1 : decpt);
f1=FLAG(flag!=0);
f2=FLAG(isdigit((unsigned)(sig[0]))!=0);
memmove(c_addr,sig,u);

>float	( c_addr u -- flag )	float	to_float
""Actual stack effect: ( c_addr u -- r t | f ).  Attempt to convert the
character string @i{c-addr u} to internal floating-point
representation. If the string represents a valid floating-point number
@i{r} is placed on the floating-point stack and @i{flag} is
true. Otherwise, @i{flag} is false. A string of blanks is a special
case and represents the floating-point number 0.""
/* real signature: c_addr u -- r t / f */
Float r;
char *number=cstr(c_addr, u, 1);
char *endconv;
int sign = 0;
if(number[0]=='-') {
   sign = 1;
   number++;
   u--;
}
while(isspace((unsigned)(number[--u])) && u>0);
switch(number[u])
{
   case 'd':
   case 'D':
   case 'e':
   case 'E':  break;
   default :  u++; break;
}
number[u]='\0';
r=strtod(number,&endconv);
if((flag=FLAG(!(Cell)*endconv)))
{
   IF_fpTOS(fp[0] = fpTOS);
   fp += -1;
   fpTOS = sign ? -r : r;
}
else if(*endconv=='d' || *endconv=='D')
{
   *endconv='E';
   r=strtod(number,&endconv);
   if((flag=FLAG(!(Cell)*endconv)))
     {
	IF_fpTOS(fp[0] = fpTOS);
	fp += -1;
	fpTOS = sign ? -r : r;
     }
}

fabs	( r1 -- r2 )	float-ext	f_abs
r2 = fabs(r1);

facos	( r1 -- r2 )	float-ext	f_a_cos
r2 = acos(r1);

fasin	( r1 -- r2 )	float-ext	f_a_sine
r2 = asin(r1);

fatan	( r1 -- r2 )	float-ext	f_a_tan
r2 = atan(r1);

fatan2	( r1 r2 -- r3 )	float-ext	f_a_tan_two
""@i{r1/r2}=tan(@i{r3}). ANS Forth does not require, but probably
intends this to be the inverse of @code{fsincos}. In gforth it is.""
r3 = atan2(r1,r2);

fcos	( r1 -- r2 )	float-ext	f_cos
r2 = cos(r1);

fexp	( r1 -- r2 )	float-ext	f_e_x_p
r2 = exp(r1);

fexpm1	( r1 -- r2 )	float-ext	f_e_x_p_m_one
""@i{r2}=@i{e}**@i{r1}@minus{}1""
#ifdef HAVE_EXPM1
extern double
#ifdef NeXT
              const
#endif
                    expm1(double);
r2 = expm1(r1);
#else
r2 = exp(r1)-1.;
#endif

fln	( r1 -- r2 )	float-ext	f_l_n
r2 = log(r1);

flnp1	( r1 -- r2 )	float-ext	f_l_n_p_one
""@i{r2}=ln(@i{r1}+1)""
#ifdef HAVE_LOG1P
extern double
#ifdef NeXT
              const
#endif
                    log1p(double);
r2 = log1p(r1);
#else
r2 = log(r1+1.);
#endif

flog	( r1 -- r2 )	float-ext	f_log
""The decimal logarithm.""
r2 = log10(r1);

falog	( r1 -- r2 )	float-ext	f_a_log
""@i{r2}=10**@i{r1}""
extern double pow10(double);
r2 = pow10(r1);

fsin	( r1 -- r2 )	float-ext	f_sine
r2 = sin(r1);

fsincos	( r1 -- r2 r3 )	float-ext	f_sine_cos
""@i{r2}=sin(@i{r1}), @i{r3}=cos(@i{r1})""
r2 = sin(r1);
r3 = cos(r1);

fsqrt	( r1 -- r2 )	float-ext	f_square_root
r2 = sqrt(r1);

ftan	( r1 -- r2 )	float-ext	f_tan
r2 = tan(r1);
:
 fsincos f/ ;

fsinh	( r1 -- r2 )	float-ext	f_cinch
r2 = sinh(r1);
:
 fexpm1 fdup fdup 1. d>f f+ f/ f+ f2/ ;

fcosh	( r1 -- r2 )	float-ext	f_cosh
r2 = cosh(r1);
:
 fexp fdup 1/f f+ f2/ ;

ftanh	( r1 -- r2 )	float-ext	f_tan_h
r2 = tanh(r1);
:
 f2* fexpm1 fdup 2. d>f f+ f/ ;

fasinh	( r1 -- r2 )	float-ext	f_a_cinch
r2 = asinh(r1);
:
 fdup fdup f* 1. d>f f+ fsqrt f/ fatanh ;

facosh	( r1 -- r2 )	float-ext	f_a_cosh
r2 = acosh(r1);
:
 fdup fdup f* 1. d>f f- fsqrt f+ fln ;

fatanh	( r1 -- r2 )	float-ext	f_a_tan_h
r2 = atanh(r1);
:
 fdup f0< >r fabs 1. d>f fover f- f/  f2* flnp1 f2/
 r> IF  fnegate  THEN ;

sfloats	( n1 -- n2 )	float-ext	s_floats
""@i{n2} is the number of address units of @i{n1}
single-precision IEEE floating-point numbers.""
n2 = n1*sizeof(SFloat);

dfloats	( n1 -- n2 )	float-ext	d_floats
""@i{n2} is the number of address units of @i{n1}
double-precision IEEE floating-point numbers.""
n2 = n1*sizeof(DFloat);

sfaligned	( c_addr -- sf_addr )	float-ext	s_f_aligned
""@i{sf-addr} is the first single-float-aligned address greater
than or equal to @i{c-addr}.""
sf_addr = (SFloat *)((((Cell)c_addr)+(sizeof(SFloat)-1))&(-sizeof(SFloat)));
:
 [ 1 sfloats 1- ] Literal + [ -1 sfloats ] Literal and ;

dfaligned	( c_addr -- df_addr )	float-ext	d_f_aligned
""@i{df-addr} is the first double-float-aligned address greater
than or equal to @i{c-addr}.""
df_addr = (DFloat *)((((Cell)c_addr)+(sizeof(DFloat)-1))&(-sizeof(DFloat)));
:
 [ 1 dfloats 1- ] Literal + [ -1 dfloats ] Literal and ;

\ The following words access machine/OS/installation-dependent
\   Gforth internals
\ !! how about environmental queries DIRECT-THREADED,
\   INDIRECT-THREADED, TOS-CACHED, FTOS-CACHED, CODEFIELD-DOES */

\ local variable implementation primitives
\+
\+glocals

@local#	( #noffset -- w )	gforth	fetch_local_number
w = *(Cell *)(lp+noffset);

@local0	( -- w )	new	fetch_local_zero
w = *(Cell *)(lp+0*sizeof(Cell));

@local1	( -- w )	new	fetch_local_four
w = *(Cell *)(lp+1*sizeof(Cell));

@local2	( -- w )	new	fetch_local_eight
w = *(Cell *)(lp+2*sizeof(Cell));

@local3	( -- w )	new	fetch_local_twelve
w = *(Cell *)(lp+3*sizeof(Cell));

\+floating

f@local#	( #noffset -- r )	gforth	f_fetch_local_number
r = *(Float *)(lp+noffset);

f@local0	( -- r )	new	f_fetch_local_zero
r = *(Float *)(lp+0*sizeof(Float));

f@local1	( -- r )	new	f_fetch_local_eight
r = *(Float *)(lp+1*sizeof(Float));

\+

laddr#	( #noffset -- c_addr )	gforth	laddr_number
/* this can also be used to implement lp@ */
c_addr = (Char *)(lp+noffset);

lp+!#	( #noffset -- )	gforth	lp_plus_store_number
""used with negative immediate values it allocates memory on the
local stack, a positive immediate argument drops memory from the local
stack""
lp += noffset;

lp-	( -- )	new	minus_four_lp_plus_store
lp += -sizeof(Cell);

lp+	( -- )	new	eight_lp_plus_store
lp += sizeof(Float);

lp+2	( -- )	new	sixteen_lp_plus_store
lp += 2*sizeof(Float);

lp!	( c_addr -- )	gforth	lp_store
lp = (Address)c_addr;

>l	( w -- )	gforth	to_l
lp -= sizeof(Cell);
*(Cell *)lp = w;

\+floating

f>l	( r -- )	gforth	f_to_l
lp -= sizeof(Float);
*(Float *)lp = r;

fpick	( u -- r )		gforth
""Actually the stack effect is @code{ r0 ... ru u -- r0 ... ru r0 }.""
r = fp[u+1]; /* +1, because update of fp happens before this fragment */
:
 floats fp@ + f@ ;

\+
\+

\+OS

#line 2096

#line 2099

\ argflist(argnum): Forth argument list
#line 2103

\ argdlist(argnum): declare C's arguments
#line 2107

\ argclist(argnum): pass C's arguments
#line 2111

\ icall(argnum)
#line 2117

#line 2122


\ close ' to keep fontify happy

open-lib	( c_addr1 u1 -- u2 )	gforth	open_lib
#if defined(HAVE_LIBDL) || defined(HAVE_DLOPEN)
#ifndef RTLD_GLOBAL
#define RTLD_GLOBAL 0
#endif
u2=(UCell) dlopen(cstr(c_addr1, u1, 1), RTLD_GLOBAL | RTLD_LAZY);
#else
#  ifdef _WIN32
u2 = (Cell) GetModuleHandle(cstr(c_addr1, u1, 1));
#  else
#warning Define open-lib!
u2 = 0;
#  endif
#endif

lib-sym	( c_addr1 u1 u2 -- u3 )	gforth	lib_sym
#if defined(HAVE_LIBDL) || defined(HAVE_DLOPEN)
u3 = (UCell) dlsym((void*)u2,cstr(c_addr1, u1, 1));
#else
#  ifdef _WIN32
u3 = (Cell) GetProcAddress((HMODULE)u2, cstr(c_addr1, u1, 1));
#  else
#warning Define lib-sym!
u3 = 0;
#  endif
#endif

icall0	( u -- uret )	gforth
#line 2153
uret = (SYSCALL(Cell(*)())u)();
#line 2153

#line 2153
icall1	( u1 u -- uret )	gforth
#line 2153
uret = (SYSCALL(Cell(*)(Cell))u)(u1);
#line 2153

#line 2153
icall2	( u1 u2 u -- uret )	gforth
#line 2153
uret = (SYSCALL(Cell(*)(Cell, Cell))u)(u1, u2);
#line 2153

#line 2153
icall3	( u1 u2 u3 u -- uret )	gforth
#line 2153
uret = (SYSCALL(Cell(*)(Cell, Cell, Cell))u)(u1, u2, u3);
#line 2153

#line 2153
icall4	( u1 u2 u3 u4 u -- uret )	gforth
#line 2153
uret = (SYSCALL(Cell(*)(Cell, Cell, Cell, Cell))u)(u1, u2, u3, u4);
#line 2153

#line 2153
icall5	( u1 u2 u3 u4 u5 u -- uret )	gforth
#line 2153
uret = (SYSCALL(Cell(*)(Cell, Cell, Cell, Cell, Cell))u)(u1, u2, u3, u4, u5);
#line 2153

#line 2153
icall6	( u1 u2 u3 u4 u5 u6 u -- uret )	gforth
#line 2153
uret = (SYSCALL(Cell(*)(Cell, Cell, Cell, Cell, Cell, Cell))u)(u1, u2, u3, u4, u5, u6);
#line 2153

#line 2153

icall20	( u1 u2 u3 u4 u5 u6 u7 u8 u9 u10 u11 u12 u13 u14 u15 u16 u17 u18 u19 u20 u -- uret )	gforth
#line 2154
uret = (SYSCALL(Cell(*)(Cell, Cell, Cell, Cell, Cell, Cell, Cell, Cell, Cell, Cell, Cell, Cell, Cell, Cell, Cell, Cell, Cell, Cell, Cell, Cell))u)(u1, u2, u3, u4, u5, u6, u7, u8, u9, u10, u11, u12, u13, u14, u15, u16, u17, u18, u19, u20);
#line 2154

#line 2154

fcall0	( u -- rret )	gforth
#line 2155
rret = (SYSCALL(Float(*)())u)();
#line 2155

#line 2155
fcall1	( u1 u -- rret )	gforth
#line 2155
rret = (SYSCALL(Float(*)(Cell))u)(u1);
#line 2155

#line 2155
fcall2	( u1 u2 u -- rret )	gforth
#line 2155
rret = (SYSCALL(Float(*)(Cell, Cell))u)(u1, u2);
#line 2155

#line 2155
fcall3	( u1 u2 u3 u -- rret )	gforth
#line 2155
rret = (SYSCALL(Float(*)(Cell, Cell, Cell))u)(u1, u2, u3);
#line 2155

#line 2155
fcall4	( u1 u2 u3 u4 u -- rret )	gforth
#line 2155
rret = (SYSCALL(Float(*)(Cell, Cell, Cell, Cell))u)(u1, u2, u3, u4);
#line 2155

#line 2155
fcall5	( u1 u2 u3 u4 u5 u -- rret )	gforth
#line 2155
rret = (SYSCALL(Float(*)(Cell, Cell, Cell, Cell, Cell))u)(u1, u2, u3, u4, u5);
#line 2155

#line 2155
fcall6	( u1 u2 u3 u4 u5 u6 u -- rret )	gforth
#line 2155
rret = (SYSCALL(Float(*)(Cell, Cell, Cell, Cell, Cell, Cell))u)(u1, u2, u3, u4, u5, u6);
#line 2155

#line 2155

fcall20	( u1 u2 u3 u4 u5 u6 u7 u8 u9 u10 u11 u12 u13 u14 u15 u16 u17 u18 u19 u20 u -- rret )	gforth
#line 2156
rret = (SYSCALL(Float(*)(Cell, Cell, Cell, Cell, Cell, Cell, Cell, Cell, Cell, Cell, Cell, Cell, Cell, Cell, Cell, Cell, Cell, Cell, Cell, Cell))u)(u1, u2, u3, u4, u5, u6, u7, u8, u9, u10, u11, u12, u13, u14, u15, u16, u17, u18, u19, u20);
#line 2156

#line 2156


\+

up!	( a_addr -- )	gforth	up_store
UP=up=(char *)a_addr;
:
 up ! ;
Variable UP

wcall	( u -- )	gforth
IF_fpTOS(fp[0]=fpTOS);
FP=fp;
sp=(Cell*)(SYSCALL(Cell(*)(Cell *, void *))u)(sp, &FP);
fp=FP;
IF_spTOS(spTOS=sp[0];)
IF_fpTOS(fpTOS=fp[0]);

\+file

open-dir	( c_addr u -- wdirid wior )	gforth	open_dir
""Open the directory specified by @i{c-addr, u}
and return @i{dir-id} for futher access to it.""
wdirid = (Cell)opendir(tilde_cstr(c_addr, u, 1));
wior =  IOR(wdirid == 0);

read-dir	( c_addr u1 wdirid -- u2 flag wior )	gforth	read_dir
""Attempt to read the next entry from the directory specified
by @i{dir-id} to the buffer of length @i{u1} at address @i{c-addr}. 
If the attempt fails because there is no more entries,
@i{ior}=0, @i{flag}=0, @i{u2}=0, and the buffer is unmodified.
If the attempt to read the next entry fails because of any other reason, 
return @i{ior}<>0.
If the attempt succeeds, store file name to the buffer at @i{c-addr}
and return @i{ior}=0, @i{flag}=true and @i{u2} equal to the size of the file name.
If the length of the file name is greater than @i{u1}, 
store first @i{u1} characters from file name into the buffer and
indicate "name too long" with @i{ior}, @i{flag}=true, and @i{u2}=@i{u1}.""
struct dirent * dent;
dent = readdir((DIR *)wdirid);
wior = 0;
flag = -1;
if(dent == NULL) {
  u2 = 0;
  flag = 0;
} else {
  u2 = strlen(dent->d_name);
  if(u2 > u1) {
    u2 = u1;
    wior = -512-ENAMETOOLONG;
  }
  memmove(c_addr, dent->d_name, u2);
}

close-dir	( wdirid -- wior )	gforth	close_dir
""Close the directory specified by @i{dir-id}.""
wior = IOR(closedir((DIR *)wdirid));

filename-match	( c_addr1 u1 c_addr2 u2 -- flag )	gforth	match_file
char * string = cstr(c_addr1, u1, 1);
char * pattern = cstr(c_addr2, u2, 0);
flag = FLAG(!fnmatch(pattern, string, 0));

\+

newline	( -- c_addr u )	gforth
""String containing the newline sequence of the host OS""
char newline[] = {
#if defined(unix) || defined(__MACH__)
/* Darwin/MacOS X sets __MACH__, but not unix. */
'\n'
#else
'\r','\n'
#endif
};
c_addr=newline;
u=sizeof(newline);
:
 "newline count ;
Create "newline e? crlf [IF] 2 c, $0D c, [ELSE] 1 c, [THEN] $0A c,

\+os

utime	( -- dtime )	gforth
""Report the current time in microseconds since some epoch.""
struct timeval time1;
gettimeofday(&time1,NULL);
dtime = timeval2us(&time1);

cputime ( -- duser dsystem ) gforth
""duser and dsystem are the respective user- and system-level CPU
times used since the start of the Forth system (excluding child
processes), in microseconds (the granularity may be much larger,
however).  On platforms without the getrusage call, it reports elapsed
time (since some epoch) for duser and 0 for dsystem.""
#ifdef HAVE_GETRUSAGE
struct rusage usage;
getrusage(RUSAGE_SELF, &usage);
duser = timeval2us(&usage.ru_utime);
dsystem = timeval2us(&usage.ru_stime);
#else
struct timeval time1;
gettimeofday(&time1,NULL);
duser = timeval2us(&time1);
#ifndef BUGGY_LONG_LONG
dsystem = (DCell)0;
#else
dsystem=(DCell){0,0};
#endif
#endif

\+

\+floating

v*	( f_addr1 nstride1 f_addr2 nstride2 ucount -- r ) gforth v_star
""dot-product: r=v1*v2.  The first element of v1 is at f_addr1, the
next at f_addr1+nstride1 and so on (similar for v2). Both vectors have
ucount elements.""
for (r=0.; ucount>0; ucount--) {
  r += *f_addr1 * *f_addr2;
  f_addr1 = (Float *)(((Address)f_addr1)+nstride1);
  f_addr2 = (Float *)(((Address)f_addr2)+nstride2);
}
:
 >r swap 2swap swap 0e r> 0 ?DO
     dup f@ over + 2swap dup f@ f* f+ over + 2swap
 LOOP 2drop 2drop ; 

faxpy	( ra f_x nstridex f_y nstridey ucount -- )	gforth
""vy=ra*vx+vy""
for (; ucount>0; ucount--) {
  *f_y += ra * *f_x;
  f_x = (Float *)(((Address)f_x)+nstridex);
  f_y = (Float *)(((Address)f_y)+nstridey);
}
:
 >r swap 2swap swap r> 0 ?DO
     fdup dup f@ f* over + 2swap dup f@ f+ dup f! over + 2swap
 LOOP 2drop 2drop fdrop ;

\+

\+file

(read-line)	( c_addr u1 wfileid -- u2 flag u3 wior )	file	paren_read_line
Cell c;
flag=-1;
u3=0;
for(u2=0; u2<u1; u2++)
{
   c = getc((FILE *)wfileid);
   u3++;
   if (c=='\n') break;
   if (c=='\r') {
     if ((c = getc((FILE *)wfileid))!='\n')
       ungetc(c,(FILE *)wfileid);
     else
       u3++;
     break;
   }
   if (c==EOF) {
	flag=FLAG(u2!=0);
	break;
     }
   c_addr[u2] = (Char)c;
}
wior=FILEIO(ferror((FILE *)wfileid));

\+

(listlfind)	( c_addr u longname1 -- longname2 )	new	paren_listlfind
for (; longname1 != NULL; longname1 = (struct Longname *)(longname1->next))
  if ((UCell)LONGNAME_COUNT(longname1)==u &&
      memcasecmp(c_addr, longname1->name, u)== 0 /* or inline? */)
    break;
longname2=longname1;
:
    BEGIN  dup WHILE  (findl-samelen)  dup  WHILE
	>r 2dup r@ cell+ cell+ capscomp  0=
	IF  2drop r>  EXIT  THEN
	r> @
    REPEAT  THEN  nip nip ;
: (findl-samelen) ( u longname1 -- u longname2/0 )
    BEGIN  2dup cell+ @ lcount-mask and <> WHILE  @  dup 0= UNTIL  THEN ;

\+hash

(hashlfind)	( c_addr u a_addr -- longname2 )	new	paren_hashlfind
struct Longname *longname1;
longname2=NULL;
while(a_addr != NULL)
{
   longname1=(struct Longname *)(a_addr[1]);
   a_addr=(Cell *)(a_addr[0]);
   if ((UCell)LONGNAME_COUNT(longname1)==u &&
       memcasecmp(c_addr, longname1->name, u)== 0 /* or inline? */)
     {
	longname2=longname1;
	break;
     }
}
:
 BEGIN  dup  WHILE
        2@ >r >r dup r@ cell+ @ lcount-mask and =
        IF  2dup r@ cell+ cell+ capscomp 0=
	    IF  2drop r> rdrop  EXIT  THEN  THEN
	rdrop r>
 REPEAT nip nip ;

(tablelfind)	( c_addr u a_addr -- longname2 )	new	paren_tablelfind
""A case-sensitive variant of @code{(hashfind)}""
struct Longname *longname1;
longname2=NULL;
while(a_addr != NULL)
{
   longname1=(struct Longname *)(a_addr[1]);
   a_addr=(Cell *)(a_addr[0]);
   if ((UCell)LONGNAME_COUNT(longname1)==u &&
       memcmp(c_addr, longname1->name, u)== 0 /* or inline? */)
     {
	longname2=longname1;
	break;
     }
}
:
 BEGIN  dup  WHILE
        2@ >r >r dup r@ cell+ @ lcount-mask and =
        IF  2dup r@ cell+ cell+ -text 0=
	    IF  2drop r> rdrop  EXIT  THEN  THEN
	rdrop r>
 REPEAT nip nip ;

\+

\+peephole

\g peephole

primtable	( -- wprimtable )	new
""wprimtable is a table containing the xts of the primitives indexed
by sequence-number in prim (for use in prepare-peephole-table).""
wprimtable = (Cell)primtable(symbols+DOESJUMP+1,MAX_SYMBOLS-DOESJUMP-1);

prepare-peephole-table	( wprimtable -- wpeeptable ) new prepare_peephole_opt
""wpeeptable is a data structure used by @code{peephole-opt}; it is
constructed by combining a primitives table with a simple peephole
optimization table.""
wpeeptable = prepare_peephole_table((Xt *)wprimtable);

peephole-opt	( xt1 xt2 wpeeptable -- xt )	new	peephole_opt
""xt is the combination of xt1 and xt2 (according to wpeeptable); if
they cannot be combined, xt is 0.""
xt = peephole_opt(xt1, xt2, wpeeptable);

call	( #a_callee -- R:a_retaddr )	new
""Call callee (a variant of docol with inline argument).""
#ifdef DEBUG
    {
      CFA_TO_NAME((((Cell *)a_callee)-2));
      fprintf(stderr,"%08lx: call %08lx %.*s\n",(Cell)ip,(Cell)a_callee,
	      len,name);
    }
#endif
a_retaddr = (Cell *)IP;
SET_IP((Xt *)a_callee);

useraddr	( #u -- a_addr )	new
a_addr = (Cell *)(up+u);

compile-prim ( xt1 -- xt2 )	new	compile_prim
xt2 = (Xt)compile_prim((Label)xt1);

lit@ / lit_fetch = lit @

lit-perform	( #a_addr -- )	new	lit_perform
ip=IP;
SUPER_END;
EXEC(*(Xt *)a_addr);

lit+ / lit_plus = lit +

does-exec ( #a_cfa -- R:nest a_pfa )	new	does_exec
a_pfa = PFA(a_cfa);
nest = (Cell)ip;
IF_spTOS(spTOS = sp[0]);
#ifdef DEBUG
    {
      CFA_TO_NAME(a_cfa);
      fprintf(stderr,"%08lx: does %08lx %.*s\n",
	      (Cell)ip,(Cell)a_cfa,len,name);
    }
#endif
SET_IP(DOES_CODE1(a_cfa));

register-compile ( a_addr -- ) new register_compile
register_compile((Label *)a_addr);

bbend ( -- ) new
bbend();

#line 1 "peeprules.vmg"
super1 = lit @ 
super2 = lit + 
super3 = 2drop drop 
super4 = 2drop drop lit 
super5 = cells lit 
super6 = cells lit + 
super7 = 2drop drop lit ;s 
super8 = 2@ rot 
super9 = + @ 
super10 = 0= ?branch 
super11 = lit + @ 
super12 = cells lit + @ 
super13 = @ = 
super14 = lit @ = 
super15 = 1- call 
super16 = @ lit 
super17 = @ and 
super18 = lit @ and 
super19 = + dup 
super20 = @ ;s 
super21 = 2! ;s 
super22 = dup lit 
super23 = lit ! 
super24 = @ lit @ 
super25 = 2dup < 
super26 = 2dup < ?branch 
super27 = 2drop ;s 
super28 = 2dup call 
super29 = 2@ call 
super30 = 2@ rot xor 
super31 = 2@ rot xor >r 
super32 = 2@ rot xor >r xor 
super33 = 2@ rot xor >r xor r> 
super34 = 2@ rot xor >r xor r> lit 
super35 = 2@ rot xor >r xor r> lit 2! 
super36 = 2@ rot xor >r xor r> lit 2! ;s 
super37 = @ + 
super38 = @ lit @ and 
super39 = lit @ + 
super40 = + ;s 
super41 = and lit 
super42 = lit call 
super43 = dup lit @ 
super44 = + @ lit 
super45 = lit + @ lit 
super46 = over cells 
super47 = cells lit + @ lit 
super48 = lit @ lit 
super49 = + @ lit @ 
super50 = lit + @ lit @ 
super51 = cells lit + @ lit @ 
super52 = + @ lit @ and 
super53 = cells lit + @ lit @ and 
super54 = lit + @ lit @ and 
super55 = @ and lit 
super56 = lit @ and lit 
super57 = 1+ lit 
super58 = and lit @ 
super59 = @ and lit @ 
super60 = lit @ and lit @ 
super61 = = ;s 
super62 = dup @ 
super63 = i call 
super64 = + dup @ 
super65 = @ and lit @ = 
super66 = and lit @ = 
super67 = lit @ and lit @ = 
super68 = ! ;s 
super69 = cells + 
super70 = @ = ;s 
super71 = lit @ = ;s 
super72 = 0= lit 
super73 = @ * 
super74 = lit @ * 
super75 = swap lit 
super76 = 1+ lit ! 
super77 = @ ?branch 
super78 = @ lit @ and lit 
super79 = @ lit @ and lit @ 
super80 = + @ lit @ and lit 
super81 = + @ lit @ and lit @ 
super82 = cells lit + @ lit @ and lit 
super83 = cells lit + @ lit @ and lit @ 
super84 = lit + @ lit @ and lit 
super85 = lit + @ lit @ and lit @ 
super86 = + @ lit @ and lit @ = 
super87 = + @ lit @ and lit @ = ;s 
super88 = @ and lit @ = ;s 
super89 = @ lit @ and lit @ = 
super90 = @ lit @ and lit @ = ;s 
super91 = and lit @ = ;s 
super92 = cells lit + @ lit @ and lit @ = 
super93 = cells lit + @ lit @ and lit @ = ;s 
super94 = lit + @ lit @ and lit @ = 
super95 = lit + @ lit @ and lit @ = ;s 
super96 = lit @ and lit @ = ;s 
super97 = useraddr @ 
super98 = over + 
super99 = over + dup 
super100 = cells + dup 
super101 = cells + dup @ 
super102 = over cells + 
super103 = over cells + dup 
super104 = over cells + dup @ 
super105 = useraddr @ ;s 
super106 = swap lit @ 
super107 = 0= ;s 
super108 = 0= lit ! 
super109 = dup cells 
super110 = dup cells lit 
super111 = dup cells lit + 
super112 = dup cells lit + @ 
super113 = ! lit 
super114 = 0= lit ! lit 
super115 = 0= lit ! lit @ 
super116 = 0= lit ! lit @ negate 
super117 = 0= lit ! lit @ negate lit 
super118 = 0= lit ! lit @ negate lit ! 
super119 = 0= lit ! lit @ negate lit ! ;s 
super120 = 2@ lit 
super121 = 2@ d= 
super122 = 2@ lit 2@ 
super123 = 2@ lit 2@ d= 
super124 = 2@ d0= 
super125 = 2@ d0= ?branch 
super126 = lit ! lit 
super127 = @ call 
super128 = + lit 
super129 = + dup cells 
super130 = + dup cells lit 
super131 = + dup cells lit + 
super132 = + dup cells lit + @ 
super133 = 2@ rot @ 
super134 = 2@ rot @ ;s 
super135 = lit @ call 
super136 = 0= swap 
super137 = 0= swap 0= 
super138 = 0= swap 0= xor 
super139 = 0= swap 0= xor ;s 
super140 = 0= xor 
super141 = 0= xor ;s 
super142 = over + dup cells 
super143 = over + dup cells lit 
super144 = over + dup cells lit + 
super145 = over + dup cells lit + @ 
super146 = = or 
super147 = dup lit @ = 
super148 = @ = or 
super149 = lit @ = or 
super150 = lit @ lit @ 
super151 = 2@ rot 2@ 
super152 = 2@ ;s 
super153 = 2@ rot 2@ ;s 
super154 = and dup 
super155 = @ and dup 
super156 = lit @ and dup 
super157 = 2@ d= ?branch 
super158 = 2@ lit 2@ d= ?branch 
super159 = @ @ 
super160 = lit @ @ 
super161 = @ @ ;s 
super162 = lit @ @ ;s 
super163 = @ + ;s 
super164 = 2>r call 
super165 = lit @ ?branch 
super166 = dup call 
super167 = 2drop drop lit @ 
super168 = 2r> lit 
super169 = 1+ lit ! ;s 
super170 = * ;s 
super171 = @ * ;s 
super172 = lit @ * ;s 
super173 = lit @ + ;s 
super174 = drop lit 
super175 = or ?branch 
super176 = swap lit @ = 
super177 = swap lit @ = or 
super178 = 2>r dup 
super179 = and dup lit 
super180 = @ and dup lit 
super181 = lit @ and dup lit 
super182 = 2@ rot lit 
super183 = 2@ rot lit + 
super184 = 2@ rot lit + @ 
super185 = 2@ rot lit + @ ;s 
super186 = = swap 
super187 = = swap lit 
super188 = 1- lit 
super189 = = swap lit @ 
super190 = = swap lit @ = 
super191 = = swap lit @ = or 
super192 = @ = swap 
super193 = @ = swap lit 
super194 = @ = swap lit @ 
super195 = @ = swap lit @ = 
super196 = @ = swap lit @ = or 
super197 = dup lit @ = swap 
super198 = dup lit @ = swap lit 
super199 = dup lit @ = swap lit @ 
super200 = dup lit @ = swap lit @ = 
super201 = dup lit @ = swap lit @ = or 
super202 = lit @ = swap 
super203 = lit @ = swap lit 
super204 = lit @ = swap lit @ 
super205 = lit @ = swap lit @ = 
super206 = lit @ = swap lit @ = or 
super207 = @ and dup lit @ 
super208 = @ and dup lit @ = 
super209 = @ and dup lit @ = swap 
super210 = @ and dup lit @ = swap lit 
super211 = @ and dup lit @ = swap lit @ 
super212 = @ and dup lit @ = swap lit @ = 
super213 = @ and dup lit @ = swap lit @ = or 
super214 = and dup lit @ 
super215 = and dup lit @ = 
super216 = and dup lit @ = swap 
super217 = and dup lit @ = swap lit 
super218 = and dup lit @ = swap lit @ 
super219 = and dup lit @ = swap lit @ = 
super220 = and dup lit @ = swap lit @ = or 
super221 = lit @ and dup lit @ 
super222 = lit @ and dup lit @ = 
super223 = lit @ and dup lit @ = swap 
super224 = lit @ and dup lit @ = swap lit 
super225 = lit @ and dup lit @ = swap lit @ 
super226 = lit @ and dup lit @ = swap lit @ = 
super227 = 2r> ;s 
super228 = = or ?branch 
super229 = @ = or ?branch 
super230 = lit @ = or ?branch 
super231 = = ?branch 
super232 = @ + dup 
super233 = lit @ + dup 
super234 = lit and 
super235 = lit ! ;s 
super236 = lit lit 
super237 = @ cells 
super238 = lit @ cells 
super239 = and ;s 
super240 = ! lit @ 
super241 = dup lit @ and 
super242 = @ cells lit 
super243 = @ cells lit + 
super244 = lit @ cells lit 
super245 = lit @ cells lit + 
super246 = 1- lit ! 
super247 = 1- lit ! ;s 
super248 = * + 
super249 = + lit call 
super250 = swap lit @ = or ?branch 
super251 = @ = ?branch 
super252 = lit @ = ?branch 
super253 = over cells lit 
super254 = over cells lit + 
super255 = over cells lit + @ 
super256 = 2@ 2>r 
super257 = 2@ 2>r call 
super258 = dup lit @ and dup 
super259 = dup lit @ and dup lit 
super260 = lit ! lit @ 
super261 = lit ;s 
super262 = 2>r lit 
super263 = = swap lit @ = or ?branch 
super264 = @ = swap lit @ = or ?branch 
super265 = and dup lit @ = swap lit @ = or ?branch 
super266 = dup lit @ = swap lit @ = or ?branch 
super267 = lit @ = swap lit @ = or ?branch 
super268 = dup lit @ and dup lit @ 
super269 = dup lit @ and dup lit @ = 
super270 = dup lit @ and dup lit @ = swap 
super271 = dup lit @ and dup lit @ = swap lit 
super272 = dup lit @ and dup lit @ = swap lit @ 
super273 = 2r> 2r> 
super274 = + @ ?branch 
super275 = cells lit + @ ?branch 
super276 = lit + @ ?branch 
super277 = >r over 
super278 = @ r> 
super279 = r> + 
super280 = + @ r> 
super281 = + @ r> + 
super282 = + @ r> + over 
super283 = + @ r> + over cells 
super284 = + @ r> + over cells lit 
super285 = + @ r> + over cells lit + 
super286 = + @ r> + over cells lit + @ 
super287 = + @ r> + over cells lit + @ ?branch 
super288 = + dup cells lit + @ r> 
super289 = + dup cells lit + @ r> + 
super290 = + dup cells lit + @ r> + over 
super291 = + dup cells lit + @ r> + over cells 
super292 = + dup cells lit + @ r> + over cells lit 
super293 = + over 
super294 = + over cells 
super295 = + over cells lit 
super296 = + over cells lit + 
super297 = + over cells lit + @ 
super298 = + over cells lit + @ ?branch 
super299 = >r over + 
super300 = >r over + dup 
super301 = >r over + dup cells 
super302 = >r over + dup cells lit 
super303 = >r over + dup cells lit + 
super304 = >r over + dup cells lit + @ 
super305 = >r over + dup cells lit + @ r> 
super306 = >r over + dup cells lit + @ r> + 
super307 = >r over + dup cells lit + @ r> + over 
super308 = @ r> + 
super309 = @ r> + over 
super310 = @ r> + over cells 
super311 = @ r> + over cells lit 
super312 = @ r> + over cells lit + 
super313 = @ r> + over cells lit + @ 
super314 = @ r> + over cells lit + @ ?branch 
super315 = cells lit + @ r> 
super316 = cells lit + @ r> + 
super317 = cells lit + @ r> + over 
super318 = cells lit + @ r> + over cells 
super319 = cells lit + @ r> + over cells lit 
super320 = cells lit + @ r> + over cells lit + 
super321 = cells lit + @ r> + over cells lit + @ 
super322 = dup cells lit + @ r> 
super323 = dup cells lit + @ r> + 
super324 = dup cells lit + @ r> + over 
super325 = dup cells lit + @ r> + over cells 
super326 = dup cells lit + @ r> + over cells lit 
super327 = dup cells lit + @ r> + over cells lit + 
super328 = lit + @ r> 
super329 = lit + @ r> + 
super330 = lit + @ r> + over 
super331 = lit + @ r> + over cells 
super332 = lit + @ r> + over cells lit 
super333 = lit + @ r> + over cells lit + 
super334 = lit + @ r> + over cells lit + @ 
super335 = lit + @ r> + over cells lit + @ ?branch 
super336 = over + dup cells lit + @ r> 
super337 = over + dup cells lit + @ r> + 
super338 = over + dup cells lit + @ r> + over 
super339 = over + dup cells lit + @ r> + over cells 
super340 = over cells lit + @ ?branch 
super341 = r> + over 
super342 = r> + over cells 
super343 = r> + over cells lit 
super344 = r> + over cells lit + 
super345 = r> + over cells lit + @ 
super346 = r> + over cells lit + @ ?branch 
super347 = drop lit ;s 
super348 = lit - 
super349 = + call 
super350 = lit cells 
super351 = 0< ?branch 
super352 = 2r> lit ! 
super353 = 2r> lit ! lit 
super354 = 2r> lit ! lit ! 
super355 = @ and ;s 
super356 = @ lit @ and ;s 
super357 = lit @ and ;s 
super358 = + @ lit @ and ;s 
super359 = cells lit + @ lit @ and ;s 
super360 = lit + @ lit @ and ;s 
super361 = 2dup 2>r 
super362 = lit 2@ 
super363 = over lit 
super364 = - ;s 
super365 = lit - ;s 
super366 = dup 1- 
super367 = + dup 1- 
super368 = + dup 1- call 
super369 = dup 1- call 
super370 = within lit 
super371 = within lit and 
super372 = 2dup lit 
super373 = 2dup lit @ 
super374 = 2dup lit @ lit 
super375 = 2dup lit @ lit @ 
super376 = 2dup lit @ lit @ d= 
super377 = 2dup lit @ lit @ d= ?branch 
super378 = 2>r dup cells 
super379 = 2>r dup cells lit 
super380 = 2>r dup cells lit + 
super381 = 2>r dup cells lit + @ 
super382 = 2>r dup cells lit + @ 2r> 
super383 = 2>r dup cells lit + @ 2r> ;s 
super384 = 2dup 2>r dup 
super385 = 2dup 2>r dup cells 
super386 = 2dup 2>r dup cells lit 
super387 = 2dup 2>r dup cells lit + 
super388 = 2dup 2>r dup cells lit + @ 
super389 = 2dup 2>r dup cells lit + @ 2r> 
super390 = 2dup 2>r dup cells lit + @ 2r> ;s 
super391 = @ * + 
super392 = lit @ * + 
super393 = drop ;s 
super394 = @ lit ! 
super395 = ! call 
super396 = 2@ r> 
super397 = 2@ r> 2! 
super398 = lit ! call 
super399 = dup lit ! 
super400 = nip nip 
super401 = * lit 
super402 = 2over rot 
super403 = 2over rot ;s 
super404 = @ dup 
super405 = lit @ lit @ and 
super406 = 2@ d= and 
super407 = 2@ d= and ;s 
super408 = 2@ lit 2@ d= and 
super409 = 2@ lit 2@ d= and ;s 
super410 = ! lit ! 
super411 = lit ! lit ! 
super412 = and + 
super413 = 2>r = 
super414 = 2>r = -rot 
super415 = 2>r = -rot 2r> 
super416 = 2>r = -rot 2r> d= 
super417 = 2>r = -rot 2r> d= and 
super418 = 2>r = -rot 2r> d= and ;s 
super419 = 2r> d= 
super420 = 2r> d= and 
super421 = 2r> d= and ;s 
super422 = * + ;s 
super423 = @ * + ;s 
super424 = lit @ * + ;s 
super425 = swap lit @ * 
super426 = swap lit @ * + 
super427 = swap lit @ * + ;s 
super428 = @ + call 
super429 = lit @ + call 
super430 = 1+ lit @ 
super431 = @ dup lit 
super432 = dup lit cells 
super433 = + dup lit 
super434 = lit lit ! 
super435 = 0< 0= 
super436 = @ 0= 
super437 = 0< 0= ?branch 
super438 = 2>r dup call 
super439 = lit * 
super440 = 0= dup 
super441 = 0= dup lit 
super442 = 0= dup lit ! 
super443 = 0= dup lit ! ;s 
super444 = + dup call 
super445 = @ + dup call 
super446 = lit @ + dup call 
super447 = over + dup @ 
super448 = cells lit @ 
super449 = cells lit @ + 
super450 = cells lit @ + ;s 
super451 = 2r> lit 2! 
super452 = 2r> lit 2! ;s 
super453 = 2dup > 
super454 = 2dup > 0= 
super455 = 2dup > 0= ?branch 
super456 = + @ ;s 
super457 = 2>r lit @ 
super458 = < ?branch 
super459 = 2drop branch 
super460 = @ * lit 
super461 = lit @ * lit 
super462 = + lit @ 
super463 = lit 2! 
super464 = @ and lit @ = ?branch 
super465 = and lit @ = ?branch 
super466 = lit @ and lit @ = ?branch 
super467 = 2! drop 
super468 = 2! drop ;s 
super469 = 2>r lit ! 
super470 = 2>r lit ! lit 
super471 = 2>r lit ! lit 2@ 
super472 = 2>r lit ! lit 2@ 2>r 
super473 = 2>r lit ! lit 2@ 2>r call 
super474 = 2>r lit @ >r 
super475 = 2>r lit @ >r >r 
super476 = 2>r lit @ >r >r 2dup 
super477 = 2>r lit @ >r >r 2dup call 
super478 = 2@ r> 2! drop 
super479 = 2@ r> 2! drop ;s 
super480 = 2r> 2r> 2r> 
super481 = 2r> 2r> 2r> ;s 
super482 = 2r> 2r> ;s 
super483 = 1+ lit ! >r 
super484 = 1+ lit ! >r lit 
super485 = 1+ lit ! >r lit 2@ 
super486 = 1+ lit ! >r lit 2@ r> 
super487 = 1+ lit ! >r lit 2@ r> 2! 
super488 = 1+ lit ! >r lit 2@ r> 2! drop 
super489 = 1+ lit ! >r lit 2@ r> 2! drop ;s 
super490 = 2! call 
super491 = @ + dup lit 
super492 = lit @ + dup lit 
super493 = = and 
super494 = lit cells + 
super495 = or ;s 
super496 = 0= lit or 
super497 = 0= lit or lit 
super498 = 0= lit or lit ! 
super499 = 0= lit or lit ! ;s 
super500 = 2r> = 
super501 = 2r> = and 
super502 = 2r> = and ;s 
super503 = + 2@ 
super504 = + 2@ rot 
super505 = drop (loop) 
super506 = r> lit 
super507 = @ 2dup 
super508 = @ 2dup < 
super509 = @ 2dup < ?branch 
super510 = nip ;s 
super511 = lit 2! ;s 
super512 = @ and cells 
super513 = @ and cells lit 
super514 = @ and cells lit + 
super515 = and cells 
super516 = and cells lit 
super517 = and cells lit + 
super518 = lit @ and cells 
super519 = lit @ and cells lit 
super520 = lit @ and cells lit + 
super521 = lit + @ ;s 
super522 = @ = and 
super523 = lit @ = and 
super524 = lit + ;s 
super525 = 2r> lit ! lit ! ;s 
super526 = 1- dup 
super527 = and + ;s 
super528 = 1+ lit @ + 
super529 = 1+ lit @ + ;s 
super530 = 1- dup lit 
super531 = 1- dup lit @ 
super532 = 1- dup lit @ > 
super533 = 1- dup lit @ > 0= 
super534 = 1- dup lit @ > 0= ?branch 
super535 = 2r> lit ! lit ! 2r> 
super536 = 2r> lit ! lit ! 2r> lit 
super537 = 2r> lit ! lit ! 2r> lit ! 
super538 = 2r> lit ! lit ! 2r> lit ! lit 
super539 = 2r> lit ! lit ! 2r> lit ! lit ! 
super540 = 2r> lit ! lit ! 2r> lit ! lit ! ;s 
super541 = lit /mod 
super542 = - swap 
super543 = 1+ lit ! call 
super544 = 2over d= 
super545 = 2over d= i 
super546 = 2over d= i call 
super547 = nip nip ;s 
super548 = >r xor 
super549 = >r xor r> 
super550 = >r xor r> lit 
super551 = >r xor r> lit 2! 
super552 = >r xor r> lit 2! ;s 
super553 = lit 2@ rot 
super554 = lit 2@ rot xor 
super555 = lit 2@ rot xor >r 
super556 = lit 2@ rot xor >r xor 
super557 = lit 2@ rot xor >r xor r> 
super558 = lit 2@ rot xor >r xor r> lit 
super559 = lit 2@ rot xor >r xor r> lit 2! 
super560 = lit 2@ rot xor >r xor r> lit 2! ;s 
super561 = r> lit 2! 
super562 = r> lit 2! ;s 
super563 = rot xor 
super564 = rot xor >r 
super565 = rot xor >r xor 
super566 = rot xor >r xor r> 
super567 = rot xor >r xor r> lit 
super568 = rot xor >r xor r> lit 2! 
super569 = rot xor >r xor r> lit 2! ;s 
super570 = xor >r 
super571 = xor >r xor 
super572 = xor >r xor r> 
super573 = xor >r xor r> lit 
super574 = xor >r xor r> lit 2! 
super575 = xor >r xor r> lit 2! ;s 
super576 = xor r> 
super577 = xor r> lit 
super578 = xor r> lit 2! 
super579 = xor r> lit 2! ;s 
super580 = lit * + 
super581 = = and + 
super582 = = and + ;s 
super583 = @ = and + 
super584 = @ = and + ;s 
super585 = lit @ = and + 
super586 = lit @ = and + ;s 
super587 = drop lit @ 
super588 = 2drop call 
super589 = over lit @ 
super590 = + ! 
super591 = rot lit 
super592 = lit @ lit ! 
super593 = * lit + 
super594 = * lit + ;s 
super595 = @ * lit + 
super596 = @ * lit + ;s 
super597 = lit @ * lit + 
super598 = lit @ * lit + ;s 
super599 = @ over 
super600 = lit @ over 
super601 = + @ dup 
super602 = + @ dup lit 
super603 = cells lit + @ dup 
super604 = cells lit + @ dup lit 
super605 = lit + @ dup 
super606 = lit + @ dup lit 
super607 = cells - 
super608 = lit cells - 
super609 = or lit 
super610 = 0<> lit 
super611 = 0<> lit @ 
super612 = 0<> lit @ and 
super613 = @ lit ! lit 
super614 = dup lit call 
super615 = or lit ! 
super616 = lit @ 0= 
super617 = swap call 
super618 = cells + 2@ 
super619 = dup lit cells + 
super620 = cells + 2@ rot 
super621 = dup lit cells + 2@ 
super622 = dup lit cells + 2@ rot 
super623 = lit cells + 2@ 
super624 = lit cells + 2@ rot 
super625 = over lit @ = 
super626 = 2drop drop ;s 
super627 = @ 1+ 
super628 = lit @ 1+ 
super629 = lit + ! 
super630 = @ 1+ lit 
super631 = lit @ 1+ lit 
super632 = @ swap 
super633 = 2drop lit 
super634 = 2drop lit @ 
super635 = = or ;s 
super636 = 2drop lit @ ;s 
super637 = or over 
super638 = or over lit 
super639 = @ negate 
super640 = = over 
super641 = @ branch 
super642 = lit @ branch 
super643 = @ and 0= 
super644 = and 0= 
super645 = lit @ and 0= 
super646 = @ negate lit 
super647 = negate lit 
super648 = lit @ negate 
super649 = lit @ negate lit 
super650 = 1+ lit ! lit 
super651 = 1+ lit ! lit @ 
super652 = 1+ lit ! lit @ ;s 
super653 = @ lit @ lit 
super654 = lit @ lit @ lit 
super655 = @ = over 
super656 = dup lit @ = over 
super657 = lit @ = over 
super658 = * lit @ 
super659 = 2drop drop lit @ ;s 
super660 = over call 
super661 = lit @ lit ! lit 
super662 = @ lit call 
super663 = lit @ lit call 
super664 = - @ 
super665 = - @ lit 
super666 = - @ lit ! 
super667 = cells - @ 
super668 = cells - @ lit 
super669 = cells - @ lit ! 
super670 = lit cells - @ 
super671 = lit cells - @ lit 
super672 = lit cells - @ lit ! 
super673 = over lit @ = or 
super674 = dup lit ! ;s 
super675 = ! lit ! lit 
super676 = lit ! lit ! lit 
super677 = 0= and 
super678 = + cells 
super679 = + cells lit 
super680 = + cells lit + 
super681 = dup 2@ 
super682 = @ dup lit @ 
super683 = - call 
super684 = lit @ dup 
super685 = cells lit + ! 
super686 = and ?branch 
super687 = or lit ! lit 
super688 = or lit ! lit @ 
super689 = @ 1+ lit ! 
super690 = lit @ 1+ lit ! 
super691 = @ negate lit ! 
super692 = lit @ negate lit ! 
super693 = negate lit ! 
super694 = @ and 0= ?branch 
super695 = and 0= ?branch 
super696 = lit @ and 0= ?branch 
super697 = + dup lit call 
super698 = + lit @ = 
super699 = + lit @ = ?branch 
super700 = @ + dup lit call 
super701 = lit @ + dup lit call 
super702 = @ lit @ and 0= 
super703 = @ lit @ and 0= ?branch 
super704 = lit @ lit @ and 0= 
super705 = lit @ lit @ and 0= ?branch 
super706 = ! cells 
super707 = ! cells lit 
super708 = ! cells lit + 
super709 = ! cells lit + @ 
super710 = dup lit ! cells 
super711 = dup lit ! cells lit 
super712 = dup lit ! cells lit + 
super713 = dup lit ! cells lit + @ 
super714 = lit ! cells 
super715 = lit ! cells lit 
super716 = lit ! cells lit + 
super717 = lit ! cells lit + @ 
super718 = nip lit 
super719 = nip lit @ 
super720 = dup lit + 
super721 = dup lit cells - 
super722 = dup lit cells - @ 
super723 = dup lit cells - @ lit 
super724 = dup lit cells - @ lit ! 
super725 = @ swap lit 
super726 = 0<> lit @ and + 
super727 = 0<> lit @ and + ;s 
super728 = + ! ;s 
super729 = + @ dup lit ! 
super730 = @ dup lit ! 
super731 = cells lit + @ dup lit ! 
super732 = lit + @ dup lit ! 
super733 = swap cells 
super734 = swap cells lit 
super735 = swap cells lit + 
super736 = @ = or ;s 
super737 = lit @ = or ;s 
super738 = = swap lit @ = or ;s 
super739 = @ = swap lit @ = or ;s 
super740 = dup lit @ = swap lit @ = or ;s 
super741 = lit @ = swap lit @ = or ;s 
super742 = swap lit @ = or ;s 
super743 = and dup lit @ = swap lit @ = or ;s 
super744 = cells swap 
super745 = cells swap cells 
super746 = cells swap cells lit 
super747 = cells swap cells lit + 
super748 = nip lit @ and 
super749 = nip lit @ and dup 
super750 = nip lit @ and dup lit 
super751 = nip lit @ and dup lit @ 
super752 = nip lit @ and dup lit @ = 
super753 = nip lit @ and dup lit @ = swap 
super754 = nip lit @ and dup lit @ = swap lit 
super755 = nip lit @ and dup lit @ = swap lit @ 
super756 = nip nip lit 
super757 = nip nip lit @ 
super758 = nip nip lit @ and 
super759 = nip nip lit @ and dup 
super760 = nip nip lit @ and dup lit 
super761 = nip nip lit @ and dup lit @ 
super762 = nip nip lit @ and dup lit @ = 
super763 = nip nip lit @ and dup lit @ = swap 
super764 = nip nip lit @ and dup lit @ = swap lit 
super765 = + dup lit + 
super766 = @ + dup lit + 
super767 = lit @ + dup lit + 
super768 = lit lit @ 
super769 = + @ call 
super770 = + @ 0= 
super771 = + @ 0= ;s 
super772 = @ 0= ;s 
super773 = cells lit + @ 0= 
super774 = cells lit + @ 0= ;s 
super775 = lit + @ 0= 
super776 = lit + @ 0= ;s 
super777 = + does-exec 
super778 = * + does-exec 
super779 = @ and lit * 
super780 = @ and lit * + 
super781 = @ and lit * + does-exec 
super782 = and lit * 
super783 = and lit * + 
super784 = and lit * + does-exec 
super785 = lit * + does-exec 
super786 = lit @ and lit * 
super787 = lit @ and lit * + 
super788 = lit @ and lit * + does-exec 
super789 = @ + lit 
super790 = lit @ + lit 
super791 = lit @ dup lit 
super792 = d= ?branch 
super793 = ! lit ! lit @ 
super794 = lit ! lit ! lit @ 
super795 = ! lit ! lit @ 0= 
super796 = ! lit ! lit @ 0= lit 
super797 = ! lit ! lit @ 0= lit ! 
super798 = ! lit ! lit @ 0= lit ! lit 
super799 = ! lit ! lit @ 0= lit ! lit @ 
super800 = ! lit ! lit @ 0= lit ! lit @ negate 
super801 = ! lit @ 0= 
super802 = ! lit @ 0= lit 
super803 = ! lit @ 0= lit ! 
super804 = ! lit @ 0= lit ! lit 
super805 = ! lit @ 0= lit ! lit @ 
super806 = ! lit @ 0= lit ! lit @ negate 
super807 = ! lit @ 0= lit ! lit @ negate lit 
super808 = ! lit @ 0= lit ! lit @ negate lit ! 
super809 = ! lit @ negate 
super810 = ! lit @ negate lit 
super811 = ! lit @ negate lit ! 
super812 = ! lit @ negate lit ! ;s 
super813 = @ 0= lit 
super814 = @ 0= lit ! 
super815 = @ 0= lit ! lit 
super816 = @ 0= lit ! lit @ 
super817 = @ 0= lit ! lit @ negate 
super818 = @ 0= lit ! lit @ negate lit 
super819 = @ 0= lit ! lit @ negate lit ! 
super820 = @ 0= lit ! lit @ negate lit ! ;s 
super821 = @ lit ! lit ! 
super822 = @ lit ! lit ! lit 
super823 = @ lit ! lit ! lit @ 
super824 = @ lit ! lit ! lit @ 0= 
super825 = @ lit ! lit ! lit @ 0= lit 
super826 = @ lit ! lit ! lit @ 0= lit ! 
super827 = @ lit ! lit ! lit @ 0= lit ! lit 
super828 = @ lit @ lit ! 
super829 = @ lit @ lit ! lit 
super830 = @ lit @ lit ! lit ! 
super831 = @ lit @ lit ! lit ! lit 
super832 = @ lit @ lit ! lit ! lit @ 
super833 = @ lit @ lit ! lit ! lit @ 0= 
super834 = @ lit @ lit ! lit ! lit @ 0= lit 
super835 = @ negate lit ! ;s 
super836 = lit ! lit ! lit @ 0= 
super837 = lit ! lit ! lit @ 0= lit 
super838 = lit ! lit ! lit @ 0= lit ! 
super839 = lit ! lit ! lit @ 0= lit ! lit 
super840 = lit ! lit ! lit @ 0= lit ! lit @ 
super841 = lit ! lit @ 0= 
super842 = lit ! lit @ 0= lit 
super843 = lit ! lit @ 0= lit ! 
super844 = lit ! lit @ 0= lit ! lit 
super845 = lit ! lit @ 0= lit ! lit @ 
super846 = lit ! lit @ 0= lit ! lit @ negate 
super847 = lit ! lit @ 0= lit ! lit @ negate lit 
super848 = lit ! lit @ negate 
super849 = lit ! lit @ negate lit 
super850 = lit ! lit @ negate lit ! 
super851 = lit ! lit @ negate lit ! ;s 
super852 = lit @ 0= lit 
super853 = lit @ 0= lit ! 
super854 = lit @ 0= lit ! lit 
super855 = lit @ 0= lit ! lit @ 
super856 = lit @ 0= lit ! lit @ negate 
super857 = lit @ 0= lit ! lit @ negate lit 
super858 = lit @ 0= lit ! lit @ negate lit ! 
super859 = lit @ 0= lit ! lit @ negate lit ! ;s 
super860 = lit @ lit ! lit ! 
super861 = lit @ lit ! lit ! lit 
super862 = lit @ lit ! lit ! lit @ 
super863 = lit @ lit ! lit ! lit @ 0= 
super864 = lit @ lit ! lit ! lit @ 0= lit 
super865 = lit @ lit ! lit ! lit @ 0= lit ! 
super866 = lit @ lit @ lit ! 
super867 = lit @ lit @ lit ! lit 
super868 = lit @ lit @ lit ! lit ! 
super869 = lit @ lit @ lit ! lit ! lit 
super870 = lit @ lit @ lit ! lit ! lit @ 
super871 = lit @ lit @ lit ! lit ! lit @ 0= 
super872 = lit @ negate lit ! ;s 
super873 = negate lit ! ;s 
super874 = = over lit 
super875 = lit 2@ d= 
super876 = r> - 
super877 = ! cells lit + @ dup 
super878 = ! cells lit + @ dup lit 
super879 = ! cells lit + @ dup lit ! 
super880 = ! cells lit + @ dup lit ! ;s 
super881 = + @ dup lit ! ;s 
super882 = @ dup lit ! ;s 
super883 = cells lit + @ dup lit ! ;s 
super884 = dup lit ! cells lit + @ dup 
super885 = dup lit ! cells lit + @ dup lit 
super886 = dup lit ! cells lit + @ dup lit ! 
super887 = dup lit ! cells lit + @ dup lit ! ;s 
super888 = lit ! cells lit + @ dup 
super889 = lit ! cells lit + @ dup lit 
super890 = lit ! cells lit + @ dup lit ! 
super891 = lit ! cells lit + @ dup lit ! ;s 
super892 = lit + @ dup lit ! ;s 
super893 = dup 2@ lit 
super894 = dup 2@ lit 2@ 
super895 = dup 2@ lit 2@ d= 
super896 = swap >r 
super897 = swap r> 
super898 = swap ;s 
super899 = dup lit @ call 
super900 = - swap ;s 
super901 = - swap r> 
super902 = - swap r> - 
super903 = - swap r> - swap 
super904 = - swap r> - swap ;s 
super905 = /mod rot 
super906 = /mod rot lit 
super907 = /mod rot lit /mod 
super908 = /mod rot lit /mod swap 
super909 = /mod rot lit /mod swap >r 
super910 = /mod rot lit /mod swap >r - 
super911 = /mod rot lit /mod swap >r - swap 
super912 = /mod rot lit /mod swap >r - swap r> 
super913 = /mod rot lit /mod swap >r - swap r> - 
super914 = /mod rot lit /mod swap >r - swap r> - swap 
super915 = /mod swap 
super916 = /mod swap >r 
super917 = /mod swap >r - 
super918 = /mod swap >r - swap 
super919 = /mod swap >r - swap r> 
super920 = /mod swap >r - swap r> - 
super921 = /mod swap >r - swap r> - swap 
super922 = /mod swap >r - swap r> - swap ;s 
super923 = >r - 
super924 = >r - swap 
super925 = >r - swap r> 
super926 = >r - swap r> - 
super927 = >r - swap r> - swap 
super928 = >r - swap r> - swap ;s 
super929 = lit /mod rot 
super930 = lit /mod rot lit 
super931 = lit /mod rot lit /mod 
super932 = lit /mod rot lit /mod swap 
super933 = lit /mod rot lit /mod swap >r 
super934 = lit /mod rot lit /mod swap >r - 
super935 = lit /mod rot lit /mod swap >r - swap 
super936 = lit /mod rot lit /mod swap >r - swap r> 
super937 = lit /mod rot lit /mod swap >r - swap r> - 
super938 = lit /mod swap 
super939 = lit /mod swap >r 
super940 = lit /mod swap >r - 
super941 = lit /mod swap >r - swap 
super942 = lit /mod swap >r - swap r> 
super943 = lit /mod swap >r - swap r> - 
super944 = lit /mod swap >r - swap r> - swap 
super945 = lit /mod swap >r - swap r> - swap ;s 
super946 = r> - swap 
super947 = r> - swap ;s 
super948 = rot lit /mod 
super949 = rot lit /mod swap 
super950 = rot lit /mod swap >r 
super951 = rot lit /mod swap >r - 
super952 = rot lit /mod swap >r - swap 
super953 = rot lit /mod swap >r - swap r> 
super954 = rot lit /mod swap >r - swap r> - 
super955 = rot lit /mod swap >r - swap r> - swap 
super956 = rot lit /mod swap >r - swap r> - swap ;s 
super957 = swap >r - 
super958 = swap >r - swap 
super959 = swap >r - swap r> 
super960 = swap >r - swap r> - 
super961 = swap >r - swap r> - swap 
super962 = swap >r - swap r> - swap ;s 
super963 = swap r> - 
super964 = swap r> - swap 
super965 = swap r> - swap ;s 
super966 = swap lit + 
super967 = @ or 
super968 = + cells lit + @ 
super969 = ! dup 
super970 = ! dup lit 
super971 = = over lit @ 
super972 = = over lit @ = 
super973 = = over lit @ = or 
super974 = @ = over lit 
super975 = @ = over lit @ 
super976 = @ = over lit @ = 
super977 = @ = over lit @ = or 
super978 = dup lit @ = over lit 
super979 = dup lit @ = over lit @ 
super980 = dup lit @ = over lit @ = 
super981 = dup lit @ = over lit @ = or 
super982 = lit @ = over lit 
super983 = lit @ = over lit @ 
super984 = lit @ = over lit @ = 
super985 = lit @ = over lit @ = or 
super986 = @ and dup does-exec 
super987 = and dup does-exec 
super988 = dup does-exec 
super989 = lit @ and dup does-exec 
super990 = d0= ?branch 
super991 = @ over or 
super992 = @ over or lit 
super993 = @ over or lit ! 
super994 = @ over or lit ! lit 
super995 = @ over or lit ! lit @ 
super996 = lit @ over or 
super997 = lit @ over or lit 
super998 = lit @ over or lit ! 
super999 = lit @ over or lit ! lit 
super1000 = lit @ over or lit ! lit @ 
super1001 = over or 
super1002 = over or lit 
super1003 = over or lit ! 
super1004 = over or lit ! lit 
super1005 = over or lit ! lit @ 
super1006 = abs = 
super1007 = abs = ;s 
super1008 = abs swap 
super1009 = abs swap abs 
super1010 = abs swap abs = 
super1011 = abs swap abs = ;s 
super1012 = swap abs 
super1013 = swap abs = 
super1014 = swap abs = ;s 
super1015 = @ and ?branch 
super1016 = lit @ and ?branch 
super1017 = lit + @ call 
super1018 = @ lit ! dup 
super1019 = @ lit ! dup lit 
super1020 = lit ! dup 
super1021 = lit ! dup lit 
super1022 = + 2@ rot @ 
super1023 = + 2@ rot @ ;s 
super1024 = cells + 2@ rot @ 
super1025 = cells + 2@ rot @ ;s 
super1026 = dup lit cells + 2@ rot @ 
super1027 = dup lit cells + 2@ rot @ ;s 
super1028 = lit cells + 2@ rot @ 
super1029 = lit cells + 2@ rot @ ;s 
super1030 = rot @ 
super1031 = rot @ ;s 
super1032 = ! branch 
super1033 = lit ! branch 
super1034 = ! lit lit 
super1035 = lit ! lit lit 
super1036 = lit + ! ;s 
super1037 = lit lit ! lit 
super1038 = lit lit ! call 
super1039 = cells lit + ! ;s 
super1040 = @ lit and 
super1041 = @ lit and lit 
super1042 = lit and lit 
super1043 = cells lit + @ call 
super1044 = + @ swap 
super1045 = + @ swap lit 
super1046 = + @ swap lit + 
super1047 = @ swap lit + 
super1048 = lit + @ swap 
super1049 = lit + @ swap lit 
super1050 = lit + @ swap lit + 
super1051 = swap 0= 
super1052 = swap 0= xor 
super1053 = swap 0= xor ;s 
super1054 = xor ;s 
super1055 = @ lit @ and cells 
super1056 = @ lit @ and cells lit 
super1057 = @ lit @ and cells lit + 
super1058 = ! lit lit ! 
super1059 = lit ! lit lit ! 
super1060 = - lit 
super1061 = ! dup lit cells 
super1062 = ! dup lit cells - 
super1063 = ! dup lit cells - @ 
super1064 = ! dup lit cells - @ lit 
super1065 = ! dup lit cells - @ lit ! 
super1066 = - @ lit ! dup 
super1067 = - @ lit ! dup lit 
super1068 = - @ lit ! dup lit cells 
super1069 = - @ lit ! dup lit cells - 
super1070 = - @ lit ! dup lit cells - @ 
super1071 = - @ lit ! dup lit cells - @ lit 
super1072 = - @ lit ! dup lit cells - @ lit ! 
super1073 = @ lit ! dup lit cells 
super1074 = @ lit ! dup lit cells - 
super1075 = @ lit ! dup lit cells - @ 
super1076 = @ lit ! dup lit cells - @ lit 
super1077 = @ lit ! dup lit cells - @ lit ! 
super1078 = cells - @ lit ! dup 
super1079 = cells - @ lit ! dup lit 
super1080 = cells - @ lit ! dup lit cells 
super1081 = cells - @ lit ! dup lit cells - 
super1082 = cells - @ lit ! dup lit cells - @ 
super1083 = cells - @ lit ! dup lit cells - @ lit 
super1084 = dup lit cells - @ lit ! dup 
super1085 = dup lit cells - @ lit ! dup lit 
super1086 = dup lit cells - @ lit ! dup lit cells 
super1087 = dup lit cells - @ lit ! dup lit cells - 
super1088 = lit ! dup lit cells 
super1089 = lit ! dup lit cells - 
super1090 = lit ! dup lit cells - @ 
super1091 = lit ! dup lit cells - @ lit 
super1092 = lit ! dup lit cells - @ lit ! 
super1093 = lit cells - @ lit ! dup 
super1094 = lit cells - @ lit ! dup lit 
super1095 = lit cells - @ lit ! dup lit cells 
super1096 = lit cells - @ lit ! dup lit cells - 
super1097 = lit cells - @ lit ! dup lit cells - @ 
super1098 = 1- lit @ 
super1099 = 1- lit @ * 
super1100 = 1- lit @ * move 
super1101 = 1- lit @ * move ;s 
super1102 = ! lit @ over 
super1103 = ! lit @ over or 
super1104 = ! lit @ over or lit 
super1105 = ! lit @ over or lit ! 
super1106 = ! lit @ over or lit ! lit 
super1107 = ! lit @ over or lit ! lit @ 
super1108 = @ over or lit ! lit @ over 
super1109 = @ over or lit ! lit @ over or 
super1110 = @ over or lit ! lit @ over or lit 
super1111 = @ over or lit ! lit @ over or lit ! 
super1112 = lit ! lit @ over 
super1113 = lit ! lit @ over or 
super1114 = lit ! lit @ over or lit 
super1115 = lit ! lit @ over or lit ! 
super1116 = lit ! lit @ over or lit ! lit 
super1117 = lit ! lit @ over or lit ! lit @ 
super1118 = lit @ over or lit ! lit @ over 
super1119 = lit @ over or lit ! lit @ over or 
super1120 = lit @ over or lit ! lit @ over or lit 
super1121 = or lit ! lit @ over 
super1122 = or lit ! lit @ over or 
super1123 = or lit ! lit @ over or lit 
super1124 = or lit ! lit @ over or lit ! 
super1125 = or lit ! lit @ over or lit ! lit 
super1126 = or lit ! lit @ over or lit ! lit @ 
super1127 = over or lit ! lit @ over 
super1128 = over or lit ! lit @ over or 
super1129 = over or lit ! lit @ over or lit 
super1130 = over or lit ! lit @ over or lit ! 
super1131 = over or lit ! lit @ over or lit ! lit 
super1132 = lit * lit 
super1133 = ! lit @ and 
super1134 = lit ! lit @ and 
super1135 = @ + dup cells 
super1136 = @ + dup cells lit 
super1137 = @ + dup cells lit + 
super1138 = @ + dup cells lit + @ 
super1139 = lit @ + dup cells 
super1140 = lit @ + dup cells lit 
super1141 = lit @ + dup cells lit + 
super1142 = lit @ + dup cells lit + @ 
super1143 = @ < 
super1144 = or swap 
super1145 = lit @ < 
super1146 = lit (?do) 
super1147 = or swap lit 
super1148 = @ lit (?do) 
super1149 = lit @ lit (?do) 
super1150 = lit * lit @ 
super1151 = ! lit @ lit 
super1152 = rot 2@ 
super1153 = lit and + 
super1154 = over lit @ = or ?branch 
super1155 = lit @ ;s 
super1156 = 2r> rot 
super1157 = 2r> rot ?branch 
super1158 = + 2@ rot 2@ 
super1159 = + 2@ rot 2@ ;s 
super1160 = cells + 2@ rot 2@ 
super1161 = cells + 2@ rot 2@ ;s 
super1162 = dup lit cells + 2@ rot 2@ 
super1163 = dup lit cells + 2@ rot 2@ ;s 
super1164 = lit cells + 2@ rot 2@ 
super1165 = lit cells + 2@ rot 2@ ;s 
super1166 = rot 2@ ;s 
super1167 = lit = 
super1168 = dup 2@ lit 2@ d= ?branch 
super1169 = lit 2@ d= ?branch 
super1170 = dup cells lit + @ call 
super1171 = i cells 
super1172 = i cells lit 
super1173 = i cells lit + 
super1174 = i cells lit + @ 
super1175 = 2over 2over 
super1176 = 2over 2over ;s 
super1177 = 2over ;s 
super1178 = i cells lit + @ dup 
super1179 = i cells lit + @ dup lit 
super1180 = lit @ swap 
super1181 = over = 
super1182 = over = ?branch 
super1183 = + lit @ * 
super1184 = + lit @ * lit 
super1185 = + lit @ * lit + 
super1186 = + lit @ * lit + ;s 
super1187 = @ does-exec 
super1188 = @ i 
super1189 = lit @ i 
super1190 = lit @ dup lit @ 
super1191 = @ i call 
super1192 = lit @ i call 
super1193 = @ 1+ lit ! ;s 
super1194 = lit @ 1+ lit ! ;s 
super1195 = 2@ r> 2! ;s 
super1196 = lit @ does-exec 
super1197 = + i 
super1198 = + i call 
super1199 = ! lit cells 
super1200 = over lit call 
super1201 = @ * lit @ 
super1202 = lit @ * lit @ 
super1203 = @ 2>r 
super1204 = @ lit @ 2>r 
super1205 = lit @ 2>r 
super1206 = lit @ lit @ 2>r 
super1207 = lit + cells 
super1208 = lit + cells lit 
super1209 = lit + cells lit + 
super1210 = lit + cells lit + @ 
super1211 = and call 
super1212 = @ and call 
super1213 = lit @ and call 
super1214 = lit - call 
super1215 = lit + call 
super1216 = + 2@ rot lit 
super1217 = + 2@ rot lit + 
super1218 = + 2@ rot lit + @ 
super1219 = + 2@ rot lit + @ ;s 
super1220 = + dup lit + 2@ 
super1221 = + dup lit + 2@ rot 
super1222 = + dup lit + 2@ rot lit 
super1223 = + dup lit + 2@ rot lit + 
super1224 = + dup lit + 2@ rot lit + @ 
super1225 = + dup lit + 2@ rot lit + @ ;s 
super1226 = @ + dup lit + 2@ 
super1227 = @ + dup lit + 2@ rot 
super1228 = @ + dup lit + 2@ rot lit 
super1229 = @ + dup lit + 2@ rot lit + 
super1230 = @ + dup lit + 2@ rot lit + @ 
super1231 = @ + dup lit + 2@ rot lit + @ ;s 
super1232 = dup lit + 2@ 
super1233 = dup lit + 2@ rot 
super1234 = dup lit + 2@ rot lit 
super1235 = dup lit + 2@ rot lit + 
super1236 = dup lit + 2@ rot lit + @ 
super1237 = dup lit + 2@ rot lit + @ ;s 
super1238 = lit + 2@ 
super1239 = lit + 2@ rot 
super1240 = lit + 2@ rot lit 
super1241 = lit + 2@ rot lit + 
super1242 = lit + 2@ rot lit + @ 
super1243 = lit + 2@ rot lit + @ ;s 
super1244 = lit @ + dup lit + 2@ 
super1245 = lit @ + dup lit + 2@ rot 
super1246 = lit @ + dup lit + 2@ rot lit 
super1247 = lit @ + dup lit + 2@ rot lit + 
super1248 = lit @ + dup lit + 2@ rot lit + @ 
super1249 = rot lit + 
super1250 = rot lit + @ 
super1251 = rot lit + @ ;s 
super1252 = lit @ lit @ and cells 
super1253 = lit @ lit @ and cells lit 
super1254 = lit @ lit @ and cells lit + 
super1255 = @ lit @ * 
super1256 = lit @ lit @ * 
super1257 = @ lit @ * lit 
super1258 = @ lit @ * lit + 
super1259 = @ lit @ * lit + ;s 
super1260 = lit @ lit @ * lit 
super1261 = lit @ lit @ * lit + 
super1262 = lit @ lit @ * lit + ;s 
super1263 = lit lit call 
super1264 = ! lit ! call 
super1265 = lit ! lit ! call 
super1266 = over lit + 
super1267 = > 0= 
super1268 = > 0= ?branch 
super1269 = @ lit * 
super1270 = d= and 
super1271 = d= and ;s 
super1272 = dup lit + @ 
super1273 = ! lit @ over or lit ! lit @ over 
super1274 = ! lit @ over or lit ! lit @ over or 
super1275 = lit ! lit @ over or lit ! lit @ over 
super1276 = + @ dup lit @ 
super1277 = cells lit + @ dup lit @ 
super1278 = lit + @ dup lit @ 
super1279 = @ lit lit 
super1280 = @ lit lit @ 
super1281 = @ lit lit @ * 
super1282 = lit @ lit lit 
super1283 = lit @ lit lit @ 
super1284 = lit @ lit lit @ * 
super1285 = lit lit @ * 
super1286 = lit fill 
super1287 = fill ;s 
super1288 = lit fill ;s 
super1289 = @ 1- 
super1290 = @ 0= ?branch 
super1291 = lit @ 0= ?branch 
super1292 = lit @ or 
super1293 = i cells lit + @ dup lit @ 
super1294 = @ over = 
super1295 = @ over = ?branch 
super1296 = lit @ over = 
super1297 = lit @ over = ?branch 
super1298 = lit @ 1- 
super1299 = @ or lit 
super1300 = lit @ or lit 
super1301 = >r call 
super1302 = @ 1- lit 
super1303 = lit @ 1- lit 
super1304 = @ 1- lit ! 
super1305 = @ 1- lit ! ;s 
super1306 = lit @ 1- lit ! 
super1307 = lit @ 1- lit ! ;s 
super1308 = + ! lit 
super1309 = lit + ! lit 
super1310 = + ! lit @ 
super1311 = lit + ! lit @ 
super1312 = lit swap 
super1313 = + lit + 
super1314 = + lit + @ 
super1315 = + lit + @ ;s 
super1316 = @ + lit + 
super1317 = @ + lit + @ 
super1318 = @ + lit + @ ;s 
super1319 = lit @ + lit + 
super1320 = lit @ + lit + @ 
super1321 = lit @ + lit + @ ;s 
super1322 = @ lit * lit 
super1323 = lit @ lit * 
super1324 = lit @ lit * lit 
super1325 = * call 
super1326 = @ * call 
super1327 = lit @ * call 
super1328 = dup lit @ = ?branch 
super1329 = + @ dup lit @ = 
super1330 = + @ dup lit @ = ?branch 
super1331 = @ dup lit @ = 
super1332 = @ dup lit @ = ?branch 
super1333 = cells lit + @ dup lit @ = 
super1334 = cells lit + @ dup lit @ = ?branch 
super1335 = i cells lit + @ dup lit @ = 
super1336 = i cells lit + @ dup lit @ = ?branch 
super1337 = lit + @ dup lit @ = 
super1338 = lit + @ dup lit @ = ?branch 
super1339 = @ lit lit @ * call 
super1340 = lit @ lit lit @ * call 
super1341 = lit lit @ * call 
super1342 = + @ swap lit + @ 
super1343 = + @ swap lit + @ ;s 
super1344 = + dup lit + @ 
super1345 = + dup lit + @ swap 
super1346 = + dup lit + @ swap lit 
super1347 = + dup lit + @ swap lit + 
super1348 = + dup lit + @ swap lit + @ 
super1349 = + dup lit + @ swap lit + @ ;s 
super1350 = @ + dup lit + @ 
super1351 = @ + dup lit + @ swap 
super1352 = @ + dup lit + @ swap lit 
super1353 = @ + dup lit + @ swap lit + 
super1354 = @ + dup lit + @ swap lit + @ 
super1355 = @ + dup lit + @ swap lit + @ ;s 
super1356 = @ swap lit + @ 
super1357 = @ swap lit + @ ;s 
super1358 = dup lit + @ swap 
super1359 = dup lit + @ swap lit 
super1360 = dup lit + @ swap lit + 
super1361 = dup lit + @ swap lit + @ 
super1362 = dup lit + @ swap lit + @ ;s 
super1363 = lit + @ swap lit + @ 
super1364 = lit + @ swap lit + @ ;s 
super1365 = lit @ + dup lit + @ 
super1366 = lit @ + dup lit + @ swap 
super1367 = lit @ + dup lit + @ swap lit 
super1368 = lit @ + dup lit + @ swap lit + 
super1369 = lit @ + dup lit + @ swap lit + @ 
super1370 = swap lit + @ 
super1371 = swap lit + @ ;s 
super1372 = swap lit call 
super1373 = + swap 
super1374 = @ + swap 
super1375 = lit @ + swap 
super1376 = * lit @ * 
super1377 = * lit @ * lit 
super1378 = * lit @ * lit + 
super1379 = * lit @ * lit + ;s 
super1380 = @ lit * lit @ 
super1381 = @ lit * lit @ * 
super1382 = @ lit * lit @ * lit 
super1383 = @ lit * lit @ * lit + 
super1384 = @ lit * lit @ * lit + ;s 
super1385 = lit * lit @ * 
super1386 = lit * lit @ * lit 
super1387 = lit * lit @ * lit + 
super1388 = lit * lit @ * lit + ;s 
super1389 = lit @ lit * lit @ 
super1390 = lit @ lit * lit @ * 
super1391 = lit @ lit * lit @ * lit 
super1392 = lit @ lit * lit @ * lit + 
super1393 = lit @ lit * lit @ * lit + ;s 
super1394 = lit and ;s 
super1395 = + lit @ lit 
super1396 = tuck ! 
super1397 = + 2! 
super1398 = + 2! ;s 
super1399 = cells + 2! 
super1400 = cells + 2! ;s 
super1401 = lit cells + 2! 
super1402 = lit cells + 2! ;s 
super1403 = ! lit cells + 
super1404 = ! lit cells + 2! 
super1405 = ! lit cells + 2! ;s 
super1406 = tuck ! lit 
super1407 = tuck ! lit cells 
super1408 = tuck ! lit cells + 
super1409 = tuck ! lit cells + 2! 
super1410 = tuck ! lit cells + 2! ;s 
super1411 = lit 2@ 2>r 
super1412 = lit 2@ 2>r call 
super1413 = lit = or 
super1414 = + swap 2@ 
super1415 = + swap 2@ d0= 
super1416 = + swap 2@ d0= ?branch 
super1417 = @ + swap 2@ 
super1418 = @ + swap 2@ d0= 
super1419 = @ + swap 2@ d0= ?branch 
super1420 = lit @ + swap 2@ 
super1421 = lit @ + swap 2@ d0= 
super1422 = lit @ + swap 2@ d0= ?branch 
super1423 = swap 2@ 
super1424 = swap 2@ d0= 
super1425 = swap 2@ d0= ?branch 
super1426 = swap lit @ + 
super1427 = swap lit @ + swap 
super1428 = swap lit @ + swap 2@ 
super1429 = swap lit @ + swap 2@ d0= 
super1430 = swap lit @ + swap 2@ d0= ?branch 
super1431 = @ lit ! lit @ 
super1432 = + @ + 
super1433 = + @ + ;s 
super1434 = cells lit + @ + 
super1435 = cells lit + @ + ;s 
super1436 = lit + @ + 
super1437 = lit + @ + ;s 
super1438 = * lit @ lit 
super1439 = * lit @ lit and 
super1440 = * lit @ lit and lit 
super1441 = * lit @ lit and lit @ 
super1442 = * lit @ lit and lit @ + 
super1443 = * lit @ lit and lit @ + + 
super1444 = * lit @ lit and lit @ + + ;s 
super1445 = + + 
super1446 = + + ;s 
super1447 = @ * lit @ lit 
super1448 = @ * lit @ lit and 
super1449 = @ * lit @ lit and lit 
super1450 = @ * lit @ lit and lit @ 
super1451 = @ * lit @ lit and lit @ + 
super1452 = @ * lit @ lit and lit @ + + 
super1453 = @ * lit @ lit and lit @ + + ;s 
super1454 = @ + + 
super1455 = @ + + ;s 
super1456 = @ lit and lit @ 
super1457 = @ lit and lit @ + 
super1458 = @ lit and lit @ + + 
super1459 = @ lit and lit @ + + ;s 
super1460 = and lit @ + 
super1461 = and lit @ + + 
super1462 = and lit @ + + ;s 
super1463 = lit @ * lit @ lit 
super1464 = lit @ * lit @ lit and 
super1465 = lit @ * lit @ lit and lit 
super1466 = lit @ * lit @ lit and lit @ 
super1467 = lit @ * lit @ lit and lit @ + 
super1468 = lit @ * lit @ lit and lit @ + + 
super1469 = lit @ + + 
super1470 = lit @ + + ;s 
super1471 = lit @ lit and 
super1472 = lit @ lit and lit 
super1473 = lit @ lit and lit @ 
super1474 = lit @ lit and lit @ + 
super1475 = lit @ lit and lit @ + + 
super1476 = lit @ lit and lit @ + + ;s 
super1477 = lit and lit @ 
super1478 = lit and lit @ + 
super1479 = lit and lit @ + + 
super1480 = lit and lit @ + + ;s 
super1481 = @ or lit ! 
super1482 = lit @ or lit ! 
super1483 = dup 0< 
super1484 = @ lit @ and ?branch 
super1485 = lit @ lit @ and ?branch 
super1486 = lit ! lit @ lit 
super1487 = 2@ over 
super1488 = 2@ over <> 
super1489 = 2@ over <> ?branch 
super1490 = 2dup i 
super1491 = 2dup i 2@ 
super1492 = 2dup i 2@ over 
super1493 = 2dup i 2@ over <> 
super1494 = 2dup i 2@ over <> ?branch 
super1495 = lit within 
super1496 = over lit - 
super1497 = over lit - call 
super1498 = < 0= 
super1499 = cells lit + @ ;s 
super1500 = or over lit + 
super1501 = or over lit + call 
super1502 = over lit + call 
super1503 = + lit ! 
super1504 = + lit ! ;s 
super1505 = @ + lit ! 
super1506 = @ + lit ! ;s 
super1507 = lit @ + lit ! 
super1508 = lit @ + lit ! ;s 
super1509 = = or over 
super1510 = = or over lit 
super1511 = ! (loop) 
super1512 = 0= and ;s 
super1513 = and + i 
super1514 = and + i call 
super1515 = lit and + i 
super1516 = lit and + i call 
super1517 = r> 2r> 
super1518 = * + lit 
super1519 = * + lit @ 
super1520 = * + lit @ * 
super1521 = * + lit @ * lit 
super1522 = * + lit @ * lit + 
super1523 = * + lit @ * lit + ;s 
super1524 = @ * + lit 
super1525 = @ * + lit @ 
super1526 = @ * + lit @ * 
super1527 = @ * + lit @ * lit 
super1528 = @ * + lit @ * lit + 
super1529 = @ * + lit @ * lit + ;s 
super1530 = @ dup lit @ * 
super1531 = @ dup lit @ * + 
super1532 = @ dup lit @ * + lit 
super1533 = @ dup lit @ * + lit @ 
super1534 = @ dup lit @ * + lit @ * 
super1535 = @ dup lit @ * + lit @ * lit 
super1536 = @ dup lit @ * + lit @ * lit + 
super1537 = dup lit @ * 
super1538 = dup lit @ * + 
super1539 = dup lit @ * + lit 
super1540 = dup lit @ * + lit @ 
super1541 = dup lit @ * + lit @ * 
super1542 = dup lit @ * + lit @ * lit 
super1543 = dup lit @ * + lit @ * lit + 
super1544 = dup lit @ * + lit @ * lit + ;s 
super1545 = lit @ * + lit 
super1546 = lit @ * + lit @ 
super1547 = lit @ * + lit @ * 
super1548 = lit @ * + lit @ * lit 
super1549 = lit @ * + lit @ * lit + 
super1550 = lit @ * + lit @ * lit + ;s 
super1551 = lit @ dup lit @ * 
super1552 = lit @ dup lit @ * + 
super1553 = lit @ dup lit @ * + lit 
super1554 = lit @ dup lit @ * + lit @ 
super1555 = lit @ dup lit @ * + lit @ * 
super1556 = lit @ dup lit @ * + lit @ * lit 
super1557 = swap lit @ and 
super1558 = + @ or 
super1559 = + @ or ?branch 
super1560 = + dup lit @ 
super1561 = + dup lit @ = 
super1562 = + dup lit @ = over 
super1563 = + dup lit @ = over cells 
super1564 = + dup lit @ = over cells lit 
super1565 = + dup lit @ = over cells lit + 
super1566 = + dup lit @ = over cells lit + @ 
super1567 = + dup lit @ = over cells lit + @ or 
super1568 = = over cells 
super1569 = = over cells lit 
super1570 = = over cells lit + 
super1571 = = over cells lit + @ 
super1572 = = over cells lit + @ or 
super1573 = = over cells lit + @ or ?branch 
super1574 = @ = over cells 
super1575 = @ = over cells lit 
super1576 = @ = over cells lit + 
super1577 = @ = over cells lit + @ 
super1578 = @ = over cells lit + @ or 
super1579 = @ = over cells lit + @ or ?branch 
super1580 = @ or ?branch 
super1581 = cells lit + @ or 
super1582 = cells lit + @ or ?branch 
super1583 = dup lit @ = over cells 
super1584 = dup lit @ = over cells lit 
super1585 = dup lit @ = over cells lit + 
super1586 = dup lit @ = over cells lit + @ 
super1587 = dup lit @ = over cells lit + @ or 
super1588 = dup lit @ = over cells lit + @ or ?branch 
super1589 = lit + @ or 
super1590 = lit + @ or ?branch 
super1591 = lit @ = over cells 
super1592 = lit @ = over cells lit 
super1593 = lit @ = over cells lit + 
super1594 = lit @ = over cells lit + @ 
super1595 = lit @ = over cells lit + @ or 
super1596 = lit @ = over cells lit + @ or ?branch 
super1597 = over + dup lit 
super1598 = over + dup lit @ 
super1599 = over + dup lit @ = 
super1600 = over + dup lit @ = over 
#line 2457 "./prim"


\+
