\ brew-basics.fs
\ 	$Id: brew-basics.fs,v 1.40 2001/02/14 23:58:35 f Exp $	

\ doing some very basic definitions first makes life easier...
\ compile options:
\ preset-run-mode
\
decimal

\ basics (that couldn't be defined in 'basics.fs'):

\ bit masks that can be written to files as readable text:
\ Puts the xt of each named mask in a list.
: LISTED-MASK: ( list bit-position -- list bit-position+1 )
    >r
    save-input
    r>
    MASK:
    >r
    restore-input
    ABORT" LISTED-MASK: Troubles restoring input source"
    bl word find IF
	over  >list		\ put xt into list
    ELSE bell THEN
    r> ;

\ from brew.fs

variable run-mode	run-mode off		\ mode like recording and so
[DEFINED] preset-run-mode [IF]
    preset-run-mode run-mode !
[THEN]

LIST: run-mode-masks
run-mode-masks
0
LISTED-MASK: mutating
LISTED-MASK: writing-code	\ writing code *only*
LISTED-MASK: recording
LISTED-MASK: write-diff	\ writing diff files (for recording or logging)
LISTED-MASK: playback
2drop

\ do something not recording it
: not-recording ( xt -- )
    run-mode dup @
    dup >r			\ preserve run-mode
    recording invert and swap !
    EXECUTE
    r> run-mode ! ;

2variable spot					\ field index of actual spot
spot cell+ CONSTANT step	-1 step !	\ actual time
\ to get both use:  spot 2@ ( -- step spot )

variable child-spot			\ handy ;-)     spot of an actual child

64 constant max-stack-effect	\ no genom should ever go beyond that
max-stack-effect cell mod [IF]
    cr cr .( Compile time error: 'max-stack-effect CELL MOD' must be zero! ) cr
    quit
[THEN]

\ dummy function to indicate missing selection in some xt lists
: (none) ;

\ I/O stuff:
[UNDEFINED] .tab [IF]	: .tab	9 emit ;	[THEN]	\ prints a tab
[UNDEFINED] .bs  [IF]	: .bs	8 emit ;	[THEN]	\ prints a backspace

: at-x ( n -- )	 at? nip at-xy ;

\ put cursor on last line of screen
l-s 1- CONSTANT	last-line
: last-left ( -- )   0 last-line at-xy ;

\ set cursor to 'column' of 'columns' screen columns.
: screen-column ( column columns -- )  c-l swap /  *  at-x ;

\ erase line end:
: clear-line-to-end ( -- )   c-l  at? drop - spaces ;
\
lower-right-scrolls [IF]			\ there are some terminals that do
    : clear-line-to-end ( -- )		\ automatic scrolling when writing
	at? nip last-line <> IF		\ to the very last screen position.
	    clear-line-to-end
	ELSE
	    c-l 1-  at? drop -  spaces	\ don't touch last position.
	THEN ;
[THEN]

\ wait some time (at least) if key? gets true
\ on my system it waits about 3 times as long, but I don't care
: await-key? ( milliseconds -- true|false )
    BEGIN	( milliseconds )
	key? 0= IF  2 ms THEN
	key? 0= IF  3 ms THEN
	key? 0= IF  5 ms THEN
	key? IF drop true EXIT THEN
	10 -
	dup 1 < IF drop  false EXIT THEN
    AGAIN ;

: .ON-off ( flag -- )   IF ." is ON " ELSE ." is off" THEN ;

: .YES-NO ( flag -- )   IF ." YES" ELSE ." NO" THEN ;

: (do-FORTH)				\ FORTH interpreter for menus
   last-left clear-line-to-end
   last-left ." FORTH: "
   pad dup 80 7 - accept  evaluate ;
: do-FORTH
    ['] (do-FORTH) CATCH IF bell THEN
    key
    last-left clear-line-to-end
    bl = IF RECURSE THEN ;

c-l STRINGBUF-HANDLE: (message)		\ messages to be desplayed in info line
variable message-count			\ how many steps to display message
: >message ( addr count n -- )
    message-count !
    (message) dup stringbuf-empty cat ;

variable message-fg-color-xt	message-fg-color-xt off
variable message-bg-color-xt	message-bg-color-xt off

: |>message| ( -- )
    last-left
    pad 80 accept ?dup IF
\	cursor-visible
\	."  # "
	pad swap 20 >message
    THEN ;

variable (manually-selected-cell)	(manually-selected-cell) off

\ brew will put all it's output relative to out-dir, if the user does not
\ select another directory. Make the string end with a directory separator.
[UNDEFINED] out-dir [IF]   : out-dir ( -- addr count )   s" OUTPUT/" ;	[THEN]

\ word to put out-dir in front of a subdirectory name:
\ attention: there is only *one* buffer used, don't use multiple strings.
file-names-length# S-BUF: (out-sub-dir)
: out-sub-dir ( addr count -- addr' count' )
    (out-sub-dir) >r
    r@ s-buf-clear
    out-dir r@ s-buf-cat
    ( addr count ) r@ s-buf-cat
    r> s-buf>string ;

: rec-play-dir ( -- addr count )	s" rec-play/"	out-sub-dir ;
: log-dir ( -- addr count )		s" log/"	out-sub-dir ;
: FORTH-dir ( -- addr count )		s" FORTH/"	out-sub-dir ;
: tmp-dir ( -- addr count )		s" tmp/"	out-sub-dir ;

\ Marcel Hendrix didn't like that and I think he has a point.
\ I guess I can throw it out without problems.
\
\  s" bigFORTH" environment? [IF] 2drop \ drop version number
\      cr .( Make output directories for bigforth version not done yet. )
\      cr 400 ms
\  [ELSE]
\  \ make sure directories exists
\  MARKER forget-this

\  : ?make-directory ( addr count -- )
\      128 stringbuf-open >r
\      s" test -e "	r@ cat
\      2dup		r@ cat
\      s"  || mkdir "	r@ cat
\      ( addr count )	r@ cat
\      r@ buffer-as-string system
\      r> stringbuf-close ;

\  out-dir		?make-directory
\  tmp-dir		?make-directory
\  rec-play-dir	?make-directory
\  log-dir		?make-directory
\  FORTH-dir	?make-directory

\  forget-this
\  [THEN]


\ cat *two* strings to a filename and handle a buffer.
\ Please close the buffer yourself.
: file-name-cat ( addr1 count1 addr2 count2 -- handle )
    file-names-length# stringbuf-open >r
    2swap r@ cat r@ cat
    r> ;
