\  compile-options.fs
\ 	$Id: compile-options.fs,v 1.43 2001/02/15 21:10:37 f Exp $	

\ This file is included early, but after 'system-dependent.fs'

\ You can change this file or make a 'my-compile-options.fs' to change
\ compile time options.  It gets included just after this one.

\ Note that some switches can be compile time switches *and* run time switches.
\ Compile swich policy:
\ * CREATEd names are switching by their existence.
\   Be careful with names that could be predefined in a system...
\ * CONSTANTs are pure compile time switches too.
\ * a VALUE can be used instead, in cases where it might be reset.
\ * If a VARIABLE is defined it gets compiled as run time switch.
\   It's value gets the default run time value.
\   (There are exceptions, see 'ekey-cursor-support')
\ * Words that do something can be defined system dependent,
\   if the source checks with [UNDEFINED]
\   Tell me, if other words should be checked for.  epprecht@sunweb.ch


\ KEYS:

\ VARIABLE use-ekey
\ If 'use-ekey' is defined at compile time it gets compiled as run time switch
\ Define it in the forth system specific files.

\ scope of ekey and ekey?   Both depend on 'use-ekey', see below.
\ I use a variable here, because it might get reset.
[UNDEFINED] ekey-cursor-support [IF]
    variable ekey-cursor-support	ekey-cursor-support off
[THEN]
[UNDEFINED] ekey-function-keys-support [IF]
    variable ekey-function-keys-support	ekey-function-keys-support off
[THEN]
\ some switches depend on 'use-ekey':
[DEFINED] use-ekey [IF]
    use-ekey @ 0= [IF]
	ekey-cursor-support off
	ekey-function-keys-support off
    [THEN]
[ELSE]
    ekey-cursor-support off
    ekey-function-keys-support off
[THEN]


\ MOUSE:

\ Switch for real mouse support or only cursor keys and <return>?
\ The switch is here, but no mouse support yet.  Help appreciated.
FALSE CONSTANT mouse-supported


\ DISPLAY:
1 5 2CONSTANT default-horizontal-zoom-scale
1 4 2CONSTANT default-vertical-zoom-scale

\ there are systems where writing to the very last screen position does
\ automatic scrolling.  Set 'lower-right-scrolls' to true in that case.
[UNDEFINED] lower-right-scrolls [IF]  false CONSTANT lower-right-scrolls [THEN]

\ where to switch in arbitrary situations, spot or step display?
TRUE CONSTANT spot-display-prefered
\ COLORS:

\ for systems where colors don't work:
\ CREATE never-use-colors

\ if colors don't start at zero (but have continuous range)
\ ( lowest-color-code ) constant color-offset

\ 'title-colors-xt' for menu title colors. See 'brew.fs'

\ FILES:
decimal 256 constant file-names-length#	\ defaults to &256

TRUE CONSTANT flush-files			\ flush files after writing?

\ brew will put all it's output relative to out-dir, if the user does not
\ select another directory. Make the string end with a directory separator.
\ : out-dir ( -- addr count )    s" OUTPUT/" ;	\ see brew-basics.fs

[UNDEFINED] use-fileselect [IF] \ work in progress
    false CONSTANT use-fileselect
[THEN]


\ BREW:

\ normally run-mode defaults to zero.
\ sometimes it can make sense to change that.
\ ( give run-mode as a number ) CONSTANT preset-run-mode
\ 1 CONSTANT preset-run-mode	\ record

\ if 'future-quality-change' is on, cells can change future qualities
\ this can happen individually or after a step from 'world-do'
\ TRUE CONSTANT future-change-individal
FALSE CONSTANT future-change-individal

\ legacy support for single cells outside the world's field wanted?
FALSE CONSTANT support-single-individuals?

\ maximal number of scans displayed in step display.
\ (on a bigger screen it could make sense to increase that).
\ 12 constant max-step-scans#

\ DEBUG:

\ If 'log-mask' is true at compile time there will be more places
\ where it is possible to log the inner working of mutation.
\ It will still be run time switchable, but brew will run a bit slower.
\ Setting it to a positive value will log some extra information.
\ This makes only sense for heavy debugging and is only occasionally supported.
VARIABLE log-mask	log-mask ON	\ compile time *and* run time switch

log-mask ON		\ possibility to log more mutation infos. a bit slower
\ log-mask OFF		\ possibility to log usual amount of infos	fastest
\ 1 log-mask !		\ extra info.				  could be slow
\ log-mask get's reset after compiling.

VARIABLE debugging	debugging ON	\ 'debug' as name is taken


\ EXPERIMENTAL:
\ Work in progress, not all combinations of values are possible yet.

7 VALUE nuc-organs#	\ compile time VALUE
\ parameters, read only variant of organs:
\ seems ok, but mutation get's out of balance with current defaults.
\ switch it off.
7 VALUE nuc-parameters#		\ # of read only nuc parameters
\ invisible, diversified:
0 CONSTANT nuc-invisible#	\ planed to become a compile time VALUE
\ invisible, not diversified:
0 CONSTANT nuc-secrets#		\ planed to become a compile time VALUE

\ these are planed to become compile time VALUEs:
3 VALUE spot-qualities#		\ # of different qualities
0 CONSTANT spot-secrets#		\ # of hidden spot variables
\ read only equivalent of qualities:
3 VALUE spot-properties#		\ # of read only spot properties

\ use new random gene selecting pool lists with relative probability:
TRUE VALUE use-pool-select

\ Include pointer plane in spot scans?
\ 0 VALUE p0		\ watch memory allocation ;-)
1 VALUE p0		\ normal usage

\ Linux console can use color code 9 and gets a default color.
\ This works on both, background and foreground, giving two different colors.
\ Marcel Hendrix reportet that this is not portable. 
\ On systems where this mechanism does not work do the following:
\ * Define 'default-background' and 'default-foreground'
\   I would take black and white.
\ * Define 'default-color' 'color-foreground' and 'color-background'
\   in a system dependent way, such that they will do The Right Thing.
\   I'd be interested to get a mail with your adaptions epprecht@sunweb.ch


\ fake conservative settings:
TRUE [IF] \ with these the old demos and benchmarks should still run.
    TRUE TO use-pool-select
    7 TO nuc-organs#
    0 TO nuc-parameters#
    3 TO spot-qualities#
    0 TO spot-properties#
    1 TO p0
[THEN]

FALSE [IF] \ current guess-A experiment
    3 TO nuc-organs#
    1 TO spot-properties#
    0 TO nuc-parameters#
[THEN]

FALSE [IF] \ just trying out somethings...
    1 TO nuc-organs#		\ ##### not possible yet
    3 TO nuc-organs#
    0 TO nuc-parameters#
    1 TO spot-qualities#
    0 TO spot-properties#
[THEN]
