\ lists.fs
\ 	$Id: lists.fs,v 1.9 2001/02/18 19:02:55 f Exp $	
\ version with list reference to list descriptor data field
\ (so next-node works also on the descriptor, giving node zero)

\ nodes are referenced by the data's address

false				\ true for testing, false normally
dup [if]
    cr .( testing new-lists.fs ) cr
    include keybuf.fs
    include basics.fs
[then]

\ structure of each node:
0
OFFSET: >link	\ *must* be first, because it is used in nodes *and* descriptor
OFFSET: >flags	\ a cell for flags ( I'm generous ;-)
dup cells constant node-descriptor-length
( count )	\  for the list descriptor offsets

\ flag masks
0
MASK: is-descriptor
MASK: is-list
drop

: >node-descriptor ( node-addr -- node-descriptor-addr )
    node-descriptor-length - ;

\ the list descriptor has the same structure as a node with two data cells
\ the 1st data is the number of cells for the other nodes
\ the 2nd is the number of nodes, not counting the descriptor node
\ flag is-descriptor describes it as list descriptor

\ list descriptor structure:
\ offset in cells is on the stack from the node-descriptor offsets
\ >link  already compiled
\ >flags already compiled
( count )
OFFSET: >data-cells#	\ number of data cells in each node
OFFSET: >nodes#		\ element counter
cells constant list-descriptor-length

: >list-descriptor ( list-addr -- list-descriptor-addr )
    POSTPONE >node-descriptor ; IMMEDIATE	\ it's just an alias

: list>flags ( list-addr -- list-flags-addr )
    >list-descriptor >flags ;	\ encapsulate >flags

\ list descriptors and nodes can be allocated, in buffers, in the data space...
defer <list-allocate> ( u -- addr )	\ allocate a list buffer
: here-list-allocate ( u -- addr )
    dup              ( u u )
    here swap allot  ( u addr )
    dup rot erase ;  ( addr )
' here-list-allocate IS <list-allocate>	\ at least for testing
defer <list-free> ( addr -- )		\ free a list buffer
' drop IS <list-free>			\ cheating a bit...

: allocated-list-allocate ( u -- addr )
    dup allocate ABORT" Couldnt allocate list."	( u addr )
    dup rot erase ;				( addr )
' allocated-list-allocate IS <list-allocate>
: free-allocated-list ( addr -- )
    free ABORT" free-allocated-list: Couldn't free list." ;
' free-allocated-list IS <list-free>

\ word to define and initiate a list:
\ deflist ( data-cells -- list-addr )	\ list-addr = addr of zero node
\ the list is referenced by the address of the descriptors data field
: deflist ( u -- list-addr )
    dup 1 < ABORT" deflist: lists must have at least one data cell."
    list-descriptor-length <list-allocate> ( u list-descriptor-addr )
    dup  >link		off
    dup  >flags is-descriptor is-list or swap !
    tuck >data-cells#	!
    dup  >nodes#	off
    node-descriptor-length + ;	\ same structure as nodes

\ word to define a named list:
: nLIST: ( u -- )
    deflist
    CREATE ,
    DOES> @ ;
: LIST: ( -- )   1 nLIST: ;

: nodes ( list-addr -- u ) >list-descriptor >nodes# @ ;

\ (next-node applied to the list-address gives node zero)
: next-node ( node-addr -- node-addr'|0)   >node-descriptor >link @ ;

: next-node-and-value ( node-addr -- node-addr' n true | false )
    next-node  dup IF dup @ true THEN  ;

: n'th-node ( u list-addr -- node-addr )
    >list-descriptor
    2dup >nodes# @ 1- > ABORT" n'th-node: node does not exist."
    >link @   swap 0 ?DO next-node LOOP ;

: allocate-node ( data-cells -- node-addr )
    cells node-descriptor-length + <list-allocate> node-descriptor-length + ;

: new-node ( list-addr -- node-addr)
    dup >list-descriptor		( list-addr list-descriptor-addr )
    dup >nodes# dup >r @	 	( r: addr-nodes# )
    ?dup IF				\ not zero node
	1-
	2 pick
	n'th-node >node-descriptor	( l-addr ld-addr last-node-dscr-addr )
    ELSE				\ zero node
	 dup				( list-addr ld-addr ld-addr )
    THEN
    >link				( list-addr ld-addr link-addr )
    swap >data-cells# @ allocate-node	( list-addr link-addr new-node-addr )
    dup rot ! nip ( new-node-addr )	\ set link
    1 r> +! ;				\ increase node count

: next-or-new-node ( node list-addr -- )
    >r
    next-node				\ switch to next node
    dup 0= IF
	drop
	r@ new-node			\ adding one if none left
    THEN
    rdrop ;

: >list ( n list-addr --)   new-node ! ;	\ makes a new node and stores n

\ word to do 'something ( node --)' on all nodes of a list.
: do-with-all-nodes ( xt list -- )
    dup nodes 0 ?DO ( xt actual-node )
	next-node
	2dup swap execute
    LOOP
    2drop ;

: copy-simple-list-elements ( from-list to-list -- )
    >r
    BEGIN
	next-node-and-value
    WHILE
	 r@ >list
    REPEAT
    rdrop ;
    
\ lists as list nodes:
: node-is-n-list ( node-addr u -- list-addr' )
    >r
    dup >node-descriptor
    >flags dup @ is-list or swap !
    r> deflist dup rot ! ;

: new-node-as-n-list ( list-addr u -- list-addr' )
    swap new-node
    swap node-is-n-list ;

: new-node-as-list ( list-addr -- list-addr' )
    dup >list-descriptor >data-cells# @ new-node-as-n-list ;

: node-is-list? ( node-addr - list-addr'|0)
    dup >node-descriptor >flags @ is-list and IF
	@
    ELSE
	drop false
    THEN ;

dup [if]
    1 deflist
    dup new-node 0 swap !
    dup new-node 1 swap !
    dup new-node 2 swap !
    dup 3 swap >list
    dup nodes cr .( nodes: ) .
    cr .( data: )
    0 over n'th-node @ .
    1 over n'th-node @ .
    2 over n'th-node @ .
    3 over n'th-node @ .
    cr .( adding a new node as list: data: )
    dup new-node-as-list	dup new-node 7 swap !
				dup new-node 14 swap !
				0 over n'th-node @ .
				1 swap n'th-node @ .
    cr .( adding another node )
    dup new-node 5 swap !
    
    dup nodes cr .( nodes: ) .

: .all-data ( list-addr -- )
    dup nodes 0 ?DO
	i over n'th-node
	dup @
	swap node-is-list?
	?dup IF
	    cr ." list as element " i . ." :" cr
	    RECURSE drop
	    cr ." continue with old list:" cr
	ELSE
	    .
	THEN
    LOOP drop cr ." done" ;
    
    cr cr .( .all-data) cr
    dup .all-data cr

    drop

    cr .( testing LIST: and >list : )
    LIST: testlist
    0  testlist >list
    10 testlist >list
    20 testlist >list
    30 testlist >list
    testlist .all-data
[then]

drop
