\ menu.fs
\ 	$Id: menu.fs,v 1.19 2001/02/11 09:52:46 f Exp $	
\ work in progress...

\ ****************************************************************
\ Tool to build a kind of active text menu as user interface.

\ User input like pressing a key or selecting a location can trigger actions.
\ Locations can be selected by moving the cursor and pressing <enter>.
\ If 'mouse-supported' is on, the mouse can be used too.

\ Menus can be nested.

\ There are two types of menus: reactions on keys, and reactions on
\ choosing locations (by the mouse, or with cursor keys and RETURN)
\ Other menus, like reacting on MIDI input could be constructed easily.

\ See the tests at the end of the file for a usage example.

\ ****************************************************************
\ compile options:	Normally set by 'compile-options.fs'.
\
\ Test switch: Not all dependencies resolved any more. Use it with 'brew.fs'
FALSE		\ true for testing, false normally
dup [IF]
    [UNDEFINED] allowed-key-codes# [IF]
	decimal 276 constant allowed-key-codes#
    [THEN]
    INCLUDE system-dependent.fs
[THEN]
\
\ Real mouse support or only cursor keys and <return> ?
\ The switch is here, but mouse support is not done yet.  Help appreciated.
[UNDEFINED] mouse-supported [IF]
    FALSE CONSTANT mouse-supported
[THEN]
\
\ scope of 'ekey' and 'ekey?'
\ ekey-cursor-support
[UNDEFINED] ekey-cursor-support [IF]
    variable ekey-cursor-support	ekey-cursor-support off
[THEN]
\
\ ekey-function-keys-support
[UNDEFINED] ekey-function-keys-support [IF]
    variable ekey-function-keys-support	ekey-function-keys-support off
[THEN]

\ ****************************************************************
[UNDEFINED] cursor-visible [IF]	   include console-codes.fs 	[THEN]

[UNDEFINED] c-l [IF]		   include basics.fs 		[THEN]
\ ****************************************************************

\ The idea is to have an array entry for each key and each location
\ describing the action to take if the user selects it.
\ I call such an entry record a menu cell.
\ Data common to all cells of a menu are stored in the menu descriptor.

\ Define menu *cell* structure:
0
OFFSET:	>menu-xt
OFFSET:	>menu-flags
OFFSET:	>menu-parameter-1
OFFSET:	>menu-parameter-2
OFFSET:	>menu-input-at		\ x-y of input (and 16 bit unused)
OFFSET: >menu-any-data	\ can be used for anything (like a string pointer)
OFFSET: >menu-any-data2
cells CONSTANT menu-cell-length#

\ Defining menu entries takes too many parameters to pass them all on stack.
\ So we put them in a scratch menu cell.  Used for prefix notation.
\ It's also used to store location of user input.
CREATE (menu-scratch-cell)   menu-cell-length# allot

: menu-clear-prefixes ( -- )  (menu-scratch-cell) menu-cell-length# erase ;
menu-clear-prefixes

\ Store data in menu cell at 'addr' respecting prefixes.
: menu! ( xt addr -- )		\ uses and clears prefixes
    >r					( xt  r: addr )
    (menu-scratch-cell) tuck >menu-xt !
    r> menu-cell-length# move
    menu-clear-prefixes ;

\ menu@ reads menu cell at addr, puts xt on stack, sets the prefixes
: menu@ ( addr -- xt )
    >r						( r: addr )
    r@ (menu-scratch-cell) menu-cell-length# move
    r> >menu-xt	@ ;

0 value (actual-menu)	\ address of actual menu descriptor
			\ or zero if no menu is actual

: MDE: \ 'menu-descriptor-entry'
    CREATE ( offset-in-cells - offset+1 ) 
	dup cells , 1+
    DOES> ( -- addr )
	@ (actual-menu) + ;

\ each menu descriptor starts with the following structure

0 \ offset in cells
MDE: menu-display-xt	\ xt of display-routine which should also set the keys
MDE: clear-screen-xt	\ xt of clear-screen action
MDE: to-do-before-xt	\ xt of action to be taken in advance
MDE: to-do-after-xt	\ xt of action to be made after
MDE: to-do-after-2-xt	\ xt of second action to be made after
MDE: menu-exit		\ shall we exit menu after this action?
MDE: menu-redisplay	\ must the display be renewed after this action?
MDE: menu-selected	\ data of selected item is copied here
\ you can never know if an action selected by the user (i.e. a submenu ) did
\ not corrupt the menu data in the field.
\ So I save all data of the selected item here.
\
\ 'menu-selected' is also used during construction of the menu entries
\ as each menu-entry does 'menu-selected menu!' the datas of the
\ last menu entry remains there and can be reused

\ we need one menu cell length: add extra length:
menu-cell-length# cell / 1- +

\ after that there is a structure like this for the screen-menu:
\ MDE: screen-menu-items#	\ not really needed yet
0 value screen-menu-items#	\ so we have it like this for now
\
MDE: menu-2-at			\ double cell storage of cursor position
1+				\ second cell
MDE: menu-scroll-lines		\ defined items, switches scrolling too.
MDE: menu-scrolled		\ display offset in scrolling menus
MDE: screen-menu-defaults	\ (one menu-cell-length# length)
menu-cell-length# cell / 1- +		\ add extra length

\ if there's a key-menu, then there is a structure like this:

MDE: key-menu-defaults		\ (one menu-cell-length# length)
menu-cell-length# cell / 1- +		\ add extra length

cells CONSTANT menu-descriptor-length 

variable screen-menu-only	screen-menu-only off
variable key-menu-only \ very likely to make problems with nested menus! DADA##
         key-menu-only off

variable screen-menu-array	screen-menu-array off	\ array base pointer

: screen-menu-free  ( -- ) 
    screen-menu-array @ ?dup IF 
	free  IF
	    bell  cr ." screen-menu-free: Couldn't free screen menu "	800 ms
	THEN
	screen-menu-array off 
    THEN ;

variable key-menu-array		key-menu-array off	\ array base pointer

: key-menu-free  ( -- )
    key-menu-array @ ?dup IF
	free  IF bell cr ." key-menu-free: Couldn't free key menu " 800 ms THEN
	key-menu-array off 
    THEN ;

: free-menus ( -- )   screen-menu-free  key-menu-free ;

: ?menus-clear
    screen-menu-array @ ?dup IF
	screen-menu-items# menu-cell-length# * erase
    THEN
    key-menu-array @ ?dup IF
	allowed-key-codes# menu-cell-length# * erase
    THEN ;

: ?menus-allocate
    screen-menu-array @ 0= IF	\ screen array not yet allocated?
	key-menu-only @ 0= IF	\ but needed
	    c-l l-s *		\ for each character on the screen
	    dup menu-cell-length# *		\ is a menu item needed
	    allocate   abort" no memory for menus "	\ allocate
	    swap to screen-menu-items#
	    screen-menu-array !
	THEN
    THEN
    key-menu-array @ 0= IF	\ key array not yet allocated?
	screen-menu-only @ 0= IF	\ but needed
	    allowed-key-codes#  menu-cell-length# *
	    allocate   abort"  no memory for menus "	\ allocate
	    key-menu-array !
	THEN
    THEN 
    ?menus-clear ;

: xy># ( -- i# ) at? c-l * + ;
: #>addr ( i#-- addr ) menu-cell-length# *   screen-menu-array @  + ;

2variable	(2last-entry-range)			\ range of last entry

\ sometimes we want the active range start earlier on screen as the entry:
variable	(from-here)	-1 (from-here) !	\ start point
: from-here ( -- )   xy># (from-here) ! ;		\ use this to set start

\ sometimes we want the active range expand after the menu entry:
: up-to-here ( -- )
    (2last-entry-range) cell+ @ xy># swap ?DO
	menu-selected  i #>addr  menu-cell-length# move
    LOOP ;

: start-xy># ( -- i# )
    (from-here) @ -1 = IF
	xy>#
    ELSE
	(from-here) @
	-1 (from-here) !
    THEN ;

: menu-entry ( addr count xt -- )
    menu-selected menu!
    start-xy># >r   type   xy>#   r>
    2dup (2last-entry-range) 2! ?DO
	menu-selected  i #>addr  menu-cell-length#  move
    LOOP ;

: menu-entry-value ( addr count n xt -- )
    swap >r		( addr count xt	R: n )
    menu-selected menu!	( addr count	R: n )
    r>			( addr count n )
    start-xy># >r >r  type r> .  xy>#   r>
    2dup (2last-entry-range) 2! ?DO
	menu-selected  i #>addr  menu-cell-length#  move
    LOOP ;

: name-menu-entry ( xt -- )	\ name of word as visible menu entry
    dup >r
    xt>name
    r> menu-entry ;

: menu-default ( xt -- )
    screen-menu-defaults menu! 
    screen-menu-items# 0 DO
	screen-menu-defaults   i #>addr  menu-cell-length#  move
    LOOP ;

: char>addr ( c -- addr )
    dup allowed-key-codes# >=  over 0<  OR
    ABORT" char>addr: Key out of range."

    menu-cell-length# *   key-menu-array @ + ;

: menu-key-default ( xt -- )
    key-menu-defaults menu!
    key-menu-defaults
    allowed-key-codes# 0 DO
	dup   i char>addr  menu-cell-length#  move 
    LOOP drop ;

: menu-key-entry ( addr count xt -- )
    menu-selected menu!
    bounds DO menu-selected  i c@ char>addr  menu-cell-length#  move LOOP ;

: menu-same-key-entry ( addr count -- )	\ same menu entries as entered before
					\ for the keys in the string
    menu-selected menu@ menu-key-entry ;

: #key-menu-entry ( key% xt -- )
    menu-selected menu!
    menu-selected swap char>addr menu-cell-length# move ;

: #key-same-entry ( key% -- )  menu-selected menu@  #key-menu-entry ;

: menus-default-restore		\ restores menu and menu-keys defaults
    screen-menu-defaults  menu@  menu-default
    key-menu-defaults     menu@  menu-key-default ;

\ Bit masks to switch a couple of actions executed when a menu is selected
0
MASK: |don't
MASK: |cls
MASK: |do-before
MASK: |do-after
MASK: |do-after-2
MASK: |ping
MASK: |redisplay
MASK: |menu-wait
MASK: |wait-n'go
MASK: |put-cursor
MASK: |menu-done
MASK: |>stack
MASK: |>stack-2
drop


: m-prefix 
    CREATE ( mask -- ) ,
    DOES>  ( -- )
	(menu-scratch-cell) >menu-flags >r
	@  r@ @  or r> ! ;

\ Prefixes to set menu-flags, (they are off by default).
\ Use these prior to a menu entry.

|don't		m-prefix don't
|>stack		m-prefix >stack)	\ see below
|>stack-2	m-prefix >stack-2)	\ see below
|cls		m-prefix cls
|do-before	m-prefix do-before
|do-after	m-prefix do-after
|do-after-2	m-prefix do-after-2
|ping		m-prefix ping
|redisplay	m-prefix redisplay
|menu-wait	m-prefix menu-wait
|wait-n'go	m-prefix wait-n'go	\ waits for key? doesn't destroy input
|put-cursor	m-prefix put-cursor
|menu-done	m-prefix menu-done

\ Use these two for stack entries (prior to a menu entry).
: >stack   ( parameter-1 -- )
    (menu-scratch-cell) >menu-parameter-1 !   >stack) ;
: >stack-2 ( parameter-2 -- )
    (menu-scratch-cell) >menu-parameter-2 !   >stack-2) ;

\ cursor control, where to put the cursor when user input will be required
: >xy ( -- )	\ save current cursor position
    put-cursor	\ it will be remembered in the *coming* menu entry
    (menu-scratch-cell) >menu-input-at >r
    at?   r@ 1+ c!   r> c! ;

: >last-xy	\ current cursor position is remembered in the entries of
		\ the *preceeding* menu-entry
    at?
    menu-selected >menu-input-at 1+ c! 
    menu-selected >menu-input-at    c!
    menu-selected >menu-flags dup @ |put-cursor or swap ! 
    (2last-entry-range) 2@ DO
	menu-selected  i #>addr  menu-cell-length#  move
    LOOP ;

\ menu stack to make menus nestable
hex 12 constant menus#		\ how many menus can be nested
\ cr .( ==> allows menu nestings through ) menus# . .( levels) cr

variable menus-on-stack       \ stackpointer
       0 menus-on-stack !

CREATE menu-stack
   menus# cells allot		\ for the stack, menus and menu-2-at

: push-menu ( -- )            \ pushes menu on the stack
    menus-on-stack @ menus# < 0=
    ABORT" push-menu: menu stack full, increase 'menus#' in 'menu.fs'"

    (actual-menu) ?dup IF	\ noop if there was no menu
	menu-stack menus-on-stack @ cells + !
	1 menus-on-stack +!
	0 to (actual-menu)	\ (actual-menu) is not initialized any more
    THEN ;

: pop-menu ( -- )             \ goes back to previos menu
    menus-on-stack @
    dup IF
	1-
	dup menus-on-stack !
	cells menu-stack + @ to (actual-menu)	\ set menu descriptor  
	(actual-menu) IF ?menus-allocate THEN
	?menus-clear
	menus-default-restore
	menu-redisplay on			\ to have the screen displayed
    ELSE
	( 0 )  to (actual-menu)			\ hmmm?
    THEN ;

: empty-actions
    menu-clear-prefixes			\ just in case
    ['] noop to-do-before-xt !		\ more for aesthetics
    ['] noop to-do-after-xt  !		\ ditto
    ['] noop to-do-after-2-xt  !	\ ditto
    ['] page clear-screen-xt !		\ no crashes please  
    ['] noop menu-display-xt !		\ ditto
    -1 -1 menu-2-at 2! ;		\ -1 -1 means no cursor position saved

: noopmenus
    empty-actions
    ['] noop menu-default
    ['] noop menu-key-default ;

\ MENU:  creates a named menu descriptor.
\ the descriptor name also manages allocation of the arrays
: MENU: ( -- )
    CREATE
	menu-descriptor-length allot
    DOES> ( -- )
	push-menu		\ pushes the old menu
	to (actual-menu)	\ sets (actual-menu)
	?menus-allocate		\ do allocation if needed
	empty-actions ;		\ don't crash please...

\ If there are troubles moving the cursor with the arrow key's
\ try setting the following variable. decimal 120 cursor-code-escape-wait !
variable cursor-code-esc-wait		decimal 18 cursor-code-esc-wait ! hex
variable cur-esc-wait-calibrate		cur-esc-wait-calibrate off


\ function keys
[UNDEFINED] LIST: [IF]		include lists.fs 	[THEN]
LIST: function-key-actions
' noop function-key-actions >list

variable F1-xt		' noop F1-xt !
variable F2-xt		' noop F2-xt !
variable F3-xt		' noop F3-xt !
variable F4-xt		' noop F4-xt !
variable F5-xt		' noop F5-xt !
variable F6-xt		' noop F6-xt !
variable F7-xt		' noop F7-xt !
variable F8-xt		' noop F8-xt !
variable F9-xt		' noop F9-xt !
variable F10-xt		' noop F10-xt !
variable F11-xt		' noop F11-xt !
variable F12-xt		' noop F12-xt !
variable shift-F1-xt	' noop shift-F1-xt !
variable shift-F2-xt	' noop shift-F2-xt !
variable shift-F3-xt	' noop shift-F3-xt !
variable shift-F4-xt	' noop shift-F4-xt !
variable shift-F5-xt	' noop shift-F5-xt !
variable shift-F6-xt	' noop shift-F6-xt !
variable shift-F7-xt	' noop shift-F7-xt !
variable shift-F8-xt	' noop shift-F8-xt !
\  these do not exist on my system:
\  variable shift-F9-xt	' noop shift-F9-xt !
\  variable shift-F10-xt	' noop shift-F10-xt !
\  variable shift-F11-xt	' noop shift-F11-xt !
\  variable shift-F12-xt	' noop shift-F12-xt !

\ 'menu-preset-xt' is meant to set menu prefixes.
\ 'false' is allowed as ' noop default-function-keys gives a wrong impression.
: default-function-keys ( menu-preset-xt -- )
    >r

    F1%		F1-xt @		r@ ?EXECUTE  #key-menu-entry
    F2%		F2-xt @		r@ ?EXECUTE  #key-menu-entry
    F3%		F3-xt @		r@ ?EXECUTE  #key-menu-entry
    F4%		F4-xt @		r@ ?EXECUTE  #key-menu-entry
    F5%		F5-xt @		r@ ?EXECUTE  #key-menu-entry
    F6%		F6-xt @		r@ ?EXECUTE  #key-menu-entry
    F7%		F7-xt @		r@ ?EXECUTE  #key-menu-entry
    F8%		F8-xt @		r@ ?EXECUTE  #key-menu-entry
    F9%		F9-xt @		r@ ?EXECUTE  #key-menu-entry
    F10%	F10-xt @	r@ ?EXECUTE  #key-menu-entry
    F11%	F11-xt @	r@ ?EXECUTE  #key-menu-entry
    F12%	F12-xt @	r@ ?EXECUTE  #key-menu-entry
    shift-F1%	shift-F1-xt @	r@ ?EXECUTE  #key-menu-entry
    shift-F2%	shift-F2-xt @	r@ ?EXECUTE  #key-menu-entry
    shift-F3%	shift-F3-xt @	r@ ?EXECUTE  #key-menu-entry
    shift-F4%	shift-F4-xt @	r@ ?EXECUTE  #key-menu-entry
    shift-F5%	shift-F5-xt @	r@ ?EXECUTE  #key-menu-entry
    shift-F6%	shift-F6-xt @	r@ ?EXECUTE  #key-menu-entry
    shift-F7%	shift-F7-xt @	r@ ?EXECUTE  #key-menu-entry
    shift-F8%	shift-F8-xt @	r@ ?EXECUTE  #key-menu-entry
    rdrop ;

mouse-supported [IF]
: release-mousek ( -- )
    BEGIN get-mouse >r 2drop r> 0= IF EXIT THEN bell AGAIN ;
[ELSE]
    \ define some mouse words as noops:
    : mousekey 0 ;
    : release-mousek ;	IMMEDIATE
    : showmouse	 ;	IMMEDIATE
    : hidemouse	 ;	IMMEDIATE
[THEN]

\ Clear all pending keys and force the user to release keys.
: release-key ( -- )
    BEGIN
	BEGIN is-key? WHILE get-key drop REPEAT
	[ decimal ] 10 ms		\ estimated value
	is-key?
    WHILE
	get-key drop
    REPEAT

    BEGIN
	[ decimal ] 70 ms		\ estimated value
	is-key?
    WHILE bell			\ ring the bell if there's still a key pressed
	get-key drop
	BEGIN is-key? WHILE get-key drop bell REPEAT
    REPEAT ;
hex

: release-keys ( -- )
    release-key release-mousek
    release-key release-mousek ;

\ Wait for key or mousekey and drop it.
: wait ( -- )
    BEGIN
	mousekey IF 2drop    EXIT THEN
	is-key?  IF get-key drop EXIT THEN
    AGAIN ;

[UNDEFINED] mousek! [IF]	include mouse.fs	[THEN]

\ In some cases the key was pressed *before* entering the menu loop
\ so we need a switch to shut off key canceling:
\       (could be put in 'release-key' instead)
\ Also used for cursor arrow keys.
variable do-not-cancel-keys	do-not-cancel-keys off

\ Words to actually move the cursor, keeping menus happy. Beware of key repeat.
: menu-cursor-move:
    CREATE ( xt-of-display-cursor-move -- )
	,
    DOES> ( -- )
	do-not-cancel-keys on	\ beware of key repeat
	@ execute		\ acually move the cursor on the display
	-1 -1 menu-2-at 2!	\ sometimes cursor would be set back again
    ;

' cursor-up	menu-cursor-move: |cursor-up|    
' cursor-down	menu-cursor-move: |cursor-down| 
' cursor-left	menu-cursor-move: |cursor-left| 
' cursor-right	menu-cursor-move: |cursor-right|

variable (total-scroll-range)
: scroll-sanity ( -- )
    menu-scrolled dup @ 0 max swap !	\ stop at zero
    menu-scrolled dup @			\ end not beyond range:
    menu-scroll-lines @ (total-scroll-range) @ swap - min swap !
    menu-scroll-lines @ (total-scroll-range) @ >= IF	\ whole range visible
	menu-scrolled off
    THEN ;

: cursor-up-?scroll ( -- )
    menu-scroll-lines @ 0= IF |cursor-up| EXIT THEN	\ menu does not scroll
    at? nip IF
	|cursor-up|
    ELSE
	-1 menu-scrolled +!
	scroll-sanity
	menu-redisplay on
	at? menu-2-at 2!
    THEN ;

: cursor-down-?scroll ( -- )
    menu-scroll-lines @ 0= IF |cursor-down| EXIT THEN	\ menu does not scroll
    at? nip last-line < IF
	|cursor-down|
    ELSE
	1 menu-scrolled +!
	scroll-sanity
	menu-redisplay on
	at? menu-2-at 2!
    THEN ;

: scroll-page-up ( -- )
    menu-scroll-lines @
    dup 0= IF drop EXIT THEN		\ menu does not scroll
    negate menu-scrolled +!
    scroll-sanity
    menu-redisplay on
    at? menu-2-at 2! ;

: scroll-page-down ( -- )
    menu-scroll-lines @
    dup 0= IF drop EXIT THEN		\ menu does not scroll
    menu-scrolled +!
    scroll-sanity
    menu-redisplay on
    at? menu-2-at 2! ;

: scroll-home ( -- )
    menu-scroll-lines @ 0= IF EXIT THEN	\ menu does not scroll
    menu-scrolled off
    menu-redisplay on
    at? menu-2-at 2! ;

: scroll-end ( -- )
    menu-scroll-lines @
    dup 0= IF drop EXIT THEN		\ menu does not scroll

    (total-scroll-range) @ swap -  menu-scrolled !
    scroll-sanity
    menu-redisplay on
    at? menu-2-at 2! ;

\ make key menu entries for cursor movement words.
\ the scrolling version does not bite when not scrolling, so it's used always.
: cursor-entries ( -- )
    up%		['] cursor-up-?scroll	#key-menu-entry
    down%	['] cursor-down-?scroll	#key-menu-entry
    left%	['] |cursor-left|	#key-menu-entry
    right%	['] |cursor-right|	#key-menu-entry
    page-up%	['] scroll-page-up	#key-menu-entry
    page-down%	['] scroll-page-down	#key-menu-entry
    home%	['] scroll-home		#key-menu-entry
    end%	['] scroll-end		#key-menu-entry ;

\ get normalized scroll offset
: scrolled ( range -- +n )
    >r		( r: range )
    r@ 0> IF			\ range defined?
	menu-scrolled @  r@ mod		\ within range
	dup 0< IF r@ + THEN		\ positive
    ELSE
	false
    THEN
    rdrop ;

\ loop parameter mapping from possible range and displayed lines
\ to loop parameters for display.

: scrolled-range ( upper lower -- upper' lower' )
    2dup - (total-scroll-range) !	\ range for scroll-end
    2dup -				( upper lower range	 )
    scrolled				( upper lower scrolled	 )
    +					( upper lower'		 )
    swap >r				( lower'   r: upper )
    dup menu-scroll-lines @ + r> min	( lower' upper' )
    swap ;				( upper' lower' )

\ cyclic-range would be nice to have too ############


\ Try to get escape sequences not captured by ekey:
\ Returns a (mapped) key% or zero.  Zero means 'what' waits for other input.
true [IF] \ not echoing escape sequences. new version for scrolling.

: .unknown-escape-sequence ( last-key addr count -- FALSE )
    bell cr ." escape-sequence: Unknown escape sequence: esc "
    type	\ key string of preceeding characters, if any.
    emit	\ last key
    BEGIN
	cursor-code-esc-wait @ await-key?
    WHILE
	get-key emit
    REPEAT
    1000 ms
    FALSE ;

\ Map escape sequences to key menu % codes.	New version for scrolling.
\ No echoing of escape sequences.
\ Displays unrecognized escape sequences.
\
\ sorry for the horrible design...
: escape-sequence ( -- key-code|0 )
    \ first $1B already received ?
    is-key? IF get-key
	[char] [ over = IF drop
	    get-key CASE \ key
		[char] A OF up%    ENDOF	\ cursor up
		[char] B OF down%  ENDOF	\ cursor down
		[char] C OF right% ENDOF	\ cursor right
		[char] D OF left%  ENDOF	\ cursor left
		[char] [ OF			\ F1,F2,F3,F4,F5
		    get-key CASE \ next key code
			[char] A OF F1% ENDOF
			[char] B OF F2% ENDOF
			[char] C OF F3% ENDOF
			[char] D OF F4% ENDOF
			[char] E OF F5% ENDOF
			s" [[" .unknown-escape-sequence
			FALSE
		    ENDCASE
		ENDOF
		[char] 1 OF			\ F6,F7,F8
		    get-key CASE \ next key code
			[char] 7 OF
			    get-key CASE \ next key code
				[char] ~ OF F6% ENDOF
				s" [17" .unknown-escape-sequence
				FALSE
			    ENDCASE
			ENDOF
			[char] 8 OF
			    get-key CASE \ next key code
				[char] ~ OF F7% ENDOF
				s" [18" .unknown-escape-sequence
				FALSE
			    ENDCASE
			ENDOF
			[char] 9 OF
			    get-key CASE \ next key code
				[char] ~ OF F8% ENDOF
				s" [19" .unknown-escape-sequence
				FALSE
			    ENDCASE
			ENDOF
			[char] ~ OF home% ENDOF
			s" [1" .unknown-escape-sequence
			FALSE
		    ENDCASE
		ENDOF
		[char] 2 OF	\ F9,F10,F11,F12 <shift-F1> to <shift F4> 
		    get-key CASE \ next key code
			[char] 0 OF
			    get-key CASE \ next key code
				[char] ~ OF F9% ENDOF
				s" [20" .unknown-escape-sequence
				FALSE
			    ENDCASE
			ENDOF
			[char] 1 OF
			    get-key CASE \ next key code
				[char] ~ OF F10% ENDOF
				s" [21" .unknown-escape-sequence
				FALSE
			    ENDCASE
			ENDOF
			[char] 3 OF
			    get-key CASE \ next key code
				[char] ~ OF F11% ENDOF
				s" [23" .unknown-escape-sequence
				FALSE
			    ENDCASE
			ENDOF
			[char] 4 OF
			    get-key CASE \ next key code
				[char] ~ OF F12% ENDOF
				s" [24" .unknown-escape-sequence
				FALSE
			    ENDCASE
			ENDOF
			[char] 5 OF		\ <shift-F1>
			    get-key CASE \ next key code
				[char] ~ OF shift-F1% ENDOF
				s" [25" .unknown-escape-sequence
				FALSE
			    ENDCASE
			ENDOF
			[char] 6 OF		\ <shift-F1>
			    get-key CASE \ next key code
				[char] ~ OF shift-F2% ENDOF
				s" [26" .unknown-escape-sequence
				FALSE
			    ENDCASE
			ENDOF
			[char] 8 OF		\ <shift-F1>
			    get-key CASE \ next key code
				[char] ~ OF shift-F3% ENDOF
				s" [28" .unknown-escape-sequence
				FALSE
			    ENDCASE
			ENDOF
			[char] 9 OF		\ <shift-F1>
			    get-key CASE \ next key code
				[char] ~ OF shift-F4% ENDOF
				s" [29" .unknown-escape-sequence
				FALSE
			    ENDCASE
			ENDOF
			s" [2" .unknown-escape-sequence
			FALSE
		    ENDCASE
		ENDOF
		[char] 3 OF		\ <shift-F5> to <shift F8> 
		    get-key CASE \ next key code
			[char] 1 OF
			    get-key CASE \ next key code
				[char] ~ OF shift-F5% ENDOF
				s" [31" .unknown-escape-sequence
				FALSE
			    ENDCASE
			ENDOF
			[char] 2 OF
			    get-key CASE \ next key code
				[char] ~ OF shift-F6% ENDOF
				s" [32" .unknown-escape-sequence
				FALSE
			    ENDCASE
			ENDOF
			[char] 3 OF
			    get-key CASE \ next key code
				[char] ~ OF shift-F7% ENDOF
				s" [33" .unknown-escape-sequence
				FALSE
			    ENDCASE
			ENDOF
			[char] 4 OF
			    get-key CASE \ next key code
				[char] ~ OF shift-F8% ENDOF
				s" [34" .unknown-escape-sequence
				FALSE
			    ENDCASE
			ENDOF
			s" [3" .unknown-escape-sequence
			FALSE
		    ENDCASE
		ENDOF
		[char] 4 OF
		    get-key CASE \ next key code
			[char] ~ OF end% ENDOF
			s" [4" .unknown-escape-sequence
			FALSE
		    ENDCASE
		ENDOF
		[char] 5 OF
		    get-key CASE \ next key code
			[char] ~ OF page-up% ENDOF
			s" [5" .unknown-escape-sequence
			FALSE
		    ENDCASE
		ENDOF
		[char] 6 OF
		    get-key CASE \ next key code
			[char] ~ OF page-down% ENDOF
			s" [6" .unknown-escape-sequence
			FALSE
		    ENDCASE
		ENDOF
		s" [" .unknown-escape-sequence
		FALSE
	    ENDCASE

\  	    BEGIN \ ???? ##########
\  		cursor-code-esc-wait @ await-key? \ maybe not necessary?
\  		( key? ) WHILE
\  		get-key emit
\  	    REPEAT
	ELSE \ not a '<esc> [' sequence
	    s" " .unknown-escape-sequence
\	    FALSE
	THEN
    ELSE \ escape only
	1B
    THEN

    cur-esc-wait-calibrate @ IF	\ to calibrate the wait time
	drop
	bell
	cursor-code-esc-wait @ 1+
	dup .
	cursor-code-esc-wait !
	false
    THEN
;

[ELSE] \ echoing escape sequences. No 'ekey' or such.  No scrolling done!
: escape-sequence ( -- key-code|0 )
    \ first $1B already received
    is-key? IF get-key
	dup [char] [ = IF 1B emit emit	\ output cursor codes
	    get-key dup CASE \ key
		[char] A OF emit 0 ENDOF	\ cursor up
		[char] B OF emit 0 ENDOF	\ cursor down
		[char] C OF emit 0 ENDOF	\ cursor right
		[char] D OF emit 0 ENDOF	\ cursor left
		[char] [ OF			\ F1,F2,F3,F4,F5
		    emit
		    get-key dup CASE \ next key code
			[char] A OF emit F1% ENDOF
			[char] B OF emit F2% ENDOF
			[char] C OF emit F3% ENDOF
			[char] D OF emit F4% ENDOF
			[char] E OF emit F5% ENDOF
			emit
		    ENDCASE
		ENDOF
		[char] 1 OF			\ F6,F7,F8
		    emit
		    get-key dup CASE \ next key code
			[char] 7 OF emit
			    get-key dup CASE \ next key code
				[char] ~ OF emit F6% ENDOF
				emit
			    ENDCASE
			ENDOF
			[char] 8 OF emit
			    get-key dup CASE \ next key code
				[char] ~ OF emit F7% ENDOF
				emit
			    ENDCASE
			ENDOF
			[char] 9 OF emit
			    get-key dup CASE \ next key code
				[char] ~ OF emit F8% ENDOF
				emit
			    ENDCASE
			ENDOF
			emit
		    ENDCASE
		ENDOF
		[char] 2 OF	\ F9,F10,F11,F12 <shift-F1> to <shift F4> 
		    emit
		    get-key dup CASE \ next key code
			[char] 0 OF emit
			    get-key dup CASE \ next key code
				[char] ~ OF emit F9% ENDOF
				emit
			    ENDCASE
			ENDOF
			[char] 1 OF emit
			    get-key dup CASE \ next key code
				[char] ~ OF emit F10% ENDOF
				emit
			    ENDCASE
			ENDOF
			[char] 3 OF emit
			    get-key dup CASE \ next key code
				[char] ~ OF emit F11% ENDOF
				emit
			    ENDCASE
			ENDOF
			[char] 4 OF emit
			    get-key dup CASE \ next key code
				[char] ~ OF emit F12% ENDOF
				emit
			    ENDCASE
			ENDOF
			[char] 5 OF emit	\ <shift-F1>
			    get-key dup CASE \ next key code
				[char] ~ OF emit shift-F1% ENDOF
				emit
			    ENDCASE
			ENDOF
			[char] 6 OF emit	\ <shift-F1>
			    get-key dup CASE \ next key code
				[char] ~ OF emit shift-F2% ENDOF
				emit
			    ENDCASE
			ENDOF
			[char] 8 OF emit	\ <shift-F1>
			    get-key dup CASE \ next key code
				[char] ~ OF emit shift-F3% ENDOF
				emit
			    ENDCASE
			ENDOF
			[char] 9 OF emit	\ <shift-F1>
			    get-key dup CASE \ next key code
				[char] ~ OF emit shift-F4% ENDOF
				emit
			    ENDCASE
			ENDOF
			emit
		    ENDCASE
		ENDOF
		[char] 3 OF		\ <shift-F5> to <shift F8> 
		    emit
		    get-key dup CASE \ next key code
			[char] 1 OF emit
			    get-key dup CASE \ next key code
				[char] ~ OF emit shift-F5% ENDOF
				emit
			    ENDCASE
			ENDOF
			[char] 2 OF emit
			    get-key dup CASE \ next key code
				[char] ~ OF emit shift-F6% ENDOF
				emit
			    ENDCASE
			ENDOF
			[char] 3 OF emit
			    get-key dup CASE \ next key code
				[char] ~ OF emit shift-F7% ENDOF
				emit
			    ENDCASE
			ENDOF
			[char] 4 OF emit
			    get-key dup CASE \ next key code
				[char] ~ OF emit shift-F8% ENDOF
				emit
			    ENDCASE
			ENDOF
			emit
		    ENDCASE
		ENDOF
		emit
	    ENDCASE

	    BEGIN
		cursor-code-esc-wait @ await-key? \ maybe not necessary?
		( key? ) WHILE
		get-key emit
	    REPEAT
	ELSE
	    bell cr ." unknown escape sequence: esc " emit
	    BEGIN
		cursor-code-esc-wait @ await-key?
	    WHILE
		get-key emit
	    REPEAT
	    1000 ms
	    FALSE
	THEN
    ELSE \ escape only
	1B
    THEN

    cur-esc-wait-calibrate @ IF	\ to calibrate the wait time
	drop
	bell
	cursor-code-esc-wait @ 1+
	dup .
	cursor-code-esc-wait !
	false
    THEN
;
[THEN]

true [IF] \ passing % codes for cursor movement, not calling cursor movement words

\ passing % codes for cursor movement, not calling cursor movement words here
: what ( -- x y mousekey | normalized-key )
    do-not-cancel-keys @ IF		\ see above
	release-mousek			\ only
    ELSE	
	release-keys			\ keyboard and mouse
    THEN

    BEGIN
	mousekey ?dup IF hidemouse EXIT THEN
	is-key? IF
	    get-key
	    dup CASE
		0  OF drop FALSE ENDOF	\ no keys < 4	(reserved for mouse)
		1  OF drop FALSE ENDOF	\ no keys < 4	(reserved for mouse)
		2  OF drop FALSE ENDOF	\ no keys < 4	(reserved for mouse)
		3  OF drop FALSE ENDOF	\ no keys < 4	(reserved for mouse)
		1B OF drop escape-sequence ?dup ENDOF

		dup	\ ordinary keys including cursor array codes
	    ENDCASE
	ELSE false
	THEN
    UNTIL

    dup D = IF drop		\ catch return and fake a mouse event
	at?
	2dup menu-2-at 2!
	SWAP  c-x *  swap  c-y *   1	\ fake x y mouse-key=1
	mousek! mousek@
    THEN

    do-not-cancel-keys off	\ here because of calibrate-cur-esc-wait
    hidemouse ;

[ELSE] \ doing cursor moving here.
: what ( -- x y mousekey | normalized-key )
    do-not-cancel-keys @ IF		\ see above
	release-mousek			\ only
    ELSE	
	release-keys			\ keyboard and mouse
    THEN

    BEGIN
	mousekey ?dup IF hidemouse EXIT THEN
	is-key? IF
	    get-key
	    dup CASE
		0 OF drop FALSE ENDOF	\ don't accept keys < 4
		1 OF drop FALSE ENDOF	\ don't accept keys < 4
		2 OF drop FALSE ENDOF	\ don't accept keys < 4
		3 OF drop FALSE ENDOF	\ don't accept keys < 4
	[ ekey-cursor-support @ ] [IF]
		up%	OF drop cursor-up	0 ENDOF
		down%	OF drop cursor-down	0 ENDOF
		left%	OF drop cursor-left	0 ENDOF
		right%	OF drop cursor-right	0 ENDOF
	[THEN]
		1B	OF drop escape-sequence ?dup ENDOF
		dup	\ ordinary keys
	    ENDCASE
	ELSE false
	THEN
    UNTIL

    dup D = IF drop		\ catch return and fake a mouse event
	at?
	2dup menu-2-at 2!
	SWAP  c-x *  swap  c-y *   1	\ fake x y mouse-key=1
	mousek! mousek@
    THEN


    do-not-cancel-keys off	\ here because of calibrate-cur-esc-wait
    hidemouse ;
[THEN]

: calibrate-cur-esc-wait
    cur-esc-wait-calibrate on	\ switch calibration on
    cursor-code-esc-wait off	\ start at zero
    do-not-cancel-keys on	\ calibration doesnt work else
    page ." Calibrating wait time for cursor movement escapes. "
    cr
    cr ." Press a cursor array until the beeping stops."
    cr ." Make sure it only beeps when you press or release a arrow key."
    cr ." Then press <return>."
    cr
    what
    clearstack
    cur-esc-wait-calibrate off ;

: xy-pix>c-cells ( px py -- cx cy )   >r c-x /   r> c-y / ;

: ?perform  ( addr -- )   @ dup IF execute ELSE drop THEN ;

: ?at-xy ( d -- )    dup -1 = IF 2drop ELSE at-xy THEN ;

hex
: do-menu ( -- )
    BEGIN	\ wait for a user input that needs a reaction
	cursor-entries
	what
	dup 0= IF
	    TRUE		\ TRUE, continue
	ELSE
	    dup 4 < IF
		drop  
		xy-pix>c-cells   c-l * +   #>addr
	    ELSE
		dup IF
		    char>addr
		ELSE
		    drop EXIT
		THEN 
	    THEN
	    dup menu@ menu-selected menu! \ save selected data
	    >menu-flags @       \ get flags and check
	    dup |don't and
	THEN
    WHILE
	drop	\ bad choice or noop 
    REPEAT	( flags )
    dup |cls		and IF clear-screen-xt ?perform THEN
    dup |do-before	and IF to-do-before-xt ?perform THEN
    dup |ping		and IF bell         THEN
    dup |redisplay	and menu-redisplay swap IF on ELSE off THEN
    dup |put-cursor	and IF		\ set cursor?
	menu-selected >menu-input-at dup c@ swap 1+ c@ at-xy
    THEN

    dup |>stack and IF
	>r menu-selected >menu-parameter-1 @ r>	\ parameter-1 on stack
    THEN
    |>stack-2 and IF
	menu-selected >menu-parameter-2 @		\ parameter-2 on stack
    THEN

    menu-selected >menu-xt ?perform

    menu-selected >menu-flags @			\ get flags again
    dup	|menu-wait and IF wait THEN			\ now *before* do-after
    dup |wait-n'go and IF BEGIN is-key? UNTIL THEN	\ key still available
    dup |do-after and IF to-do-after-xt ?perform THEN
    dup |do-after-2 and IF to-do-after-2-xt ?perform THEN
    |menu-done and IF menu-exit on ELSE menu-exit off THEN	\ set it here
    menu-2-at 2@ ?at-xy ;

: do-menu-loop ( -- )
    menu-redisplay on
    cursor-visible
    BEGIN
	menu-redisplay @ IF 
	    menu-redisplay off  
	    ?menus-allocate        \ a submenu might have free'd it...
	    menus-default-restore
	    clear-screen-xt ?perform
	    menu-display-xt ?perform
	    menu-2-at 2@ ?at-xy
	THEN
	showmouse do-menu 
    menu-exit @ UNTIL
    pop-menu ;

: (num-in) ( default -- n true | false ) \ allows using FORTH
    depth >r >r				\ remembering depth and default
    0 0 0 0 0 0	r>			\ buffering with zeros, default on top
    depth >r				\ remembering paded depth
    [ decimal ]
    pad dup 80  accept  evaluate	\ evaluating FORTH
    [ hex ]
    depth r> < IF 1 THROW THEN		\ less stack items? error
    r> swap >r >r			\ swap value and depth on return stack
    BEGIN r@ depth < WHILE drop REPEAT rdrop	\ restoring stack
    r> ;				\ result

: num-in ( default -- n true | false )
    ['] (num-in) CATCH IF bell drop false ELSE true THEN ;

: num-in-to-addr ( addr -- )   dup @ num-in IF swap ! ELSE drop THEN ;

: (2num-in) ( default-d -- d true | false ) \ allows using FORTH
    depth >r 2>r			\ remembering depth and default
    0 0 0 0 0 0	2r>			\ buffering with zeros, default on top
    depth >r				\ remembering paded depth
    [ decimal ]
    pad dup 80  accept  evaluate	\ evaluating FORTH
    [ hex ]
    depth r> < IF 1 THROW THEN		\ less stack items? error
    r> -rot 2>r >r			\ swap value and depth on return stack
    BEGIN r@ depth 1+ < WHILE drop REPEAT rdrop	\ restoring stack
    2r> ;				\ result

: 2num-in ( default-d -- d true | false )
    ['] (2num-in) CATCH IF bell 2drop false ELSE true THEN ;

: 2num-in-to-addr ( addr -- )   dup 2@ 2num-in IF rot 2! ELSE drop THEN ;

false [if]
: change-value-at-addr ( addr -- )
    at? 2>r
    last-left c-l spaces
    last-left ." change value from " dup @ .
    ." to new value: " num-in-to-addr
    last-left c-l spaces 2r> at-xy ;
[else]
: change-value-at-addr ( addr -- )
    ." new:       " .bs .bs .bs .bs .bs .bs
    num-in-to-addr ;
[then]

: change-2value-at-addr ( addr -- )
    ." new:       " .bs .bs .bs .bs .bs .bs
    2num-in-to-addr ;

: change-scale-at-addr ( addr -- )	\ checks for divider zero
    >r				( r: addr )
    r@ 2@			( initial-d-value  r: addr )
    r@ change-2value-at-addr
    r@ @ IF
	2drop		\ divider ok
    ELSE		\ zero,
\	r@ 2!		\ restore old value
\	bell ." divider zero." 600 ms
	2drop 0 1 r@ 2!	\ set ratio to 0/1  ;-)
	." ratio set to zero." 300 ms
    THEN rdrop ;

\ factoring out a common case: give a variable and a string
: simple-menu-entry-value ( addr u addr2 -- )
    dup >stack		\ addr2 of value >stack
    redisplay		\ will show the changed value
    @			\ value on top of string
    ['] change-value-at-addr menu-entry-value >last-xy ;

: char-to-addr ( addr -- )
    ( menu-2-at 2@ at-xy ) ." :"
    get-key [ hex ] FF and 
    dup D = IF   \ 'RETURN' is a noop
	2drop
    ELSE
	swap !
    THEN ;

: simple-menu-entry-char ( addr -- )	\ note that the char is stored as cell
    dup @ pad c!
    >stack  redisplay	pad 1	['] char-to-addr  menu-entry  >last-xy ;

\ words to deal with scales (double cells to be used with '*/' for scaling)
: .scale ( d -- )   swap . .bs ." /" . ;

: menu-entry-scale ( addr count d xt -- )
    -rot 2>r			( addr count xt	R: d )
    menu-selected menu!	( addr count	R: d )
    2r>				( addr count d )
    start-xy># >r 2>r  type 2r> .scale  xy>#   r>
    2dup (2last-entry-range) 2! DO
	menu-selected  i #>addr  menu-cell-length#  move
    LOOP ;

: simple-menu-entry-scale ( addr u addr2 -- )
    dup >stack		\ addr2 of value >stack
    redisplay		\ will show the changed value
    2@			\ double value on top of string
    ['] change-scale-at-addr menu-entry-scale >last-xy ;

: toggle-addr ( addr -- )  dup @ -1 xor swap ! ;

: xor! ( mask addr -- )  dup @ rot xor swap ! ;

dup [IF] \ test and usage example
decimal

: .something ." *SOMETHING*" ;				\ dummy action

\ display routine that displays the menu and sets actions and the like

MENU: (testmenu)
: .testmenu ( -- )
    cr s" Here's something hidden"	['] .something		menu-entry
    cr s" Here something hides"		['] .something
					menu-wait redisplay	menu-entry
    cr s" Something Ringing The Bell"   ring
    ['] .something	menu-entry
    cr s" Cute Bell"   	7 >stack	['] emit		menu-entry
    cr 			redisplay	['] words menu-wait	name-menu-entry
    cr					['] quit		name-menu-entry
    cr s" leave the menu" menu-done	['] noop		menu-entry cr
    F1%		['] .something		#key-menu-entry
    s" bB"		ring		['] noop		menu-key-entry
    s" sShH"				['] .something		menu-key-entry
    s" qQxX"		menu-done	['] noop		menu-key-entry
    s" 1"  1 >stack	menu-wait	['] .			menu-key-entry
    s" 2"		menu-wait	['] .s			menu-key-entry
;

: test-menu ( -- )
    (testmenu)
    ['] .testmenu menu-display-xt !
    do-menu-loop
; \    free-menus ;
\ test-menu

: adding-an-entry ( -- )
    .tab .tab .tab .tab s" something is here now"
    ['] .something menu-entry ;

variable testvariable	999 testvariable !
variable testvariable2	888 testvariable2 !
MENU: top-test-men
: .top-test-menu
    s" Here you go to a lower level"	['] test-menu  redisplay menu-entry cr
    s" lLdD" menu-same-key-entry
    s" Here you go and leave then"	['] test-menu  menu-done menu-entry cr
    s" mM" menu-same-key-entry
    s" Klick here to add something"	['] adding-an-entry	 menu-entry cr
    s" Want to change this number? " testvariable   simple-menu-entry-value cr
    from-here >xy testvariable2 @ .	redisplay
    s" or this one?" testvariable2 >stack
    ['] change-value-at-addr menu-entry cr
\    s" X" testvariable >stack redisplay
\    ['] change-value-at-addr menu-entry testvariable . up-to-here cr
    s" Crazy number " testvariable2 >stack redisplay
    ['] change-value-at-addr menu-entry testvariable2 @ 2/ 17 + . up-to-here cr
    s" Leave the menu"	menu-done	['] noop	menu-entry cr
    cls ['] bye name-menu-entry
    F1%		['] .something		#key-menu-entry
    [char] s	['] .something		#key-menu-entry
    s" qQxX"		menu-done	['] noop	menu-key-entry ;

: top-test-menu
    top-test-men
    ['] .top-test-menu menu-display-xt !
    do-menu-loop
    free-menus ;

top-test-menu
cr cr .( You left the menu now) cr

[THEN]
drop \ test switch

decimal

\ work in progress... ############
: show-key-bindings ( -- )
    key-menu-array @ dup IF
	key-menu-defaults >menu-xt @  swap	( default-xt key-menu-array )
	page
	." default: " over xt>name type 
	allowed-key-codes# bl ?DO
	    dup  menu-cell-length# i *  + >menu-xt @
	    2 pick over <> IF
		cr i emit bl emit xt>name .s key drop ( 2drop \ ) type
	    THEN
	LOOP
    THEN drop ;
