\ mutation-0.2.fs
\ 	$Id: mutation-0.2.fs,v 1.51 2001/02/15 16:42:01 f Exp $	

\ simple mutation mechanisms.
\ third version to be used with the new symbol stack version
\ fourth version with nested 'IF ... THEN ... ELSE'es implemented
\ with gene-follow and frames.

\ I try to make these words reentrant.
\ Stringbuffers and symbol stacks are now identified by handles.

\ Works on a string basis, not directly on the xt's.
\ It's a bit complicated, but easy to see what's happening...
\ Thats good for a first experimental implementation, I think.
\ it's not as slow as I was afraid of at least.

\ My main reason for choosing the strings to work on
\ was that I want to keep mutated genes as strings first.
\ I call this trial phase.
\ If they survive a minimum number of generations,
\ they can be compiled into the genepool later on.

\ The strings are handy to write to log files or code files.

decimal

\ I need a good tool to debug mutation ;-)
\ I make it optional in the sense of a compile time option.
\	Often it just redifines the interesting words adding logging.
\	It hooks into logging by kidnapping the '0 log' channel.
\ *If* it's compiled in, it's runtime switchable like logging.
[UNDEFINED] log-mask [IF]  	\ if not defined in compile-options.fs
    VARIABLE log-mask		\ compile time switch
    log-mask OFF
[THEN]


\ variables to store the handles of buffers for mutation to work on:
decimal
2048  STRINGBUF-HANDLE: (head)		\ the unchanged head of the genom
2048  STRINGBUF-HANDLE: (tail)		\ the unchanged tail
2048  STRINGBUF-HANDLE: (segment)	\ part that will be changed
4096  STRINGBUF-HANDLE: (new-gene)	\ spliced head, new-segment and tail

\ this word cuts a string after the n'th space into head and tail
\ *no checks at all!*
: cut-bl-n ( addr u n -- addr-tail u-tail addr u-head )
    >r swap 0 r>	( u addr 0 n )		\ zero is the initial offset
    0 ?DO		( u addr offset )	\ how many words to skip 
	BEGIN
	    2dup + c@ bl <> WHILE	\ scan for space
	    1+				\ increase offset
	REPEAT
	1+				\ skip space
    LOOP
    2dup 2>r				\ head 2>r
    dup >r				\ head length >r
    +					\ start of tail
    swap r> -				\ length of tail, now tail is on stack
    2r> ;				\ head on top of stack

\ cut a gene in head, segment and tail.
\ initialize (head) (segment) (tail) and put the FORTH code there
: xt>cutten-strings ( internal-xt head-items segment-items -- )
    >r >r
    internal-xt>string			\ disassembles a gene to a string
    r> cut-bl-n						\ cut head
    (head) dup stringbuf-empty   cat			\ buffer head
    r> cut-bl-n						\ cut segment
    (segment) dup stringbuf-empty cat			\ buffer segment
    (tail) dup stringbuf-empty   cat			\ buffer tail

    log-mask @ log-m-some and IF
	s" gene cutten in (3 lines: head, segment, tail)"	0 log
	(head)    		buffer-as-string		0 log
	(segment) 		buffer-as-string		0 log
	(tail)    		buffer-as-string		0 log
    THEN ;


\ stack control:
\ I see two good ways to control the stack effect of mutated genes:

\ No control at all at mutation time, combined with a more or less strict
\ control at run time. Stack violations could be made mortal, or be tolerated
\ to a certain degree.

\ Control at mutation time. This allowes some type checking too.
\ I don't like type checking too much, but I want my mutations to be save.
\ I wouldn't like my cells to be able to try changing random addresses
\ or the like.  So I want to treat at least addresses and other data different
\ Some more distinctions between data types might accelerate evolution,
\ like having a special type for flags or directions.
\ In these cases type checking could be made tolerant in an adjustable degree.

\ I like to use the stack for stack type checking:
\ Genes can be executed in a special way, that they work on symbols on the
\ stack in analogy they would work normally.  So a gene taking two numbers
\ from the stack and leaving a flag there, would check and swallow two
\ number symbols ( which are the ASCII code of the char 'n') and leave a
\ flag symbol there.

[UNDEFINED] max-stack-effect [IF]
    &64 constant max-stack-effect	\ no genom should ever go beyond that
[THEN]

\ to use instack" and outstack" it's convenient to have a switch
\ between skipping the data or compiling them
\ variable (compiling-gene-internals)	(compiling-gene-internals) off
variable (stack-entries)	(stack-entries) off

\ always call instack" *before* outstack"
: instack"
    [char] " parse
    (compiling-gene-internals) @ IF	\ compiling?
	-1 (stack-entries) +!		\ control recursive compilation level
	(stack-entries) @ 0= IF		\ on top level?
	    string>stack-in EXIT	\ done
	THEN
    THEN
    2drop ; IMMEDIATE	\ just skip unless compiling top level

: outstack"
    [char] " parse
    (compiling-gene-internals) @  (stack-entries) @ 0=  AND IF
	string>stack-out
    ELSE  2drop  THEN ; IMMEDIATE	\ just skip unless compiling top level

: start" ( -- )				\ inserted by the mutation process
    [char] " parse
    (compiling-gene-internals) @ IF
	1 (stack-entries) +!		\ recursive compilation, increase level
    THEN  2drop ; IMMEDIATE

: end" ( -- )   [char] " parse  2drop  ; IMMEDIATE

: insert-start-mark ( addr count handle -- )
    >r

    \ do we need string twice?
    write-code-file  code-file-mask @  AND IF
	file-mutating  code-file-mask @  AND IF
	    file-start&stop-entries code-file-mask @  AND IF
		[ decimal ] 64 stringbuf-open >r
	    THEN
	THEN
    THEN

    s" start"	r@ cat
    [char] "	r@ char-cat
    bl		r@ char-cat
		r@ cat
    [char] "	r@ char-cat
    bl		r@ char-cat

    write-code-file  code-file-mask @  AND IF
	file-mutating  code-file-mask @  AND IF
	    file-start&stop-entries code-file-mask @  AND IF
		indent-code
		r> dup buffer-as-string
		2dup cat>code-file code-file-write-line
		r@ cat
		stringbuf-close
	    THEN
	THEN
    THEN
    rdrop ;

: snipped ; IMMEDIATE	\ just a marker for my eye

\ the mutation process uses special words in the evaluated strings
\ that do things like storing stack informations during internals compilation
\ or give the structure of the sequence of words,
\ which can be used for subgene compilation and such.
\ as i use them in GENE-INTERNALS: it was deferred there:
\ DEFER <internal-compiling-extensions> ( xt -- xt flag )
\ the flag says if the xt should be compiled into internals.
\ ( might have been changed by the deferred word called)
:NONAME
    dup CASE
	['] start"	OF dup EXECUTE false ENDOF
	['] instack"	OF dup EXECUTE false ENDOF
	['] outstack"	OF dup EXECUTE false ENDOF
	['] end"	OF dup EXECUTE false ENDOF
	['] snipped	OF false ENDOF
	TRUE
    ENDCASE ; IS <internal-compiling-extensions>

\ gene-follow is called sucessively on each xt of a gene sequence
\ it can be used reading out genes or during the mutation process.
\ what it does:
\ - keep track of stack symbols, or abort if they do not fit.
\ - do depth-watch.
\ - follow IF-ELSE-THEN structures keeping track of nesting levels,
\   stack information and stack requirements of the two branches.
\ - follow any type of sequences, as long as they are contained in each other.
\ - let's add a frame type tag.  it is a one word string.
\	we could even execute the string during gene compilation... ;-)
32 CONSTANT frame-tag-max-length#

VARIABLE (follow-frames)	32 (follow-frames) !	\ default
0
OFFSET: >stack-aim	( addr-of-count )	\ first, because it's fastest
1+			( pointer )
max-stack-effect +	\ this reserves in fact four times as much, intended.

OFFSET: >initial-stack	( addr-of-count )
1+			( pointer )
max-stack-effect +	\ this reserves in fact four times as much, intended.

OFFSET: >frame-tag	( addr-of-count )
1+			( pointer )
frame-tag-max-length# cell / +		\ here the tag gets stored

OFFSET: >minimal-depth
OFFSET: >peak-level
OFFSET: >frame-cost
cells CONSTANT frame-size#
frame-size# cell mod [IF]
    cr cr .( Compile time error: 'frame-size# CELL MOD' must be zero! ) cr
    quit
[THEN]

frame-size# (follow-frames) @ * STRINGBUF-HANDLE: (follow-data)

VARIABLE frame#		frame# off


0 VALUE (frame)		\ gives startaddress of actual frame

: aim ( -- addr )		(frame) >stack-aim ;
: initial ( -- addr )		(frame) >initial-stack ;
: frame-tag ( -- addr )		(frame) >frame-tag ;
: minimal-depth ( -- addr)	(frame) >minimal-depth ;
: peak-level ( -- addr) 	(frame) >peak-level ;
: frame-cost ( -- addr) 	(frame) >frame-cost ;

DEFER <all-frames-init-pointers>
: set-frame-actual ( level -- )
    dup 0< ABORT" set-frame-actual: Level negative."

    dup frame# !

    (follow-data) handle>stringbuf-descriptor >stringbuf-pointer @
    frame# @ frame-size# *  +  to (frame)

    (follow-frames) @ < 0= IF			\ enough (follow-frames)?
	(follow-data) handle>stringbuf-descriptor stringbuf-double
	(follow-frames) dup @ 2* swap !		\ double if not enough

	(follow-data) handle>stringbuf-descriptor >stringbuf-pointer @
	frame# @ frame-size# *  +  to (frame)	\ (frame) must be reset
	<all-frames-init-pointers>		\ initialize stringpointers
    THEN
;

\ I use some stringpoiners of the form
\ addr			holds count	\ 2@ on addr gives ( string-addr count)
\ addr cell+		holds pointer
\ addr cell+ cell+	start of string
\ 'init-stringpointer'  initializes the pointer, but does *not* change count.
: init-stringpointer ( addr -- )
    cell+	\ address of stringpointer
    dup cell+	\ string starts just behind
    swap ! ;	\ initialize pointer
    
: frame-init-stringpointers ( -- )
    aim		init-stringpointer
    initial	init-stringpointer
    frame-tag	init-stringpointer ;

: all-frames-init-stringpointers ( -- )
    frame# @			\ to restore it when done
    (follow-frames) @ 0 DO
	i set-frame-actual  frame-init-stringpointers
    LOOP
    set-frame-actual  ;			\ restore frame
' all-frames-init-stringpointers IS <all-frames-init-pointers>
all-frames-init-stringpointers

: frame-store-string ( from-addr count to-address -- )
	>r
	dup r@ !	\ store length
	r> cell+ cell+	\ skip count and pointer
	swap move ;	\ move string into frame data

3 CONSTANT frame-indent#
: indent-frame ( -- )
    frame# @  frame-indent# *  0 ?DO s"  " cat-log LOOP ;

: log-frame-name&stack ( addr count -- )
    indent-frame
    ( addr count -- )	cat-log
    s"  "		cat-log
    frame# @ num>string	cat-log
    s" 	( "		cat-log
    symbols-as-string	cat-log
    s"  )"		0 log-it ;

: ?stack>code-file ( -- )
    file-stack  code-file-mask @  AND IF
	code-next-2-tab
	s" ( "			cat>code-file
	symbols-as-string	cat>code-file
	s" )"			cat>code-file
    THEN ;

VARIABLE (item-number)
: ?item-number>code-file ( -- )
    file-item-number  code-file-mask @  AND IF
	(item-number) dup @ 1+ swap !
	s"  	\ item number "		cat>code-file
	(item-number) @ num>string	cat>code-file
    THEN ;

: ?depth&cost>code-file ( -- )
    file-depth&cost code-file-mask @ AND IF
	indent-code
	s" \ depth-min: "		cat>code-file
	minimal-depth @ num>string	cat>code-file
	s"  max: "			cat>code-file
	peak-level @ num>string		cat>code-file
	s"  cost: "			cat>code-file
	frame-cost @ num>string		cat>code-file
	code-file-write-line
    THEN ;

: frame-tag&stack&depth&cost>file ( addr count -- )
    indent-code
    s" \ "			cat>code-file
    ( addr count )		cat>code-file
    frame-tag 2@		cat>code-file
    s"  "			cat>code-file
    frame# @ num>string		cat>code-file
    ?stack>code-file
    code-file-write-line

    ?depth&cost>code-file ;

: push-frame ( tag-addr tag-count -- )	\ stack symbols must be there already
    frame# @  1+  set-frame-actual	\ new horizon

    [ log-mask @ ] [IF]	\ conditional compiling
	log-mask @ log-m-much and IF
	    s" PUSHing frame "	cat-log
	    2dup log-frame-name&stack

	    frame# @ ?dup IF
		s"     frames below: " cat-log
		0 BEGIN
		    2dup <> WHILE
		    dup set-frame-actual
		    frame-tag 2@ cat-log s"  " cat-log
		    1+
		REPEAT
		drop s"  " 0 log-it
		set-frame-actual
	    THEN

	THEN
    [THEN]

    frame-tag frame-store-string	\ store tag in frame data
    symbols-as-string initial frame-store-string \ store initial stack symbols
    #symbols-on-stack minimal-depth !	\ init depth and peak watch
    #symbols-on-stack peak-level !
    frame-cost off

    write-code-file  code-file-mask @  AND IF
	file-mutating  code-file-mask @  AND IF
	    file-frames code-file-mask @ AND IF
		s" push frame: " frame-tag&stack&depth&cost>file
	    THEN
	THEN
    THEN ;

VARIABLE (no-start&stop-entries)	(no-start&stop-entries) off
: push-frame-with-entry ( tag-addr tag-count handle -- )
    >r
    (no-start&stop-entries) @ IF
	(no-start&stop-entries) off
	2drop s" NO-ENTRIES" push-frame
	rdrop
    ELSE
	2dup push-frame
	r> insert-start-mark
    THEN ;

: pop-frame ( -- )	\ we could check the tag here, I don't for efficiency.
    minimal-depth @
    peak-level @
    frame-cost @
    frame# @  1-  set-frame-actual
    frame-cost +!				\ add cost
    #symbols-on-stack max			\ is this needed? #######
    peak-level >r r@ @ max r> !
    #symbols-on-stack min			\ is this needed? #######
    minimal-depth >r r@ min r> !

    [ log-mask @ ] [IF]	\ conditional compiling
	log-mask @ log-m-much and IF
	    s" POPping frame "		cat-log
	    frame-tag 2@		log-frame-name&stack

	    s"               "		cat-log
	    s" new depth-min: "		cat-log
	    minimal-depth @ num>string	cat-log
	    s"  max:"			cat-log
	    peak-level @ num>string	0 log-it
	THEN
    [THEN]

    write-code-file  code-file-mask @  AND IF
	file-mutating  code-file-mask @  AND IF
	    file-frames code-file-mask @ AND IF
		s" pop frame:  " frame-tag&stack&depth&cost>file
	    THEN
	THEN
    THEN ;

\ store frame#, all frame data and actual stack symbols:
frame-size# cell / OFFSET: >preserved-stack drop
frame-size# cell+ cell+ max-stack-effect + cell / OFFSET: >stored-frame# drop
\ ==> please don't forget to 'free' it later!
: store-actual-stack-and-frame ( -- data-address ) \ 'FREE' it yourself.
    [ frame-size#				\ space for frame data
    cell+ cell+ max-stack-effect +		\ count addr stack-symbol strng
    cell+ ] literal				\ frame#
    allocate	( addr ior )			\ 'free' it later, please.
    ABORT" frame-preserve-entry: Couldn't allocate."
    (frame) over frame-size# move

    \ initialized stringpointers let us read strings from here, if we need that
    dup >stack-aim	init-stringpointer
    dup >initial-stack	init-stringpointer
    dup >frame-tag	init-stringpointer

    \ store actual stack
    symbols-as-string >r	( data-address addr-symbols  r: count-symbols )
    over >preserved-stack r@ swap !		\ store symbols count
    over >preserved-stack init-stringpointer	\ init symbols pointer
    over >preserved-stack cell+ cell+ r> move	\ store actual stack symbols

    frame# @  over >stored-frame# ! ;		\ store frame#

\ restore frame#, all frame data and stack symbols
: restore-stack-and-frame ( data-pointer -- )
    dup >stored-frame# @ set-frame-actual	\ reset old frame# actual

    dup (frame) frame-size# move		\ reset frame data
    frame-init-stringpointers			\ init stringpointers

    clear-symbols
    >preserved-stack 2@ push-symbol-string	\ restore stack symbols
;

: depth-min-watch ( count -- )   minimal-depth >r r@ @ min r> ! ;

: depth-max-watch ( count -- )   peak-level    >r r@ @ max r> ! ;

\  : depth-watch ( count -- )
\      dup minimal-depth >r r@ @ min r> !
\          peak-level    >r r@ @ max r> ! ;

: code-entry ( internals-xt -- )
    write-code-file  code-file-mask @  AND 0= IF drop EXIT THEN	 \ not filing

    file-mutating file-end-trial OR code-file-mask @ AND   \ time for an entry?
    0= IF drop EXIT THEN				   \ else nothing to do

    run-mode @ writing-code and 0<>		\ making code entries?
						\ or
    run-mode @ mutating and 0<>			\ are we mutating?
    code-file-mask @ file-mutating and 0<> AND	\ and want entries then?
    OR 0= IF drop EXIT THEN			\ no?  nothing to do

    file-code  code-file-mask @  AND		\ we want code entries?

    dup 0= IF					\ no? maybe structures?
	file-structure code-file-mask @  AND IF	\ yes
	    over CASE
		['] [IF]   OF drop true ENDOF
		['] [ELSE] OF drop true ENDOF
		['] [THEN] OF drop true ENDOF
		['] ;gene  OF drop true ENDOF
	    ENDCASE
	THEN
    THEN

    IF
	indent-code
	dup CASE
	    ['] [IF]   OF drop s" IF"   ENDOF
	    ['] [ELSE] OF drop s" ELSE" ENDOF
	    ['] [THEN] OF drop s" THEN" ENDOF
	    >r xt>name r>
	ENDCASE
	cat>code-file
	?stack>code-file
	writing-code run-mode @ AND 0= IF	\ not for trial entries
	    ?item-number>code-file
	THEN
	code-file-write-line

	writing-code run-mode @ AND 0= IF	\ not for trial entries
	    ?depth&cost>code-file
	THEN

    ELSE drop THEN ;

: new-gene-name ( -- addr count )
    (name-buf) stringbuf-empty
    s" g-" (name-buf) cat
    genome-id @ num>string (name-buf) cat
    bl (name-buf) char-cat
    (name-buf) buffer-as-string ;

false [if] \ not used any more
: next-gene-name-will-be ( -- addr count )
    genome-id dup @ >r r@ 1+ swap !
    new-gene-name
    r> genome-id ! ;
[then]

: step&spot-name ( -- addr count )
    (name-buf) >r
    r@ stringbuf-empty
    s" mutation."	r@ cat
    step @ num>string	r@ cat
    [char] :		r@ char-cat
    spot @ num>string	r@ cat
    bl			r@ char-cat
    r> buffer-as-string ;

: start-to-follow ( -- )
    (no-start&stop-entries) off		\ just in case...

    write-code-file  code-file-mask @  AND IF
	file-mutating  code-file-mask @  AND IF
	    file-code  file-structure OR code-file-mask @  AND IF
		s" \ mutated genome, set on trial:" cat>code-file
		code-file-write-line
	    THEN
	THEN
    THEN

    -1 frame# !
    s" BOTTOM" push-frame

    (code-indent) off
    (item-number) off
    write-code-file  code-file-mask @  AND IF
	file-mutating  code-file-mask @  AND IF
	    file-code  file-structure OR code-file-mask @  AND IF
		s" : "			cat>code-file
		step&spot-name		cat>code-file
		s"  ( -- ) "		cat>code-file

		file-step&spot&id  code-file-mask @  AND IF
		    s" \ spot:"		cat>code-file
		    spot 2@ num>string	cat>code-file
		    s"  step:"		cat>code-file
		    num>string		cat>code-file
		    s"  based on ID:"	cat>code-file
		    id @ num>string	cat>code-file
		THEN

		code-file-write-line
		1 (code-indent) +!
	    THEN
	THEN
    THEN ;

: gene-follow ( internal-xt -- )
    dup
    CASE
	['] [IF] OF
	    [char] C symbols-tos-match? 0=
	    ABORT" gene-follow: input stack symbols mismatch, flag expected."

	    s" CONDITION" push-frame
	    conditional-token-price @ frame-cost +!
	    drop-symbol  #symbols-on-stack depth-min-watch	\ drop 'C'

	    ( xt ) code-entry
	    1 (code-indent) +!
	ENDOF

	['] [ELSE] OF
	    conditional-token-price @ frame-cost +!

	    \ push-aim:		actual symbols as aim for else branche
	    symbols-as-string aim frame-store-string
	    \ aim s-buf-clear   symbols-as-string aim s-buf-cat \ same as above

	    \ dup initial	reset stack symbols to the state saved by 'IF'
	    clear-symbols  initial 2@ 1- push-symbol-string	\ skip 'C'

	    -1 (code-indent) +!
	    ( xt ) code-entry
	    1 (code-indent) +!

	ENDOF

	['] [THEN] OF
	    conditional-token-price @ frame-cost +!
	    -1 (code-indent) +!
	    ( xt ) code-entry

	    pop-frame
	    \ 'last-frame-insert-end-markings' called by 'build-if-else-then'
	ENDOF

	['] ;gene OF
	    (default-gene-cost#) frame-cost +!
	    -1 (code-indent) +!
	    ( xt ) code-entry
	ENDOF

	\ default: normal genes
	dup >body		( xt body )	\ for gene-stack-out

	\ get input symbols and make sure they match:
	dup dup gene-body>stack-in-string
	dup >r			( xt body body addr count -- r: in-count)
	symbols-match		( xt body r: in-count )
	0= IF
	    cr ." gene-follow: Input stack symbols mismatch in "
		  over xt>name type
	    cr ." on stack: " symbols-as-string type
	    cr ." expected: " gene-body>stack-in-string type
	    true ABORT" gene-follow: Input stack symbols mismatch."
	THEN

	dup >gene-cost @  frame-cost +!

	\ cut input symbols away:
	symbols-as-string			\ addr count of stack symbols
	r> -					\ cut consumed symbols

	\ How deep did we ever go within this sequence?
	\ (we need to know to determine input requirements of the sequence).
	dup depth-min-watch

	\ store intermediate stack in stack symbol stringbuf
	clear-symbols push-symbol-string

	\ now add output symbols.		( xt body )
	gene-body>stack-out-string  push-symbol-string
	#symbols-on-stack depth-max-watch

	( xt ) code-entry

    ENDCASE ;

: gene-follow-mute ( internal-xt -- )
    code-file-mask dup dup @ 2>r off
    gene-follow
    r> r> ! ;

: cut-to-used-symbols ( addr count -- addr' count' )
    minimal-depth @
    over min >r			\ length got negative without 'min' sometimes
    swap r@ +			\ start of used part
    swap r> - ;			\ used length

: complete-symbols-info ( addr count handle -- )
    >r
    [char] "  		r@ char-cat
    bl	  		r@ char-cat
    cut-to-used-symbols r@ cat
    [char] "		r@ char-cat
    bl			r> char-cat ;

\ given the symbol string of initial stack these two words append
\ the actually used part as xxstack" ..." to the handled buffer
\ always call insert-in-symbols-info *before* insert-out-symbols-info
: insert-in-symbols-info ( handle -- )
    >r
    s" instack" r@ cat
    initial 2@  r> complete-symbols-info ;

: insert-out-symbols-info ( handle -- )
    >r
    s" outstack" r@ cat
    symbols-as-string r> complete-symbols-info ;

: frame-insert-end-markings ( handle -- )
    >r

    \ do we want an entry in the genes code?
    s" NO-ENTRIES"  frame-tag 2@  compare		( gene-code-entry? )

    \ do we want an entry in the code file?
    write-code-file  code-file-mask @  AND dup IF drop
	file-mutating  code-file-mask @  AND dup IF drop
	    file-start&stop-entries code-file-mask @  AND	( code? file? )
	    dup IF
		[ decimal ] 64 stringbuf-open >r
	    THEN
	THEN
    THEN

    2dup OR IF						\ cat the string
	r@ insert-in-symbols-info
	r@ insert-out-symbols-info

	s" end" r@ cat		[char] " r@ char-cat	bl r@ char-cat
	frame-tag 2@ r@ cat	[char] " r@ char-cat	bl r@ char-cat

	( code? file? ) IF		( code? )		
	    indent-code
	    r> dup buffer-as-string	( code? scratch-handle addr count )
	    2dup cat>code-file
	    write-code-file  code-file-mask @  AND IF
		file-mutating  code-file-mask @  AND IF
		    code-file-write-line
		THEN
	    THEN
	    clear-code-file-buffer
	    3 pick IF
		r@ cat
	    ELSE 2drop THEN		( code? scratch-handle )
	    stringbuf-close		( code? )
	\ ELSE the string is already in the handled buffer
	THEN drop
    ELSE 2drop
    THEN rdrop ;

: last-frame-insert-end-markings ( handle -- )
    write-code-file  code-file-mask @  AND IF
	file-mutating  code-file-mask @  AND IF
	    file-start&stop-entries code-file-mask @  AND IF
		indent-code
		s" \ next line inserted by 'last-frame-insert-end-markings'"
		cat>code-file code-file-write-line
	    THEN
	THEN
    THEN

    frame# @  1+  set-frame-actual
    frame-insert-end-markings
    frame# @  1-  set-frame-actual ;

: symbols-play ( start-addr n -- )
    cells over + swap	( stop-addr start-addr )
    BEGIN	( stop-addr actual-addr )
	2dup = IF 2drop EXIT THEN	\ end of chain reached? done
	dup @				\ get xt
	dup [internal'] gene-literal = IF  \ special treatment of gene-literal's
	    drop				\ drop xt
	    s" n"  push-symbol-string		\ symbol 'n'
	    cell+				\ skip value
	ELSE				\ no literal
	    gene-follow			\ do the play
	THEN
	dup @ ['] ;gene <>		\ end of gene reached?
    WHILE				\ then end
	cell+
    REPEAT
    2drop ;


\ do it without code file entries:
: symbols-play-mute ( start-addr n -- )
    code-file-mask dup dup @ 2>r off	\ no code file entries, please
    symbols-play
    r> r> ! ;				\ restore code-file-mask

: play-symbols-to-end ( tail-addr )	\ xt chain, ['] ;gene terminated

    [ log-mask @ ] [IF]
	log-mask @ log-m-more and IF
	    s" playing tail symbols:"	0 log
	THEN
    [THEN]

    highest-integer# cell / symbols-play ;

\ output the code of xt to the code file:
: xt-write-code ( internals-xt -- )
    run-mode dup @ writing-code or swap !
    1 (code-indent) !
    s" : "		cat>code-file
    dup xt>name		cat>code-file
    s"  ( -- ) "	cat>code-file	code-file-write-line
    \ the ingredients:
    >body >gene-start  highest-integer# cell /  symbols-play
    code-file-write-line
    run-mode dup @ writing-code invert and swap ! ;

use-pool-select [IF] \ new pool with relative probability

: random-gene-picking ( -- internal-xt)
    actual-genepool-xt @ execute	\ get list pointer
    random-pool-gene ;

[ELSE]

: random-gene-picking ( -- internal-xt)
    actual-genepool-xt @ execute	\ get list pointer
    dup nodes				\ get number of elements
    random-ranged			\ pick one

    [ log-mask @ 0> ] [IF]	\ conditional compiling
	log-mask @ log-m-extra and IF
	    s" random-gene-picking picks number "	cat-log
	    dup num>string				cat-log
	    s"  : "					cat-log
	THEN
    [THEN]

    swap n'th-node @			\ get xt

    [ log-mask @ 0> ] [IF]	\ conditional compiling
	log-mask @ log-m-extra and IF
	    dup xt>name					0 log-it
	THEN
    [THEN]

;
[THEN]

VARIABLE (depth-should-change)		\ to avoid depth drifting off too far
VARIABLE resolve-flags	1 resolve-flags !

: done-with-depth-change? ( body -- flag )
    (depth-should-change) @ >r

    \ if (depth-should-change) is off, we're done
    r@ 0= IF  rdrop drop TRUE EXIT THEN

    \ don't care about (depth-should-change) if depth does not change
    dup gene-stack-effect 0= IF drop rdrop TRUE EXIT THEN

    \ if (depth-should-change) is negative we only care if there are
    \ at least three symbols on stack
    r@ 0< IF	#symbols-on-stack 3 < IF

	[ log-mask @ ] [IF]	\ conditional compiling
	    log-mask @ log-m-more and IF
		s" selected-gene-picking: "		cat-log
		s" too few symbols to care about depth"	0 log-it
	    THEN
	[THEN]

	rdrop drop TRUE EXIT
    THEN THEN

    \ now (depth-should-change) must be respected
    gene-stack-effect r> * 0>		\ goes in the right direction?

    [ log-mask @ ] [IF]	\ conditional compiling
	log-mask @ log-m-more and IF
	    dup IF		\ goes in the right direction?
		s" selected-gene-picking: trying to guide stack back"
		0 log
	    THEN
	THEN
    [THEN]

    dup 0=   [char] C symbols-tos-match?  AND IF
	0=			\ accept conditions when depth should change

	[ log-mask @ ] [IF]	\ conditional compiling
	    log-mask @ log-m-more and IF
		s" selected-gene-picking: condition accepted, might fix depth"
		0 log
	    THEN
	[THEN]

    THEN
;

\ pick a random, but preselected gene matching the symbol stack
\ addresses on tos will always get matched first
\ otherwise (depth-should-change) is respected if
\   * depth changes at all
\   * it's positive, or it's negative and at least two symbols on stack
: selected-gene-picking ( -- internal-xt)
    BEGIN
	random-gene-picking dup >body		( internal-xt body )

	dup dup gene-body>stack-in-string symbols-match IF	\ input match?

	    \ addresses get matched first
	    [char] a symbols-tos-match? IF	\ address to match?
		\ addresses on tos must always be matched
		dup >gene-stack-in @			\ gen requires input?
		over done-with-depth-change? AND IF	\ and depth change OK?

		    [ log-mask @ ] [IF]	\ conditional compiling
			log-mask @ log-m-much and IF
			    s" selected-gene-picking: address resolved"
			    0 log
			THEN
		    [THEN]

		    drop EXIT		\ address matched, done

		ELSE			\ gen requires no input,
		    2drop RECURSE EXIT	\ start again
		THEN
	    THEN
	    \ no address to match

	    [char] C symbols-tos-match? IF
		resolve-flags @ random-ranged 0= IF
		    dup >gene-stack-in @		\ gen requires input?
		    over done-with-depth-change? AND IF	\ and depth change OK?

			[ log-mask @ ] [IF]	\ conditional compiling
			    log-mask @ log-m-much and IF
				s" selected-gene-picking: flag resolved"
				0 log
			    THEN
			[THEN]

			drop EXIT		\ flag resolved, done
		    ELSE
			2drop RECURSE EXIT	\ start again
		    THEN

		    [ log-mask @ ] [IF]	\ conditional compiling
			log-mask @ log-m-much and IF
			    s" selected-gene-picking: flag resolving delayed"
			    0 log
			THEN
		    [THEN]

		THEN
	    THEN

	    dup done-with-depth-change? IF drop EXIT THEN

	THEN	\ stack mismatch

	2drop		\ did not fit
    AGAIN		\ try again
;


log-mask @ [IF]		\ I just redefine selected-gene-picking ;-)
: selected-gene-picking ( -- internal-xt)
    selected-gene-picking ( -- internal-xt)
    log-mask @ log-m-much and IF
	s" selected-gene-picking picked: "	cat-log
	dup xt>name				0 log-it
    THEN ;
[THEN]

\ words to deal with provisoric gene sequences kept as string in an
\ evaluation buffer ['eb']
0
OFFSET:	eb>length		\ length of the string to evaluate
OFFSET:	eb>compiled		\ zero or xt of compiled version
OFFSET:	eb>counter		\ how many living cells use this sequence
\ eb>sequence *must* be last!   It *is* the starting address, not a pointer.
OFFSET:	eb>sequence		\ start of sequence string
constant eb-entries#		\ how many entries at the beginning of buffer

: decrease-eb-count ( "xt" -- )	\ the "xt" is the allocation address
    dup >r
    eb>counter				\ decrease counter
    dup @ 1- ?dup IF			\ if >0
	swap !				\ store it
    ELSE
	drop
	r@ free				\ else free memory
	IF
	    cr ." problem free'ing trial gene sequence."
	    bell 1500 ms
	THEN
    THEN rdrop ;

log-mask @ [IF]
\ logs stack symbols after a starting string.  log-mask is checked elsewhere.
: log-stack-symbols  ( addr count -- )
    #symbols-on-stack over + stringbuf-open >r r@ cat	\ r: handle
    symbols-as-string r@ cat
    r@ buffer-as-string 0 log
    r> stringbuf-close ;
[THEN]

VARIABLE mutations-threshold \ after so many mutations we try to come to an end
1 mutations-threshold !

VARIABLE (mutated)

\ if stack depth is more then that off, only genes that won't make it worse
\ will be inserted. Strong influence on the complexity of mutations.
VARIABLE stack-turning-point	3 stack-turning-point !

DEFER build-if-else-then ( handle -- )

: follow-&-add ( xt handle -- )
    >r

    [ also gene-internals ] [DEFINED] g-IF-ELSE-THEN [ previous ] [IF]
	dup [internal'] g-IF-ELSE-THEN = IF
	    drop r> build-if-else-then	EXIT
	THEN
    [THEN]

    1 (mutated) +!

    dup gene-follow

    xt>name 2dup		( addr-name count-name addr-name count-name)
    r@ cat  bl r> char-cat

    \ if an organ is used set my-diversification-bitmap accordingly:
    \ this works on the nuc
    2dup s" organ-" dup >r search nip IF	\ beware of new names
	r> + c@ [char] A - 1 swap lshift
	my-diversification-bitmask dup @ rot or swap !
\ we could diversify here ############ no, not here.
    ELSE drop rdrop THEN

[ nuc-parameters# ] [IF]
    \ if a parameter is used set my-diversification-bitmap accordingly:
    \ this works on the nuc
    s" parameter-" dup >r search nip IF		\ beware of new names
	r> + c@ [char] A - nuc-parameters nuc-organs - + 1 swap lshift
	my-diversification-bitmask dup @ rot or swap !
\ we could diversify here ############ no, not here.
    ELSE drop rdrop THEN
[ELSE]
    2drop
[THEN]
;


\ Build a '[IF] ... [ELSE] ... [THEN]' structure in the buffer handled.
\ Start with a 'start" CONDITION"' mark.
\ Respect the type of stack data using the actual symbols stack,
\ and make sure, both parts deliver the symbol stack in the same state.
\ The symbols stack will contain this state after execution.
\
\ The sequence might use only a part of input data, so:
\ Keep track of the input requirements of the sequence and the output produced,
\ put results as comment in the code string of the form
\ ... ... CODE ... ... instack" nc" outstack" nna"
\ These can be used to compile the code as a separate gene when combining
\ sequences to longer sequences.
\ Do depth-watch-restart with the minimum of the previous and the actual value,
\ to be reused by the calling word.
DEFER <build-new-segment>	\ defined afterwards
\ up to how many items shall we put in the first if branch?
VARIABLE max-if-items		4 max-if-items !
\ build-if-else-then ( handle -- )
:NONAME ( handle -- )
    >r						\ handle of working buffer

    [ log-mask @ ] [IF]
	log-mask @ log-m-more and IF
	    s" build-if-else-then starting..."	0 log
	    s" stack at start: "		log-stack-symbols
	    s" building 'IF' branch:"		0 log
	THEN
	r@ buffered-length (log-offset) !	\ still ok? ##########
    [THEN]

    s" CONDITION" r@ insert-start-mark

    ['] [IF]  r@  follow-&-add

    max-if-items @ random-ranged   0  BEGIN
	selected-gene-picking r@ follow-&-add
	2dup <> WHILE
	1+
    REPEAT 2drop

    [ log-mask @ ] [IF]			\ conditional compiling
	log-mask @ log-m-much and IF
	    s" building 'ELSE' branch:" 0 log
	THEN
    [THEN]

    ['] [ELSE]  r@ follow-&-add
    (no-start&stop-entries) on
    aim 2@ r@ <build-new-segment>	\ build '[ELSE]' branch

    ['] [THEN]  r@ follow-&-add
    r@ last-frame-insert-end-markings

    [ log-mask @ ] [IF]
	log-mask @ log-m-more and IF
	    s" ==> conditional branching built:"	0 log
	    r@ buffer-as-string  offset>log		0 log
	THEN
    [THEN]

    rdrop ; IS build-if-else-then

\ Build a new segment and append it to the given string buffer.
\ The buffer holds the Forth code as a string
\ (that can be evaluate'd and compiled).
\ Start with a 'start" SEQUENCE"' mark unless (no-start&stop-entries) is set.
\
\ Respect the type of stack data using the actual symbols stack.
\ Guide the stack data flow to the state given in addr-aim count-aim.
\ The symbols stack will contain this string after execution.
\
\ The sequence might use only a part of input data, so:
\ Keep track of the input requirements of the sequence and the output produced,
\ put results as comment in the code string of the form
\ ... ... CODE ... ... instack" nna" outstack" nn",
\ unless (no-start&stop-entries) was set.
\ These can be used to compile the code as a separate gene when combining
\ sequences to longer sequences, doing conditionals or so.
\ Adjust depth-watch to be reused by the calling word.
: build-new-segment ( addr-aim count-aim handle -- )

    s" build-new-segment starting..."	log-m-some log
    dup buffered-length (log-offset) !

    
    >r						\ handle of working buffer

    s" SEGMENT" r@ push-frame-with-entry

    \ preserve desired stack condition (aim)
    ( addr-aim count-aim ) aim frame-store-string

    [ log-mask @ ] [IF]			\ conditional compiling
	log-mask @ log-m-more and IF
	    s" stack at start: "	log-stack-symbols
	    s" stack aim     : "	cat-log
	    aim 2@			0 log-it	    
	THEN
    [THEN]

    \			( r: handle of the buffer to put the FORTH words in )

    (depth-should-change) off
    BEGIN	( old-depth-watch initial-stack-handle  aim-stack-handle )
	(mutated) @ mutations-threshold @ > IF
	    \ set desired direction of stack depth change:
	    aim @				( aim-depth )
	    #symbols-on-stack - (depth-should-change) !

	    (depth-should-change) @ abs  stack-turning-point @  - 0> 0= IF
		(depth-should-change) off
		[ log-mask @ ] [IF]		\ conditional compiling
		    log-mask @ log-m-more and IF
			s" stack-turning-point not reached "
			stack-turning-point @	0 log-string-and-number
		    THEN
	    ELSE
		    log-mask @ log-m-more and IF
			s" stack-turning-point passed "	0 log
		    THEN
		[THEN]
	    THEN

	    [ log-mask @ ] [IF]		\ conditional compiling
		log-mask @ log-m-more and IF
		    s" (depth-should-change) set to "
		    (depth-should-change) @    0 log-string-and-number
		THEN
	    [THEN]

	THEN

	selected-gene-picking r@ follow-&-add
	[ log-mask @ ] [IF]		\ conditional compiling
	    log-mask @ log-m-more and IF
		s" new stack: " log-stack-symbols
		s" aim is   : " cat-log
		aim 2@		0 log-it
	    THEN
	[THEN]

	aim 2@  symbols-as-string  compare 0=	\ aim reached?
	[ log-mask @ ] [IF]		\ conditional compiling
	    log-mask @ log-m-more and IF
		dup IF
		    s" aim reached :-)"		0 log
		THEN
	    THEN
	[THEN]

    UNTIL

    r@ frame-insert-end-markings
    pop-frame

    log-mask @ log-m-some and IF
	s" ==> new-segment built:"	0 log
	r@ buffer-as-string  offset>log 0 log
    THEN

    rdrop ;
' build-new-segment IS <build-new-segment>

\ set up everything for trial phase including changes in the nuc
\ the handled buffer has the FORTH words
: set-up-trial ( addr-to-store-in handle )
    >r						( r: handle )
    r@ buffered-length eb-entries# cells +	\ length of buffer needed 
    allocate
    IF	." could'nt allocate memory for newly mutated gene."	rdrop
    ELSE ( addr-to-store-in addr-allocated )

	\ store at addr-to-store-in (probably at wake-me-xt in a nuc).
	( addr-to-store-in addr-allocated ) dup rot ! ( addr-allocated )

	\ set up entries in the evaluation buffer:
	dup eb>length r@ buffered-length swap ! \ store string length
	dup eb>compiled off		\ not compiled yet
	dup eb>counter 1 swap !		\ init counter to 1

	\ move string there, just after the eb-entries
	eb>sequence		( to )	\ start address of sequence string
	r> buffer-as-string	( to from count )
	>r swap r>		( from to count )	move

	\ setting up the nuc
	new-genome-id genome-id !
	nuc-flags @ wake-me-trial or nuc-flags !
    THEN ;

: initialize-mutation ( -- )
[ use-pool-select ] [IF] \ new pool with relative probability
    actual-genepool-xt @ execute ?pool-setup
[THEN]
    (depth-should-change) off
    clear-symbols			\ opens symbols stack, if needed
    start-to-follow
    (new-gene) >r r@ stringbuf-empty	\ prepare buffer to store in
    s" TOP" r> push-frame-with-entry ;

: top-level-insertion ( xt -- )

    s" top-level-insertion STARTING...	top-level-insertion STARTING"
    log-m-type log

    gene-xt>internal-xt >r		\ r: internal-xt
    r@ >body >gene-tokens# @ random-ranged	( head-items )

    log-mask @ log-m-some and IF
	s" top-level-insertion from:"		0 log
	r@ internal-xt>string			0 log
	s" insertion after item "		cat-log
	dup num>string				cat-log
	s"  of "				cat-log
	r@ >body >gene-tokens# @ num>string	0 log-it
    THEN

    dup
    0		( head-items head-items segment-items=0 )  \ 0 segment items
    r@ -rot	( head-items xt head-items segment-items=0 )
    xt>cutten-strings	\ initializes (head) (segment) (tail)
		( head-items )

    r> >body >gene-start ( head-items start-addr)   \ addr of first sub gene

    \ get tails start to play tail later on:
    over cells over +  >r			\ start of tail

    \ play-head symbols:
    [ log-mask @ ] [IF]
	log-mask @ log-m-more and IF
	    s" playing head symbols..."	0 log
	THEN
    [THEN]

    clear-symbols				\ open an empty symbol stack
    swap ( start head-items ) symbols-play	\ put symbols on symbol stack

    symbols-as-string			\ give same symbols as stack aim

    (new-gene) >r			\ buffer to store in

    (head)	buffer-as-string	r@ cat	\ head

    r@ build-new-segment ( a u handle -- )	\ build matching gene sequence

    \ splice the genes
    (tail)	buffer-as-string	r> cat	\ tail

    r> play-symbols-to-end ;


\ Word to skip a structure and return skipped items.
\ addr points to a chain of xt's,
\ 1'st xt is 'addr start-index cells + @'
\
\ does:	'IF ... ELSE ... THEN's
\ Aborts if called starting at a xt, that unnests a structure.
\ Should not reach ';gene', checks for that.
\ no other structural checks yet.
\
\ If called with a starting xt that does'nt change nesting structure,
\ it returns with index unchanged.
: skip-structures ( start-index addr -- last-skipped-item )
    >r			( start-index             r: addr )
    0 swap		( nesting=0 start-index   r: addr )
    cells		( nesting=0 start-offset  r: addr )

    [ log-mask @ ] [IF]
	log-mask @ log-m-much and IF
	    s" Skipping structures starting:" 0 log
	THEN
    [THEN]

    BEGIN		( current-nesting current-offset  r: addr )
	\ get xt:
	dup r@ + @	( current-nesting current-offset xt r: addr )

	[ log-mask @ ] [IF]
	    log-mask @ log-m-much and IF
		over cell / num>string	cat-log
		s"  	"		cat-log
		s" checking "		cat-log
		dup xt>name		cat-log
		s" ... "		cat-log
	    ELSE
		dup xt>name 2drop	\ does NOT work without! ???
		\ on rare occasions there's a crash without xt>name
		\ i have no idea why... ###########
	    THEN
	[ELSE]
	    dup xt>name 2drop	\ does NOT work without! ???
	[THEN]

	CASE	\ on xt		( current-nesting current-offset  r: addr )
	    \ check if it is a structure starting gene:
	    ['] [IF] OF
		>r 1+ r>		\ increase nesting level

		[ log-mask @ ] [IF]
		    log-mask @ log-m-much and IF
			s"  We skip from 'IF' to 'THEN' level: "  cat-log
			over num>string				  0 log-it
		    THEN
		[THEN]

	    ENDOF

	    \ check if it is a structure terminating gene:
	    ['] [THEN] OF

		[ log-mask @ ] [IF]
		    log-mask @ log-m-much and IF
			s"  'THEN' reached, level: "	cat-log
			over num>string			0 log-it
		    THEN
		[THEN]

		>r 1- r>		\ decrease nesting level
		over 0= IF		\ level zero reached?
		    cell+		\ skip last end token too
		THEN
	    ENDOF

	    ['] ;gene OF		\ savety check
		cr ." Level: " over .
		true ABORT"  skip-structures: ';gene' reached!"
	    ENDOF

	    
	    \ ordinary xt's

	    [ log-mask @ ] [IF]
		log-mask @ log-m-much and IF
		    s" skipped." 0 log-it
		THEN
	    [THEN]

	ENDCASE	  ( current-nesting current-offset  r: addr )

	over 0= IF			\ zero nesting?
	    nip  cell /	( skipped-items )

	    [ log-mask @ ] [IF]
		log-mask @ log-m-much and IF
		    s" Skipping structures completed: "	cat-log
		    dup num>string			cat-log
		    s"  items skipped."			0 log-it
		THEN
	    [THEN]

	    rdrop EXIT
	THEN				\ done

	cell+				\ next item
    AGAIN ;

VARIABLE (nesting)
: check-items-to-replace ( max-items-to-replace start-addr -- items-ok )
    >r			( max-items-to-replace  r: start-addr )

    0	\ count checked items
    BEGIN ( max-items-to-replace  checked-items  r: start-addr )
	2dup > 0= IF nip rdrop EXIT THEN
	dup cells r@ + @	( max-items-to-replace  checked-items xt )

	[ log-mask @ ] [IF]
	    log-mask @ log-m-much and IF
		over num>string		cat-log
		s" 	checking "	cat-log
		dup xt>name		cat-log
		s" : "			cat-log
	    THEN
	[THEN]

	CASE
	    ['] [ELSE] OF

		[ log-mask @ ] [IF]
		    log-mask @ log-m-much and IF
			s" we stop at 'ELSE' " 0 log-it
		    THEN
		[THEN]

		nip  rdrop EXIT	\ '[ELSE]' is EXcluded
	    ENDOF

	    ['] [THEN] OF

		[ log-mask @ ] [IF]
		    log-mask @ log-m-much and IF
			s" we stop before 'THEN' " 0 log-it
		    THEN
		[THEN]

		nip  rdrop EXIT	\ '[THEN]' gets EXcluded
	    ENDOF

	    ['] [IF] OF	\ Let's skip the whole condition branch
		r@ skip-structures ( start-index addr -- last-skipped-item )
		1-			\ will be added again
	    ENDOF

	    \ default case: ordinary genes
	    [ log-mask @ ] [IF]
		log-mask @ log-m-much and IF
		    s"    	skipped."	0 log-it
		THEN
	    [THEN]

	ENDCASE
	1+
    AGAIN ;

: top-level-replacement ( xt -- )

    s" top-level-replacement STARTING...	top-level-replacement"
    log-m-type log

    gene-xt>internal-xt >r		( r: internal-xt )

    r@ >body >gene-tokens# @ dup
    random-ranged swap random-ranged
    2dup > IF swap THEN		( head-items tail-start-item )
    over -			( head-items max-items-to-replace )

    log-mask @ log-m-some and IF
	s" top-level-replacement from: "	cat-log
	r@ xt>name				cat-log
	s"  items: "				cat-log
	r@ >body >gene-tokens# @ num>string	0 log-it
	r@ internal-xt>string			0 log
	over >r s" replacement after item " r>	0 log-string-and-number

	log-mask @ log-m-more and IF
	    s" trying to replace up to "	cat-log
	    dup num>string			cat-log
	    s"  items. Let's check:"		0 log-it
	THEN

    THEN

    r@ >body 2 pick gene-body>n'th-xt-addr	( head-i to-replace start-a )
    ( head-items max-items-to-replace start-addr )
    check-items-to-replace ( head-items items-ok )

    log-mask @ log-m-some and IF
	s" I will replace "		cat-log
	dup num>string			cat-log
	s"  items."			0 log-it
    THEN

    \ get tails start to play tail later on:
    2dup + cells  r@ >body >gene-start +	\ start of tail
    -rot			( tail-addr head-items items-ok-segment )

    \ remember segment-items to play segment later on:
    tuck ( tail-addr segment-items head-items items-ok-segment )

    \ cut strings:
    r@ -rot	( tail-addr segment-items xt head-items segment-items )
    \ remember head items to play head after cutting strings:
    over >r
    ( tail-addr segment xt head-items segment-items  r: internal-xt head-items)

    xt>cutten-strings	\ initializes (head) (segment) (tail)
    log-mask @ log-m-some and IF
	s" (Segment will be replaced)."  0 log
    THEN

    r>		( tail-addr segment-items head-items  r: internal-xt )

    r@ >body >gene-start ( tail segment head-items start-addr)

    \ follow head genes:
    clear-symbols				\ open an empty symbol stack
    \ put symbols on symbol stack:
    swap ( tail segment-items start head-items ) symbols-play

    \ remember frame and stack as start of new segment to be built:
    store-actual-stack-and-frame ( tail segment-items data-addr )

    \ play old segments items to get the aim:
    over cells >r 2 pick r> -			\ start address of segment
    ( tail segment-items data-addr seg-start)
    rot symbols-play-mute			( tail data-addr )
    symbols-as-string				\ symbols as stack aim
    dup stringbuf-open >r r@ cat		\ preserved in a buffer
    dup restore-stack-and-frame free ( tail)	\ restore start
    ABORT" top-level-replacement: Couldn't free stored frame data."

    \ now construct the code:
    r@ buffer-as-string			\ segment's end given as aim
    (new-gene) >r    ( r: xt handle )	\ buffer to store in
    (head) buffer-as-string r@ cat	\ cat head

    write-code-file  code-file-mask @  AND IF
	file-mutating  code-file-mask @  AND IF
	    file-frames code-file-mask @ AND IF
		indent-code
		s" \ replacement begins" cat>code-file code-file-write-line
	    THEN
	THEN
    THEN

    \ check if the aim is not already reached:
    2dup symbols-as-string compare ?dup 0= IF	\ aim not reached?
	(segment) buffered-length		\ yes: is segment there?

	log-mask @ log-m-some and IF
	    dup IF
		s" stack matches: Segment just snipped."  0 log
	    THEN
	THEN

	0=					\ (otherwise we'll make one)
    THEN

    \ build a new segment if appropriate:
    IF r@ build-new-segment ( ... a u handle -- )
    ELSE 2drop THEN

    write-code-file  code-file-mask @  AND IF
	file-mutating  code-file-mask @  AND IF
	    file-frames code-file-mask @ AND IF
		indent-code
		s" \ replacement ended." cat>code-file code-file-write-line
	    THEN
	THEN
    THEN

    \ splice the genes
    (tail)  buffer-as-string  r> cat	\ cat tail

    r> stringbuf-close			\ close intermediate stack buffer
    rdrop

    ( tail-addr ) play-symbols-to-end ;

\ 'follow-item-or-structure' follows next item, or structure.
\ Word to skip a structure and return skipped items.
\ addr points to a chain of xt's,
\ 1'st xt is 'addr start-index cells + @'
\
\ does:	'IF ... ELSE ... THEN's
\ If a unnesting word or ';gene' is encountered it returns items=0
\ no other structural checks yet.
\ Does *not* consume but *change* stack symbols, to be used in a loop.
: follow-item-or-structure ( start-addr current-offset -- a o' followed-items )
    dup 0 2>r	( start-addr current-offset   r: start-offset nesting=0 )
    BEGIN		( addr current-offset   r: start-offset nesting )
	2dup + @	( addr current-offset xt  r: start-offset nesting )

	[ log-mask @ ] [IF]
	    log-mask @ log-m-much and IF
		s" follow-item-or-structure: level="	cat-log
		r@ num>string				cat-log
		s" 	( "				cat-log
		symbols-as-string			cat-log
		s"  ) 	"				cat-log
		dup xt>name				0 log-it
	    THEN
	[THEN]

	CASE	\ on xt
	    \ check if it is a structure starting gene:
	    ['] [IF] OF	r> 1+ >r ENDOF		\ increase nesting level

	    \ check if it is a structure terminating gene:
	    ['] [THEN] OF
		r> 1- >r			\ decrease nesting level
		r@ CASE \ on nesting
		    -1 OF			\ 'THEN' without 'IF'
			2rdrop FALSE EXIT	\ no luck
		    ENDOF
		ENDCASE
	    ENDOF

	    \ check if it is a structure dividing gene:
	    ['] [ELSE] OF
		r@ 0= IF			\ level zero?
		    2rdrop FALSE EXIT		\ 'ELSE' without 'IF'
		THEN
	    ENDOF

	    ['] ;gene OF	( addr current-offset  r: start-offset nesting)
		2r> nip		( addr current-offset nesting )
		?dup IF
		    cr
		    ." follow-item-or-structure: ';gene' encountered on level "
		    . cr
		    ABORT" follow-item-or-structure: gave up..."
		THEN
		FALSE EXIT 
	    ENDOF
	ENDCASE

	2dup + @ gene-follow-mute	\ follow
	cell+				\ one item done

	\ check nesting:
	r@ 0=	( addr current-offset flag  r: start-offset nesting )
    UNTIL				\ done if nesting is zero

    dup 2r> drop	( addr current-offset current-offset start-offset )
    - cell / ;	( addr current-offset followed-items )

\ word to follow a chain of xt's until stack reaches the given aim,
\ or give up, if ';gene' is reached, or enter level structure is left.
: next-stack-match ( start-addr addr-aim count-aim -- snipped-items )
    2>r			( start-addr   r: addr-aim count-aim )
    0			( start-addr offset=0  r: addr-aim count-aim )
    BEGIN		( start-addr current-offset  r: addr-aim count-aim )
	\ follow current item or structure increasing current-offset:
	follow-item-or-structure   ( start-addr current-offset followed-items) 
	0= IF				\ ';gene' or leaving entering level

	    [ log-mask @ ] [IF]
		log-mask @ log-m-more and IF
		    s" next-stack-match: no match found till end."
		    0 log
		THEN
	    [THEN]

	    2drop 2rdrop FALSE EXIT	\ no luck
	THEN

	2r@ symbols-as-string compare 0=	\ stack match?
    UNTIL		( start-addr current-offset  r: addr-aim count-aim )
    2rdrop nip cell /	( followed-items )

    [ log-mask @ ] [IF]
	log-mask @ log-m-more and IF
	    s" next-stack-match: next stack match after "	cat-log
	    dup num>string					cat-log
	    s"  tokens."					0 log-it
	THEN
    [THEN]
;

\ Search all matches in structure and stack given as 'addr-aim count-aim'.
\ 'addr' points into a xt chain.
\ Return the handle of a buffer with addresses of the matches and their number.
\ Please *close the buffer* in the calling word.
: all-stack-matches ( start-addr addr-aim count-aim -- handle matches# )
    2>r				( start-addr   r: addr-aim count-aim )
    [ decimal ]
    1024 stringbuf-open swap	( handle start-addr   r: addr-aim count-aim )
    BEGIN
	dup 2r@ next-stack-match ( handle start-addr snipped-items   r: a u )
    dup WHILE
	cells +
	over >r sp@ cell r> cat		\ cat address of match to buffer
    REPEAT
    2drop 2rdrop		( handle   r:-- )
    dup buffered-length cell /	( handle matches# )

    [ log-mask @ ] [IF]
	log-mask @ log-m-more and IF
	    s" all-stack-matches: found "	cat-log
	    dup num>string			cat-log
	    s"  matches."			0 log-it
	THEN
    [THEN] ;	( handle matches# )	\ please close buffer..

\ snip out a random sequence that does not harm structure or stack matching
\ and omit it
0 CONSTANT random-snip
1 CONSTANT short-snip
2 CONSTANT long-snip
VARIABLE (snip-type)		random-snip (snip-type) !
\
: top-level-snip ( xt -- u )		\ returns number of snipped items
    gene-xt>internal-xt >r		( r: internal-xt )
    r@ >body >gene-start		( addr-of-xt-chain  r: internal-xt )

    r@ >body >gene-tokens# @
    dup 2 < IF				\ nothing to snip?
	s" top-level-snip: There's nothing to skip." log-m-some log
	2drop rdrop FALSE EXIT		\ nothing to snip, done
    THEN

    s" top-level-snip STARTING...	top-level-snip" log-m-type log

    random-ranged			( addr start-token  r: internal-xt )

    log-mask @ log-m-some and IF
	s" top-snip type "			cat-log
	(snip-type) @ CASE
	    random-snip	OF s" random"	ENDOF
	    short-snip	OF s" short"	ENDOF
	    long-snip	OF s" long"	ENDOF
	ENDCASE					cat-log
	s"  from: "				cat-log
	r@ xt>name				cat-log
	s"  items: "				cat-log
	r@ >body >gene-tokens# @ num>string	0 log-it
	r@ internal-xt>string			0 log
	dup >r s" try snipping after item " r>	0 log-string-and-number

	[ log-mask @ ] [IF]
	    log-mask @ log-m-more and IF
		s" following head genes to get stack and frame:" 0 log
	    THEN
	[THEN]

    THEN

    \ follow head genes:
    clear-symbols				\ open an empty symbol stack
    ( start head-items ) 2dup symbols-play-mute	\ put symbols on symbol stack

    [ log-mask @ ] [IF]
	log-mask @ log-m-more and IF
	    s" top-level-snip: Stack at start of snip is "	cat-log
	    symbols-as-string					0 log-it
	THEN
    [THEN]

    \ seek stack matches
    2dup cells +	( start head-items first-snip-addr )
    symbols-as-string				\ symbols as stack aim
    dup stringbuf-open >r r@ cat		\ preserved in a buffer
    r@ buffer-as-string	( start head-items snip-addr addr-aim count-aim )
    (snip-type) @ short-snip = IF		\ we need only first match
	next-stack-match		( start head-items followed-items )
    ELSE					\ we need all matches
	all-stack-matches ( start head handle matches# )	\ opens buffer
	dup IF					\ matches found
	    \ what type?
	    (snip-type) @ random-snip = IF		\ random match
		random-ranged
	    ELSE 1-				\ longest match
	    THEN		( start head handle selected-match# )

	    [ log-mask @ ] [IF]
		log-mask @ log-m-much and IF
		    s" I take match "	cat-log
		    dup num>string	0 log-it
		THEN
	    [THEN]		( start head handle selected-match# )

	    over buffer-as-string drop	( start head handle match# data-addr )
	    swap cells + @		( start head# handle address-of-match )
	    swap stringbuf-close	\ close buffer from 'all-stack-matches'
	    \	( start head# address-of-match)
	    >r 2dup cells + r> swap - cell / ( start head-items snipped-items )
	ELSE				\ no match found
	    ( start head handle 0 ) swap stringbuf-close \ close buffer
	THEN
    THEN	( start head-items snipped-items )

    r> stringbuf-close				\ close stack buffer

    ?dup IF					\ match found?

	log-mask @ log-m-some and IF
	    dup num>string	cat-log
	    s"  items snipped."	0 log-it
	THEN

	\ cut strings:		( start head-items followed-items  r: xt )
	2dup r> -rot		( start head followed xt head followed  r:-- )
	xt>cutten-strings	( start head-items followed-items )

	log-mask @ log-m-some and IF
	    s" (Segment gets snipped)."  0 log
	THEN

	\ follow head again (last time was mute)
	( start head-items followed-items )
	>r			( start head-items  r: followed-items )
	2dup symbols-play			\ follow head

	write-code-file  code-file-mask @  AND IF
	    file-mutating  code-file-mask @  AND IF
		file-frames code-file-mask @ AND IF
		    indent-code
		    s" \ "			cat>code-file
		    r@ num>string		cat>code-file
		    s"  words snipped."	cat>code-file  code-file-write-line
		THEN
	    THEN
	THEN

	\ foolow tail		( start head-items   r: followed-items )
	r@ + cells + play-symbols-to-end
	r>			( followed-items )

	\ splice the gene:
	(new-gene) >r	( r: followed handle )		\ buffer to store gene
	(head) buffer-as-string	r@ cat		\ cat head
	s" SNIPPED "			r@ cat
	(tail) buffer-as-string	r> cat		\ cat tail

    ELSE					\ no match found
	2drop rdrop FALSE
    THEN ;

: top-level-random-snip ( xt -- u )  random-snip (snip-type) ! top-level-snip ;
: top-level-short-snip ( xt -- u )   short-snip (snip-type) !  top-level-snip ;
: top-level-long-snip ( xt -- u )    long-snip (snip-type) !   top-level-snip ;

\ real or faked internal (no internals data)?
: real-internal? ( internals-xt -- flag )
    CASE
	['] [IF]	OF FALSE ENDOF
	['] [ELSE]	OF FALSE ENDOF
	['] [THEN]	OF FALSE ENDOF
	TRUE swap
    ENDCASE ;
	
: do-address-replacement ( token internals-xt  -- )	\ no checks
    >r			( token  r: internals-xt )

    log-mask @ log-m-some and IF
	s" replacing token "		cat-log
	dup num>string			cat-log
	s"  (replacing segment)."	0 log-it
    THEN

    r@ over  1 xt>cutten-strings
    r@ >body >gene-start over symbols-play
    BEGIN
	random-gene-picking
	dup >body >gene-flags @ out-address and IF
	    dup >body
	    dup >gene-stack-in @ 0=
	    swap >gene-stack-out 2@ s" a" compare 0= AND
	    dup 0= IF nip THEN	    
	ELSE drop false THEN
    UNTIL	( token replacement-xt   r: internals-xt )

    log-mask @ log-m-some and IF
	s" replacing by "	cat-log
	dup xt>name		0 log-it
    THEN

    \ splice the gene:
    (new-gene) >r	( r: followed handle )		\ buffer to store gene
    (head) buffer-as-string	r@ cat		\ cat head
    ( replacement-xt ) r@ follow-&-add
    (tail) buffer-as-string	r> cat		\ cat tail

    ( token   r: internals-xt )
    r> >body >gene-start swap 1+ cells + play-symbols-to-end ;

: top-level-address-replacement ( xt -- flag )
    gene-xt>internal-xt >r		( r: internal-xt )
    r@ >body >gene-tokens# @		( tokens   r: internal-xt )
    
    dup 2 < IF
	s" top-level-address-replacement: ';gene' only." log-m-type log
	drop rdrop FALSE EXIT		\ nothing to do, done
    THEN

    log-mask @ log-m-type and IF
	s" top-level-address-replacement: "	cat-log
    THEN

				( tokens   r: internal-xt )
    random-ranged		( random-token   r: internal-xt )
    r@ >body >gene-start	( random-token start-addr     r: i-xt )
    >r				( random-t    r: i-xt start-addr)
    \ search from random token to the end
    dup
    BEGIN			( random-t actual-token     r: i-xt start-addr)
	dup cells r@ + @	( random-t actual-token act-xt  r: xt start-a )
	dup ['] ;gene <>
    WHILE			( random-t actual-token act-xt  r: xt start-a )
	dup real-internal? IF	\ ordinary genes with internals data
	    >body		( random-t actual-token body  r: xt start-addr)
	    dup >gene-flags @ out-address and IF
		dup >gene-stack-in @ 0=
		( random-t actual-token body flag-zero-in r: xt start-addr)
		swap >gene-stack-out 2@ s" a" compare 0= AND IF
		    nip rdrop r> do-address-replacement	\ address giver found
		    TRUE EXIT
		THEN
	    ELSE drop
	    THEN
	ELSE drop THEN		\ conditionals and the like
	1+			( random-t actual-token   r: xt start-addr)
    REPEAT
    2drop			( random-token   r: internal-xt start-addr )

    \ no address giver from random token 'til ;gene
    \ so we search from start to random
    0				( random-token 0       r: xt start-addr )
    BEGIN			( end-token actual-token    r: xt start-addr )
	2dup >
    WHILE
	dup cells r@ + @	( end-token actual-token xt   r: xt start-addr)
	dup real-internal? IF	\ ordinary genes with internals data
	    dup >body >gene-flags @ out-address and IF
		dup >body	( end-token actual-token xt body r: xt start-a)
		dup >gene-stack-in @ 0=
		( end-token actual-token xt body zero-in-flag   r: xt start-a)
		swap >gene-stack-out 2@ s" a" compare 0= AND IF		\ found
		    ( end-token actual-token xt   r: xt start-address )
		    drop nip  rdrop r>  do-address-replacement
		    TRUE EXIT
		THEN
	    THEN
	    drop
	ELSE drop THEN		( end-token actual-token    r: xt start-addr )
	1+
    REPEAT 2rdrop 2drop		( -- )

    s" no address found."	log-m-type log-it

    FALSE ;

: (do-token-replacement) ( replacement-xt token internals-xt -- )
    >r		( replacement-xt token   r: internals-xt )

    log-mask @ log-m-some and IF
	s" replacing token "		cat-log
	dup num>string			cat-log
	s"  replacing segment by "	cat-log
	over xt>name			0 log-it
    THEN

    ( replacement-xt token  r: internals-xt )
    r@ over  1 xt>cutten-strings
    r@ >body >gene-start over symbols-play

    \ splice the gene:
    (new-gene) >r	( r: followed handle )		\ buffer to store gene
    (head) buffer-as-string	r@ cat		\ cat head
    swap ( token replacement-xt ) r@ follow-&-add
    (tail) buffer-as-string	r> cat		\ cat tail

    ( token   r: internals-xt )
    r> >body >gene-start swap 1+ cells + play-symbols-to-end ;

: do-token-replacement ( token internals-xt  -- flag )	\ no checks
    >r				( token  r: internals-xt )

    dup cells r@ >body >gene-start + @	( token xt    r: internals-xt )
    >body				( token body  r: internals-xt )
    dup >gene-stack-in 2@ 2>r	( token body  r: internals-xt addr-in count-in)
    >gene-stack-out 2@ 2r>	( tok a-out c-out addr-in count-in  r: ints-xt)
    10000 0 DO
	random-gene-picking >r	( tok a-out c-out a-in c-in  r: i-xt xt' )
	r@ >body >r		( tok a-out c-out a-in c-in  r: i-xt xt' body)
	2dup r@ >gene-stack-in 2@ compare 0= IF		\ stack-in match?
	    2over r@ >gene-stack-out 2@ compare 0= IF	\ stack-out match?
		2drop 2drop rdrop r> swap	( xt' token r: xt loop )
		unloop r> (do-token-replacement) TRUE EXIT 
	    THEN
	THEN
	2rdrop
    LOOP
    2drop 2drop drop  2rdrop

    s" no replacement found. "	log-m-some log-it

    FALSE ;

: top-level-token-replace ( xt -- flag )
    gene-xt>internal-xt >r		( r: internal-xt )
    r@ >body >gene-tokens# @		( tokens   r: internal-xt )

    dup 2 < IF
	log-mask @ log-m-some and IF
	    s" top-level-token-replace: ';gene' only." 0 log
	THEN

	drop rdrop FALSE EXIT		\ nothing to do, done
    THEN

    log-mask @ log-m-type and IF
	s" top-level-token-replace: "		cat-log
    THEN

				( tokens   r: internal-xt )
    random-ranged		( random-token   r: internal-xt )
    r@ >body >gene-start	( random-token start-addr     r: i-xt )
    >r				( random-t    r: i-xt start-addr)
    \ search from random token to the end
    dup
    BEGIN			( random-t actual-token     r: i-xt start-addr)
	dup cells r@ + @	( random-t actual-token act-xt  r: xt start-a )
	dup ['] ;gene <>
    WHILE			( random-t actual-token act-xt  r: xt start-a )
	dup real-internal? IF	\ ordinary genes with internals data
	    drop nip  rdrop r> do-token-replacement EXIT
	ELSE drop THEN		\ conditionals and the like
	1+			( random-t actual-token   r: xt start-addr)
    REPEAT
    2drop			( random-token   r: internal-xt start-addr )

    \ no address giver from random token 'til ;gene
    \ so we search from start to random
    0				( random-token 0       r: xt start-addr )
    BEGIN			( end-token actual-token    r: xt start-addr )
	2dup >
    WHILE
	dup cells r@ + @	( end-token actual-token xt   r: xt start-addr)
	dup real-internal? IF	\ ordinary genes with internals data
	    drop nip  rdrop r> do-token-replacement EXIT
	ELSE drop THEN		( end-token actual-token    r: xt start-addr )
	1+
    REPEAT 2rdrop 2drop		( -- )

    s" no address found."		log-m-some log-it

    FALSE ;

VARIABLE mutations					\ counter

\ word that ignores the current genome, but builds a fresh one from scatch.
\ give new designs a chance to pop up in a later state of an evolution.
\ used to let new genes pop up sporadically just like in the very beginning
\ of an evolution.

: restart-from-scratch ( dummy-xt -- flag )
    drop	\ dummy xt to let it look like the other mutation types

    s" restart-from-scratch: insert into empty gene: " log-m-type log

    1 mutations +!
    genom-generation off
    [ also genes ] ['] g-noop [ previous ] top-level-insertion TRUE ;

2VARIABLE mutation-rate		1 100 mutation-rate 2!	\ how often to mutate
2VARIABLE top-level-insert	1 2 top-level-insert 2!	\ insert in top level?
2VARIABLE top-level-replace	1 3 top-level-replace 2! \ replace in top level
2VARIABLE top-snipping		1 3 top-snipping 2!	\ omit code snippets?
\ these are relative probabilities of snip type:
VARIABLE top-random-snipping	1 top-random-snipping !	\  random code snippets
VARIABLE top-short-snipping	1 top-short-snipping !	\  shortest snippet
VARIABLE top-long-snipping	1 top-long-snipping !	\  longest snippet

2VARIABLE top-address-replace	1 3 top-address-replace 2! \ replace an address
2VARIABLE top-token-replace	1 3 top-token-replace 2!   \ replace a token?
2VARIABLE top-from-scratch	1 60 top-from-scratch 2! \ restart from scratch
: (mutate) ( xt -- )
    BEGIN

	top-level-insert rated-flag IF		\ top level insertion?
	    1 mutations +!
	    genom-generation off
	    ( xt ) top-level-insertion EXIT
	THEN

	top-level-replace rated-flag IF		\ top level replacement?
	    1 mutations +!
	    genom-generation off
	    ( xt ) top-level-replacement EXIT
	THEN

	top-snipping rated-flag IF			\ top level snip?

	    \ select type based on relative probabilities:
	    top-random-snipping @ dup	( random )
	    top-short-snipping @ + dup	( random rnd+short rnd+short )
	    top-long-snipping @ +	( rnd rnd+short rnd+short+long)
	    random-ranged >r	( rnd rnd+short   r: selected-position)
	    r@ > IF
		r@ > IF		\ type random
		    ( xt ) dup top-level-random-snip ( u )
		ELSE 		\ type short
		    ( xt ) dup top-level-short-snip ( u )
		THEN
	    ELSE drop			\ type long
		( xt ) dup top-level-long-snip ( u )
	    THEN rdrop

	    IF
		1 mutations +!
		genom-generation off
		drop EXIT
	    THEN
	THEN

	top-address-replace rated-flag IF
	    ( xt ) dup top-level-address-replacement
	    IF
		1 mutations +!
		genom-generation off
		drop EXIT
	    THEN
	THEN	    

	top-token-replace rated-flag IF
	    ( xt ) dup top-level-token-replace
	    IF
		1 mutations +!
		genom-generation off
		drop EXIT
	    THEN
	THEN

	top-from-scratch rated-flag IF
	    ( xt ) dup restart-from-scratch
	    IF
		1 mutations +!
		genom-generation off
		drop EXIT
	    THEN
	THEN	    

    AGAIN ;

VARIABLE (mutated-max)	\ just to locate fat genes in the log file
(mutated-max) off

: mutate ( addr-of-xt -- )
    (mutated) off
    nuc-flags @ wake-me-trial and IF drop EXIT THEN \ trial time: no mutation

    log-mask @ log-mutation and IF
	s" MUTATION:  based on "	cat-log
	log-cat-id
	s"  at spot "			cat-log
	spot 2@ num>string		cat-log
	s"   in step "			cat-log
	num>string			0 log-it
    THEN

    initialize-mutation
    >r					\ addr-of-xt of the gene to mutate
    run-mode dup @ mutating or swap !		\ set run mode on mutating
    r@ @ (mutate)
    run-mode dup @ mutating invert and swap !	\ mutating off

    (new-gene) frame-insert-end-markings
    pop-frame

    \ copy cost to the nuc-var including cost of ;gene
    frame-cost @ code-cost !

    write-code-file  code-file-mask @  AND IF
	file-mutating  code-file-mask @  AND IF
	    file-code file-structure OR  code-file-mask @  AND IF
		code-file-write-line			\ add an empty line
	    THEN
	THEN
    THEN

    log-mask @ log-mutation and IF
	s" ==> new-gene:"		0 log
	(new-gene) buffer-as-string	0 log
	s" code-cost: "			cat-log
	code-cost @ num>string		0 log-it
    THEN

    (mutated) @ (mutated-max) @ > IF
	(mutated) @ (mutated-max) !

	[ log-mask @ ] [IF]
	    log-mask @ log-m-more and IF
		s" new maximal mutated gene items= "
		(mutated) @	0 log-string-and-number
	    THEN
	[THEN]
    THEN

    r>  (new-gene)  ( addr-to-store-in handle )  set-up-trial ;

: mutate? ( -- )			\ shall we mutate?
    mutation-rate cell+ @ IF
	mutation-rate rated-flag IF
	    wake-me-xt mutate

	    (mutated) @ IF	\ don't log if not mutated i.e. trial phase
		log-mask @ log-mutation and IF
		    s" MUTATION: step "		cat-log
		    spot 2@ swap num>string	cat-log
		    s" : nuc at spot "		cat-log
		    num>string			cat-log
		    s" 	  mutated to child at "	cat-log
		    child-spot @ num>string	cat-log
		    s"  as:"			0 log-it

		    wake-me-xt @ dup eb>sequence swap eb>length @  0 log

		    log-mask @ log-mutation and IF
			s" " 0 log			\ empty line
		    THEN
		THEN
	    THEN
	THEN
    THEN ;
defer <mutate>
' mutate? IS <mutate>

\ replace substrings *of equal length*
\ string1 string to work on	string2 substring to be replaced by string3
: replace ( addr1 u1  addr2 u2  addr3 u3 -- )
    [ debugging @ ] [IF]
	dup 3 pick = 0= IF
	    cr type bl emit type
	    true ABORT" replace: substrings have different length"
	THEN
    [THEN]

    drop >r			\ replacement start on return stack
    2swap			\ to be replaced string under
    BEGIN
	2over search WHILE
	r@ 2 pick 4 pick move	\ do a replacement
    REPEAT
    rdrop 2drop 2drop ;

\ some words in the trial string must be replaced before compiling the gene.
\ 'replacements' holds the xt's of words that do this replacement.
\ These words get the string to work on as ( addr u --)
LIST: replacements

: replace-[if][else][then] ( addr u -- )
    2dup  s" [IF]"   s"   IF"   replace
    2dup  s" [ELSE]" s"  ELSE " replace
	  s" [THEN]" s" THEN  " replace ;
' replace-[if][else][then] replacements >list

: do-replacements ( addr u -- )
    2>r
    replacements
    BEGIN
	next-node-and-value
    WHILE
	2r@ rot EXECUTE
    REPEAT
    2rdrop ;

\ how many generations is trial phase?
VARIABLE trial-phase	10 trial-phase !

\ how many genes have survived trial phase and have been compiled
VARIABLE compiled-genes	compiled-genes off

decimal
36 STRINGBUF-HANDLE: (get-xt)

\ compile genes on trial and put them into the gene pool
: end-trial-phase ( -- )
    log-cat-id
    s" end of trial phase " log-trial log-it

    wake-me-xt @ eb>compiled @ ?dup IF		\ buffer already compiled?
	nuc-flags dup @ wake-me-trial xor swap ! \ end trial phase
	wake-me-xt @ decrease-eb-count		\ decrease counter
	wake-me-xt !				\ set wake-me-xt
	s" xt set to compiled gene" log-trial log
	EXIT					\ nothing else to do
    THEN

    \ create internals word and compile the gene:
    current @					\ remember current wordlist
    also genes definitions previous
    [ decimal ] 4096 stringbuf-open >r
    new-gene-name r@ cat					\ name
    wake-me-xt @ dup eb>sequence swap eb>length @  r@ cat	\ and code
    r@ buffer-as-string compile-internals	\ create internals word
    r@ buffer-as-string				\ string with name and code
    dup 8 + stringbuf-open >r			\ hack: open a second buffer
    s" GENE: " r@ cat
    ( addr count ) r@ cat			\ name and code
    r@ buffer-as-string do-replacements
    r@ buffer-as-string EVALUATE		\ now the gene is compiled
    r> stringbuf-close
    r> stringbuf-close
    current !					\ restore current wordlist
    1 compiled-genes +!				\ bookkeeping

    (get-xt)	stringbuf-empty			\ we need it's xt:
    s" ' "	(get-xt) cat
    (name-buf) buffer-as-string (get-xt) cat

    also genes					\ we want the *genes* xt now
    (get-xt) buffer-as-string EVALUATE >r	\ get xt
    previous

    wake-me-xt @
    dup eb>compiled r@ swap !			\ mark buffer as compiled (xt)
    decrease-eb-count				\ decrease counter
    nuc-flags dup @ wake-me-trial xor swap !	\ end trial phase

    r> wake-me-xt !				\ and set wake-me-xt

\	last-gene-into-pool

    log-mask @ log-trial and IF
	s" compiled as: "		cat-log
	(name-buf) buffer-as-string	log-trial log-it
    THEN

    write-code-file  code-file-mask @  AND IF
	file-end-trial code-file-mask @ AND IF
	    s" \ survived trial phase:" cat>code-file code-file-write-line
	    (name-buf) buffer-as-string name>internal-xt xt-write-code
	THEN
    THEN ;
