\ stringbuf-0.3.fs
\ 	$Id: stringbuf-0.3.fs,v 1.21 2001/02/15 16:47:01 f Exp $	

\ allocated string buffers referred to by handles.
\ dynamic doubling of size and number of buffers when needed.

\ cat
\ char-cat

[UNDEFINED] OFFSET:  [IF]	include basics.fs 			[THEN]
[UNDEFINED] debugging [IF]	variable debugging	debugging off	[THEN]

0
OFFSET: >stringbuf-count   ( addr -- addr-of-count )
OFFSET: >stringbuf-pointer ( addr -- addr-of-allocated-memory )
OFFSET: >stringbuf-size    ( addr -- addr-of-size )
CONSTANT stringbuf-descriptor-items#

\ The descriptors belonging to the different string buffers are stored
\ in a string buffer itself.
\ This makes it easy to increase the number of buffers,
\ if more buffers are needed.

: stringbuf-descriptor-size# ( -- u )
    [ stringbuf-descriptor-items# cells ] literal ;

\ pointer to the buffer with descriptors field
\ the other item holds number of descriptors
2variable (stringbuf-descriptors)	0. (stringbuf-descriptors) 2!
\ (stringbuf-descriptors) 2@ ( -- allocated|false stringbufs )

: stringbufs ( -- addr ) (stringbuf-descriptors) cell+ ;

decimal 128  constant default-string-buffers#

: ?allocate-descriptors ( -- )
    (stringbuf-descriptors) @ 0= IF
	\ size will be dynamically increased, if needed.
	default-string-buffers# stringbuf-descriptor-size# * allocate
	ABORT" ?allocate-descriptors: Couldn't allocate. "
	(stringbuf-descriptors) !
	default-string-buffers# stringbufs !
    THEN ;

\  \ resizes an allocated area at addr of old-length to the double
\  \ and erases new upper half.
\  : double-allocated ( addr old-length -- addr ior )
\      dup >r
\      2* resize
\      dup 0= IF
\  	over r@ + r@ erase	\ erase new area
\      THEN rdrop ;

: double-descriptors ( -- )
    (stringbuf-descriptors) @				( addr )
    stringbufs @   stringbuf-descriptor-size# cells *	( addr old-size )
    double-allocated
    ABORT" double-descriptors: Couldn't resize."
    (stringbuf-descriptors) !
    stringbufs @  2*  stringbufs ! ;

: handle>stringbuf-descriptor ( u -- addr )

    [ debugging @ ] [if]
	-1  stringbufs @  2 pick  within		\ handle within limits?
	0= ABORT" Stringbuf handle out of range."
    [then]

    stringbuf-descriptor-size# *
    (stringbuf-descriptors) @ + ;

\ a word to double stringbuf size
: stringbuf-double ( descriptor-addr -- )
    dup >stringbuf-pointer @
    over >stringbuf-size @ 2* dup >r resize
    ABORT" stringbuf-double: Couldn't resize."
    over >stringbuf-pointer !
    >stringbuf-size r> swap ! ;

\ open a string buffer of initial size 'size' and give it's handle.
\ take care that the descriptor field is allocated and big enough.
: stringbuf-open ( size -- handle )
    4 max	\ savety net if size is zero... you can't double that...

    \ if the buffer for the descriptors is not initialized, do it. 
    ?allocate-descriptors
    \ size will be dynamically increased, if needed.

    0
    BEGIN
	dup handle>stringbuf-descriptor >stringbuf-pointer @
    WHILE	\ buffer address?
	1+
	dup stringbufs @  =
    UNTIL
	double-descriptors drop RECURSE
    ELSE
	over allocate
	ABORT" open-stringbuf: Couldn't allocate buffer."
	over handle>stringbuf-descriptor >r		\ start of descriptor
	r@ >stringbuf-count off				\ count off
	r@ >stringbuf-pointer !				\ store address
	swap
	r> >stringbuf-size !				\ store size 
    THEN ;

: stringbuf-close ( handle -- )
    handle>stringbuf-descriptor >stringbuf-pointer
    dup @ 0= ABORT" stringbuf-close: Buffer not opened."
    dup @ free
    ABORT" stringbuf-close: Couldn't free buffer."
    off ;

: stringbuf-empty ( handle -- )
    handle>stringbuf-descriptor >stringbuf-count off ;

: buffer-pointer ( handle -- addr )	\ you can do 2@ on this
    handle>stringbuf-descriptor >stringbuf-count ;

: buffered-length ( handle -- u ) buffer-pointer @ ;

: buffer-as-string ( handle -- addr count )   buffer-pointer 2@ ;

\  concatenation with automatic doubling of size, if needed
: cat ( addr count handle -- )
    >r

    \ make sure size is big enough:
    BEGIN				  ( addr count )
	r@ handle>stringbuf-descriptor	  ( addr count descriptor-addr )
	dup >stringbuf-count @  2 pick  + ( addr count descrpt-addr new-count )
	over >stringbuf-size @ > WHILE	  ( addr count descriptor-addr )
	stringbuf-double		  ( addr count )
    REPEAT drop

    \ preparing arguments to 'move'	( from-addr count )
    r@ buffer-as-string +		( from-addr count to-addr )
    swap  dup >r			( from-addr to-addr count )
    move
    r>					( count )
    r> handle>stringbuf-descriptor >stringbuf-count +! ;


variable (scratch)	\ don't want to rely on r@, thats why i use (scratch)

: char-cat ( c handle -- )
    >r
    (scratch) c!
    (scratch) 1 r> cat ;

\ Word to create a named stringbuf handle with automatic
\ opening of a buffer when called for the first time.
\ The created word puts the handle on stack.
: STRINGBUF-HANDLE: ( size -- handle )
    CREATE
	-1 ,			\ -1 means not initialized
	,			\ store initial size
    DOES> ( -- handle )
	dup @ -1 = IF		\ if no buffer opened, do that
	    dup cell+ @ stringbuf-open
	    over !
	THEN

	\ double check if buffer is allocated, as in very rare cases it wasn't
	dup @ handle>stringbuf-descriptor >stringbuf-pointer @ 0= IF
	    dup cell+ @ stringbuf-open
	    over !
	THEN

	@ ; ( handle )

: close-named-handle ( xt -- )
    dup execute stringbuf-close
    >body -1 swap ! ;

: n>buf ( n handle -- )
    (scratch) !
    (scratch) cell r> cat ;

\ testing a little bit:
false [if]
    cr .( Testing string buffers, cat and char-cat: ) cr
    
    variable smallbuf
    1 stringbuf-open smallbuf !
    
    s" Hallo!" smallbuf @ cat
    smallbuf @ buffer-as-string cr type
    s"  Here I am ;-)" smallbuf @ cat
    smallbuf @ handle>stringbuf-descriptor stringbuf-double
    smallbuf @ buffer-as-string cr type

    smallbuf @ buffer-pointer 2@ cr type
    
    smallbuf @ stringbuf-empty
    cr .( Gone after stringbuf-empty: )
    smallbuf @ buffer-as-string type


    char " smallbuf @ char-cat
    s" Here again..." smallbuf @ cat
    char " smallbuf @ char-cat
    smallbuf @ buffer-as-string cr type

    variable second-buf
    16 stringbuf-open second-buf !
    s"  a second time." second-buf @ cat
    second-buf @ buffer-as-string smallbuf @ cat
    smallbuf @ buffer-as-string cr type
    
    16 STRINGBUF-HANDLE: wow
    wow cr .
    ' wow close-named-handle

    smallbuf @ stringbuf-close
    second-buf @ stringbuf-close

    cr cr .( And this should fail:) cr
    99 stringbuf-close

    cr cr bye
[then]

\ hmm, this buf's are handy, but...

\ mutation is slowed down a lot using the handled buffers for it
\ i need a simplified version and tried it out in symbols-stack.fs, it helped.

\ i want a simplified, fast set of words, without handles, length checking etc.
\ i could reintroduce named buffers and run it on it's own,
\ but as i use named buffers anyway, i can also base it on them,
\ giving the address of the double cell containing count and address.
\ hmm, so i could use them on their own anyway...	we'll see.

\ naming: prefix 's-' like 'simple', or 'symbols', too.

\ descriptor:	stack-pointer-as-offset  stack-bottom

\ Attention: because of the IMMEDIATE statements
\ some s-buf words are *compile only* if compiled with debugging off!

debugging @ [IF]
: s-buf-count ( addr-of-descriptor -- count )   @ ;
[ELSE]
: s-buf-count ( addr-of-descriptor -- count )   POSTPONE @ ; IMMEDIATE
[THEN]

: s-buf>addr ( addr-of-descriptor -- addr )   cell+ @ ;

: s-buf>size ( addr-of-descriptor -- size )   [ 2 cells ] literal + @ ;

debugging @ [IF]
: s-buf>string ( addr-of-descriptor -- addr count )   2@ ;
[ELSE]
: s-buf>string ( addr-of-descriptor -- addr count )   POSTPONE 2@ ; IMMEDIATE
[THEN]

debugging @ [IF]
: s-buf-clear ( addr-of-descriptor -- )   off ;
[ELSE]
: s-buf-clear ( addr-of-descriptor -- )   POSTPONE off ; IMMEDIATE
[THEN]

\ remember: it's your job to make sure the buffer is allocated and big enough!
\ this version does *not* do any size or allocation control. 
: s-buf-cat ( addr count addr-of-descriptor )
    >r
    r@ 2@ + swap  dup >r  move
    r> r>  +! ;

: s-buf-char-cat ( c addr-of-descriptor -- )
    >r
    (scratch) c!
    (scratch) 1 r> s-buf-cat ;

\ why not having named s-bufs?
\ calling them by name *does* allocation if needed and gives the descriptor
\ (the descriptor is at body address)
\ after that the pointer can be kept without calling the name again.
\ or you can work directly on the string, if you want to
: S-BUF: ( fixed-size -- )
    CREATE
	0 ,		\ descriptor: count
	0 ,		\ not allocated
	,		\ size
    DOES> ( -- addr-of-descriptor )
	dup cell+ @	IF EXIT THEN	\ done, if already allocated
	>r		( r: body )
	r@ cell+ cell+ @ allocate
	ABORT" named s-buf couldn't allocate."
	r@ cell+ !		\ allocated address
	r@ off			\ count off
	r> ;	( descriptor-address=body )

\ always close s-bufs before savesystem
: s-buf-close ( addr-of-descriptor -- )
    dup cell+ @  ?dup IF
	free ABORT" s-buf-close: Couldn't free."
	>r 0. r> 2!
    ELSE
	drop
    THEN ;

false [if]	cr .( testing s-buf's...)
    debugging @ 0= [if]
	cr bell
	cr .( ATTENTION: because of the IMMEDIATE statements)
	cr .( some s-buf words are *compile only* if compiled with debugging off.)
	cr .( turn it on before compiling!)
	cr
    [then]

    &80 S-BUF: s-test

: t  s-test s" |<-- count= " type s-buf-count . ;
    cr .( opened:) 
    s-test s-buf>string cr type t
    s" s-buf-cat: first" s-test s-buf-cat
    s-test s-buf>string cr type t
    s"  second" s-test s-buf-cat
    s-test s-buf>string cr type t
    s-test s-buf-close cr .( closed:) cr t
    cr bye
[then]
