
\ now we need a playground for our beeings: space and time

\ This version can cope with several worlds each one having her own
\ set of variables like dimensions, time-planes, qualities, etc.

\ the actual world is pointed to by a variable:
variable world
world off

\ we need a word to create world's variables:
: world-var ( n -- n+1 )
    CREATE dup cells , 1+
    DOES> @ world @ + ;


\ ok, lets build the blueprint for worlds:
5 constant max-dimensions		\ seems plenty
0					\ counter must be on stack
\ if we want to save worlds it's convenient to have the length right here:
world-var world-length		\ total length including this variables, bytes
world-var w-version		\ for saving, too
world-var w-max-dimensions	\ 
world-var dimensions   		\ this worlds dimension count
world-var spots-this-dimension	\ depth, an entry for each dimension
max-dimensions 1- +	\ world-var 2nd-dimension-spots
			\ world-var 3rd-dimension-spots
			\ world-var 4th-dimension-spots
			\ world-var 5th-dimension-spots
world-var time-planes		\ past-planes (for undoing) ... present, future
world-var list-planes		\ 1 plane pointers to the nuc's, and qualities
				\ (how many of them)
world-var step			\ current time (counter)
world-var spot			\ might be convenient on rare occasions

\ the following variables can be calculated from the previous data, but it's
\ convenient to store them here:
world-var plane-length          \ bytes for one plane of reality ( spots cells)
world-var time-plane-length     \ all diffrent reality planes of one moment
world-var total-list-length     \ for all times (but excluding these var's)
world-var spots                 \ how many spots in the universe
\ store the length of this field that every world starts with
cells CONSTANT world-variables-length

\ Many of these field variables have a corresponding 'normal' variable
\ which is used while in a certain world for speed reasons.
\ If you enter or leave a world these variables must be copied into
\ each other which is done by 'enter-world' and 'leave-world'.

\ The offsets to reach the different planes get initialized by the same time.
0 VALUE offset-cell-pointers
0 VALUE offset-food
0 VALUE offset-A-quality
0 VALUE offset-B-quality
0 VALUE offset-C-quality
0 VALUE world-data-start

: enter-world ( addr -- )
    world !
    world-variables-length	TO world-data-start
    \ The offsets to the different planes get stored as values
    \ (On my system values are faster than 'variable @')
    world-variables-length	dup TO offset-cell-pointers
    plane-length @ +		dup TO offset-food
    plane-length @ +		dup TO offset-A-quality
    plane-length @ +		dup TO offset-B-quality
    plane-length @ +		dup TO offset-C-quality
    drop

;

\ preparing the big bang:	\ not very FORTH like internal variables...
variable (dimensions)
CREATE (dim-spots)		\ field for depth of each dimension
5 cells allot			\ 5 dimensions seem enough ;-)
DOES> ( -- addr ) swap cells + ;
variable (time-planes)
variable (list-planes)


: (big-bang) ( list-planes time-planes dim-sizeN ... dim-size1 dimensions -- )
    \ calculate size, temporarily storing intermediate results:
    dup (dimensions) !
    0 DO				\ loops over dimensions
	i (dim-spots) !			\ stores # of spots of the dimension
    LOOP
    (time-planes) !
    (list-planes) !
    
    \ now calculate:
    1
    (dimensions) @ 0 DO  i (dim-spots) @ *  LOOP
    dup >r
    cells
    (list-planes) @ *		dup >r
    (time-planes) @ *		dup >r
    world-variables-length +	dup >r
    allocate ( -- adr ior ) 0= IF
	dup world !
	r@ erase
	r> world-length !
	r> total-list-length !
	r> time-plane-length !
	r@ cells plane-length !
	r> spots !
	(dimensions) @  dup dimensions !
	0 DO
	    i (dim-spots) @
	    i cells  spots-this-dimension + !
	LOOP
	(time-planes) @  time-planes !	\ # of different times: now, later
	(list-planes) @  list-planes !	\ # of different planes af the world
					\ like cp or food

    ELSE rdrop rdrop rdrop rdrop
	drop   false world !		\ uups, no space there?
	bell   true ABORT" Big bang failure. nothing to do..."
    THEN ;

false [if]	cr .( a small big bang test ;-)
\      3  2 80 1 dbg (big-bang)
    5  2 24 80 2 dbg (big-bang)
\ 5 2 24 24 80 3 dbg (big-bang)

cr  ." world-length     "	world-length      @ .
cr  ." plane-length     "	plane-length      @ .
cr  ." time-plane-length"	time-plane-length @ .
cr  ." total-list-length"	total-list-length @ .
cr  ." spots            "	spots             @ .
cr
key drop
[then]
false [if]	cr .( timing test)
: t-var
    8 0 DO
    bell
    1000000 0 DO
	(list-planes) drop
	(list-planes) drop
	(list-planes) drop
	(list-planes) drop
    LOOP LOOP bell ;
: t-fieldvar
    8 0 DO
    bell
    1000000 0 DO
	list-planes drop
	list-planes drop
	list-planes drop
	list-planes drop
    LOOP LOOP bell ;
t-var  t-fieldvar
[then]

\ this artificial world has several planes of reality:
\ the first one is always for the cp's
\ the others are variables like food or so.

\ the 'spot' is an index in thouse fields

\ there are time-planes structures like this for the different times:
\ one for now, one for the next time, ...

\ the idea is to work from present to future:
\ if something is changing, it changes in the future plain
\ a time-step makes it appear, then
\ time-step is switching a cyclic pointer

\ spot >spot! initializes the connections with the different planes


: big-bang
    depth ?dup IF
	>r
	?dup 0= IF
	    rdrop true
	ELSE
	    dup 4 + r> < IF
		false
	    ELSE
		true
	    THEN
	THEN
    ELSE true
    THEN

    IF
	." input"
    THEN
    
    ." unfertig" ; \ interactive version
quit

\ now let's create a world:


variable spot					\ list index, determing the
						\ actual spot
variable spot-offset				\ same, as offset in bytes

\ time:
variable step
-1 step !				\ actual time
variable time-index	-1 time-index !	\ cyclic index of the present
					\ *ONLY* time-step should change it
					\ for internal use *ONLY* ( tp )

: time-step ( -- )			\ step from present to future
    time-index @ 1+ time-planes mod time-index !  \ time as a cycle, dizzy?
    1 step +! ;
time-step				\ start the time

\ how to find the other times:
variable time-offset			\ internal use only
: present ( -- )   0 time-offset ! ;	\ makes the present be the time seen
present					\ initialize to the present
: future  ( -- )   1 time-offset ! ;	\ makes the future be the time seen

false [IF]
\ tp = time pointer points to the actual time-plane
CREATE tp				\ interpreting!
\ we're going to construct the offsets in an interpreting loop
time-planes 0 [DO] time-plane-length [i] * world @ + , [LOOP]
DOES> ( -- adr )
    @					\ pointer to first entry
    time-index @ time-offset @ +	\ index
    time-planes mod			\ normalized index
    cells + @ ;				\ returns adress of actual time-plane

[ELSE]

\ .( compiling provisoric version of tp) cr \ ##################

: tp ( -- adr )
    time-index @ time-offset @ +
    time-planes mod
    time-plane-length *
    world-data-start + ;

[THEN]

: present-initializes-future ( -- )	\ initializes the future plain
    present tp				\ with the data from the present
    future  tp  time-plane-length move	\ with the data from the present
    present ;				\ stay tuned to the present

\ variable world?	world? off	\ list or individual?
\ world? was compiled earlier

\ fcp = 'field cell pointer'
\ word returning the adress in the field, where the pointer to the cell is.
\ If it's an individual, cp is returned for convenience, (but I don't know,
\ if I'll ever use this).
\ hmm, yes and it makes troubles adding things to cp! #######
: fcp ( -- adr-of-the-pointer-to-the-nuc )	\ 'field cell pointer'
    world? @ IF				\ pointer in a field
	tp offset-cell-pointers +	\ start of the pointer plane
	spot-offset @ +			\ go to the right spot
    ELSE cp THEN ;			\ special case: individual


\ let the spot have another quality:
variable single-A-quality	\ A-quality for individuals
single-A-quality off	\ no 'A' quality for singles?	no, I'm just kidding!
: A-quality ( -- adr)
    world? @ IF
	tp offset-A-quality +		\ A-quality plane
	spot-offset @ +			\ at the right spot
    ELSE
	single-A-quality
    THEN ;

\ wanted to have more qualities for experiments with symbiotic biotops:
variable single-B-quality	\ B-quality for individuals
single-B-quality off
: B-quality ( -- adr)
    world? @ IF
	tp offset-B-quality +		\ B-quality plane
	spot-offset @ +			\ at the right spot
    ELSE
	single-B-quality
    THEN ;

variable single-C-quality	\ C-quality for individuals
single-C-quality off
: C-quality ( -- adr)
    world? @ IF
	tp offset-C-quality +		\ C-quality plane
	spot-offset @ +			\ at the right spot
    ELSE
	single-C-quality
    THEN ;


: >spot! ( i -- )			\ sets spot and spot-offset
    dup spot !
    cells spot-offset ! ;
0 >spot!				\ initialize it

: erase-list ( -- )			\ guess what it does ;-)
    world @ IF				\ for convenience
	world @ list-length erase	\ clean up the whole universe
	step off			\ hmm? not really sure about that
	living off			\ programmers aesthetics
    ELSE ." there's no world to clean up!" cr THEN ;	\ looks like an error

: someone-here? ( i -- adr|false)	\ checks if a spot is occupied
    spot @ swap	( spot i )		\ remember spot
    >spot! fcp @  ( spot a|false )	\ return value on TOS
    swap >spot! ;			\ go back to actual spot

\ set list element i to a cloned nuc of the actual one
: set-list-element ( i -- adr=flag )	\ returns adr of the clone or 0
    dup someone-here?			\ is it free?
    IF drop false			\ no, sorry
    ELSE				\ ( i )
	spot @ >r    >spot!		\ ( --   R:old-cp@, spot)
	clone				\ ( adr flag | false  R:old-cp@, spot)
	IF dup fcp !			\ cloning was possible
	ELSE false THEN			\ cloning didn't work
	r> >spot!
    THEN ;

: nuc-unlink ( -- )			\ remove actual cell from the world
    world? @ IF				\ this trick makes it work as a noop
	future				\ if it's called with an individual
	fcp off
	present
	fcp off		\ hmm, I'm not sure about this... ######
    THEN ;

' nuc-unlink is <disappear>		\ unlinks aborted cells from the list

: free-spots ( -- n )			\ how many spots are not inhabited?
    0					\ counter
    spots 0 DO				\ loop over all spots
	i someone-here? 0= IF 1+ THEN	\ count
    LOOP ;

: free-list ( -- )			\ lets everybody die and cleans up
    spots 0 DO				\ everywhere
	i >spot!  fcp @			\ who's here?
	?dup IF				\ somebody?
	    cp!				\ set as actual nuc
	    die				\ sorry
	THEN
    LOOP
    erase-list				\ clean everything up
    living off    false cp! ;		\ programmers aesthetics

\ orientation in space
: i-normalize ( i offset -- i' )	\ normalizes the indizes
    + spots mod ;			\ this universe wraps at it's borders

: left-spot  ( i -- i' )   -1 i-normalize ;	\ one step to the left
: right-spot ( i -- i' )    1 i-normalize ;	\ one step to the right
: upper-spot ( i -- i' )    c-l negate i-normalize ;	\ one step up
: lower-spot ( i -- i' )    c-l        i-normalize ;	\ one step down

true [IF]				\ 4 or 8 directions?

8 CONSTANT directions			\ we have 8 directions
: direction>index ( direction -- i')	\ spot must be set
    directions mod			\ ( normalized-direction )
    spot @ swap				\ ( actual-index normalized-direction )
    CASE
	0 OF upper-spot			ENDOF
	1 OF upper-spot	right-spot	ENDOF
	2 OF right-spot			ENDOF
	3 OF right-spot	lower-spot	ENDOF
	4 OF lower-spot			ENDOF
	5 OF lower-spot	left-spot	ENDOF
	6 OF left-spot			ENDOF
	     left-spot	upper-spot
    ENDCASE ;

[ELSE]
4 CONSTANT directions			\ we have only 4 directions right now
: direction>index ( direction -- i')	\ spot must be set
    directions mod			\ ( normalized-direction )
    spot @ swap				\ ( actual-index normalized-direction )
    dup 0=  IF drop upper-spot	EXIT THEN
    dup 1 = IF drop right-spot	EXIT THEN
    dup 2 = IF drop lower-spot	EXIT THEN
	3 = IF      left-spot        THEN ;
[THEN]

: free-neighbor-spot? ( -- i' true | false )	\ free neighboring spot?
    directions random-ranged >r			\ ( R: random-direction )
    0						\ ( z=0 )
    BEGIN					\ ( z ) test all directions
	dup r@ +				\ ( z z+random-direction )
	direction>index				\ ( z i' )
	dup someone-here?			\ ( z i' flag )
	0= IF					\ ( z i'ok )
	    rdrop nip true EXIT			\ RETURN ( -- i' true )
	THEN drop				\ ( z )
	1+ dup directions < WHILE		\ ( z+1 )
    REPEAT drop
    rdrop false ;				\ RETURN ( -- false )

: sow ( n -- n' )	\ sow n clones of the actual nuc into the field
\ like in real life sow does *not* insist if the spot is not ok
\ it sows n seeds, and dosn't care, if they do not fall on fertile ground
\ it returns n', the number of succesfully sowed nucs
    world @ 0= IF
	." sorry, no gardening before the big bang " cr	drop EXIT
    THEN
    dup >r				\ ( R: n )
    world? on
    future				\ we're working always for the future
    dup 0 DO			( n )		\ try it n times, no more
	spots random-ranged	( n i )		\ random spot
	set-list-element	( n flag )	\ set if possible
	IF 1- THEN		( n' )		\ count decreasing n
    LOOP			( n' )
    present				\ back in the present
    r> swap - ;				\ calculate success number

: .world ( -- )
    world @ IF				\ check if there's something
	world? @ >r			\ save world?
	world? on			\ could be needed
	0 0 at-xy			\ puts cursor in the upper left corner
	spots 0 DO			\ for each and every spot in this world
	    i dup >spot!		\ look at the spot 
	    someone-here? ?dup IF	\ if there's someone
		cp!			\ make it the actual one
		show-me-xt @ EXECUTE	\ and show his face
	    ELSE
		show-background		\ empty
	    THEN
	LOOP
	r> world? !			\ restore world?
    ELSE ." there's nothing to look at" cr THEN ;    
