\ gene-pool.fs
\ 	$Id: gene-pool.fs,v 1.6 2001/02/18 19:04:29 f Exp $	

\ List of (internals) gene xt's and relative probabilities to be selected.

\  Hope 'require' is defined.  It's only for testing, comment it out.
\  require system-dependent.fs
\  require basics.fs
\  require lists.fs
\  require random.fs


\ Defining node structure
\ Includes redundant data for efficiency of random gene selecting.
0
\ Summed up probabilities of all prior nodes including this one
\ Any word that changed probabilities must recompute it.
OFFSET: >selection-limit	\ summed up probability, redundant
OFFSET: >probability		\ relative probability
OFFSET: >xt			\ internals xt
\ OFFSET: >pooled-flags		\ we could copy gene flags here
constant pool-node-items#

: make-pool ( -- addr )   pool-node-items# deflist ;

: GENE-POOL: ( <name>   -- )   make-pool CONSTANT ;

GENE-POOL: gene-primitives

\ pool descriptor flags
0 MASK: dirty
drop

: dirty? ( pool -- flag )   list>flags @ dirty and 0<> ;

: dirty! ( pool -- )   list>flags dup @ dirty or swap ! ;

\ Recompute redundant data in all nodes, if needed
\ Check for negative probabilities and fix them.
: ?pool-setup ( pool -- )
    dup dirty? 0= IF drop EXIT THEN

    0 swap			( sum=0 pool )
    dup nodes 0 ?DO
	next-node >r		( sum		r: actual-node )
	r@ >probability @
	dup 0< IF drop 0  r@ >probability off THEN	\ probability positive
	+			( actual-sum	r: actual-node )
	dup r@ >selection-limit !
	r>
    LOOP
    2drop ;

\ Put a new xt that's not there already in the pool
: new-xt-to-pool ( xt probability pool -- )
    dup dirty!
    new-node >r
    r@ >probability !
    r> >xt ! ;

\ Is the xt already in the pool?
: xt-in-pool? ( xt pool -- node-addr|0 )
    dup nodes 0 ?DO
	next-node >r	( xt  r: actual-node )
	dup r@ >xt @ = IF  drop r> unloop EXIT THEN
	r>
    LOOP
    2drop FALSE ;

\ Set probability of an xt (inserting xt's if they are new).
: set-xt-probability ( xt probability pool -- )
    >r			( xt probability    r: pool )
    over r@ xt-in-pool?	( xt probability node|0 )
    dup IF
	>probability !
	drop
	r> dirty!
    ELSE drop
	r> new-xt-to-pool
    THEN ;

\ Change probability of a existing or new xt.  New xt's get added.
\ (a resulting negative probability would be fixed by '?pool-setup')
: change-xt-probability ( xt probability pool -- )
    >r			( xt probability    r: pool )
    over r@ xt-in-pool?	( xt probability node|0 )
    dup IF
	>probability +!
	drop
	r> dirty!
    ELSE drop
	r> new-xt-to-pool
    THEN ;

\ Zero all probabilities of a pool.
: (nul-prob) ( node -- )   >probability off ;
: nul-all-probabilities ( pool -- )   >r ['] (nul-prob) r> do-with-all-nodes ;

: probability? ( xt pool -- probability )
    xt-in-pool? dup IF
	>probability @
    THEN ;

: random-pool-gene ( pool -- xt )
    dup nodes 1- over n'th-node >selection-limit @ ( pool top-selection-limit )
    dup 0= ABORT" random-pool-gene: No probabilities set."
    random-ranged >r			( pool   r: selected )
    BEGIN				( actual-node   r: selected )
	next-node
	r@  over >selection-limit @ <
    UNTIL
    rdrop
    >xt @

    [ log-mask @ 0> ] [IF]	\ conditional compiling
	log-mask @ log-m-extra and IF
	    s" random-pool-gene picked: "	cat-log
	    dup xt>name				0 log-it
	THEN
    [THEN]
;

LIST: gene-pools

\ analyse input and output stack requirements of a gene pool:
: count-chars ( addr-of-result-array addr-string count-string factor -- )
    swap		( addr-of-result-array addr-string factor count-string)

    0 ?DO		( addr-of-result-array addr-string factor )
	over i + c@	( addr-of-result-array addr-string factor char )
	cells 3 pick + >r dup r> +!
    LOOP drop 2drop ;

: pool-stack-type-scan ( pool -- pool-stackdata )
    [ decimal 256 2* cells ] literal >r r@ allocate
    ABORT" pool-stack-type-scan:  Couldn't allocate."
    dup r> erase
    swap ( pool-stackdata pool )
    dup nodes 0 ?DO
	next-node
	dup >probability @ dup IF	\ skip probability zero
	    >r			( pool-stackdata actual-node  r: probability )
	    2dup	    >xt @ >body gene-body>stack-in-string
	    r@ count-chars
	    over [ 256 cells ] literal +
	    over >xt @ >body gene-body>stack-out-string
	    r> count-chars
	ELSE drop THEN
    LOOP drop ;

: .pool-stackdata ( pool -- )
    pool-stack-type-scan	( pool-stackdata )
    page
    [ decimal ] 256 0 DO	( pool-stackdata )
	dup i cells + dup @
	swap [ 256 cells ] literal + @	( pool-stackdata in out )
	2dup OR IF
	    swap
	    cr ." stack data type: " i emit
	    1 3 screen-column ." INPUT: " .
	    2 3 screen-column ." OUTPUT: " .
	ELSE 2drop THEN
    LOOP cr
    free ABORT" .pool-stackdata: Couldn't free memory." ;

\ : ttt actual-genepool-xt @ execute .pool-stackdata ;

false [if] \ testing

    ' noop gene-primitives xt-in-pool? cr .
    ' noop 1 gene-primitives change-xt-probability
    ' noop gene-primitives probability? cr . 
    ' noop gene-primitives xt-in-pool? >xt @ xt>name cr type
    ' swap 2 gene-primitives change-xt-probability

    ' ! 1 gene-primitives change-xt-probability
    ' @ 2 gene-primitives change-xt-probability
    ' - 3 gene-primitives change-xt-probability
    ' + 4 gene-primitives change-xt-probability
    ' . 5 gene-primitives change-xt-probability

    \ gene-primitives nul-all-probabilities

    gene-primitives ?pool-setup

   : t   1000 0 DO gene-primitives random-pool-gene xt>name type bl emit LOOP ;
   cr t bye

[then]
