\ brew.fs			started: in the beginning of march 2000

\ 	$Id: brew.fs,v 1.100 2001/02/15 21:48:14 f Exp $	

page
cr .( ==> 'Cells' alias 'brew' <== )
cr .( 'The Evolutionary Programmers Playground' )
cr

\ this comes *before* compile time options.
include system-dependent.fs

\ compile time options:
' include catch compile-options.fs
?dup [IF]
    -38 = [IF]
	.( compile-options.fs not present. ) cr
    [ELSE]
	7 emit .( Error loading compile-options.fs ) . cr
    [THEN]
[THEN]

\ private options
' include catch my-compile-options.fs
?dup [IF]
    -38 <> [IF]
	7 emit .( Error loading my-compile-options.fs ) . cr
    [THEN]
[THEN]

[UNDEFINED] support-single-individuals? [IF]
    FALSE CONSTANT support-single-individuals?
[THEN]

VOCABULARY brew-words
also brew-words

include basics.fs
include key-mapping.fs

\ simple key buffering:
[UNDEFINED] push-key [IF]	include keybuf.fs 			[THEN]

include lists.fs 

[UNDEFINED] at? [IF]		include console-codes.fs 		[THEN]
\ get cursor position, colors.		\ '[UNDEFINED] at?' could go wrong...
[UNDEFINED] cursor-visible [IF]	include console-codes.fs 		[THEN]
[UNDEFINED] black [IF]		include console-codes.fs 		[THEN]

[UNDEFINED] color-list [IF]
    LIST: color-list
    ' black color-list >list
    ' red color-list >list
    ' green color-list >list
    ' brown color-list >list
    ' blue color-list >list
    ' magenta color-list >list
    ' cyan color-list >list
    ' white color-list >list
    ' default-color color-list >list
[THEN]

\ screen size
[UNDEFINED] c-l [IF]		include screen-size.fs 			[THEN]

include stringbuf-0.3.fs 	\ allocated string buffers, 'cat' strings
[UNDEFINED] (scratch-buf) [IF]
    decimal 128 STRINGBUF-HANDLE: (scratch-buf)
[THEN]

\ doing some very basic definitions first makes life easier...
include brew-basics.fs

include display.fs

include debuging.fs 		\ debugging help

brew-words definitions
include reporting.fs 		\ writing logs to the hd

also FORTH definitions previous

include mouse.fs 		\ faked mouse interface ##########
include menu.fs 		\ user interaction interface

definitions

include random.fs 		\ random functions
include genes-0.2.fs

\ include mutation-0.2.fs 	\ later on...
include statistics.fs
decimal

: goodbye ( -- )   cursor-visible bye ;

\ and now the real stuff:

0 VALUE cp			\ values are faster than variables
\ VARIABLE cp		cp off	\ 'cell pointer' points to the actual cell
\ the 'cell pointer' variable cp is very basic:
\ it's value is the base address of the actual nucleus
\ that means, that a whole group of operations doing something with a cell
\ will work on the one cp@ points to
\ it makes the nuc pointed to to the center of attention
\ i mean center of attention for nuc based operations

\ : cp@ ( -- addr-of-actual-cell )  cp ;
: cp@ ( -- addr-of-actual-cell )  POSTPONE cp ;  IMMEDIATE
: |cp@| ( -- addr-of-actual-cell )  cp ;	\ use this when interpreting
\ : cp@ ( -- addr-of-actual-cell )  cp @ ;
\ cp@ gives the address of the cells nucleus
\ each cell is represented by it's nucleus, containing his genes.
\ each nucleus starts with the nuc. The nuc is 'The Hard Core' of each nucleus
\ the structure of the nuc is the same for all nuclei
\ it contains nuc-var's and execution tokens for different actions
\ after the nuc there can be a structure containing other genes

: cp! ( addr -- )	to cp ;
\ : cp! ( addr -- )	cp ! ;
\ storing the address of a nuc in cp makes it the actual one
\ using cp! leaves the option to add other things here

: NUC-VAR: ( n -- n+1 )		\ creates a named variable in the nuc
    CREATE dup cells , 1+
    DOES> @ cp@ + ;

\ ****************************************************************
\  Changed nuc structure:
\
\  Define named variables: 'define-nuc-organs' 'define-nuc-parameters'
\  action pointers first: (see source)
\  genes
\  CONSTANT nuc-genes-limit	nuc indizes
\  CONSTANT nuc-xt's
\  xt's
\  CONSTANT nuc-xt-limit
\  CONSTANT nuc-variables
\  nuc intern data
\  CONSTANT nuc-diversificable-area
\  CONSTANT nuc-organs       \ index of first organ
\  named organs configurable by VALUE nuc-organs#
\  the cells can read and write them
\  CONSTANT nuc-parameters   \ base index
\  named parameters configurable by VALUE nuc-parameters#
\  the cells can read but not change them
\  0 CONSTANT nuc-invisible#     \ planed invisible, diversified
\  ( # ) CONSTANT nuc-diversificable-items
\  0 CONSTANT nuc-secrets#       \ planed invisible, not diversified
\  others			\ planed
\  cells CONSTANT nuc-length#
\ ****************************************************************

0				\ counter, must be on stack at compile time

\ action pointers first		\ each one containing an xt
\ genes:
\ 'wake-me-xt' holds a gene.  It's the genome of the cell.  It get's mutated.
NUC-VAR: wake-me-xt		\ main entry point: do a life action. genome
dup CONSTANT nuc-genes-limit	\ index limit

\ these hold xt's of actions that can't be mutated:
dup CONSTANT nuc-xt's
NUC-VAR: eat-xt			\ the task, the genome should solve
NUC-VAR: reproduce-xt		\ how to reproduce
NUC-VAR: show-me-xt		\ how to show up in world view (spot display)
dup CONSTANT nuc-xt-limit	\ index limit

\ some variables
dup CONSTANT nuc-variables	\ base index of cell sized nuc variables
NUC-VAR: length			\ holds length of whole nucleus in bytes
NUC-VAR: nuc-flags
0 MASK: wake-me-trial		\ wake me action is freshly mutated, on trial
  MASK: nuc-is-word		\ nucleus is in dictionary
				\ must be cleared when cloning a nuc.
drop
NUC-VAR: nuc-supplements		\ how many cells data follow the nuc
NUC-VAR: id			\ big brother is watching
NUC-VAR: genome-id		\ big brother is watching
NUC-VAR: age ( -- addr )		\ how many times was the cell awakened
NUC-VAR: generation ( -- addr )	\ counts generations since start of breeding
NUC-VAR: genom-generation	\ counts generations since last mutation
NUC-VAR: code-cost		\ sum of code cost of all subgenes
NUC-VAR: energy			\ storage of energy, food or whatever
\ 'energy' can only be stored or moved, transformed, used, or even be lost,
\ but never be produced by any action of a cell.
\ it is something like a 'real' stuff.
\ this reality only applies to the cells,
\ the universe can produce any reallity it happens to enjoy ;-)
\ (and you can break this rule if you feel like ;-)

\ hmm, we could make the (over) next 32 cells subject to diversification
\ and storing a mask which addresses to diversify in the cell called
\ my-diversification-bitmask
\ (even this one could be diversified xor'ing a few bits.)
\ (there can be more than one bitmask for different diversification strategies)
\
\ take care to limit the range of addresses that are diversified...

\ here we have negative offsets to nuc-diversificable-area
\ it's a good place for nuc vars not commonly but possibly diversified.
NUC-VAR: reproduction-threshold	\ how much energy to trigger reproduction?
NUC-VAR: age-threshold ( --addr)	\ does't get older than this
NUC-VAR: appearance		\ how it looks
NUC-VAR: my-diversification-bitmask

dup CONSTANT nuc-diversificable-area	\ base index

\ some named organs
\ the most archaic 'organs' are part of the nuc
\ these 'organs' can be used in any way by wake-me
\ they are hereditary, get passed on to the children
\ they can be made subject to diversification
dup CONSTANT nuc-organs	\ index of first organ
\ 7 VALUE nuc-organs#	\ compile time VALUE

TRUE [IF] \ new stile defining nuc-organs# named organs
    nuc-organs# +

\ I define them as named words
\ It seems more efficient and more readable
: define-nuc-organs ( -- )
    nuc-organs		( base-index )
    nuc-organs# 0 ?DO	( current-nuc-var-index )
	(scratch-buf) stringbuf-empty
	s" NUC-VAR: organ-" 	(scratch-buf) cat
	i [char] A +		(scratch-buf) char-cat
	(scratch-buf) buffer-as-string EVALUATE
    LOOP drop ;

define-nuc-organs
    
[ELSE] \ old stile, defining seven named organs
    
NUC-VAR: organ-A
NUC-VAR: organ-B
NUC-VAR: organ-C
NUC-VAR: organ-D
NUC-VAR: organ-E
NUC-VAR: organ-F
NUC-VAR: organ-G

[THEN]

dup CONSTANT nuc-parameters	\ base index of nuc-parameters
nuc-parameters# [IF] \ experimental. work in progress...
    \ read only variant of organs:

    TRUE [IF] \ new stile, defining nuc-parameters# parameters.
	nuc-parameters# +

	: define-nuc-parameters ( -- )
	    nuc-parameters		( base-parameter-index )
	    nuc-parameters# 0 ?DO	( current-nuc-var-index )
		(scratch-buf) stringbuf-empty
		s" nuc-var nuc-parameter-" (scratch-buf) cat
		i [char] A +		   (scratch-buf) char-cat
		(scratch-buf) buffer-as-string EVALUATE
	    LOOP drop ;

	define-nuc-parameters

    [ELSE]
	nuc-parameters# 7 <> [IF]	\ planed to become a compile time VALUE
	    bell cr .( nuc-parameters# must be 7 or zero ) cr quit
	[THEN]

	NUC-VAR: nuc-parameter-A		\ I use 7 parameters ;-)
	NUC-VAR: nuc-parameter-B		\ not elegant, but faster.
	NUC-VAR: nuc-parameter-C
	NUC-VAR: nuc-parameter-D
	NUC-VAR: nuc-parameter-E
	NUC-VAR: nuc-parameter-F
	NUC-VAR: nuc-parameter-G
    [THEN]

\ dup nuc-parameters - CONSTANT nuc-parameters#

: n'th-nuc-parameter ( u - a )
    dup 0<  over nuc-parameters# < 0= or
    ABORT" n'th-nuc-parameter: Index out of range."
    cells nuc-parameter-A + ;

[THEN]

dup CONSTANT nuc-invisible	\ base index
\ 0 CONSTANT nuc-invisible#	\ planed to become a compile time VALUE

\ the diversificable area stops here
dup nuc-diversificable-area - CONSTANT nuc-diversificable-items

\ 0 CONSTANT nuc-secrets#	\ planed to become a compile time VALUE

\ potencies
\ special organs, which are a bit 'more real' than the ordinary organs
\ this variables contain energy, which can be moved from one place to the other
\ we can make exceptions for reproduction, if we want ... hmm
\ not used in the moment

\ cell-variables
\ nuc-var cell-variables	\ pointer to other data or zero 
\ these datas can be anywhere
\ good place to have shared datas
\ not used in the moment

dup CONSTANT nuc-scan-limit

\ nuc-var nuc-link		\ link to other cells on the same spot
\ makes multi-celled organismes possible
\ not used yet

\ now the structure of the nuc is complet:
cells CONSTANT nuc-length#	\ eats counter up


: nuc-addr ( i -- a )           \ gets the address of the i'th nuc entry
    cells cp@ + ;

\ an xt list would be better, but I'll change nuc structure anyway.
: nuc-var-name ( index -- addr count )
    CASE
	 0 OF s" wake-me-xt"			ENDOF
	 1 OF s" eat-xt"			ENDOF
	 2 OF s" reproduce-xt"			ENDOF
	 3 OF s" show-me-xt"			ENDOF
	 4 OF s" length"			ENDOF
	 5 OF s" nuc-flags"			ENDOF
	 6 OF s" nuc-supplements"		ENDOF
	 7 OF s" id"				ENDOF
	 8 OF s" genome-id"			ENDOF
	 9 OF s" age"				ENDOF
	10 OF s" generation"			ENDOF
	11 OF s" genom-generation"		ENDOF
	12 OF s" code-cost"			ENDOF
	13 OF s" energy"			ENDOF
	14 OF s" reproduction-threshold"	ENDOF
	15 OF s" age-threshold"			ENDOF
	16 OF s" appearance"			ENDOF
	17 OF s" my-diversification-bitmask"	ENDOF
	dup nuc-organs < ABORT" nuc-var-name: Index negative."

	\ default case
	\ other nuc variables

	(scratch-buf) stringbuf-empty		\ buffer for result
	\ organs:
	nuc-organs -
	dup nuc-organs# < IF			\ organs
	    s" organ-"		(scratch-buf) cat
	    [char] A +	(scratch-buf) char-cat
	    (scratch-buf) buffer-as-string
	    false	\ dummy, keep ENDCASE happy
	ELSE nuc-organs# -
	    dup nuc-parameters# < IF		\ parameters
		s" nuc-parameter-"	(scratch-buf) cat
		[char] A +		(scratch-buf) char-cat
		(scratch-buf) buffer-as-string
		false	\ dummy, keep ENDCASE happy
	    ELSE nuc-parameters# -
		dup nuc-invisible# < IF		\ invisible
		    drop
		    s" nuc-invisibles not implemented yet"
		    false	\ dummy, keep ENDCASE happy
		ELSE nuc-invisible# -
		    dup nuc-secrets# < IF
			drop
			s" no secrets implemented"
			false	\ dummy, keep ENDCASE happy
		    ELSE nuc-secrets# -
			drop
			s" nothing else implemented yet"
			false	\ dummy, keep ENDCASE happy
		    THEN
		THEN
	    THEN
	THEN
	\
    ENDCASE ;

\ some action list's for selection in menus or by the mutation process:
LIST: eat-actions
LIST: reproduce-actions
LIST: show-me-actions
LIST: wake-me-actions
also genes ' g-noop wake-me-actions >list   previous

\ produce unique id's
variable (id)	(id) off
: new-id ( -- id )	(id) @ 1+ dup (id) ! ;

variable (genome-id)	(genome-id) off
: new-genome-id ( -- genome-id )	(genome-id) @ 1+ dup (genome-id) ! ;

: new-nucleus ( length -- addr flag | false )
\ allocates memory and erases it, (but does *not* set cp yet)
    dup >r		\ length is on return stack
    allocate	( addr ior )
    IF	rdrop drop cr ." couldn't allocate space for a nucleus!" 1000 ms cr
	false		\ flag
    ELSE     ( addr )
	dup r@ erase	\ initiate everything to zero
	cp@ swap	\ ( cp@ addr ) keep old cp on stack
	dup cp!		\ temporary set cp
	r> length !	\ (length works, because cp is set now)
	swap cp!	\ ( addr ) restore cp
\ 	new-id id !	\ give an individual id ??? hmm, no, not here
	true		\ ( -- addr true )
    THEN ;

: new-nucleus-as-word	( length -- addr TRUE | * ) \ always true, or ABORT
\ similar to new-nucleus, but creates the nucleus in the dictionary
    dup >r			\ length is on return stack
    here swap	( here length )
    allot	( here )
    dup r@ erase		\ initiate everything to zero
    cp@ swap	( cp@ addr ) 	\ keep old cp on stack
    dup cp!			\ temporary set cp
    r> length !			\ (length works, because cp is set now)
    nuc-flags dup @ nuc-is-word or swap !
    swap cp!	( addr )		\ restore cp
    true ;	( -- addr true )	\ true, don't get here if allot aborts

: log-cat-id ( -- )
    s" ID:"			cat-log
    id @ num>string		cat-log
    s"  GI:"			cat-log
    genome-id @ num>string	cat-log
    s"  "			cat-log ;

include symbols-stack.fs
include genes/genes-conditionals.fs
include mutation-0.2.fs

\ cloning a nucleus:
variable cloned		cloned off		\ counter
: clone ( -- addr flag | false )		\ clones the actual nucleus
    \ produces an exact clone of the current nuc. same id, (should be set)
    \ does not change cp, the actual cell stays the same

    \ genom on trial? works on the mother cell, before data is copied
    nuc-flags @ wake-me-trial and IF		\ genom on trial?
	trial-phase @ genom-generation @ > IF	\ trial continues? 
	    wake-me-xt @ eb>counter
	    dup @ 1+ swap !			\ increase counter
	ELSE
	    end-trial-phase	    
	THEN
    THEN

    length @ new-nucleus		\ tries to allocate a new nucleus
    IF ( addr )
	cp@ over length @ move		\ clone all data from actual cell
	1 cloned +!			\ increase counter

	\ we must clear the nuc-is-word flag:
	cp@ >r  dup cp!			\ temporary switch in to new nuc
	nuc-flags dup @ [ nuc-is-word invert ] literal and swap !
	r> cp!				\ switch back

	true				\ :) birth announcement
    ELSE ." couldn't clone the nucleus" cr	\ cloning failure
	false
    THEN ;


variable clone-cost	clone-cost off		\ how much energy for a new one
false [if]			\ not used any more
: cell-division-cloning ( -- )			\ cell-division by cloning
						\ starting two fresh lives
    clone					\ tries to clone
    IF ( addr )					\ ok?
	age off					\ cloning starts 2 *new* lifes
	new-id id !				\ with new individual id's
	1 generation +!				\ next generation
	1 genom-generation +!			\ next genom-generation
	energy @ clone-cost @ - energy !	\ pay a price
	energy @ 2/ dup energy !		\ share the remaining energy

	swap cp!				\ make the clone the actual one
						\ now whe're in the newly born
	energy !				\ give it it's share
	age off					\ cloning starts 2 *new* lifes
	new-id id !				\ with new individual id's
	1 generation +!				\ next generation
	1 genom-generation +!			\ next genom-generation
    ELSE ." couldn't clone the nucleus" cr 	\ cloning failure
    THEN ;
' cell-division-cloning reproduce-actions >list
[then]

: ?reproduce ( -- )				\ maybe reproduce
    energy @ reproduction-threshold @ > IF	\ enough energy for children?
	reproduce-xt @ EXECUTE			\ yes, try it
    THEN ;

\ hmm, let's prepare some food:
defer <food> ( -- addr )		\ gives addr where the food is

support-single-individuals? [IF]	\ do we still need deferring food?
variable bowl		bowl off	\ the bowl is just for one individual
' bowl is <food>			\ will work different in fields later
[THEN]

\ food that's placed on the spot not taking into account, if there's somebody:
variable food-share/spot	1000 food-share/spot !
: feed-spot ( -- )  food-share/spot @ <food> +! ; \ put on spot or in the bowl

\ food that's fed to all living cells, but not put on empty spots:
variable individual-fixed-food-share	individual-fixed-food-share off

\ food that's shared between all living beings in one world
\ (will be compiled in the world's section)
\ is good for population control
\ the amount will be *added* to the other sources of food
\ and be put in the following variable:
variable food-common-share			\ how much each one gets

: feed-individual ( -- )
    individual-fixed-food-share @
    food-common-share @   +
    <food> +! ;

\ appearance: how does a cell show on the screen
variable 2-ascii-scale	\ scaling factor for values that are displayed as ascii
decimal
1000 2-ascii-scale !

\ normalization to the continuous range of printable ASCII codes '!' to '~'
: 2-ascii ( n -- n' )
    [char] ! -		\ shift range for normalization 
    [ decimal ]
    94 mod		\ 93 printable ascii codes in a row: '!' to '~'
    dup 0< IF 94 + THEN
    [char] ! + ;	\ printable codes start with '!'
: .ascii ( n -- )   2-ascii emit ;	\ prints as normalized ASCII

\ not finally edited:
\ words to determine a string with caracters to be shown by show-me-xt:
: ranged-char ( u addr count -- c )
    rot					( addr count u )
    over mod				( addr count mod )
    dup 0< IF + ELSE nip THEN		( addr mod' )
    + c@ ;

\  : selected-chars ( u -- c)
\      s" *+.o:xذ%,-|/\~@#$&'0;<=>O[]^_`s{}-" ranged-char ;

: selected-chars ( u -- c)
    s" *+.o:xذ%,-|/\~@#$&'0;<=>O[^_`s{}-]" ranged-char ;

: pretty-char ( u - c )
    s" #$%&'()*+,-./08:;<=>@IOS[\]^_`losx{|}~-,"
    ranged-char ;

: round-char ( u - c )	s" %*+.0@O`aox" ranged-char ;

: dots ( u - c )	s" '*+,-.:;@O^`ox~" ranged-char ;

: show-ascii	appearance @					.ascii ;
' show-ascii show-me-actions >list
: show-age	age	   @			[char] 0 +	.ascii ;
' show-age  show-me-actions >list
: show-energy	energy	   @ 2-ascii-scale @ /	[char] 0 +	.ascii ;
' show-energy show-me-actions >list
: show-generation generation @			[char] 0 +	.ascii ;
' show-generation show-me-actions >list
: show-genom	genome-id   @			[char] 0 +	.ascii ;
' show-genom show-me-actions >list
: show-genom-b	genome-id   @			selected-chars	emit ;
' show-genom-b show-me-actions >list
: show-genom-generation genom-generation @	[char] 0 +	.ascii ;
' show-genom-generation show-me-actions >list
: show-code-length
    code-cost @
    (default-gene-cost#) >r
    r@ 1- +			\ to catch conditionals
    r> /
    1-				\ don't count ;gene
    [char] 0 + .ascii ;		\ start with '0'
' show-code-length show-me-actions >list

: show-A	organ-A    @ 2-ascii-scale @ /	[char] 0 +	.ascii ;
' show-A show-me-actions >list
: show-sign-A
    organ-A @ >r
    r@ -9 < IF [char] - ELSE
    r@  9 > IF [char] + ELSE
	      [char] o THEN THEN rdrop
    .ascii ;
' show-sign-A show-me-actions >list

\ <look-at> is a showing function which can be changed by the user during
\ the experiment
variable look-at-xt	' show-generation look-at-xt !
: <look-at> ( -- )	look-at-xt @ EXECUTE ;  \ works like a deferred word
' <look-at> show-me-actions >list

: show-background ( -- )	paint-background space ;


\ death:
variable leave-energy-after-death	\ what about my energy?
leave-energy-after-death off		\ switch leaving energy on spot 

defer <disappear>			\ will unlink the cell from the field

variable died					\ counter
: die ( -- )					\ death
    energy @ >r					\ my energy can survive
    nuc-is-word nuc-flags @ and IF
	." Cells in the dictionary can't really die. " cr
	0
    ELSE
	nuc-flags @ wake-me-trial and IF	\ genom on trial?
	    log-mask @ IF
		log-cat-id
		s" : died on trial "  log-death log-trial or  log-it
	    THEN
	    wake-me-xt @ decrease-eb-count	\ decrease counter
	THEN
	log-mask @ IF
	    log-cat-id
	    s" : died" log-death log-it
	THEN
	cp@ free				\ free  memory
    THEN
    0= IF					\ ready to die?
	leave-energy-after-death @ IF		\ what about my energy?
	    r@ <food> +!			\ leave it back
	THEN
	<disappear>				\ unlink from field
	false cp!				\ I'm gone
	1 died +!				\ counter
    ELSE		\ strange...
	log-cat-id
	s" cell at "			cat-log
	cp@ . num>string		cat-log
	s"  could not die in peace"	(log) buffer-as-string cr type
	log-it
    THEN
    rdrop ;

: will-die? ( -- flag )
    age @ IF
	energy @ 0> 0=				\ burnt out?
    ELSE
	false					\ we let newborns survive
    THEN
    age-threshold @ IF				\ is there a highest age?
	age @ age-threshold @ > or		\ am I too old?
    THEN ;

: ?die ( -- )					\ maybe die, maybe not
    will-die? IF die THEN ;			\ sorry, your time has come


\ one day in his life:
variable living		\ counter
variable newborn
variable trial		\ how many (mutated) cells are on trial?
variable nuc-do-cost	decimal 1000 nuc-do-cost !
variable code-price	code-price off
2variable code-price-scale	2 100 code-price-scale 2!
\ switch if cells can change the qualities of the future
variable future-quality-change		future-quality-change on
future-change-individal [IF]		\ copy quality changes individually?
    defer <copy-qualities2future>
[THEN]


: log-nuc-organs ( addr count -- )	\ not checking

    2dup			cat-log
    s" organs:	"		cat-log
    nuc-organs		( base-index )
    nuc-organs# 0 ?DO
	dup i + nuc-addr @ num>string	cat-log    s"  " cat-log
    LOOP drop

    s" 	energy: "		cat-log
    energy @ num>string		0 log-it

[ nuc-parameters# ] [IF] \ experimental. work in progress...

    				cat-log
    s" nuc-parameters:	"	cat-log
    nuc-parameters# 0 ?DO
	i n'th-nuc-parameter @ num>string cat-log  s"  " cat-log
    LOOP
    s"  " 0 log-it

[ELSE] 2drop [THEN]
;

1 value brew-depth	\ depth control in brew

defer log-spot-qualities ( addr count -- )
defer log-random-generator ( addr count -- )
: ?log-nuc-&-spot-state ( addr-preamble count-preamble -- )
    log-mask @ dup 0= IF drop 2drop EXIT THEN

    >r	( addr count  r: mask )
    r@ log-organs and IF
	2dup log-nuc-organs
    THEN
    r@ log-qualities and IF
	2dup log-spot-qualities
    THEN
    r> log-random and IF
	2dup log-random-generator
    THEN
    2drop ;

\ 'nuc-do-all' is the main routine for doing a cells action
: nuc-do-all ( -- )
    cp@ 0= IF					\ tests cp for convenience
	." sorry, there's nobody here" cr	\ cp not ok
    ELSE
	?die					\ maybe die
	cp@ IF					\ still alife?
	    1 living +!				\ count living cells

	    s" before:	" ?log-nuc-&-spot-state

	    nuc-flags @ wake-me-trial and IF	\ on trial?
		1 trial +!
		wake-me-xt @ dup
		eb>length @			\ get length
		swap eb>sequence swap		\ skip length
		also genes
		EVALUATE			\ evaluate gene-string
		previous
	    ELSE
		wake-me-xt @ EXECUTE		\ triggers the life action
	    THEN				( addr-of-cost )

[ future-change-individal ] [IF]	\ copy quality changes individually?
	    \ can cells change qualities of the future?
	    future-quality-change @ IF <copy-qualities2future> THEN
[THEN]

	    log-mask @ log-meal and >r		\ log meal?
	    r@ IF energy @ THEN

	    eat-xt @ EXECUTE			\ gets food

	    r> IF				\ log meal?
		s" Meal: "			cat-log
		energy @ swap - num>string	cat-log
		s"  	Day cost: "		cat-log
		nuc-do-cost @ num>string	0 log-it
	    THEN

	    energy dup @ nuc-do-cost @ - swap !	\ and consumes energy

	    code-price @ dup IF		\ price to pay for the code?
		code-cost @  (default-gene-cost#) */
		code-price-scale 2@ */

		log-mask @ dup IF
		    [ log-costs log-meal or ] literal and IF
			s" actual code costs : "	cat-log
			dup num>string			log-costs log-it
		    THEN
		ELSE drop THEN

		negate energy +!
	    ELSE drop THEN

	    display-switch @ spot-display-on and IF	\ spot display on?
		set-colors				\ I do like colors
		show-me-xt @ EXECUTE			\ show up
	    THEN

	    1 age +!				\ encreases the counter

	    s" after:	" ?log-nuc-&-spot-state

	    ?reproduce				\ maybe reproduce
	ELSE					\ must show background if dead
	    display-switch @ spot-display-on and IF	\ spot display on?
		show-background
	    THEN
	THEN
    THEN

    \ depth control to catch bugs:
    depth brew-depth <> IF
	bell		\ Thats not tolerated.  Warn the user.
	." stack violation " cr .s

	log-mask @ IF	\ Error logged if *anything* else get's logged.
	    s" stack violation ID:"	cat-log
	    id @ num>string		cat-log
	    s"  depth="			cat-log
	    depth num>string		cat-log
	    depth IF
		s"   tos="		cat-log
		dup num>string	cat-log
	    THEN
	    s" "			0 log-it
	THEN
    THEN
;

\ don't forget to set >gene-cost when setting wake-me-xt
: setup-wake-me ( gene-xt -- )
    dup wake-me-xt !
    gene-xt>internal-xt >body >gene-cost @ code-cost ! ;

\ word to create a named individual in the dictionary
: INDIVIDUAL:		\ 'INDIVIDUAL: name' creates a nuc named 'name'
    CREATE		\ compiling:      create the dictionary entry
    nuc-length# new-nucleus-as-word drop	\ make the nucleus
    cp!			\ compiling:      the newborn gets center of attention 
  DOES> cp! ;		\ at run time:    returns addr and does cp!

LIST: individuals


\ =======================================

\ now we need a playground for our beeings: space (2D) and time
2 CONSTANT time-planes#		\ # of different times: now, later
\ 3  VALUE spot-qualities#	\ # of different qualities	compile value
\ 7  VALUE spot-properties#	\ # of different r/o properties	compile value
\ 0  VALUE spot-secrets#	\ # of hidden spot variables	compile value

1		\ cp plane		\ compute # of planes af the world
1+		\ food plane
spot-qualities# +	\ qualities, ...
spot-properties# +
spot-secrets# +
CONSTANT		field-planes#	\ # of different planes of the world
					\ like cp and food
\ 80 CONSTANT		c-l		\ cells (or caracters) in a line
c-l l-s 1- * CONSTANT	spots		\ how many spots in the universe
					\ here for simple text mode

\ this artificial world has several planes of reality:
\ the first one is always for the cp's
\ the others are variables like food or so.

\ the 'spot' is an index in thouse fields

\ there are time-planes# structures like this for the different times:
\ one for now, one for the next time, ...

\ the idea is to work from present to future:
\ if something is changing, it changes in the future plain
\ a time-step makes it appear, then
\ time-step is switching a cyclic pointer

\ spot >index initializes the connections with the different planes

spots cells CONSTANT plane-length			\ length of one plane
plane-length field-planes# * CONSTANT time-plane-length	\ for all times being
time-plane-length time-planes# * CONSTANT field-length	\ whole field

\ now let's create a world:
variable world		world off		\ flag=addr, world pointer

support-single-individuals? [IF]
variable world?		world? off		\ field or individual?
: big-bang ( -- )				\ creates new universe
						\ sets world or aborts
    world @ 0= IF				\ if it's not here already
	field-length allocate ( -- addr ior) 0=	\ we need space
	IF   dup world !			\ thanks, got it
	    field-length erase			\ clean it up
	    world? on
	ELSE drop false world !			\ uups, no space there?
	    world? on
	    true ABORT" big bang failure. nothing to do..."
	THEN
    ELSE
	cr ." dont make big bangs in an existing world!" cr
    THEN ;
[ELSE] \ support-single-individuals? is FALSE
: big-bang ( -- )				\ creates new universe
						\ sets world or aborts
    world @ 0= IF				\ if it's not here already
	field-length allocate ( -- addr ior) 0=	\ we need space
	IF   dup world !			\ thanks, got it
	    field-length erase			\ clean it up
	ELSE drop false world !			\ uups, no space there?
	    true ABORT" big bang failure. nothing to do..."
	THEN
    ELSE
	cr ." dont make big bangs in an existing world!" cr
    THEN ;
[THEN]

\ offsets to the different planes
0			CONSTANT offset-cell-pointers	\ offset pointer plane
plane-length		CONSTANT offset-food		\ where is the food?
plane-length 2 *	CONSTANT offset-qualities

\ It's ugly, but seems more efficient.  Fixed on 3 qualities
plane-length 2 *	CONSTANT offset-A-quality
plane-length 3 *	CONSTANT offset-B-quality
plane-length 4 *	CONSTANT offset-C-quality

spot-properties# [IF]
    offset-qualities  spot-qualities# plane-length *  +
    CONSTANT offset-properties
[THEN]

[UNDEFINED] spot [IF]	\ defined in brew-basics.fs
    2variable spot				\ field index of actual spot
    spot cell+ CONSTANT step	-1 step !	\ actual time
    \ to get both use:	spot 2@ ( -- step spot )
[THEN]

variable spot-offset		\ same, as address offset

debugging @ [if]	\ used for debugging only
\ words to get extrem values of something over all spots visited
2variable spot&step-max	2variable spot&step-min
variable spot&step-max-value
variable spot&step-min-value

: init-spot&step-max-min ( -- )
    lowest-integer# spot&step-max-value !
    highest-integer# spot&step-min-value ! ;

: watch-spot&step-max-min ( value -- )
    dup spot&step-max-value @ > IF
	dup spot&step-max-value !
	spot 2@ spot&step-max 2! 
    THEN
    dup spot&step-min-value @ < IF
	dup spot&step-min-value !
	spot 2@ spot&step-min 2! 
    THEN
    drop ;

\ handle a (scratch-buf) buffer with the result as text string. log, display
: spot&step-extrems-string ( -- addr count )
    (scratch-buf) >r
    r@ stringbuf-empty
    s" max: "				r@ cat
    spot&step-max 2@ swap num>string	r@ cat
    [char] :				r@ char-cat
    num>string				r@ cat
    [char] =				r@ char-cat
    spot&step-max-value @ num>string	r@ cat
    s"    min: "			r@ cat
    spot&step-min 2@ swap num>string	r@ cat
    [char] :				r@ char-cat
    num>string				r@ cat
    [char] =				r@ char-cat
    spot&step-min-value @ num>string	r@ cat
    r> buffer-as-string ;

: log-spot&step-extrems ( mask -- )
    >r
    s" SPOT&STEP-EXTREMS: "	cat-log
    spot&step-extrems-string	r> log-it ;

: .spot&step-extrems ( -- )   spot&step-extrems-string type ;

[then]

\ time:

variable (ti)		-1 (ti) !	\ cyclic index of the present
					\ *ONLY* time-step should change it
					\ for internal use *ONLY* ( tp )
\ how to find the other times, like future etc:
variable (time-offset)			\ internal use only

0 value (tp-offset)			\ speed optimation
: adjust-tp-offset
    (ti) @ (time-offset) @ +
    time-planes# mod
    time-plane-length *
    TO (tp-offset) ;

: tp ( -- addr )   (tp-offset)  world @ + ;

: time-step ( -- )			\ step from present to future
    (ti) @ 1+ time-planes# mod (ti) !	\ time as a cycle, dizzy?
    adjust-tp-offset
    1 step +! ;
time-step				\ start the time

: present ( -- )   0 (time-offset) !	\ makes the present be the time seen
    adjust-tp-offset ;

present					\ initialize time to the present

: future  ( -- )   1 (time-offset) !	\ makes the future be the time seen
    adjust-tp-offset ;

: present-initializes-future ( -- )	\ initializes the future plain
    present tp				\ with the data from the present
    future  tp  time-plane-length move	\ with the data from the present
    present ;				\ stay tuned to the present

: n'th-spot-variable ( i - addr )
    dup 0<  over field-planes# < 0= or
    ABORT" n'th-spot-variable: Index out of range."
    plane-length *
    tp offset-cell-pointers +		\ actualtime plane 
    spot-offset @ + +	;		\ at the right spot

: spot-vars@ ( -- spot-var-n spot-var-n-1 ... spot-var0 )
    0 field-planes# 1- ?DO
	i n'th-spot-variable @
    -1 +LOOP ;

: spot-vars! ( spot-var-n spot-var-n-1 ... spot-var0  addr -- )
    >r
    r@ field-planes# cells +  r> ?DO
	i !
    cell +LOOP ;

support-single-individuals? [IF]
\ variable world?	world? off	\ field or individual?
\ world? was compiled earlier

\ fcp = 'field cell pointer'
\ word returning the address in the field, where the pointer to the cell is.
\ If it's an individual, cp is returned for convenience, (but I don't know,
\ if I'll ever use this).
\ hmm, yes and it makes troubles adding things to cp! #######
offset-cell-pointers [if]
: fcp ( -- addr-of-the-pointer-to-the-nuc )	\ 'field cell pointer'
    world? @ IF					\ pointer in a field
	tp offset-cell-pointers +		\ start of the pointer plane
	spot-offset @ +				\ go to the right spot
    ELSE cp THEN ;				\ special case: individual
[else]
: fcp ( -- addr-of-the-pointer-to-the-nuc )	\ 'field cell pointer'
    world? @ IF					\ pointer in a field
	tp spot-offset @ +
    ELSE cp THEN ;				\ special case: individual
[then]

\ let the spot have another quality:
variable single-A-quality	\ A-quality for individuals
single-A-quality off	\ no 'A' quality for singles?	no, I'm just kidding!
: A-quality ( -- addr)
    world? @ IF
	tp offset-A-quality +		\ A-quality plane
	spot-offset @ +			\ at the right spot
    ELSE
	single-A-quality
    THEN ;

\ wanted to have more qualities for experiments with symbiotic biotops:
variable single-B-quality	\ B-quality for individuals
single-B-quality off
: B-quality ( -- addr)
    world? @ IF
	tp offset-B-quality +		\ B-quality plane
	spot-offset @ +			\ at the right spot
    ELSE
	single-B-quality
    THEN ;

variable single-C-quality	\ C-quality for individuals
single-C-quality off
: C-quality ( -- addr)
    world? @ IF
	tp offset-C-quality +		\ C-quality plane
	spot-offset @ +			\ at the right spot
    ELSE
	single-C-quality
    THEN ;

: n-quality ( n -- addr )	\ general access to the different qualities
\ 0 are the pointers, 1 food, 2 A-quality, 3 B-quality, ...
    world? @ IF
	plane-length * offset-cell-pointers +	\ offset of the n'th plane
	tp +					\ n-quality in actual time
	spot-offset @ +				\ at the right spot
    ELSE
	bell cr ABORT" n-quality is not defined for single cells" \ #######
    THEN ;
[ELSE] \ support-single-individuals? is FALSE
\ fcp = 'field cell pointer'
\ word returning the address in the field, where the pointer to the cell is.
\ Does not support single individuals any more.
offset-cell-pointers [if]
: fcp ( -- addr-of-the-pointer-to-the-nuc )	\ 'field cell pointer'
    tp offset-cell-pointers +			\ start of the pointer plane
    spot-offset @ + ;				\ look at actual spot
[else]
: fcp ( -- addr-of-the-pointer-to-the-nuc )	\ 'field cell pointer'
    tp spot-offset @ + ;
[then]    

\ let the spot have another quality:
: A-quality ( -- addr)
    tp offset-A-quality +		\ A-quality plane
    spot-offset @ + ;			\ at the right spot

\ wanted to have more qualities for experiments with symbiotic biotops:
: B-quality ( -- addr)
    tp offset-B-quality +		\ B-quality plane
    spot-offset @ + ;			\ at the right spot

: C-quality ( -- addr)
    tp offset-C-quality +		\ C-quality plane
    spot-offset @ + ;			\ at the right spot

: n-quality ( n -- addr )	\ general access to the different qualities
\ 0 are the pointers, 1 food, 2 A-quality, 3 B-quality, ...
\ we could check for 'n < qualities', but I don't
    plane-length * offset-cell-pointers +	\ offset of the n'th plane
    tp +					\ n-quality in actual time
    spot-offset @ + ;				\ at the right spot

spot-properties# [IF] \ spot variables read only (for the 'cells')

: n-property ( n -- addr )
    plane-length *			\ which property
    tp + offset-properties + 		\ property plane
    spot-offset @ + ;			\ at the right spot

TRUE [IF] \ new stile defining spot-properties# named properties
    : define-properties ( -- )
	spot-properties# 0 ?DO
	    (scratch-buf) stringbuf-empty
	    s" : property-"	(scratch-buf) cat
	    i [char] A +	(scratch-buf) char-cat
	    bl			(scratch-buf) char-cat
	    i [char] 0 +	(scratch-buf) char-cat
	    s"  n-property ;"	(scratch-buf) cat
	    (scratch-buf) buffer-as-string EVALUATE
	LOOP ;

	define-properties

[ELSE] \ old stile fixed on 3 properties
    : property-A ( -- addr)   0 n-property ;
    : property-B ( -- addr)   1 n-property ;
    : property-C ( -- addr)   2 n-property ;
[THEN]

[THEN]

\ log-spot-qualities ( addr count -- )
:NONAME ( addr count -- )
    ( addr count -- )		cat-log
    s" qualities:	"	cat-log
    A-quality @ num>string	cat-log	s"  " cat-log
    B-quality @ num>string	cat-log	s"  " cat-log
    C-quality @ num>string	cat-log
    s" 	<food>: "		cat-log
    <food> @ num>string	0 log-it ; IS log-spot-qualities

\ log-random-generator ( addr count -- )
:NONAME ( addr count -- )
    ( addr count -- )	cat-log
    random-xt @ xt>name cat-log	s"  " cat-log
    random-xt @ CASE
	['] random-generalized OF
	    (random-generalized) 2@ swap
	    num>string 	cat-log	s"  " cat-log
	    num>string 	cat-log
	ENDOF
	['] random-BRODIE OF
	    seed-BRODIE @ num>string cat-log
	ENDOF
    ENDCASE
    s" " 0 log-it ; IS log-random-generator

: random-qualities ( n mask -- )
    swap  1+ >r		( mask  r: n+1)
    0
    BEGIN		( mask index  r: n+1)
	2dup rshift 1 and IF
	    r@ random-ranged
	    2 random-ranged 0= IF negate THEN
	    over 2 + n-quality !		\ 'A-quality' is '2 n-quality'
	THEN
	1+ dup spot-qualities# =
    UNTIL 2drop rdrop ;

future-change-individal [IF]		\ copy quality changes individually?
\ word to transfer quality changes by a cell to the future
:NONAME	( -- )  A-quality @  B-quality @  C-quality @
    future      C-quality !  B-quality !  A-quality !
    present ; IS <copy-qualities2future>
[THEN]

: >spot! ( i -- )			\ sets spot and spot-offset
    dup spot !
    cells spot-offset ! ;
0 >spot!				\ initialize it

\ : do-at-spot ( i xt -- )   swap >spot! EXECUTE ;	\ spot not restored

: someone-here? ( i -- addr|false)	\ checks if a spot is occupied
    spot @ swap	( spot i )		\ remember spot
    >spot! fcp @  ( spot a|false )	\ return value on TOS
    swap >spot! ;			\ go back to actual spot

\ do something on all spots (sequential)
variable (do-everywhere-xt)
: do-everywhere ( xt -- )
    (do-everywhere-xt) !
    spot @ >r				\ maybe it's worth saving?
    spots 0 DO				\ loop over all spots
	i >spot!
	(do-everywhere-xt) @ EXECUTE	\ and do your job
    LOOP
    r> >spot! ;				\ you never know what's good for...

CREATE (scan-spots)			\ result data pointer
field-planes# 4 * cells allot		\ allot space for results
\ min
\ max
\ sum (double)	later: average (single)
variable spots-scanned		spots-scanned off	\ flag and counter
: init-spot-scan ( -- )
    spots-scanned off
    (scan-spots)  field-planes# 4 * cells  erase ;
: scan-spot ( -- )
    spots-scanned @ IF		\ normal (not first time) case
	field-planes# p0 DO	\ skip pointer plane?
	    i n-quality @	\ read a quality of the actual spot
	    (scan-spots) i 2* 2* cells +	\ address of actual minimum
	    dup >r
	    @ over min r@ !			\ store new minimum
	    r> cell+ dup >r			\ address of actual maximum
	    @ over max r> !			\ store new maximum
	    s>d					\ convert to double
	    (scan-spots) i 2* 2* 2 + cells +	\ address summation double cell
	    dup >r
	    2@ d+ r> 2!				\ double precision sum
	LOOP
    ELSE			\ first time, max min must be initialized
	field-planes# p0 DO	\ skip pointer plane?
	    i n-quality @	\ read a quality of the actual spot
	    (scan-spots) i 2* 2* cells + >r	\ address of minimum
	    dup r@ !				\ store as first minimum
	    r> cell+ >r				\ address of maximum
	    dup r@ !				\ store as first maximum
	    s>d					\ convert to double
	    r> cell+ 2!				\ initialize double sum
	LOOP
    THEN
    1 spots-scanned +! ;		\ counter

: spot-scan-average ( -- )
    spots-scanned @ IF
	field-planes# p0 DO			\ skip pointer plane?
	    (scan-spots) i 2* 2* cells +	\ data for this quality
	    2 cells +	dup >r			\ summation double cell
	    2@
	    1 spots-scanned @  	m*/	d>s	\ divide by counter
	    r> !				\ store as single cell
	LOOP
    ELSE bell THEN ;

: do-spot-scan ( -- )
    init-spot-scan
    ['] scan-spot do-everywhere
    spot-scan-average ;

\ field variable *text* names.
\ spot-qualities# spot-properties# spot-secrets# must fit.
: field-variable-text ( index -- )
    dup 0=  IF drop s" pointers  " EXIT THEN
    dup 1 = IF drop s" food      " EXIT THEN
    dup 0< ABORT" field-variable-text: negative index."

    \ we must build the name in a buffer:
    (scratch-buf) >r  r@ stringbuf-empty
    2 -	\ look at ordinary spot variable base index
    dup spot-qualities# < IF
	[char] A + r@ char-cat
	s" -quality " r@ cat    
    ELSE spot-qualities# -
	dup spot-properties# < IF
	    [char] A + r@ char-cat
	    s" -property " r@ cat    
	ELSE spot-properties# -
	    dup spot-secrets# < IF
		[char] A + r@ char-cat
		s" -secret   " r@ cat
	    ELSE
		cr ." Index:" .
		true ABORT" field-variable-text: Index out of range."
	    THEN
	THEN
    THEN

    r> buffer-as-string ;

\ do something with every living cell (sequential)
variable (do-xt)
: do-with-everybody ( xt -- )
    (do-xt) !
    spot @ >r				\ maybe it's worth saving?
    cp@ >r				\ ditto
    spots 0 DO				\ loop over all spots
	i someone-here? dup IF		\ inhabited?
	    cp!				\ yes: set temporally as actual
	    i >spot!			\      same for the spot
	    (do-xt) @ EXECUTE		\      and do your job
	ELSE drop THEN
    LOOP
    r> cp!				\ restore everything
    r> >spot! ;				\ you never know what's good for...

: do-at-random-spots ( u xt -- )
    (do-xt) !
    spot @ >r				\ maybe it's worth saving?
    cp@ >r				\ ditto
    0 ?DO				\ loop
	spots random-ranged		\ random spot
	dup someone-here? dup IF		\ inhabited?
	    cp!				\ yes: set temporally as actual
	    dup >spot!			\      same for the spot
	    (do-xt) @ EXECUTE		\      and do your job
	ELSE drop THEN drop
    LOOP
    r> cp!				\ restore everything
    r> >spot! ;				\ you never know what's good for...

variable (continuous-column)	(continuous-column) off	\ global
\ to prevent menu contents (including spot display) and such to override
\ continuous display space.
\ 'continuous-display-used' was introduced to prevent the changed
\ (continuous-column) to get recorded, when not needed.
: ?reset-continuous-column ( -- )
    display-switch @ continuous-display-used and IF
	(continuous-column) off		\ get screen for cont display cleared
    THEN ;

: .menu-short-help ( -- )
    at? 2>r
    r@ last-line < IF l-s 2 - ELSE last-line THEN
    0 swap at-xy
    cyan color-foreground   blue color-background
    ." Select with cursor arrows and <RETURN>, leave with 'q' (or empty spot)"
    clear-line-to-end
    at? nip last-line <> IF
	last-left
	menu-scroll-lines @ (total-scroll-range) @ < IF
	    white color-foreground
	    ." This menu is a scrolling one.  <PageUp>  <PageDown>  <home>  <end>  <arrows>"
	ELSE
	    ." (Often there are also key bindings to the first letter of a word.  Try it.)"
	THEN
	clear-line-to-end
    THEN
    default-color color-foreground  default-color color-background
    2r> at-xy ;

[UNDEFINED] title-colors-xt [IF]
    2variable title-colors-xt
    ' default-color ' blue title-colors-xt 2!
[THEN]

: title-colors ( -- )
    title-colors-xt 2@ execute color-background execute color-foreground ;

: common-menu-entries ( -- )
    s" !" ['] do-FORTH redisplay	menu-key-entry
    .menu-short-help

    \ It's a convenient place to do the following trick
    ?reset-continuous-column ;

: end-title ( -- )   clear-line-to-end reset-default-colors cr ;

\ Set choose-xt title string.  Prefix.  Must be used *before* menu entry. 
: title-choose-xt! ( addr count -- )
    2dup menu-selected  >menu-any-data 2!	\ for direct call 
    (menu-scratch-cell) >menu-any-data 2! ;	\ for menu call

2variable (title)
MENU: choose-xt-men
: .choose-xt-menu ( addr-of-list -- addr-of-list )
    page

    title-colors  (title) 2@ type  end-title	\ title bar

    20 menu-scroll-lines !

    cr
    dup nodes 0 scrolled-range ?DO
	i over n'th-node @
	dup >stack  menu-done	xt>name  over c@ >r ['] noop  menu-entry cr
	r> #key-same-entry    
    LOOP

    common-menu-entries ;

: choose-xt-menu ( addr-of-list addr-to-store-xt -- )
    menu-selected >menu-any-data 2@ (title) 2!

    >r			\ ( addr-of-list   R: addr-to-store-xt)
    choose-xt-men
    ['] .choose-xt-menu menu-display-xt !
    0 >stack menu-done	['] noop	menu-default
    0 >stack menu-done	['] noop	menu-key-default
    do-menu-loop	\ ( addr-of-list -- addr-of-list xt|false )
    ?dup IF
	r> ! drop
    ELSE rdrop drop THEN
    \    free-menus	\ I had crashes with this ########
;


\ step display: scanning nuc or spot data or doing continuous display
\ This version allows multiple sub screens.
[UNDEFINED] max-step-display-items# [IF]
    12 constant max-step-display-items#
[THEN]
variable (scan-index)	0 (scan-index) !	\ index of actual scan display

: scan-ARRAY: ( -- )
    CREATE
	max-step-display-items# cells allot
    DOES> ( -- addr )
	(scan-index) @ cells + ;

: scan-2ARRAY: ( -- )
    CREATE
	max-step-display-items# 2* cells allot
    DOES> ( -- addr )
	(scan-index) @ 2* cells + ;

scan-ARRAY: (scan-xt)			\ display xt or zero
scan-ARRAY: (scan-detail-index)		\ index of the detail to be displayed
scan-ARRAY: (scan-lines)		\ displayed lines including status line
scan-2ARRAY: (scan-min-max)
scan-2ARRAY: (last-scan-min-max)
scan-2ARRAY: horizontal-zoom-scale
scan-2ARRAY: vertical-zoom-scale
' vertical-zoom-scale is <vertical-zoom-scale>
scan-ARRAY: (vertical-display-range)
' (vertical-display-range) is <vertical-display-range>
scan-ARRAY: scan-background-xt
scan-ARRAY: scan-foreground-xt
scan-ARRAY: (scan-flags)		\ used in continuous-display

LIST: scan-background-colors
color-list scan-background-colors copy-simple-list-elements
LIST: scan-foreground-colors
color-list scan-foreground-colors copy-simple-list-elements

\ startup defaults of first two indizes
0 (scan-index) !
8 (scan-detail-index) ! 
l-s 1- 2/ (scan-lines) !
' blue scan-background-xt !
' default-color scan-foreground-xt !
1 (scan-index) !
l-s 1- 2/ (scan-lines) !
1 5 horizontal-zoom-scale 2!
' blue scan-background-xt !
' default-color scan-foreground-xt !

LIST: step-display-xt's

variable step-display-items	2 step-display-items !
: step-display ( -- )
    0	\ actual starting line
    step-display-items @ 0 ?DO
	i (scan-index) !
	(scan-xt) @ dup IF
	    over 0 swap at-xy
	    EXECUTE
	    (scan-lines) @ +
	ELSE drop THEN
    LOOP
    drop ;

: >step-display ( xt index -- )   (scan-index) !  (scan-xt) ! ;

variable (nuc-scan-data)	(nuc-scan-data) off	\ data pointer
\    scan-min			\ three pseudonuclei for data storage...
\    scan-max
\    2scan-sum	\ scanning: double sum	\ nuc-scan-average: computed results
variable nucs-scanned		nucs-scanned off	\ flag and counter
2variable nuc-scan-range				\ range to be scanned
: init-nuc-scan ( lower upper --)
    \ first free previously allocated memory
    (nuc-scan-data) @ ?dup IF free ABORT" init-nuc-scan: Couldn't free." THEN

    nucs-scanned off
    nuc-length# 4 * new-nucleus ( ... length -- ... addr flag | false )
    IF
	(nuc-scan-data) !
	nuc-scan-range 2!
    ELSE
	2drop
	bell cr ." init-nuc-scan failed" 1200 ms
    THEN
    trial off ;

: scan-nuc ( -- )	\ scanning the actual nuc
    nucs-scanned @ IF		\ normal (not first time) case
	nuc-scan-range 2@ DO
	    i nuc-addr @		\ read a nuc var in the current nuc
	    (nuc-scan-data) @ i cells +	\ address of actual minimum
	    dup >r
	    @ over min r@ !		\ store new minimum
	    r> nuc-length# + dup >r	\ address of actual maximum
	    @ over max r> !		\ store new maximum
	    s>d				\ convert to double
	    (nuc-scan-data) @ nuc-length# 2* + \ start of summation area
	    i cells 2* + dup >r		\ address of actual double sum
	    2@ d+	r> 2!		\ double precision sum
	LOOP
    ELSE	\ first time: max and min must get initialized
	nuc-scan-range 2@ DO
	    i nuc-addr @		\ read a nuc var in the current nuc
	    (nuc-scan-data) @ i cells + >r	\ address of minimum
	    dup r@ !			\ store as first minimum
	    r> nuc-length# + >r		\ address of maximum
	    dup r> !			\ store as first maximum
	    s>d				\ convert to double
	    (nuc-scan-data) @ nuc-length# 2* + \ start of summation area
	    i cells 2* +	2!	\ double precision sum
	LOOP
    THEN
    nuc-flags @ wake-me-trial and IF	\ on trial?
	1 trial +!
    THEN
    1 nucs-scanned +! ;			\ counter

: nuc-scan-average ( -- )
    nucs-scanned @ IF
	nuc-scan-range 2@ DO
	    (nuc-scan-data) @ nuc-length# 2* +  dup   \ start of summation area
	    i cells 2* +	2@		\ read double precision sum
	    nucs-scanned @  m/			\ divide by counter
	    swap i cells +	!		\ store as one cell
	LOOP		\ now we have a cell array with the averages
    ELSE bell THEN ;

: do-a-nuc-scan ( -- )
    nuc-scan-limit nuc-variables init-nuc-scan
    ['] scan-nuc do-with-everybody
    nuc-scan-average ;

2VARIABLE (scan-addr&index)
: (scan-nuc-detail) ( -- )
    (scan-addr&index) 2@   ( addr-of-result-buffer index-of-nuc-var-to-scan )
    nuc-addr @		   ( addr-of-result-buffer value-of-scanned-nuc-var )
    swap data2slice ;

: nuc-detailed-scan ( index-of-nuc-var-to-scan  addr-of-min-max -- )
    dup 2@ = IF				\ error: range is zero
	(scan-lines) @ 1- 0 ?DO 0 at-x clear-line-to-end cr LOOP
	."             scanning: " swap nuc-var-name type
	."  all values are " @ . clear-line-to-end
	EXIT
    THEN

    c-l statistic-array-size allocate			\ buffer for results
    ABORT" nuc-detailed-scan: Couldn't allocate." >r	( r: addr )
    r@ (scan-addr&index) cell+ !

    swap (scan-addr&index) !	( addr-of-min-max   r: addr-result-buffer )
    2@  c-l r@  ( min max slices# addr -- ) init-statistic-array

    ['] (scan-nuc-detail) do-with-everybody		\ reads the data in

    r@  (scan-lines) @  (scan-addr&index) @ nuc-var-name  statistic-display
    r> free ABORT" nuc-detailed-scan: Couldn't free." ;

\ use this after getting max and min values by 'do-a-nuc-scan'
: nuc-detail-extreems! ( index-of-nuc-var-to-scan -- addr )
    (scan-addr&index) !

    \ get min and max values from previous nuc scan:
    cp@ >r
    (nuc-scan-data) @ cp!		(scan-addr&index) @ nuc-addr @	\ min
    (nuc-scan-data) @ nuc-length# + cp!	(scan-addr&index) @ nuc-addr @	\ max
    r> cp!

    (scan-min-max) >r
    r@ 2!
    r> ;

\ use this after getting max and min values by 'do-a-nuc-scan'
: nuc-detailed-scan-prescanned ( index-of-nuc-var-to-scan )
    0 (scan-index) !				\ I just take the first one
    (scan-lines) @ >r			( index  r: old-lines )
    l-s (scan-lines) !			\ full screen
    dup					( index index  r: old-lines )
    (scan-detail-index) dup @ >r ( i i detail-index-addr  r: o-lin old-detail)
    !					( i  r: o-l old-detail)	\ set detail
    dup nuc-detail-extreems!	( index-of-nuc-var-to-scan addr-min-max r: l d)
    page
    nuc-detailed-scan		( r: old-lines old-detail-index )

    r> (scan-detail-index) !			\ restore index
    r> (scan-lines) ! ;				\ restore lines

: nuc-detail-2-min-max ( -- )
    (scan-detail-index) @ nuc-addr @ (scan-min-max) data2min-max ;

\ let the user know when zooming has changed
[UNDEFINED] (zoomed) [IF]
    variable (zoomed)
    (zoomed) off
[THEN]

[UNDEFINED] horizontal [IF]
    1 constant horizontal
[THEN]

: notify-zoom-change ( horizontal|vertical -- )
    [ FALSE ] [IF] \ notifying through >message
	drop
	s"        ========= < zooming! > =========      "
	1 >message
    [ELSE]
	(zoomed) dup @ rot or swap !  
    [THEN] ;
' notify-zoom-change is <notify-zoom-change>

\ use this for step display of a nuc detail
: nuc-scan-display ( -- )
    \ you want colors?
    display-switch @ step-foreground-coloring and IF
	scan-foreground-xt @ EXECUTE color-foreground \ for fun and distinction
    THEN
    display-switch @ step-background-coloring and IF
	scan-background-xt @ EXECUTE color-background \ for fun and distinction
    THEN

    (scan-min-max) min-max-init		\ init min-max
    ['] nuc-detail-2-min-max do-with-everybody	\ get min and max data

    \ horizontal zoom control: i don't want scan range to change with each step
    (scan-min-max) 2@ (last-scan-min-max) 2@ d=	\ same range?
    0= IF						\ range has changed
	\ first make sure data is within displayed range:
	\   looking at botton:
	(scan-min-max) >min @ dup	( new-minimum new-minimum )
	(last-scan-min-max)   >min @	( new-minimum new-minimum old-minimum )
	< IF						\ shrunken?
	    (last-scan-min-max) >min !
	    horizontal notify-zoom-change
	ELSE drop THEN

	\   looking at top:
	(scan-min-max) >max @ dup	( new-maximum new-maximum )
	(last-scan-min-max)   >max @	( new-maximum new-maximum old-maximum )
	> IF							\ grown?
	    (last-scan-min-max) >max !
	    horizontal notify-zoom-change
	ELSE drop THEN

	\ now check if real range has not shrunken too much:
	(last-scan-min-max)   2@  swap -	( displayed-range-size )
	(scan-min-max) 2@  swap -		( displayed real-range-size )
	swap horizontal-zoom-scale 2@ */ < IF		\ too small!
	    (scan-min-max) 2@ (last-scan-min-max) 2!	\ take real siz
	    horizontal notify-zoom-change
	THEN
    THEN

    (scan-detail-index) @ (last-scan-min-max) nuc-detailed-scan ;
' nuc-scan-display step-display-xt's >list
' nuc-scan-display 0 >step-display

\ Detailed scan of spots:
\ index of the spot variable to be displayed in spot-scan-display
\ variable (spot-index-to-scan)	1 (spot-index-to-scan) ! ##########

: (scan-spot-detail) ( -- )
    (scan-addr&index) 2@   ( addr-of-result-buffer index-of-spot-var-to-scan )
    n-quality @		   ( addr-of-result-buffer value-of-scanned-spot-var )
    swap data2slice ;

: spot-detailed-scan ( index-of-spot-var-to-scan  addr-of-min-max -- )
    dup 2@ = IF				\ error: range is zero
	(scan-lines) @ 1- 0 ?DO 0 at-x clear-line-to-end cr LOOP
	."             scanning: " swap field-variable-text type
	."  all values are " @ . clear-line-to-end
	EXIT
    THEN

    c-l statistic-array-size allocate			\ buffer for results
    ABORT" spot-detailed-scan: Couldn't allocate." >r	( r: addr )
    r@ (scan-addr&index) cell+ !

    swap (scan-addr&index) !	( addr-of-min-max   r: addr-result-buffer )
    2@  c-l r@  ( min max slices# addr -- ) init-statistic-array

    ['] (scan-spot-detail) do-everywhere		\ reads the data in

    r@ (scan-lines) @ (scan-addr&index) @ field-variable-text statistic-display
    r> free ABORT" spot-detailed-scan: Couldn't free." ;

: spot-detail-2-min-max ( -- )
    (scan-detail-index) @ n-quality @ (scan-min-max) data2min-max ;

: spot-scan-display ( -- )
    \ you want colors?
    display-switch @ step-foreground-coloring and IF
	scan-foreground-xt @ EXECUTE color-foreground \ for fun and distinction
    THEN
    display-switch @ step-background-coloring and IF
	scan-background-xt @ EXECUTE color-background \ for fun and distinction
    THEN

    (scan-min-max) min-max-init			\ init min-max
    ['] spot-detail-2-min-max do-everywhere	\ get min and max data

    \ horizontal zoom control: i don't want scan range to change with each step
    (scan-min-max) 2@ (last-scan-min-max) 2@ d=	\ same range?
    0= IF						\ range has changed
	\ first make sure data is within displayed range:
	\   looking at botton:
	(scan-min-max) >min @ dup	( new-minimum new-minimum )
	(last-scan-min-max)   >min @	( new-minimum new-minimum old-minimum )
	< IF						\ shrunken?
	    (last-scan-min-max) >min !
	    horizontal notify-zoom-change
	ELSE drop THEN

	\   looking at top:
	(scan-min-max) >max @ dup	( new-maximum new-maximum )
	(last-scan-min-max)   >max @	( new-maximum new-maximum old-maximum )
	> IF							\ grown?
	    (last-scan-min-max) >max !
	    horizontal notify-zoom-change
	ELSE drop THEN

	\ now check if real range has not shrunken too much:
	(last-scan-min-max)    2@ swap -	( displayed-range-size )
	(scan-min-max) 2@ swap -		( displayed real-range-size )
	swap horizontal-zoom-scale 2@ */ < IF			\ too small!
	    (scan-min-max) 2@ (last-scan-min-max) 2!	\ take real siz
	    horizontal notify-zoom-change
	THEN
    THEN

    (scan-detail-index) @ (last-scan-min-max) spot-detailed-scan ;
' spot-scan-display step-display-xt's >list
' spot-scan-display 1 >step-display

: |spot-scan-display| ( index -- )
    >r
    (scan-index) off		\ we take the first one...
    (scan-detail-index) @	\ preserve old values
    (scan-lines) @
    r> (scan-detail-index) !
    l-s (scan-lines) !		\ full screen display
    page spot-scan-display wait

    (scan-lines) !		\ restore old values
    (scan-detail-index) ! ;

MENU: spot-scan-men
: .menu-spot-scan ( -- )
    page

    title-colors ." Results of world scan." end-title

    cr ." spots scanned: " spots-scanned @ . cr

    21 menu-scroll-lines !	\ initialize scrolling

    cr
    field-planes# p0 scrolled-range ?DO		\ skip pointer plane?
	i field-variable-text	>r dup r>	\ 1st letter addr for key entry
	i >stack	 redisplay	['] |spot-scan-display|	menu-entry
	.tab
	(scan-spots) i 2* 2* cells + >r	\ start data for this quality
	20 at-x ." min: "		r@		@ . .tab
	40 at-x ." max: "		r@ cell+	@ . .tab
	60 at-x ." average: "	r> 2 cells +	@ . up-to-here cr
	c@ #key-same-entry			\ first letter key menu entry
    LOOP cr

    common-menu-entries ;

: spot-scan-menu ( -- )
    do-spot-scan
    spots-scanned @ IF
	spot-scan-men
	['] .menu-spot-scan menu-display-xt !
\	['] |spot-scan-display| to-do-after-xt !
	menu-done	['] noop	menu-key-default
	menu-done	['] noop	menu-default
	do-menu-loop
\	free-menus
    ELSE bell THEN ;

' spot-scan-menu function-key-actions >list
' spot-scan-menu F3-xt !


\  : detail-scan-menu-entry ( n addr count -- n+1 )
\      2 pick >stack	redisplay	menu-wait
\      ['] nuc-detailed-scan-prescanned	menu-entry
\      1+ ;


MENU: select-nuc-var-men
variable (index-selected)
: .menu-select-nuc-var ( -- )
    page

    title-colors ." Select nuc variable:" end-title

    21 menu-scroll-lines !	\ initialize scrolling

    cr
    nuc-scan-limit nuc-variables scrolled-range DO
	i dup >stack	(index-selected) >stack-2	menu-done
	nuc-var-name	over c@ >r		['] !	menu-entry cr
	r> #key-same-entry
    LOOP

    common-menu-entries ;

: menu-select-nuc-var ( -- )
    -1 (index-selected) !	\ impossible value
    select-nuc-var-men
    ['] .menu-select-nuc-var menu-display-xt !
    menu-done	['] noop	menu-key-default
    menu-done	['] noop	menu-default
    do-menu-loop ;

: |menu-select-nuc-var| ( -- )	\ sets (scan-detail-index)
    menu-select-nuc-var
    (index-selected) @ dup -1 <> IF
	(scan-detail-index) !
    ELSE drop THEN ;

: select-nuc-var-to-addr ( addr -- )
    menu-select-nuc-var
    (index-selected) @ dup -1 <> IF	\ selected something?
	swap !
    ELSE 2drop THEN ;


MENU: select-spot-var-men
: .menu-select-spot-var ( -- )
    page

    title-colors ." Select a spot variable:" end-title

    21 menu-scroll-lines !	\ initialize scrolling

    cr
    field-planes# p0 scrolled-range DO	\ skip pointer plane?
	i >stack	(index-selected) >stack-2
	i field-variable-text	>r dup r>	\ 1st letter addr for key entry
	menu-done	['] !	menu-entry cr
	c@ #key-same-entry			\ first letter key menu entry
    LOOP

    common-menu-entries ;

: menu-select-spot-var ( -- )
    select-spot-var-men
    -1 (index-selected) !		\ impossible value
    ['] .menu-select-spot-var menu-display-xt !
    menu-done	['] noop	menu-key-default
    menu-done	['] noop	menu-default
    do-menu-loop ;

: |menu-select-spot-var| ( -- )		\ sets (scan-detail-index)
    menu-select-spot-var
    (index-selected) @ dup -1 <> IF
	(scan-detail-index) !
    ELSE drop THEN ;

: select-spot-var-to-addr ( addr -- )
    menu-select-spot-var
    (index-selected) @ dup -1 <> IF	\ selected something?
	swap !
    ELSE 2drop THEN ;


\ Continuous display:  There are different types and sub types.
\ * displaying value of normal (world) variables like 'living'
\ * displaying value of values that need a nuc or spot scan with subtypes
\   minimal average and maximal
\ For each displayed item we need separate range control and zooming.
\ separate foreground color would be nice too.
\ These data (for each displayed parameter) could be stored in a list.
\ So for continuous display (scan-detail-index) could store the list xt.
\ variable (continuous-node)

\ variable (continuous-column)	(continuous-column) off	\ global

\ masks for '(scan-flags)'	one for a displayed *set* of parameters
0
MASK: cont-scan-nucs
MASK: cont-scan-spots
drop

\ structure for each displayed parameter:
0
OFFSET: >cont-xt	\ >cont-xt MUST be first. Word must be ( item -- value)
OFFSET: >cont-item
OFFSET: >cont-lower
OFFSET: >cont-upper
OFFSET: >cont-foreground-xt
OFFSET: >cont-char
nLIST: continuous-display-list		\ we have only one for the moment
decimal 12 CONSTANT max-continuous-items#

LIST: cont-variables
' living	cont-variables >list
' newborn	cont-variables >list
' cloned	cont-variables >list
' died		cont-variables >list
' nuc-do-cost	cont-variables >list
' code-price	cont-variables >list

LIST: cont-functions

\ dummy function to indicate missing selection
\ : (none) ;
' (none) cont-functions >list

: get-variable ( xt -- n )   EXECUTE @ ;
' get-variable cont-functions >list

: nuc-min ( i -- n )   (nuc-scan-data) @ cp! nuc-addr @ ;
' nuc-min cont-functions >list

: nuc-max ( i -- n )
    (nuc-scan-data) @ nuc-length# + cp!
    nuc-addr @ ;
' nuc-max cont-functions >list

: nuc-range ( i -- )   >r  r@ nuc-max  r> nuc-min - ;
' nuc-range cont-functions >list

: nuc-average ( i -- n )
    (nuc-scan-data) @ [ nuc-length# 2* ] literal + cp!
    nuc-addr @ ;
' nuc-average cont-functions >list

: spot-min ( i -- n )    2* 2* cells (scan-spots) + @ ;
' spot-min cont-functions >list

: spot-max ( i -- n )    2* 2* 1+ cells (scan-spots) + @ ;
' spot-max cont-functions >list

: spot-range ( i -- )   >r  r@ spot-max  r> spot-min  - ;
' spot-range cont-functions >list

: spot-average ( i -- n )    2* 2* 2 + cells (scan-spots) + @ ;
' spot-average cont-functions >list

: continuous-display-type ( xt -- code=0|1|2|-1 )
    CASE
	['] (none)	 OF -1 ENDOF	\ nothing 
	['] get-variable OF  0 ENDOF	\ variable
	['] nuc-min	 OF  1 ENDOF	\ nuc
	['] nuc-max      OF  1 ENDOF	\ nuc
	['] nuc-range    OF  1 ENDOF	\ nuc
	['] nuc-average  OF  1 ENDOF	\ nuc
	['] spot-min	 OF  2 ENDOF	\ spot
	['] spot-max	 OF  2 ENDOF	\ spot
	['] spot-range   OF  2 ENDOF	\ spot
	['] spot-average OF  2 ENDOF	\ spot

	ABORT" Unknown continuous display function"
    ENDCASE ;

: check-continuous-display-sanity ( -- )
    (scan-flags)
    dup @ [ cont-scan-nucs cont-scan-spots or invert ] literal and swap !

    continuous-display-list
    dup nodes 0 ?DO
	next-node >r				( r: node )

	r@ >cont-xt @ continuous-display-type	( type-code   r: node )
	CASE
	    0 OF	\ (world) variable
		BEGIN
		    r@ >cont-item @
		    cont-variables
		    dup nodes 0 DO
			next-node
			2dup @ = IF drop TRUE LEAVE THEN
		    LOOP
		    nip true <>
		WHILE
		    s" Select variable:" title-choose-xt!
		    cont-variables  r@ >cont-item  choose-xt-menu
		REPEAT

		\ range from 0 to spots:
		r@ >cont-item @
		dup  ['] living =
		dup IF
		    r@ >cont-upper dup @ 0= IF spots swap ! ELSE drop THEN
		THEN
		over ['] newborn = or
		over ['] cloned = or
		swap ['] died = or
		IF
		    r@ >cont-lower dup @ 0 max swap !
		    r@ >cont-upper dup @ spots min swap !
		THEN
	    ENDOF
	    1 OF	\ nuc variable
		(scan-flags) dup @ cont-scan-nucs or swap !
		r@ >cont-item @  nuc-variables  nuc-scan-limit 1-  within 0= IF
		    menu-select-nuc-var (index-selected) @ dup -1 <> IF
			r@ >cont-item !
		    ELSE drop THEN
		THEN
	    ENDOF
	    2 OF	\ spot variable
		(scan-flags) dup @ cont-scan-spots or swap !
		r@ >cont-item @	p0  field-planes#  within 0= IF
		    menu-select-spot-var (index-selected) @ dup -1 <> IF
			r@ >cont-item !
		    ELSE drop THEN
		THEN
	    ENDOF
	ENDCASE

	r@ >cont-xt @ dup ['] nuc-range = swap  ['] spot-range = or IF
	    r@ >cont-lower dup @ 0 max swap !
	THEN

	r@ >cont-lower @  r@ >cont-upper @ < 0= IF
	    r@ >cont-lower @ 10 + r@ >cont-upper !	\ arbitrary range
	THEN

	r@ >cont-char dup @ 0= IF [char] * swap ! ELSE drop THEN
	r@ >cont-foreground-xt dup @ 0= IF
	    ['] white swap !
	ELSE drop THEN

	r>
    LOOP drop ;

: cont-item-name ( node -- addr count )
    >r
    r@ >cont-xt @ continuous-display-type CASE
	-1 OF s" "			    ENDOF
	 0 OF r@ >cont-item @ xt>name       ENDOF
	 1 OF r@ >cont-item @ nuc-var-name  ENDOF
	 2 OF r@ >cont-item @ field-variable-text ENDOF
     ENDCASE
     rdrop ;

: -3-norm ( addr count --- addr' count' )
    s" -" search IF
	>r 1+ r> 1- 3 min
    ELSE
	8 min
    THEN ;

\ word to type but only until end of line
: ?type ( addr count -- )
    at? drop >r r@ + c-l min r> - type
    at? drop c-l < IF bl emit THEN ;

variable (cont-y)			\ start row
: .cont-status-line ( -- )
    0  (cont-y) @ (scan-lines) @ + 1-  at-xy
    continuous-display-list
    dup nodes 0 ?DO
	next-node >r					( r: node )
	r@ >cont-xt @
	dup ['] (none) <> IF
	    r@ >cont-foreground-xt @ EXECUTE color-foreground
	    r@ >cont-char @ pad c! pad 1 ?type
	    default-color color-foreground
	    xt>name -3-norm ?type
	    r@ cont-item-name ?type
	    s" " ?type
	ELSE drop THEN
	r>
    LOOP
    drop
    clear-line-to-end ;

2variable cont-zoom-up-scale		2 1 cont-zoom-up-scale 2!
: continuous-display ( -- )
    (scan-flags) @
    dup cont-scan-nucs and IF		\ we need a nuc scan?
	do-a-nuc-scan
    THEN
    cont-scan-spots and IF		\ we need a spot scan?
	do-spot-scan
    THEN

    at? (cont-y) ! drop

    (continuous-column) @ 0= IF		\ clear screen
	(cont-y) @ dup (scan-lines) @ + 1- swap ?DO
	    0 i at-xy clear-line-to-end
	LOOP
	.cont-status-line
    THEN

    continuous-display-list
    dup nodes 0 ?DO
	next-node >r					( r: node )
	r@ >cont-item @  r@ >cont-xt @
	dup ['] (none) <> IF
	    EXECUTE					( value   r: node )

	    \ check range:
	    dup r@ >cont-lower @ < IF
		r@ >cont-upper @  dup  r@ >cont-lower @ -
		cont-zoom-up-scale 2@ */  -
		over min r@ >cont-lower !
		check-continuous-display-sanity		\ range check
		(zoomed) on
	    THEN
	    dup r@ >cont-upper @ > IF
		r@ >cont-lower @  dup  r@ >cont-upper @ -
		cont-zoom-up-scale 2@ */  -
		over max r@ >cont-upper !
		check-continuous-display-sanity		\ range check
		(zoomed) on
	    THEN

	    r@ >cont-upper @  swap -
	    (scan-lines) @ 2 -
	    r@ >cont-upper @  r@ >cont-lower @  -   */
	    (cont-y) @ +  (continuous-column) @ swap at-xy

	    (zoomed) @ IF
		cyan color-foreground
		(zoomed) off
	    ELSE
		r@ >cont-foreground-xt @ EXECUTE color-foreground
	    THEN

	    r@ >cont-char @ emit

	    default-color color-foreground
	ELSE
	    2drop
	THEN

	r>
    LOOP drop
    (continuous-column) dup @ 1+ c-l mod swap ! ;
' continuous-display step-display-xt's >list

: add-continuous-entry ( -- )
    \ 'check-continuous-display-sanity' does the rest
    continuous-display-list nodes max-continuous-items# < IF
	(index-selected) off		\ abused variable (index-selected)
	s" Add a new continuous display function:" title-choose-xt!
	cont-functions (index-selected) choose-xt-menu
	(index-selected) @ dup IF
	    continuous-display-list new-node dup >r !
	    ['] white r> >cont-foreground-xt !
	ELSE drop THEN
    ELSE
	page bell cr cr green color-foreground
	." Not more than " max-continuous-items# . ." items possible."
	default-color color-foreground
	2500 ms
    THEN ;


LIST: cont-display-colors
color-list cont-display-colors copy-simple-list-elements

MENU: continuous-display-men
: .menu-continuous-display ( -- )
    check-continuous-display-sanity

    page

    title-colors ." Setup continuous display:"
    step-display-on display-switch @ and 0= IF
	green color-foreground
	s"   Step display is off, nothing will be shown!"	redisplay
	step-display-on >stack  display-switch >stack-2 ['] xor! menu-entry
	default-color color-foreground
    THEN
    end-title

    cr
\    (scan-detail-index) @ EXECUTE	\ get list addr
    continuous-display-list
    dup nodes 0 ?DO
	next-node >r			( r: node )

	s" Set display function of this item:" title-choose-xt!
	cont-functions >stack	r@ >cont-xt >stack-2
	r@ >cont-xt @ xt>name	['] choose-xt-menu	menu-entry

	r@ >cont-xt @ ['] (none) <> IF
	    1 5 screen-column
	    r@ >cont-xt @ continuous-display-type	( type-code   r: node )
	    CASE
		0 OF	\ (world) variables
		    s" Select variable:" title-choose-xt!
		    cont-variables >stack	r@ >cont-item >stack-2
		    r@ >cont-item @ xt>name	['] choose-xt-menu   menu-entry
		ENDOF
		1 OF	\ nuc variables
		    s" Select nuc variable:" title-choose-xt!
		    r@ >cont-item dup >stack	redisplay
		    @ nuc-var-name	['] select-nuc-var-to-addr   menu-entry
		ENDOF
		2 OF	\ spot variables
		    s" Select spot variable:" title-choose-xt!
		    r@ >cont-item dup >stack  redisplay   @ field-variable-text
		    ['] select-spot-var-to-addr menu-entry
		ENDOF
	    ENDCASE

	    2 5 screen-column
	    \ >xy
	    r@ >cont-char simple-menu-entry-char ."   "

	    s" Choose color for this char:" title-choose-xt!
	    cont-display-colors >stack	r@ >cont-foreground-xt dup >stack-2
	    @ xt>name	['] choose-xt-menu	menu-entry

	    3 5 screen-column s" low: "  r@ >cont-lower simple-menu-entry-value
	    4 5 screen-column s" high: " r@ >cont-upper simple-menu-entry-value
	THEN
	cr

	r>
    LOOP drop

    continuous-display-list nodes max-continuous-items# < IF
	cr
	s" add new entry"  redisplay  ['] add-continuous-entry	menu-entry cr
	s" aAnNeE" menu-same-key-entry
    THEN

    cr
    s" Zoom up scale: " redisplay cont-zoom-up-scale simple-menu-entry-scale cr
;

: menu-continuous-display ( -- )
    continuous-display-men
    ['] .menu-continuous-display menu-display-xt !
    menu-done	['] noop	menu-key-default
    menu-done	['] noop	menu-default
    do-menu-loop ;

' menu-continuous-display function-key-actions >list

true [IF] \ TEST
' continuous-display 0 >step-display
' continuous-display-list (scan-detail-index) !
l-s 1- (scan-lines) !
cont-scan-nucs (scan-flags) !

continuous-display-list new-node
' nuc-average over >cont-xt !
18 over >cont-item !
98 over >cont-lower !
160 over >cont-upper !
' red over >cont-foreground-xt !
char * over >cont-char !
drop

continuous-display-list new-node
' get-variable over >cont-xt !
' living over >cont-item !
0 over >cont-lower !
spots over >cont-upper !
' white over >cont-foreground-xt !
char x over >cont-char !
drop
[THEN]


\ step display menu:
: displayed-items-name ( -- addr count)	\ (scan-index) must be set
    (scan-xt) @ CASE
	['] spot-scan-display OF
	    (scan-detail-index) @ field-variable-text
	ENDOF
	['] nuc-scan-display  OF
	    (scan-detail-index) @ nuc-var-name
	ENDOF
    \ xt zero or not recognized:
    s" (no display)" rot
    ENDCASE ;

\ user interface to set '(scan-detail-index)' (according to displayed type)
: choose-displayed-item ( index -- )
    (scan-index) !
    (scan-xt) @ CASE
	['] spot-scan-display  OF |menu-select-spot-var|	ENDOF
	['] nuc-scan-display   OF |menu-select-nuc-var|		ENDOF
	['] continuous-display OF menu-continuous-display	ENDOF

	\ xt zero or not recognized:
	bell
    ENDCASE ;    

: step-display-lines ( -- lines ) 
    0
    step-display-items @ 0 ?DO
	i (scan-index) !
	(scan-xt) @ IF
	    (scan-lines) @ +
	THEN
    LOOP ;

: ?step-display-sanity ( -- )
    \ reset 'continuous-display-used' bit:
    display-switch dup @ continuous-display-used invert and swap !
    step-display-on display-switch @ and 0= IF EXIT THEN

    step-display-items dup @ max-step-display-items# min 1 max swap !

    0
    step-display-items @ 0 ?DO
	i (scan-index) !
	(scan-xt) @ dup IF
	    \ check if '(scan-detail-index)' is within range:
	    CASE \ on xt
		['] continuous-display OF
		    display-switch dup @ continuous-display-used or swap !
		    check-continuous-display-sanity	\ convenient place...
		    lowest-integer#	highest-integer#
		ENDOF
		['] spot-scan-display OF p0		field-planes#  1- ENDOF
		['] nuc-scan-display  OF nuc-variables	nuc-scan-limit 1- ENDOF
		\                        ^ lower limit  ^ upper limit
		ABORT" ?step-display-sanity: unknown xt."
	    ENDCASE
	    (scan-detail-index) >r r@ @ min  swap max r> !

	    \ check if '(scan-lines)' must be initialized:
	    (scan-lines) @ 0= IF
		last-line step-display-items @ / (scan-lines) !
	    THEN

	    \ check if 'horizontal-zoom-scale' must be initialized:
	    horizontal-zoom-scale 2@ 0= swap 0= or IF
		default-horizontal-zoom-scale horizontal-zoom-scale 2!
	    THEN

	    \ check if 'vertical-zoom-scale' must be initialized:
	    vertical-zoom-scale 2@ 0= swap 0= or IF
		default-vertical-zoom-scale vertical-zoom-scale 2!
	    THEN

	    \ check if 'scan-background-xt' must be initialized:
	    scan-background-xt @ 0= IF		\ we only check for zero
		['] default-color scan-background-xt !
	    THEN

	    \ check if 'scan-foreground-xt' must be initialized:
	    scan-foreground-xt @ 0= IF		\ we only check for zero
		['] default-color scan-foreground-xt !
	    THEN

	    (scan-lines) @ +	\ sum up lines
	ELSE drop THEN
    LOOP

    last-line > IF
	last-line step-display-items @ /
	step-display-items @ 0 ?DO
	    i (scan-index) !
	    (scan-xt) @ IF
		(scan-lines) @  over  min  (scan-lines) !
	    THEN
	LOOP
	drop
    THEN ;

: define-step-display-item ( index -- )
    (scan-index) !
    s" Select type of this step display item:" title-choose-xt!
    step-display-xt's  (scan-xt)  choose-xt-menu
    (scan-xt) @ IF
	s" Select displayed parameter:" title-choose-xt!
	(scan-index) @ choose-displayed-item
    THEN ;

: |define-step-display-item| ( -- )
    step-display-items @ 1 -		\ try to guess index ;-)
    define-step-display-item ;

: step-display-adapt-size ( -- )
    last-line step-display-items @ /mod
    step-display-items @ 0 ?DO
	i (scan-index) !
	dup (scan-lines) !
    LOOP

    drop
    ( mod ) 0 ?DO
	i (scan-index) !
	1 (scan-lines) +!
    LOOP ;

MENU: step-display-men
defer toggle-step-display
variable (step-more-info)	(step-more-info) off
: .menu-step-display ( -- )
    page
    ?step-display-sanity

    title-colors
    s" Step display menu:   Step display "	redisplay
    ['] toggle-step-display	menu-entry
    display-switch @ step-display-on and .ON-off
    6 spaces from-here
    (step-more-info) @ IF ." Less" ELSE ." More" THEN
    s"  infos"  (step-more-info) >stack  ['] toggle-addr redisplay  menu-entry
    s" mMlLtToO+- " menu-same-key-entry
    end-title

    cr
    s" Items displayed in step display: "
    step-display-items simple-menu-entry-value
    s" iInN" menu-same-key-entry
    cr

    (step-more-info) @ IF 5 ELSE 16 THEN menu-scroll-lines !
    cr
    false 			\ flag/counter if continuous display is used
    step-display-items @ 0 scrolled-range ?DO
	i (scan-index) !

	(scan-xt) @ dup IF
	    dup xt>name i >stack  ['] define-step-display-item	menu-entry
	    1 4 screen-column
	    dup dup ['] nuc-scan-display = swap ['] spot-scan-display = or IF
		drop
		from-here  ." scanning: " displayed-items-name  redisplay
		i >stack	['] choose-displayed-item menu-entry
	    ELSE
		['] continuous-display = IF 1+ THEN		\ set flag

		s" (edit continuous display)" ['] menu-continuous-display
		redisplay	menu-entry
		s" cCeE" menu-same-key-entry
	    THEN
	    3 5 screen-column s" lines:  " (scan-lines) simple-menu-entry-value
	    3 4 screen-column
	    s" v-range: " (vertical-display-range) simple-menu-entry-value
	    cr

	    (step-more-info) @ IF
		1 4 screen-column
		from-here
		s" zoom starts horiz: "
		horizontal-zoom-scale simple-menu-entry-scale
		s" 	vertic: "
		vertical-zoom-scale simple-menu-entry-scale
\		3 4 screen-column
\		s" v-range: " (vertical-display-range) simple-menu-entry-value
		cr
		1 4 screen-column
		s" Select background color for this item" title-choose-xt!
		s" bg color: "
		scan-background-colors >stack	scan-background-xt >stack-2
		['] choose-xt-menu	redisplay	menu-entry
		scan-background-xt @ xt>name	type	up-to-here
		3 5 screen-column
		s" Select foreground color for this item" title-choose-xt!
		s" fg: "
		scan-foreground-colors >stack	scan-foreground-xt >stack-2
		['] choose-xt-menu	menu-entry
		scan-foreground-xt @ xt>name	type	up-to-here
		cr
	    THEN
	ELSE
	    drop
	    from-here ." (undefined)"
	    1 4 screen-column s" (define new display item)"
	    i >stack	['] define-step-display-item	menu-entry cr
	THEN
    LOOP

    \ add or remove entries:
    step-display-items @ >r
    r@ max-step-display-items# <   r@ 1 > OR IF
	cr
	r@  max-step-display-items# < IF
	    s" add new item.	    "	redisplay	do-after
	    1 >stack	step-display-items >stack-2	['] +!	menu-entry
	    s" aAnN+" menu-same-key-entry
	THEN
	r@  1 > IF
	    s" remove last item."	redisplay
	    -1 >stack	step-display-items >stack-2	['] +!	menu-entry
	    s" rR-" menu-same-key-entry
	THEN
	cr
    THEN rdrop

    ( counter ) dup IF
	cr
	1 > IF
	    bell
	    green color-foreground
	    s" Please set just *one* continuous display."
	    ping	['] noop menu-entry cr
	    default-color color-foreground
	THEN
    ELSE drop THEN

    \ is the screen filled?
    step-display-lines last-line < IF
	green color-foreground
	s" fill the screen" redisplay
	['] step-display-adapt-size	menu-entry cr
	default-color color-foreground
	s" sSfF=" menu-same-key-entry
    THEN

    common-menu-entries ;

: menu-step-display ( -- )
    step-display-men
    ['] |define-step-display-item| to-do-after-xt !
    ['] .menu-step-display menu-display-xt !
    menu-done	['] noop	menu-key-default
    menu-done	['] noop	menu-default
    do-menu-loop ;
' menu-step-display function-key-actions >list

MENU: nuc-scan-men
: .menu-nuc-scan ( -- )
    page

    title-colors ." Nuc scan results." end-title

    ." Nucs scanned: " nucs-scanned @ .
    .tab ." mutated nucs on trial: " trial @ .
    .tab ." compiled genes: " compiled-genes @ . cr

    21 menu-scroll-lines !	\ initialize scrolling

    cr
    nuc-scan-range 2@ scrolled-range DO
	i >stack   redisplay	menu-wait   i nuc-var-name	over c@ >r
	['] nuc-detailed-scan-prescanned	menu-entry cr
	r> #key-same-entry
    LOOP

    cp@ >r
    3 0 DO
	0 3 at-xy
	(nuc-scan-data) @ nuc-length# i * + cp!
	nuc-scan-range 2@ scrolled-range DO
	    j 18 * 24 + at-x
	    j CASE
		0 OF ." min: " ENDOF
		1 OF ." max: " ENDOF
		2 OF ." average: " ENDOF
	    ENDCASE
	    i nuc-addr @ .
	    cr
	LOOP

    LOOP
    r> cp!

    common-menu-entries ;

: nuc-scan-menu ( -- )
    do-a-nuc-scan
    nucs-scanned @ IF
	nuc-scan-men
	['] .menu-nuc-scan menu-display-xt !
	menu-done	['] noop	menu-key-default
	menu-done	['] noop	menu-default
	do-menu-loop
\	free-menus
    ELSE bell THEN ;
' nuc-scan-menu function-key-actions >list
' nuc-scan-menu F4-xt !

\ set field element i to a cloned nuc of the actual one
: set-field-element ( i -- addr=flag )	\ returns addr of the clone or 0
    dup someone-here?			\ is it free?
    IF drop false			\ no, sorry
    ELSE				\ ( i )
	spot @ >r    >spot!		\ ( --   R:old-cp@, spot)
	clone				\ ( addr flag | false  R:old-cp@, spot)
	IF				\ cloning was possible
	    dup fcp !
	    spot @ child-spot !		\ spot of youngest child
	ELSE false THEN			\ cloning didn't work
	r> >spot!
    THEN ;

support-single-individuals? [IF]
: nuc-unlink ( -- )			\ remove actual cell from the world
    world? @ IF				\ this trick makes it work as a noop
	future				\ if it's called with an individual
	fcp off
	present
	fcp off		\ hmm, I'm not sure about this... ######
    THEN ;
[ELSE] \ support-single-individuals? is FALSE
: nuc-unlink ( -- )			\ remove actual cell from the world
    future				\ if it's called with an individual
    fcp off
    present
    fcp off ;		\ hmm, I'm not sure about this... ######
[THEN]
' nuc-unlink is <disappear>		\ unlinks aborted cells from the field

: free-spots ( -- n )			\ how many spots are not inhabited?
    0					\ counter
    spots 0 DO				\ loop over all spots
	i someone-here? 0= IF 1+ THEN	\ count
    LOOP ;

\ variable newborn		newborn off
: erase-field ( -- )
    world @ IF				\ for convenience
	world @ field-length erase	\ clean up the whole universe
    THEN

    living off				\ programmers aesthetics
    newborn off
    cloned off
    died off
    -1 step !				\ hmm? not really sure about that
    nuc-do-cost off
    code-price off
    false cp! ;

defer ?record-free-field
: free-field ( -- )			\ lets everybody die and cleans up
    world @ IF
	spots 0 DO			\ everywhere
	    i >spot!  fcp @		\ who's here?
	    dup IF			\ somebody?
		cp!			\ set as actual nuc
		die			\ sorry
	    ELSE drop THEN
	LOOP
    THEN
    erase-field				\ clean everything up
    ?record-free-field ;

\ orientation in space
: i-normalize ( i offset -- i' )	\ normalizes the indizes
    + spots mod ;			\ this universe wraps at it's borders

: left-spot  ( i -- i' )    spots 1-	i-normalize ;	\ one step to the left
: right-spot ( i -- i' )    1		i-normalize ;	\ one step to the right
: upper-spot ( i -- i' )    spots c-l -	i-normalize ;	\ one step up
: lower-spot ( i -- i' )    c-l        	i-normalize ;	\ one step down

true [IF]				\ 4 or 8 directions?

8 CONSTANT directions			\ we have 8 directions
: direction>index ( direction -- i')	\ spot must be set
    directions mod			\ ( normalized-direction )
    spot @ swap				\ ( actual-index normalized-direction )
    CASE
	0 OF upper-spot			ENDOF
	1 OF upper-spot	right-spot	ENDOF
	2 OF right-spot			ENDOF
	3 OF right-spot	lower-spot	ENDOF
	4 OF lower-spot			ENDOF
	5 OF lower-spot	left-spot	ENDOF
	6 OF left-spot			ENDOF
	     left-spot	upper-spot
    ENDCASE ;

[ELSE]
4 CONSTANT directions			\ we have only 4 directions right now
: direction>index ( direction -- i')	\ spot must be set
    directions mod			\ ( normalized-direction )
    spot @ swap				\ ( actual-index normalized-direction )
    dup 0=  IF drop upper-spot	EXIT THEN
    dup 1 = IF drop right-spot	EXIT THEN
    dup 2 = IF drop lower-spot	EXIT THEN
	3 = IF      left-spot        THEN ;
[THEN]

: free-neighbor-spot? ( -- i' true | false )	\ free neighboring spot?
    directions random-ranged >r			\ ( R: random-direction )
    0						\ ( z=0 )
    BEGIN					\ ( z ) test all directions
	dup r@ +				\ ( z z+random-direction )
	direction>index				\ ( z i' )
	dup someone-here?			\ ( z i' flag )
	0= IF					\ ( z i'ok )
	    rdrop nip true EXIT			\ RETURN ( -- i' true )
	THEN drop				\ ( z )
	1+ dup directions < WHILE		\ ( z+1 )
    REPEAT drop
    rdrop false ;				\ RETURN ( -- false )

support-single-individuals? [IF]
: .world ( -- )
    world @ IF				\ check if there's something
	world? @ >r			\ save world?
	world? on			\ could be needed
	cursor-off
	0 0 at-xy			\ puts cursor in the upper left corner
	spots 0 DO			\ for each and every spot in this world
	    i dup >spot!		\ look at the spot 
	    someone-here? dup IF	\ if there's someone
		cp!			\ make it the actual one
		set-colors
		show-me-xt @ EXECUTE	\ and show his face
	    ELSE
		drop
		show-background		\ empty
	    THEN
	LOOP
	set-default-colors
	cursor-visible
	r> world? !			\ restore world?
    ELSE ." there's nothing to look at" cr THEN ;    

\ variable bowl				\ individuals get a special bowl
: (food)  ( -- addr)			\ where can I eat?
    world? @ IF				\ field or individual?
	tp offset-food +		\ actual food plane 
	spot-offset @ +			\ at the right spot
    ELSE				\ exception: it's an individual
	bowl				\ give him his special bowl
    THEN ;
[ELSE] \ support-single-individuals? is FALSE
: .world ( -- )
    world @ IF				\ check if there's something
	cursor-off
	0 0 at-xy			\ puts cursor in the upper left corner
	spots 0 DO			\ for each and every spot in this world
	    i dup >spot!		\ look at the spot 
	    someone-here? dup IF	\ if there's someone
		cp!			\ make it the actual one
		set-colors
		show-me-xt @ EXECUTE	\ and show his face
	    ELSE
		drop
		show-background		\ empty
	    THEN
	LOOP
	set-default-colors
	cursor-visible
    ELSE ." there's nothing to look at" cr THEN ;    

: (food)  ( -- addr)			\ where can I eat?
	tp offset-food +		\ actual food plane 
	spot-offset @ +	;		\ at the right spot
[THEN]
' (food) is <food>	\ works for individuals *and* fields now

\ food that's shared between all living beings in one world
\ (will soon be made local to the world we're in)
\ that's good for population control
\ the amount will be *added* to the other sources of food
variable world-food-supply			\ worlds food supplies
world-food-supply off
\ this supply will be divided by the number of living cells and be put
\ into the following variable:
\ variable food-common-share			\ how much each one gets
: determine-food-share ( -- )			\ sets food-common-share
    living @ 0= IF
	spots free-spots - living ! 
    THEN
    living @ IF
	world-food-supply @  living @  /  food-common-share !
    \ ELSE					\ don't know what to do here
    THEN ;

\ food menu, we go to 'The Restaurant' now:
MENU: food-men
: .menu-food ( -- )
    page

    title-colors
    s" Food menu:"	['] noop		menu-entry
    end-title

    cr
    s" World food supply: " world-food-supply	simple-menu-entry-value cr
    s" wW" menu-same-key-entry

    world-food-supply @ IF
	."    (divided among all " living @ ?dup IF . THEN
	." living beings giving "
	determine-food-share food-common-share @ . ." for each one)" cr
    THEN

    cr
    s" Each living individual get's in addition: "
    individual-fixed-food-share			simple-menu-entry-value cr
    s" iI" menu-same-key-entry

    cr
    s" On each spot, inhabited or not, is further deposed: "
    food-share/spot				simple-menu-entry-value cr
    s" sS" menu-same-key-entry

    cr
    s" Actually each step in live costs: "
    nuc-do-cost					simple-menu-entry-value cr
    s" nN" menu-same-key-entry

    cr
    s" Price to pay for code usage: "
    code-price					simple-menu-entry-value
    s" cC" menu-same-key-entry

    .tab .tab
    s" rating scale "	code-price-scale	simple-menu-entry-scale cr
    s" rR" menu-same-key-entry

    common-menu-entries ;

: food-menu ( -- )
    food-men
    ['] .menu-food menu-display-xt !
    menu-done	['] noop	menu-key-default
    menu-done	['] noop	menu-default
    do-menu-loop
    free-menus ;
' food-menu function-key-actions >list

: eat-all ( -- )  <food> @ energy +!  <food> off ;
' eat-all eat-actions >list

: feed-world ( -- )		\ give food everywhere
    spots 0 DO i >spot! feed-spot LOOP ;
' feed-world function-key-actions >list

: spot>at ( -- )			\ sets the cursor depending spot
    spot @ c-l /mod at-xy ;

\ cells work on *present* food plane, 'world-do' does copy it to the future.
: food>future ( -- )			\ copy present food to future
    offset-food >r
    r@ tp +  future r> tp + present  plane-length  move ;

future-change-individal 0= [IF]		\ copy all quality changes in one move?
: qualities>future ( -- )		\ copy present qualities to future
    tp offset-A-quality +
    future tp present offset-A-quality +
    [ plane-length spot-qualities# * ] literal move ;	\ with one move
[THEN]

\ run the world trough one step in time
\ the spots are activated in selectable order
variable spot-do-xt	    ' noop spot-do-xt !
variable cell-do-before-xt  ' noop cell-do-before-xt !
variable cell-do-after-xt   ' noop cell-do-after-xt !
variable world-do-direction	1 world-do-direction !
\ 1 forward -1 backward
\ 0 random
\ 2 -2 alternate
: world-do ( -- )
[ support-single-individuals? ] [IF]
    world? on		\ flag: somethings behave different for individuals
[THEN]
    feed-world			\ puts food everywhere
    determine-food-share 	\ how much aditional food for living cells
    present-initializes-future \ if nothing changes the future, nothing changes
    living off				\ counter
    died off				\ ditto
    trial off				\ ditto
    mutations off			\ ditto
    cloned @ newborn !			\ temporally used as storage

    world-do-direction @ ?dup IF 
	0> IF
	    [ spot-display-on step-display-on or ] literal
	    display-switch @ and IF 0 0 at-xy THEN

	    spots 0 DO
		i >spot!		\ set spot actual
		spot-do-xt @ EXECUTE	\ optional actions
		fcp @ dup IF		\ hello, somebody here?
		    cp!			\ yes, set cell as actual one

		    log-mask @ IF \ checking logging (for speed reasons)
			spot 2@ swap ( spot step )
			num>string	cat-log
			s" :"		cat-log
			num>string	cat-log
			s"  "		cat-log
			log-cat-id
			nuc-flags @ wake-me-trial and IF
			    s"  on trial"
			    [ log-spot log-trial or ] literal log-it
			ELSE
			    s" " log-spot log-it
			THEN
		    THEN

		    feed-individual \ feed the food only given to living cells

		    cell-do-before-xt @ EXECUTE	\ you can put something in here
		    nuc-do-all			\ and wake her up
		    cell-do-after-xt @ EXECUTE	\ you can put something in here
		ELSE					\ noone here
		    drop
		    display-switch @ spot-display-on and IF \ spot display on?
			show-background
		    THEN
		THEN
	    LOOP
	ELSE
	    0 spots 1- DO
		i >spot!		\ set spot actual
		spot>at			\ set cursor		
		spot-do-xt @ EXECUTE	\ optional actions
		fcp @ dup IF		\ hello, somebody here?
 		    cp!			\ yes, set cell as actual one

		    log-mask @ IF \ checking logging (for speed reasons)
			spot @ num>string cat-log s" : " cat-log
			log-cat-id   s" " log-spot log-it
		    THEN

		    feed-individual \ feed the food only given to living cells

		    cell-do-before-xt @ EXECUTE	\ you can put something in here
		    nuc-do-all			\ and wake her up
		    cell-do-after-xt @ EXECUTE	\ you can put something in here
		ELSE					\ noone here
		    drop
		    display-switch @ spot-display-on and IF \ spot display on?
			show-background
		    THEN
		THEN
	   -1 +LOOP
	THEN
	world-do-direction @ dup
	abs 2 = IF negate world-do-direction ! ELSE drop THEN \ alternate
    ELSE			\ call randomly choosen spots
	spots 0 DO
	    spots random-ranged
	    >spot!			\ set spot actual
	    spot>at			\ set cursor		
	    spot-do-xt @ EXECUTE	\ optional actions
	    fcp @ dup IF		\ hello, somebody here?
		cp!			\ yes, set cell as actual one

		log-mask @ IF \ checking logging (for speed reasons)
		    spot @ num>string cat-log s" : " cat-log
		    log-cat-id   s" " log-spot log-it
		THEN

		feed-individual \ feed the food only given to living cells

		cell-do-before-xt @ EXECUTE	\ you can put something in here
		nuc-do-all			\ and wake her up
		cell-do-after-xt @ EXECUTE	\ you can put something in here
	    ELSE					\ noone here
		drop
		display-switch @ spot-display-on and IF \ spot display on?
		    show-background
		THEN
	    THEN
	LOOP	    
    THEN
    set-default-colors
    food>future
    
[ future-change-individal 0= ] [IF]	\ copy all quality changes in one move?
    future-quality-change @ IF qualities>future THEN
[THEN]

\    time-step
    cloned @ newborn @ - newborn !

    display-switch @ step-display-on AND IF	\ we want a scan display?
	step-display
    THEN

    time-step
;

\ cell division with possible diversification and mutation of one part
defer <diversify>
' noop IS <diversify>
\ defer <mutate>
\ ' noop IS <mutate>
variable cell-division-moves-both	cell-division-moves-both off
variable cell-division-diversify-both	cell-division-diversify-both off
variable cell-division-mutate-both	cell-division-mutate-both off
: cell-division ( -- )		\ mimics cell division in biology
    free-neighbor-spot? IF	\ ( i')
	dup future someone-here? present 0= IF
	    energy @ 2/ energy !	\ share energies
	    age dup @ >r off		\ remember age and reset it to zero
	    1 generation +!
	    1 genom-generation +!
	    future ( i ) set-field-element present	\ ( addr=flag )
	    ?dup IF     ( addr)	\ :-)   congratulations!
		log-mask @ IF
		    log-cat-id
		    s" : reproduction: cell division" log-birth log-it
		THEN
		cp@ >r			\ remember actual cell
		cp! <diversify>		\ diversify the newborn clone?
		<mutate>		\ mutate it?
		new-id id !
		log-mask @ IF
		    s" child as "		cat-log   log-cat-id
		    s"  to spot "		cat-log
		    child-spot @ num>string	log-birth log-it
		THEN
		r> cp!			\ restore cp
		rdrop			\ forget the age, I'm new born :)
		cell-division-moves-both @ IF	\ move mother cell too?
		    free-neighbor-spot? IF	\ ( i')
			dup future someone-here? present 0= IF
			    future ( i ) set-field-element present \ addr=flag
			    IF     		\ :-)   congratulations!
				nuc-unlink	\ remove at old spot
			    THEN
			ELSE drop
			THEN
		    THEN
		THEN
		cell-division-mutate-both @ IF
		    <mutate>
		THEN
		cell-division-diversify-both @ IF
		    <diversify>
		THEN
	    ELSE 		\ :-(	cloning didn't work, reset values
		energy @ 2* energy !	\ disregarding rounding errors
		r> age !
		-1 generation +!
		-1 genom-generation +!
	    THEN
	ELSE \ uups, somebody else wants to go here too. Let's be kind...
	    drop
	THEN
\   ELSE			\ :-( no adjacent spot free, noop
    THEN ;
' cell-division reproduce-actions >list

FALSE [IF] \ new version didn't bring much.

variable (sow-diversified)	(sow-diversified) off
defer ?record-sow ( n -- )
: sow ( n -- n' )	\ sow n clones of the actual nuc into the field
\ like in real life sow does *not* insist if the spot is not ok
\ it sows n seeds, and dosn't care, if they do not fall on fertile ground
\ it returns n', the number of succesfully sowed nucs
    world @ 0= IF
	." sorry, no gardening before the big bang " cr	drop false EXIT
    THEN
    dup 0> IF
	dup ?record-sow
	dup >r				\ ( R: n )
[ support-single-individuals? ] [IF]
	world? on
[THEN]
	dup 0 DO			( n )	\ try it n times, no more
	    spots random-ranged		( n i )			\ random spot
	    future			\ we're working always for the future
	    dup set-field-element	( n i flag=addr )   \ set if possible
	    dup IF					\ yes?
		cp@ >r   cp!	( n i  r: old-cp@ )

		new-id id !
		(sow-diversified) @ IF  <diversify>  THEN

		\ to avoid troubles in record files we set at present too
		\ we don't need a time-step any more
		present	someone-here?		\ think this shouldn't happen
		ABORT" sow: Spot occupied at present, free at future."	\ !
		cp@ fcp !

		r> cp!
		1-				\ count decreasing n
	    ELSE 2drop THEN	( n' )
	LOOP	( n' )
	present				\ back in the present
	r> swap -			\ calculate success number
	(sow-diversified) off
    ELSE drop false THEN ;		\ n < 1, error
    
[ELSE] \ old
variable (sow-diversified)	(sow-diversified) off
defer ?record-sow ( n -- )
: sow ( n -- n' )	\ sow n clones of the actual nuc into the field
\ like in real life sow does *not* insist if the spot is not ok
\ it sows n seeds, and dosn't care, if they do not fall on fertile ground
\ it returns n', the number of succesfully sowed nucs
    world @ 0= IF
	." sorry, no gardening before the big bang " cr	drop false EXIT
    THEN
    dup 0> IF
	dup ?record-sow
	dup >r				\ ( R: n )
[ support-single-individuals? ] [IF]
	world? on
[THEN]
	future				\ we're working always for the future
	dup 0 DO			( n )	\ try it n times, no more
	    spots random-ranged	( n i )		\ random spot
	    set-field-element	( n flag )	\ set if possible
	    dup IF				\ yes?
		cp@ >r   cp!	( n   r: old-cp@ )
		new-id id !
		(sow-diversified) @ IF
		    <diversify>
		THEN
		r> cp!
		1-				\ count decreasing n
	    ELSE drop THEN	( n' )
	LOOP	( n' )
	present				\ back in the present
	r> swap -			\ calculate success number
	(sow-diversified) off
    ELSE drop false THEN ;		\ n < 1, error
[THEN]

: sow-diversified ( n -- n' )	(sow-diversified) on	sow ;


\ genes
\ VOCABULARY genes
\ also genes definitions

include genes/genes-basic-stack.fs
include genes/genes-basic-arithmetics.fs
include genes/genes-organs.fs
include genes/genes-qualities.fs

include genes/genes-fetch.fs

true [if] \ just a quick fix to deal with overflow.  not more.
    include genes/genes-store-normalised.fs
[else] \ overflow is nice, the cells like it ;-)
    include genes/genes-store.fs
[then]

\ include genes/genes-conditionals.fs
\ include genes/insight.fs

: more-n-consumer-genes ( n -- )	\ puts 11*n genes in the pool
    0 ?DO
	[internal'] !(some)		1 actual-genepool-xt @ execute
	change-xt-probability
	[internal'] +!(some)	1 actual-genepool-xt @ execute
	change-xt-probability
	[internal'] -!(some)	1 actual-genepool-xt @ execute
	change-xt-probability
	[internal'] +		2 actual-genepool-xt @ execute
	change-xt-probability
	[internal'] -		2 actual-genepool-xt @ execute
	change-xt-probability
	[internal'] *		2 actual-genepool-xt @ execute
	change-xt-probability
	[internal'] ?/		2 actual-genepool-xt @ execute
	change-xt-probability
    LOOP ;

nuc-parameters# [IF] \ experimental. work in progress...
    include genes/nuc-parameters.fs
    \ quick hack to get balanced stack type probabilities:
    nuc-parameters# more-n-consumer-genes
[THEN]

spot-properties# [IF] \ experimental. work in progress...
    include genes/spot-properties.fs
    \ quick hack to get balanced stack type probabilities:
    spot-properties# more-n-consumer-genes
[THEN]

also brew-words definitions previous

INDIVIDUAL: template	' template individuals >list
also genes
' g-noop		setup-wake-me		\ he's a bit lazy
previous
' eat-all		eat-xt !		\ but very hungry
' cell-division 	reproduce-xt !
' <look-at>		show-me-xt !
100	reproduction-threshold !
10	age-threshold !


\ make another prototype to experiment with:
INDIVIDUAL: prototype
' prototype individuals >list

\ he's a bit lazy:
also genes
' g-noop		setup-wake-me
previous

\ but hungry, nevertheless
' eat-all	eat-xt !
\ there will be many soon
20	reproduction-threshold !
' cell-division reproduce-xt !
\ how he shows himself
' <look-at>	show-me-xt !
\ his lifetime is very short
3	age-threshold !

: .sorry ( -- )
    at? 2>r
    last-left clear-line-to-end last-left
    red color-foreground ." Sorry, you can't change this value..."
    1500 ms default-color color-foreground
    last-left clear-line-to-end
    2r> at-xy ;


\ .info-line with display-slots:
10 constant max-display-slots#				\ up to this many slots
CREATE display-slots					\ number and pointers
max-display-slots# 1+ cells allot			\ space for pointers
4 display-slots !					\ slots in info-line

: display-slot ( slot -- addr )
    dup max-display-slots# >= abort" Display slot out of range"
    1+ cells display-slots + ;		\ pointers start at 2'nd cell

: start-help ( -- )
    s" Start: <SPACE>   Menus: <B>   Keys: <k>   Select: cursor, <RET>   Demos: <d>"
    1 >message ;

variable single-step	single-step off
2variable (.burden-at)	0. (.burden-at) 2! \ cosmetics see automation, .burden
2variable (.code-price-at) 0. (.code-price-at) 2! \ cosmetics see automation
variable no-messages	no-messages off			\ used in benchmarks
: .info-line ( -- )					\ displays info line
    reset-default-colors
    0. (.burden-at) 2!		\ little trick to track if .burden is in a slot
    0. (.code-price-at) 2!	\ same
    message-count @ no-messages @ 0= and
    IF		\ special case: message to display
	single-step @ IF	\ in single step modus display only once
	    message-count off
	ELSE
	    -1 message-count +!	\ else as many times as specified
	THEN
	message-fg-color-xt @ dup IF EXECUTE color-foreground ELSE drop THEN
	message-bg-color-xt @ dup IF EXECUTE color-background ELSE drop THEN
	last-left clear-line-to-end last-left
	(message) buffer-as-string type
	message-fg-color-xt @ IF
	    default-color color-foreground
	THEN
	message-bg-color-xt @ IF
	    default-color color-background
	THEN
	message-count @ 0= IF
	    message-fg-color-xt off
	    message-bg-color-xt off
	THEN
    ELSE
	message-fg-color-xt off
	message-bg-color-xt off

	display-slots @ IF
	    c-l display-slots @ /		\ size of a slot
	    display-slots @ 0 DO
		dup i * last-line at-xy
		i display-slot ?perform
						\ insert spaces up to next slot
[ lower-right-scrolls ] [IF]	
		key? 0= IF			\ scroll protected version
		    dup i 1+ * 1-			\ start of next slot
		    i display-slots @ 1- <> IF 1+ THEN	\ not for last slot
		    at? drop - spaces			\ fill with spaces
		THEN
[ELSE]
		key? 0= IF			\ no scroll protection needed
		    dup i 1+ *			\ start of next slot
		    at? drop - spaces		\ fill with spaces
		THEN
[THEN]
	    LOOP drop
	ELSE
	    last-left clear-line-to-end		\ clear info line
	THEN
    THEN ;

: empty-display-slots ( -- )
    max-display-slots# 0 DO  ['] noop  i display-slot !	LOOP ;
empty-display-slots

\ display functions to put in the slots:
LIST: slot-display-words

' noop slot-display-words >list
: .step ( -- )	." step: " step @ . ;
' .step   0 display-slot !
' .step slot-display-words >list

: .cells ( -- )	." cells: " cloned @ . ;
' .cells  1 display-slot !
' .cells slot-display-words >list

: .burden ( -- )
    at? (.burden-at) 2!		."  burden: " nuc-do-cost @ . ;
' .burden 3 display-slot !
' .burden slot-display-words >list

: .code-price ( -- )
    at? (.code-price-at) 2!		."  code price: " code-price @ . ;
' .code-price slot-display-words >list

: .living ( -- ) ." living: " living @ . ;
' .living 2 display-slot !
' .living slot-display-words >list

: .newborn ( -- ) ." newborn: " newborn @ . ;
' .newborn slot-display-words >list

: .died ( -- ) ." died: " died @ . ;
' .died slot-display-words >list

: .trial ( -- ) ." trial: " trial @ . ;
' .trial slot-display-words >list

: .mutations ( -- ) ." mutated: " mutations @ . ;
' .mutations slot-display-words >list

: .compiled-genes ( -- ) ." compiled: " compiled-genes @ . ;
' .compiled-genes slot-display-words >list

: .mutation-max-items ( -- ) ." mut.items: " (mutated-max) @ . ;
' .mutation-max-items slot-display-words >list

: .age-threshold ( -- )
    ." max age: "
    cp@ IF
	age-threshold @ .
    THEN ;
' .age-threshold slot-display-words >list


\ Population control:  mean, automatic production of stress:

: ?log-nuc-do-cost ( -- )
    log-mask @ log-costs and 0= IF EXIT THEN

    s" nuc-do-cost: "		cat-log
    nuc-do-cost @ num>string	cat-log
    s" 		code-price: "	cat-log
    code-price @ num>string	log-costs log-it ;

variable additive-stress	2 additive-stress !
variable additive-release	4 additive-release !
2variable stress-rate		1 1 stress-rate 2!
variable multiplicative-release 4 multiplicative-release !
variable code-additive-stress	1 code-additive-stress !
2variable code-stress-rate	1 1 code-stress-rate 2!

variable high-water-mark	spots 3 / high-water-mark !
\ spots 4 / high-water-mark !
variable flood-mark		spots 90 100 */ flood-mark !
2variable flood-stress-rate	110 100 flood-stress-rate 2!
2variable flood-kill-rate	0 1 flood-kill-rate 2!
2variable flood-energy-rate	0 1 flood-energy-rate 2!
2variable flood-food-rate	9 10 flood-food-rate 2!
: flood-decrease-energy ( -- )  energy flood-energy-rate addr-rate ;
variable sos-mark		spots  4 100 */ sos-mark !
2variable sos-release-rate	1 2 sos-release-rate 2!
variable sos-reproduction-push	1 sos-reproduction-push ! \ energy for children
: sos-push-reproduction ( -- )	\ giving multiples of reproduction-threshold
    energy dup @
    reproduction-threshold @ 1+  sos-reproduction-push @  *
    max swap ! ;

variable sos-sow		1 sos-sow !		\ SOS: how many to sow
: (sos-do-sow) ( -- )   sos-sow @ sow drop ;
: sos-do-sow ( -- )   ['] (sos-do-sow) not-recording ;
variable low-water-mark		100 low-water-mark !
\ in the very beginning low-water-mark and sos-mark do not apply:
variable up-regulation-start	1000 up-regulation-start !  \ after which step?
variable nuc-cost-can-be-help?		nuc-cost-can-be-help? on
variable code-price-can-be-help?	code-price-can-be-help? on

: population-control ( -- )
    additive-stress @       stress-rate 2@ <>      OR
    code-additive-stress @  code-stress-rate 2@ <> OR
    OR IF				\ we want any population control?
	living @ high-water-mark @ > IF			\ too many?
	    additive-stress @ nuc-do-cost +!
	    code-additive-stress @ code-price +!
	    nuc-do-cost dup @ stress-rate 2@ */ swap !
	    code-price dup @ code-stress-rate 2@ */ swap !
	    living @ flood-mark @ > IF
[ true ] [IF] \ new version
    		s" Population flood.  Step: "	cat-log
		step @ num>string		cat-log
		s"   living: "			cat-log
		living @ num>string		log-emergency log-it
		s" Population flood." 1 >message ['] cyan message-fg-color-xt !

		\ increase stress:
		additive-stress dup @ abs flood-stress-rate 2@ */ 1+ swap !
		code-additive-stress dup @ abs flood-stress-rate 2@ */ 1+ swap !
		nuc-do-cost dup @ abs flood-stress-rate 2@ */ swap !
		code-price dup @ abs flood-stress-rate 2@ */ swap !

		\ decrease food:
		food-share/spot			flood-food-rate addr-rate
		individual-fixed-food-share	flood-food-rate addr-rate
		world-food-supply		flood-food-rate addr-rate
		
		spots living @ -		\ individuals too many
		flood-kill-rate 2@ */		\ some should go?
		leave-energy-after-death dup @ >r off	\ energy vanishes
		['] die do-at-random-spots		\ some must die
		r> leave-energy-after-death !

		flood-energy-rate 2@ <> IF
		    ['] flood-decrease-energy do-with-everybody
		THEN

[ELSE] \ old version. I keep it for the old demos.
		additive-stress dup @ abs 2* swap !
		code-additive-stress dup @ abs 2* swap !
		nuc-do-cost dup @ abs 3 2 */ swap !
		code-price dup @ abs 3 2 */ swap !
		s" flood" log-emergency log
		s" Population flood!" 1 >message
		['] cyan message-fg-color-xt !
[THEN]
	    THEN
	    (.burden-at) 2@ d0= 0= IF	  \ d0= if .burden is not active
		.info-line
		(.burden-at) 2@ at-xy ." +"
	    THEN
	    (.code-price-at) 2@ d0= 0= IF \ d0= if .code-price is not active
		.info-line
		(.code-price-at) 2@ at-xy ." +"
	    THEN
	THEN

	living @ low-water-mark  @ < IF				\ too few?
	    step @  up-regulation-start @  > IF			\ start phase?
		\ release stress some additive steps
		additive-stress @  additive-release @  *
		abs negate nuc-do-cost +!
		\ same for code stress
		code-additive-stress @  additive-release @  *
		abs negate code-price +!

		\ release stress multiplicative
		nuc-do-cost dup @ dup 0> IF
		    multiplicative-release @ 0 ?DO
			stress-rate 2@ swap */
		    LOOP
		ELSE
		    multiplicative-release @ 0 ?DO
			stress-rate 2@ */
		    LOOP
		THEN swap !
		\ same for code
		code-price dup @ dup 0> IF
		    multiplicative-release @ 0 ?DO
			code-stress-rate 2@ swap */
		    LOOP
		ELSE
		    multiplicative-release @ 0 ?DO
			code-stress-rate 2@ */
		    LOOP
		THEN swap !

		living @ sos-mark @ < IF
		    nuc-do-cost dup @ sos-release-rate 2@ */ swap !
		    additive-stress dup @ dup 0> IF
			sos-release-rate 2@ */
			1 max
		    THEN swap !

		    code-price dup @ sos-release-rate 2@ */ swap !
		    code-additive-stress dup @ dup 0> IF
			sos-release-rate 2@ */
			1 max
		    THEN swap !

		    ['] sos-push-reproduction do-with-everybody
		    sos-sow @ dup IF
			['] sos-do-sow do-with-everybody
		    THEN drop

		    s" S.O.S.  step: "	cat-log
		    step @ num>string	cat-log
		    s"   living: "	cat-log
		    living @ num>string log-emergency log-it
		    s" S.O.S. Population emergency!" 1 >message
		    ['] red message-fg-color-xt !
		THEN
		(.burden-at) 2@ d0= 0= IF	\ d0= if .burden is not active
		    .info-line
		    (.burden-at) 2@ at-xy ." -"
		THEN
		(.code-price-at) 2@ d0= 0= IF	\ d0= if .code-price not active
		    .info-line
		    (.code-price-at) 2@ at-xy ." +"
		THEN
	    THEN
	THEN

	nuc-cost-can-be-help? @ 0= IF
	    nuc-do-cost dup @  0 max  swap !
	THEN
	code-price-can-be-help? @ 0= IF
	    code-price dup @  0 max  swap !
	THEN

	?log-nuc-do-cost
    THEN ;


MENU: population-men
: .menu-population
    page

    title-colors
    ." Population control.  Mean, automatic stress producer!   "

    living @ dup IF
	>r s" (actually alive: " r> ['] .sorry menu-entry-value
	.bs ." )"
    ELSE drop THEN
    end-title

    cr
    s" Live stress, additive step: " additive-stress simple-menu-entry-value
    1 2 screen-column
    s" Stress rate:            " stress-rate simple-menu-entry-scale cr

    s" Code price, additive step:  "
    code-additive-stress simple-menu-entry-value
    1 2 screen-column
    s" Code price rate        " code-stress-rate simple-menu-entry-scale cr

    cr
    s" High water mark:     "	high-water-mark	simple-menu-entry-value
    s" hH" menu-same-key-entry

    high-water-mark @ spots < IF
	1 2 screen-column
	s" Flood-mark:             " flood-mark	simple-menu-entry-value cr
	s" fF" menu-same-key-entry

	flood-mark @ spots < IF
	    s" Flood stress rate:   " flood-stress-rate simple-menu-entry-scale
	    1 2 screen-column
	    s" Fl. energy rate:        "
	    flood-energy-rate simple-menu-entry-scale
	    s" eE" menu-same-key-entry
	    cr
	    s" Flood food rate:     " flood-food-rate simple-menu-entry-scale
	    1 2 screen-column
	    s" Flood kill rate:        "
	    flood-kill-rate simple-menu-entry-scale
	    s" kK" menu-same-key-entry
	THEN
    THEN cr

    cr
    s" Low water mark:      "	low-water-mark	simple-menu-entry-value
    s" lL" menu-same-key-entry

    low-water-mark @ 0> IF
	1 2 screen-column
	s" is active after step:   "
	up-regulation-start simple-menu-entry-value cr
	s" pP" menu-same-key-entry

	s" Additive release:    " additive-release simple-menu-entry-value
	1 2 screen-column
	s" Multiplicative release: "
	multiplicative-release	simple-menu-entry-value cr

	s" SOS mark:            "	sos-mark	simple-menu-entry-value
	s" sS" menu-same-key-entry

	sos-mark @ 0> IF
	    1 2 screen-column
	    s" SOS release rate:       "
	    sos-release-rate simple-menu-entry-scale cr
	    s" Reproduction energy: "
	    sos-reproduction-push simple-menu-entry-value
	    1 2 screen-column
	    s" SOS sow:                " sos-sow  simple-menu-entry-value
	THEN
    THEN cr

    cr
    s" Actual energy cost of each life step: "
    nuc-do-cost				simple-menu-entry-value cr
    s" nN" menu-same-key-entry

    s" Can this 'cost' become a source of help in emergency situations? "
    nuc-cost-can-be-help? >stack	['] toggle-addr	redisplay menu-entry
    nuc-cost-can-be-help? @ IF  ." YES"  ELSE  ." NO"  THEN cr
    cr

    s" Actual energy cost of code usage: "
    code-price				simple-menu-entry-value
    s" cC" menu-same-key-entry

    1 2 screen-column
    s" rating scale "	code-price-scale	simple-menu-entry-scale cr
    s" rR" menu-same-key-entry

    s" Can this 'cost' become a source of help in emergency situations? "
    code-price-can-be-help? >stack	['] toggle-addr	redisplay menu-entry
    code-price-can-be-help? @ IF  ." YES"  ELSE  ." NO"  THEN cr
    cr

    s" Food supply is another way of population control."
    ['] food-menu			menu-entry cr
    s" fF" menu-same-key-entry
    
    common-menu-entries ;

: menu-population-control
    population-men
    ['] .menu-population menu-display-xt !
    menu-done	['] noop	menu-key-default
    menu-done	['] noop	menu-default
    do-menu-loop
    free-menus ;
' menu-population-control function-key-actions >list

\ diversification
variable diversification-range		\ up to how much
2variable diversification-rate		\ how often.
1 4 diversification-rate 2!
variable diversification-closeness	\ the bigger, the less likely are
					\ big deviations
2 diversification-closeness !		\ 2 was default before
variable sporadic-value-range		sporadic-value-range off
2variable sporadic-value-rate		\ how often to set any new values
1 100 sporadic-value-rate 2!
: diversify ( addr -- )			\ diversifies contents of addr
    diversification-range @		\ never more than this
    diversification-closeness @ 0 ?DO	\ looping randomization
	1+ random-ranged		\ loop: big deviations get rarer
    LOOP
    2 random-ranged IF negate THEN	\ sign
    swap +! ;

: sporadic-replace-value ( addr -- )
    sporadic-value-range @ 1+ random-ranged
    2 random-ranged IF negate THEN	\ sign
    swap +! ;

: diversify? ( addr -- )		\ maybe diversify contents of addr
    dup diversification-rate rated-flag IF  diversify           ELSE drop THEN
    sporadic-value-rate  rated-flag IF  sporadic-replace-value  ELSE drop THEN
;

: diversify?-which ( mask -- )
    diversification-rate rated-flag IF
	my-diversification-bitmask @ and
	nuc-diversificable-area dup nuc-diversificable-items + swap DO
	    dup  1  i nuc-diversificable-area -  lshift  and IF
		i nuc-addr diversify
	    THEN
	LOOP
    THEN
    drop ;

variable diversification-mask
: init-diversification-mask ( -- )
    0
    nuc-organs# 0 ?DO
	1 i lshift OR
    LOOP
    nuc-parameters# 0 ?DO
	1 i [ nuc-parameters nuc-diversificable-area - ] literal + lshift OR
    LOOP
    nuc-invisible# 0 ?DO
	1 i [ nuc-invisible nuc-diversificable-area - ] literal + lshift OR
    LOOP
    diversification-mask ! ;
init-diversification-mask

: diversify?-some ( -- ) diversification-mask @ diversify?-which ;
' diversify?-some is <diversify>

: diversify?-A ( -- )  organ-A diversify? ;
: diversify?-B ( -- )  organ-B diversify? ;
: diversify?-C ( -- )  organ-C diversify? ;

MENU: diversification-men
: .diversification-menu
    page

    title-colors ." Diversification menu:" end-title

    ." Simple diversification: absolute addition or subtraction of a random value." cr

    cr
    s" Rate: "	diversification-rate	simple-menu-entry-scale
    s" hHoOrR"	menu-same-key-entry
    1 4 screen-column
    s" Range: "	diversification-range	simple-menu-entry-value
    s" mMdD" menu-same-key-entry
    2 4 screen-column
    diversification-closeness @ 0< IF	\ range control during redisplay ;-)
	bell cr ." Closeness can't be negative, reset." cr
	diversification-closeness off
    THEN
    s" Closeness: (fewer big deviations): "
    diversification-closeness		simple-menu-entry-value cr
    s" cC" menu-same-key-entry

\      cr
\      s" Sporadic: "	sporadic-value-rate	simple-menu-entry-scale
\      s" sS" menu-same-key-entry
\      1 4 screen-column
\      s" Range: "		sporadic-value-range	simple-menu-entry-value cr
\      2 4 screen-column

    cr
    ." Where to diversify?  (gets individually masked then) " cr
    14 menu-scroll-lines !	\ initialize scrolling
    cr
    at?						( 0 line )
    nuc-organs# ( 16 min) 0 scrolled-range ?DO
	from-here  nuc-organs i + nuc-var-name type
	s" : "
	1 i lshift >stack	diversification-mask >stack-2
	['] xor!	redisplay	menu-entry
	diversification-mask @  1 i lshift  and .ON-off
	up-to-here cr
	[char] A i + #key-same-entry
    LOOP

    at-xy
    nuc-parameters# ( 16 min ) 0 scrolled-range ?DO
	1 3 screen-column
	from-here  nuc-parameters i + nuc-var-name type
	s" : "
	1  i nuc-organs# +  lshift >stack	diversification-mask >stack-2
	['] xor!	redisplay	menu-entry
	diversification-mask @  1  i nuc-organs# + lshift  and .ON-off
	up-to-here cr
	\ [char] a i + #key-same-entry
    LOOP

    common-menu-entries ;

: diversification-menu
    diversification-men
    ['] .diversification-menu menu-display-xt !
    menu-done	['] noop	menu-key-default
    menu-done	['] noop	menu-default
    do-menu-loop
    free-menus ;
' diversification-menu function-key-actions >list

: eat-part ( -- )
    <food> @ organ-A @ min	\ doesn't eat more than organ-A
    0 max			\ no vomiting (one could leave this out)
    dup energy +!
    negate <food> +! ;
' eat-part eat-actions >list

INDIVIDUAL: Yuppie
' Yuppie individuals >list

' cell-division reproduce-xt !
' diversify?-A IS <diversify>
1 my-diversification-bitmask !
50 diversification-range !
1 4 diversification-rate 2!
' eat-part eat-xt !
100 organ-A !
' show-A show-me-xt !	100 2-ascii-scale !
1000 food-share/spot !
2099	reproduction-threshold !
20	age-threshold !
nuc-do-cost off
also genes
' g-noop setup-wake-me
previous

\ some color related stuff:
variable food>color-scale	100 food>color-scale !
: food>color ( -- col )  <food> @ food>color-scale @ / 2-color ;
' food>color background-color-xt !
' food>color x>bg-color >list

variable age>color-scale	20 age>color-scale !
: age>color ( -- col )
    age @ colors *
    age-threshold @		\ scaling: if there's an age-threshold
    ?dup 0= IF			\          take this one
	age>color-scale @	\          otherwise take age>color-scale
    THEN			\	   as length of the color cycle
    /				\ proportion: colors/cycle
    2-color ;			\ normalized
' age>color foreground-color-xt !
' age>color x>fg-color >list

: generation>color ( -- col )	generation @ 2-color ;
' generation>color  x>fg-color >list

: genome>color ( -- col )   genome-id @ 1+ 2-color ;
' genome>color x>fg-color >list

: genom-generation>color ( -- col )	genom-generation @ 2-color ;
' genom-generation>color  x>fg-color >list

: code>color ( -- col )   code-cost @ (default-gene-cost#) / 2-color ;
' code>color x>fg-color >list

: trial>color ( -- col )
    nuc-flags @ wake-me-trial and	( 0|1 )
    2* 1- 2-color ;
' trial>color x>fg-color >list

variable A-quality>color-scale	200 A-quality>color-scale !
: A-quality>color ( -- col )
    A-quality @ A-quality>color-scale @ / 2-color ;
' A-quality>color x>bg-color >list

variable B-quality>color-scale	200 B-quality>color-scale !
: B-quality>color ( -- col )
    B-quality @ B-quality>color-scale @ / 2-color ;
' B-quality>color x>bg-color >list

variable C-quality>color-scale	200 C-quality>color-scale !
: C-quality>color ( -- col )
    C-quality @ C-quality>color-scale @ / 2-color ;
' C-quality>color x>bg-color >list

variable scaling-range		1 scaling-range !
: calibrate-bg-color-scales ( -- )
    do-spot-scan
	field-planes# p0 DO			\ skip pointer plane?
	    (scan-spots) i 2* 2* cells + >r	\ start data for this quality
	    r@ cell+	@			\ max
	    r>		@	-		\ min -
	    colors	1-	/		\ divided by number of colors
	    scaling-range @	*		\ scaling-range
	    ?dup IF				\ scales should never be zero
		i CASE
		    1 OF food>color-scale	ENDOF
		    2 OF A-quality>color-scale	ENDOF
		    3 OF B-quality>color-scale	ENDOF
		    4 OF C-quality>color-scale	ENDOF
		    bell cr ." X>color-scale not defined" ABORT 
		ENDCASE
		!				\ store scale
	    THEN
	LOOP ;
: cal-bg-color-scales-full
    1 scaling-range !
    calibrate-bg-color-scales ;
: cal-bg-color-scales-half
    2 scaling-range !
    calibrate-bg-color-scales ;
: cal-bg-color-scales-two
    colors 2/ scaling-range !
    calibrate-bg-color-scales ;

\ see genome dealing with trial and (Gforth) 'see' bug
: see-genome ( -- )
    page
    wake-me-xt @
    nuc-flags @ wake-me-trial and IF		\ on trial?
	." genome is on trial." cr
	dup
	eb>length @
	swap eb>sequence swap type
    ELSE					\ compiled: try 'see'
	dup					\ save xt ('see' bug)
	[ decimal ] 48 stringbuf-open >r
	s" see " r@ cat
	xt>name  r@ cat
	s"  "    r@ cat
	also genes   r@ buffer-as-string
	['] EVALUATE CATCH  previous			\ catch see bug
	r> stringbuf-close
	IF						\ 'see' did throw,
	    gene-xt>internal-xt internal-xt>string type	\  do it yourself.
	ELSE drop THEN
    THEN ;

: is-gene? drop FALSE ;	\ heavy cheating ;-)

: page-see ( xt -- )

    dup is-gene?	\ dummy FALSE at the moment
    IF   see-genome
    ELSE
	page
	dup					\ save xt ('see' bug)
	[ decimal ] 48 stringbuf-open >r
	s" see " r@ cat
	xt>name  r@ cat
	s"  "    r@ cat
	also genes   r@ buffer-as-string
	['] EVALUATE CATCH  previous		\ catch 'see' bug
	r> stringbuf-close
	IF					\ 'see' throw
	    bell  ." page-see: Couldn't see "  xt>name type  [char] . emit
	ELSE drop THEN
    THEN ;

LIST: color-list
' black		color-list >list
' red		color-list >list   
' green		color-list >list
' brown		color-list >list
' blue		color-list >list
' magenta	color-list >list
' cyan		color-list >list
' white		color-list >list
[DEFINED] default-color [IF]
' default-color	color-list >list
[THEN]

MENU: color-men

: choose-color ( color -- )	\ what to do if the user chooses a color sample
    color-list n'th-node @	\ get xt of color

    display-switch @ step-display-on and IF
	display-switch @ step-background-coloring and IF
	    scan-background-xt !   EXIT
	THEN
	display-switch @ step-foreground-coloring and IF
	    scan-foreground-xt ! EXIT
	THEN
	display-switch dup @
	[ step-display-on step-background-coloring or ] literal or swap !
	scan-background-xt !   EXIT
    ELSE
	cursor-off   xt>name  34 2 at-xy type  600 ms   cursor-visible
    THEN ;

: .color-menu ( -- )
    page

    title-colors ." Color menu:" end-title

[UNDEFINED] never-use-colors [IF]	\ compiled with color support
    cr
    colors 0 DO
	i color-background
	i >stack redisplay	s"     "    ['] choose-color	menu-entry
    LOOP
    0 color-background cr cr

    display-switch @ spot-display-on and IF
	s" Foreground coloring: "	spot-foreground-coloring >stack
	display-switch >stack-2 ['] xor! redisplay menu-entry
	display-switch @ spot-foreground-coloring and .ON-off
	cr
	s" f" menu-same-key-entry

	s" Background coloring: "	spot-background-coloring >stack
	display-switch >stack-2 ['] xor! redisplay menu-entry
	display-switch @ spot-background-coloring and .ON-off
	cr
	s" b" menu-same-key-entry

	s" Set foreground color:" title-choose-xt!
	x>fg-color	>stack		foreground-color-xt	>stack-2
	s" Foreground color does: "	['] choose-xt-menu	menu-entry
	s" F" menu-same-key-entry
	.tab	menu-wait	redisplay	foreground-color-xt @
	dup >stack	xt>name		['] page-see		menu-entry cr

	s" Set background color:" title-choose-xt!
	x>bg-color	>stack		background-color-xt	>stack-2
	s" Background color does: "	['] choose-xt-menu	menu-entry
	s" B" menu-same-key-entry
	.tab	menu-wait	redisplay	background-color-xt @
	dup >stack	xt>name		['] page-see		menu-entry cr

	foreground-color-xt @ ['] age>color = IF
	    cr
	    s" Age  to color scale: "  age>color-scale simple-menu-entry-value
	    cr ." (only applies when no age-threshold) "
	    cr
	THEN
	background-color-xt @ ['] food>color = IF
	    cr
	    s" Food to color scale: " food>color-scale simple-menu-entry-value
	    cr cr
	THEN

	cr
	s" 'A'-Quality to color scale: "
	A-quality>color-scale			simple-menu-entry-value cr

	s" 'B'-Quality to color scale: "
	B-quality>color-scale			simple-menu-entry-value cr

	s" 'C'-Quality to color scale: "
	C-quality>color-scale			simple-menu-entry-value cr

	cr
	s" Calibrate background color scales to full scale"
	['] cal-bg-color-scales-full	redisplay	menu-entry cr
	s" cC" menu-same-key-entry
	s" Calibrate background color scales to half scale"
	['] cal-bg-color-scales-half	redisplay	menu-entry cr
	s" hH" menu-same-key-entry
	s" Calibrate background color scales to two colors"
	['] cal-bg-color-scales-two		redisplay	menu-entry cr
	s" tT2" menu-same-key-entry
    THEN

    display-switch @ step-display-on and IF
	display-switch @ step-foreground-coloring and IF
	    s" Choose step display foreground color:" title-choose-xt!
	    color-list	>stack		scan-foreground-xt	>stack-2
	    s" Foreground color: "	['] choose-xt-menu	menu-entry
	    s" F" menu-same-key-entry
	    scan-foreground-xt @ xt>name type	[ decimal ] 34 at-x
	THEN
	s" Foreground coloring: "	step-foreground-coloring >stack
	display-switch >stack-2 ['] xor! redisplay menu-entry
	display-switch @ step-foreground-coloring and .ON-off cr
	s" f" menu-same-key-entry

	cr
	display-switch @ step-background-coloring and IF
	    s" Choose step display background color:" title-choose-xt!
	    color-list	>stack		scan-background-xt	>stack-2
	    s" Background color: "	['] choose-xt-menu	menu-entry
	    s" B" menu-same-key-entry
	    scan-background-xt @ xt>name type	[ decimal ] 34 at-x
	THEN
	s" Background coloring: "	step-background-coloring >stack
	display-switch >stack-2 ['] xor! redisplay menu-entry
	display-switch @ step-background-coloring and .ON-off
	cr
	s" b" menu-same-key-entry
    ELSE display-switch @ spot-display-on and 0= IF	\ neither step or spot
	." Switch either step or spot display on, then choose colors."
    THEN THEN

[ELSE]		\ compiled without color support
    cr ." This version was compiled without color support."
    cr ." Uncomment 'never-use-colors' in 'compile-options.fs' if you want colors."
[THEN]
    common-menu-entries ;


: color-menu ( -- )
    color-men
    ['] .color-menu menu-display-xt !
    menu-done	['] noop	menu-key-default
    menu-done	['] noop	menu-default
    do-menu-loop
    free-menus ;
' color-menu function-key-actions >list

\ color plays:
: eat-red ( -- )
    <food> @ organ-A @ min	\ doesn't eat more than organ-A
    A-quality @ +		\ red makes hungry
    0 max			\ no vomiting (one could leave this out)

    \ this doesn't belong here, but it's simplest
    dup 10 / negate
    dup future A-quality +!		\ store at future *and* present to be
    present A-quality +!		\ independent of future-quality-change

    dup energy +!
    negate <food> +! ;
' eat-red eat-actions >list

: eat-blue ( -- )
    <food> @ organ-A @ min	\ doesn't eat more than organ-A
    A-quality @ -		\ blue makes hungry
    0 max			\ no vomiting (one could leave this out)

    \ this doesn't belong here, but it's simplest
    dup 10 /
    dup future A-quality +!	\ store at future *and* present to be
    present A-quality +!	\ independent of future-quality-change

    dup energy +!
    negate <food> +! ;
' eat-blue eat-actions >list

INDIVIDUAL: red-eater
' red-eater individuals >list

' cell-division reproduce-xt !

' diversify?-A IS <diversify>
1 my-diversification-bitmask !
50 diversification-range !
1 4 diversification-rate 2!
' noop is <diversify>

' eat-red eat-xt !
100 organ-A !
char * appearance !
' show-ascii show-me-xt !	100 2-ascii-scale !
200 food-share/spot !
695 reproduction-threshold !
10  age-threshold !
nuc-do-cost off
also genes
' g-noop setup-wake-me
previous

INDIVIDUAL: blue-eater
' blue-eater individuals >list

' cell-division reproduce-xt !

' diversify?-A IS <diversify>
1 my-diversification-bitmask !
50 diversification-range !
1 4 diversification-rate 2!
' noop is <diversify>

' eat-blue eat-xt !
100 organ-A !
char . appearance !
' show-ascii show-me-xt !	100 2-ascii-scale !
200 food-share/spot !
695 reproduction-threshold !
10  age-threshold !
nuc-do-cost off
also genes
' g-noop setup-wake-me
previous

spot-display-on display-switch !
1 additive-stress !

false [IF]
    INDIVIDUAL: quick-red-eater
    ' quick-red-eater individuals >list

    ' cell-division reproduce-xt !
    ' diversify?-A IS <diversify>
    1 my-diversification-bitmask !
    50 diversification-range !
    1 4 diversification-rate 2!
    ' eat-red eat-xt !
    100 organ-A !
    char * appearance !
    ' show-ascii show-me-xt !	100 2-ascii-scale !
    200 food-share/spot !
    343 reproduction-threshold !
    4  age-threshold !
    nuc-do-cost off
    also genes
    ' g-noop setup-wake-me
    previous

    INDIVIDUAL: quick-blue-eater
    ' quick-blue-eater individuals >list

    ' cell-division reproduce-xt !
    ' diversify?-A IS <diversify>
    1 my-diversification-bitmask !
    50 diversification-range !
    1 4 diversification-rate 2!
    ' eat-blue eat-xt !
    100 organ-A !
    char . appearance !
    ' show-ascii show-me-xt !	100 2-ascii-scale !
    200 food-share/spot !
    343 reproduction-threshold !
    4  age-threshold !
    nuc-do-cost off
    also genes
    ' g-noop setup-wake-me
    previous
[THEN]

false [IF]
    big-bang
    red-eater  8  sow drop
    blue-eater 12 sow drop
[THEN]

\ moving
: cell-move ( i -- flag)			\ can't call it 'move'
    dup someone-here? IF			\ shouldn't happen here!
	drop false				\ :-(
    ELSE			( i )
	\ check for dying cells first, they would make troubles if moved
	will-die? IF
	    drop false				\ I don't move dying cells
	    s" cell is about to die, not moved" log-movement log
	ELSE
	    fcp @ >r				\ current cell on return stack
	    future
	    dup someone-here? IF		\ already taken?
		drop rdrop
		present
		false				\ :-(
	    ELSE
		false fcp !			\ off
		spot @		( i old-spot )	\ cautiously remembering spot
		swap >spot!	( old-spot )
		r> fcp !			\ displacing the cell

		log-mask @ IF
		    log-cat-id
		    s" : moved to spot " cat-log
		    spot @ num>string log-movement log-it
		THEN

		present
		>spot!
		true				\ :-)
	    THEN
	THEN
    THEN ;

\  : move? ( -- )
\      free-neighbor-spot? IF		( i )
\  	cell-move drop
\      THEN ;
\  ' move? wake-me-actions >list

\ also genes definitions
\ GENE: g-move?	move?	;gene s" -"	>stackdata! ##### last-gene-into-pool
\ also brew-words definitions previous

false [IF]
    INDIVIDUAL: wanderer
    ' wanderer individuals >list

    ' move? setup-wake-me
    also genes
    ' g-noop reproduce-xt !
    previous

    ' noop is <diversify>
    ' eat-all eat-xt !
    0 my-diversification-bitmask !
    char * appearance !
    ' show-ascii show-me-xt !
    200 food-share/spot !
    343 reproduction-threshold !
    1000  age-threshold !
    nuc-do-cost off
[THEN]

false [IF]
    \ ok lets try the symbiosis with moving cells:
    red-eater
    ' move? setup-wake-me
    blue-eater
    ' move? setup-wake-me
    big-bang
    red-eater  8  sow drop
    blue-eater 12 sow drop
[THEN]

\ I wonder if this symbiotic species couldn't emerge all by themselves?
: eat-violet ( -- )
    organ-A @ 0< 2* 1+ >r	\ sign of organ-A as +1/-1 on return stack
    <food> @ organ-A @ abs min	\ doesn't eat more than abs(organ-A)
    A-quality @ r@ * -
    10 max			\ but not less than 10
    dup energy +!
    dup negate <food> +!

    \ this doesn't belong here, but it's simplest
    r> * 10 /
    dup future A-quality +!	\ store at future *and* present to be
    present A-quality +! ;	\ independent of future-quality-change
' eat-violet eat-actions >list


INDIVIDUAL: violet-eater
' violet-eater individuals >list

' show-sign-A show-me-xt !
' cell-division reproduce-xt !
' diversify?-A IS <diversify>
1 my-diversification-bitmask !
50 diversification-range !
1 4 diversification-rate 2!
' eat-violet eat-xt !
0 organ-A !
200 food-share/spot !
100 reproduction-threshold !
10  age-threshold !
nuc-do-cost off
also genes
' g-noop setup-wake-me
previous

false [IF]
    160 low-water-mark !
    big-bang
    1 sow drop
\ wow! it's working like expected! I had to increase the low water mark a bit.
[THEN]

variable (A-scale)		100 (A-scale) !
: eat-violet-remade ( -- )
    eat-all
    organ-A @  A-quality @  (A-scale) @ */
    energy +!
    organ-A @ negate
    dup future A-quality +! present A-quality +! ;	\ changes A-quality
' eat-violet-remade eat-actions >list

: eat-simple-violet-remade ( -- )
    eat-all
    organ-A @  A-quality @  *   0< 2* 1+ >r		\ same sign?
    organ-A @ abs  A-quality @ abs  min			\ the smaller one
    1 max						\ to let things start
    r> *
    energy +!
    organ-A @ negate
    dup future A-quality +! present A-quality +! ;	\ changes A-quality
' eat-simple-violet-remade eat-actions >list



variable (B-scale)		100 (B-scale) !
variable (C-scale)		100 (C-scale) !
: eat-rainbow-remade ( -- )
    eat-all
    organ-A @  A-quality @  (A-scale) @ */
    organ-B @  B-quality @  (B-scale) @ */
    organ-C @  C-quality @  (C-scale) @ */
    + + energy +!
    organ-A @ negate
    dup future A-quality +! present A-quality +!	\ changes A-quality
    organ-B @ negate
    dup future B-quality +! present B-quality +!	\ changes B-quality
    organ-C @ negate
    dup future C-quality +! present C-quality +! ;	\ changes C-quality
' eat-rainbow-remade eat-actions >list

: eat-cyclic-remade ( -- )
    eat-all
    organ-A @  A-quality @  (A-scale) @ */
    organ-B @  B-quality @  (B-scale) @ */
    organ-C @  C-quality @  (C-scale) @ */
    + + energy +!
    organ-A @ negate
    dup future B-quality +! present B-quality +!	\ changes B-quality
    organ-B @ negate
    dup future C-quality +! present C-quality +!	\ changes C-quality
    organ-C @ negate
    dup future A-quality +! present A-quality +! ;	\ changes A-quality
' eat-cyclic-remade eat-actions >list

: eat-cyclic-back-clock ( -- )
    eat-all
    organ-A @  A-quality @  (A-scale) @ */
    organ-B @  B-quality @  (B-scale) @ */
    organ-C @  C-quality @  (C-scale) @ */
    + + energy +!
    future
    organ-A @ negate
    dup future C-quality +! present C-quality +!	\ changes C-quality
    organ-B @ negate
    dup future A-quality +! present A-quality +!	\ changes A-quality
    organ-C @ negate
    dup future B-quality +! present B-quality +! ;	\ changes B-quality
' eat-cyclic-back-clock eat-actions >list

false [IF]
\ I wonder how maximal age would optimize
: diversify?-age-threshold ( -- )
    age-threshold @ >r
    age-threshold diversify?
    age-threshold @ 1 < IF	\ age-threshold shouldn't get zero
	r> age-threshold !	\ so restore it
    ELSE
	rdrop
    THEN ;

    :noname diversify?-A diversify?-age-threshold ; IS <diversify>
    big-bang
    1 sow drop
[THEN]

\ wow, this symbiotic biotop was fascinating!
\ want to try more komplex things:

: eat-rainbow ( -- )
    organ-A @ 0< 2* 1+ pad !		\ sign of organ-A as +1/-1 at pad
    organ-B @ 0< 2* 1+ pad cell+ !	\ sign of organ-B as +1/-1 at pad +cell
    organ-C @ 0< 2* 1+ pad 2 cells + !	\ sign of organ-C as +1/-1 at pad
    <food> @
    organ-A @ abs min			\ doesn't eat more than abs(organ-A)
    organ-B @ abs min			\ and not more than abs(organ-B)
    organ-C @ abs min			\ and not more than abs(organ-B)

    A-quality @ pad           @ * -
    B-quality @ pad cell+     @ * -
    C-quality @ pad 2 cells + @ * -

    10 max			\ but not less than 10
    dup energy +!
    dup negate <food> +!

    dup pad	      @ * 10 /
    dup future A-quality +! present A-quality +!
    dup pad   cell+   @ * 10 /
    dup future B-quality +! present B-quality +!
    pad 2 cells + @ * 10 /
    dup future C-quality +! present C-quality +! ;

\ to be honest: I have no idea what it does...
\ I just made it up like this. let's try:
' eat-rainbow eat-actions >list

\ hmm, we need a showing function first
: show-ABC
    organ-A @ abs organ-B @ abs > IF
	organ-A @ abs organ-C @ abs > IF
	    [char] A organ-A @
	ELSE
	    [char] C organ-C @
	THEN
    ELSE
	organ-B @ abs organ-C @ abs > IF
	    [char] B organ-B @
	ELSE
	    [char] C organ-C @
	THEN
    THEN
    dup 0= IF drop [char] 0 THEN	\ all three organs are zero
    0< IF 32 + THEN		\ capitals or not? (sign of strongest quality)
    .ascii ;
' show-ABC show-me-actions >list

\ more beauty, please
: show-ABC*.
    organ-A @ abs organ-B @ abs > IF
	organ-A @ abs organ-C @ abs > IF
	    [char] .
	    \ organ-A @
	ELSE
	    [char] *
	    \ organ-C @
	THEN
    ELSE
	organ-B @ abs organ-C @ abs > IF
	    [char] ,
	    \ organ-B @
	ELSE
	    [char] *
	    \ organ-C @
	THEN
    THEN
    dup 0= IF drop [char] 0 THEN	\ all three organs are zero
\    0< IF 32 + THEN		\ capitals or not? (sign of strongest quality)
    .ascii ;
' show-ABC*. show-me-actions >list

\ one more
: show-ABC-X|
    organ-A @ abs organ-B @ abs > IF
	organ-A @ abs organ-C @ abs > IF
	    [char] -
	    \ organ-A @
	ELSE
	    [char] +
	    \ organ-C @
	THEN
    ELSE
	organ-B @ abs organ-C @ abs > IF
	    [char] '
	    \ organ-B @
	ELSE
	    [char] X
	    \ organ-C @
	THEN
    THEN
    dup 0= IF drop [char] 0 THEN	\ all three organs are zero
\    0< IF 32 + THEN		\ capitals or not? (sign of strongest quality)
    .ascii ;
' show-ABC-X| show-me-actions >list

\ another attempt
: show-~|'.*
    organ-A @ abs organ-B @ abs > IF
	organ-A @ abs organ-C @ abs > IF
	    organ-A @ 0< IF [char] -
	    ELSE	    [char] ~
	    THEN
	ELSE
	    organ-C @ 0< IF [char] '
	    ELSE	    [char] |
	    THEN
	THEN
    ELSE
	organ-B @ abs organ-C @ abs > IF
	    organ-B @ 0< IF [char] .
	    ELSE	    [char] *
	    THEN
	ELSE
	    organ-C @ 0< IF [char] '
	    ELSE	    [char] |
	    THEN
	THEN
    THEN
    .ascii ;
' show-~|'.* show-me-actions >list

INDIVIDUAL: rainbow-eater
' rainbow-eater individuals >list

' show-ABC show-me-xt !
' cell-division reproduce-xt !
:noname diversify?-A diversify?-B diversify?-C ; IS <diversify>
7 my-diversification-bitmask !
50 diversification-range !
1 4 diversification-rate 2!
' eat-rainbow eat-xt !
also genes
' g-noop setup-wake-me
previous

0 organ-A !

false [if]
    200 food-share/spot !
    100 reproduction-threshold !
    10  age-threshold !
    nuc-do-cost off
    big-bang
[then]

: eat-cyclic-rainbow ( -- )
    organ-A @ 0< 2* 1+ pad !		\ sign of organ-A as +1/-1 at pad
    organ-B @ 0< 2* 1+ pad cell+ !	\ sign of organ-B as +1/-1 at pad +cell
    organ-C @ 0< 2* 1+ pad 2 cells + !	\ sign of organ-C as +1/-1 at pad
    <food> @
    organ-A @ abs min			\ doesn't eat more than abs(organ-A)
    organ-B @ abs min			\ and not more than abs(organ-B)
    organ-C @ abs min			\ and not more than abs(organ-B)
    
    A-quality @ pad cell+     @ * -	\ takes deliberately the wrong sign
    B-quality @ pad 2 cells + @ * -
    C-quality @ pad	      @ * -

    10 max			\ but not less than 10
    dup energy +!
    dup negate <food> +!

    dup pad 2 cells + @ * 10 /	\ and the third wrong sign
    dup future A-quality +! present A-quality +!
    dup pad	      @ * 10 /
    dup future B-quality +! present B-quality +!
    pad       cell+   @ * 10 /
    dup future C-quality +! present C-quality +! ;

\ uups, my imagination is too restricted to understand what it does ;-)
' eat-cyclic-rainbow eat-actions >list
' eat-cyclic-rainbow eat-xt !

: eat-crazy-rainbow ( -- )
    organ-A @ 0< 2* 1+ pad !		\ sign of organ-A as +1/-1 at pad
    organ-B @ 0< 2* 1+ pad cell+ !	\ sign of organ-B as +1/-1 at pad +cell
    organ-C @ 0< 2* 1+ pad 2 cells + !	\ sign of organ-C as +1/-1 at pad
    <food> @
    organ-A @ pad 2 cells + @ * +	\ just a set of crazy rules ;-)
    organ-B @ pad	    @ * +
    organ-C @ pad   cell+   @ * +
    
    A-quality @ pad cell+     @ *	pad           @ *  -	\ crazy, crazy
    B-quality @ pad 2 cells + @ *	pad   cell+   @ *  -
    C-quality @ pad	      @ *	pad 2 cells + @ *  -

    10 max			\ but not less than 10
    dup energy +!
    dup negate <food> +!

    dup pad 2 cells + @ * 10 /
    dup future A-quality +! present A-quality +!
    dup pad	      @ * 10 /
    dup future B-quality +! present B-quality +!
        pad   cell+   @ * 10 /
    dup future C-quality +! present C-quality +! ;

\ wonder what we will get!
' eat-crazy-rainbow eat-actions >list
\ ' eat-crazy-rainbow eat-xt !

: violet-eater-2
    A-quality @   dup >r   organ-A @   dup >r -
    dup future A-quality ! present A-quality !
    r> 0< 2* 1+   r> 0< 2* 1+  * ( sign )
    organ-A @ *  negate <food> @ + energy +!
    <food> off ;
' violet-eater-2 eat-actions >list

: violet-eater-3
    <food> @ energy +!
    <food> off
    A-quality @   0< 2* 1+
    organ-A @ *   energy +!
    organ-A @ negate
    dup future A-quality +! present A-quality +! ;
' violet-eater-3 eat-actions >list

: violet-eater-4
    <food> @   organ-A @ abs   -   energy +!
    organ-A @ 0< 2* 1+
    A-quality @ *  energy +!
    <food> off
    organ-A @ negate
    dup future A-quality +! present A-quality +! ;
' violet-eater-4 eat-actions >list

: violet-eater-5
    <food> @ energy +!
    <food> off
    organ-A @ abs negate energy +!
    organ-A @ 0< 2* 1+ ( signA)    A-quality @ *  energy +!
    organ-B @ abs negate energy +!
    organ-B @ 0< 2* 1+ ( signB)    B-quality @ *  energy +!
    organ-C @ abs negate energy +!
    organ-C @ 0< 2* 1+ ( signC)    C-quality @ *  energy +!

    organ-A @ negate
    dup future A-quality +! present A-quality +!
    organ-B @ negate
    dup future B-quality +! present B-quality +!
    organ-C @ negate
    dup future C-quality +! present C-quality +! ;
' violet-eater-5 eat-actions >list

' violet-eater-5 eat-xt !
\ ' diversify?-A is <diversify>
:noname diversify?-A diversify?-B diversify?-C ; IS <diversify>
0 organ-A !

false [if]
    big-bang
    1 sow drop
    10 food-share/spot !
    \ this one has a fascinating result (after a long time)
    ' generation>color foreground-color-xt !
[then]

: sow-some ( diversify? -- )
    page ." How many do you want to be sown? "
    1 num-in IF
	swap IF			\ diversify?
	    sow-diversified
	ELSE
	    sow
	THEN
	drop
	time-step
    THEN ;

MENU: mutation-men
: .mutation-menu ( -- )
    page

    title-colors ." Mutation menu:" end-title

    cr
    s" How often to mutate: "
    mutation-rate		simple-menu-entry-scale cr
    s" mMoO" menu-same-key-entry

    cr
    s" How often to insert in top level:  "
    top-level-insert		simple-menu-entry-scale cr
    s" iIlL" menu-same-key-entry
    s" How often to replace in top level: "
    top-level-replace		simple-menu-entry-scale cr
    s" rR" menu-same-key-entry
    s" How often to snip in top level:    "
    top-snipping		simple-menu-entry-scale cr
    s" s" menu-same-key-entry
    top-snipping cell+ @ IF
	."     relative probabilities of snip length: "
	s"   short "	top-short-snipping	simple-menu-entry-value
	s"   long "	top-long-snipping	simple-menu-entry-value
	s"   random "	top-random-snipping	simple-menu-entry-value cr
    THEN
    s" How often to replace addresses:    "
    top-address-replace		simple-menu-entry-scale cr
    s" How often to replace one token:    "
    top-token-replace		simple-menu-entry-scale cr
    s" How often to restart from scratch: "
    top-from-scratch		simple-menu-entry-scale cr

    cr
    s" Stack turning point: "  stack-turning-point   simple-menu-entry-value cr
    s" S" menu-same-key-entry
\   s" tT" menu-same-key-entry

    ." If stack depth is off so much, only genes that won't make it worse" cr
    ." will be insertet. Influences the possible complexity of mutations." cr

    cr
    s" Mutations threshold: "	mutations-threshold  simple-menu-entry-value cr
    ." After so many mutations we try to come to an end," cr
    ." respecting stack turning point and such." cr

    cr
    s" Trial phase: "		trial-phase	simple-menu-entry-value cr
    s" tT" menu-same-key-entry
    ." After how many generations will genes be compiled and put into the gene pool."
    cr

    cr
    s" Diversification of variables."	redisplay
    ['] diversification-menu	menu-entry cr
    s" dDvV" menu-same-key-entry

    s" qQxX" menu-done ['] noop		menu-key-entry
    common-menu-entries ;

: mutation-menu ( -- )
    mutation-men
    ['] .mutation-menu menu-display-xt !
    menu-done	['] noop	menu-key-default
    menu-done	['] noop	menu-default
    do-menu-loop
    free-menus ;
' mutation-menu function-key-actions >list

create spot-before field-planes# cells allot
create spot-after  field-planes# cells allot

MENU: edit-spot-men
defer ?record-edit-spot    
: .menu-edit-spot
    page

    title-colors ." Edit the selected spot:  # "  spot @ .  end-title
    cr

    cr
    21 menu-scroll-lines !

\   field-planes# p0 scrolled-range ?DO
    field-planes#  1 scrolled-range ?DO		\ not allowing pointers here!
	i field-variable-text  i n'th-spot-variable  simple-menu-entry-value cr
    LOOP

    common-menu-entries ;

: menu-edit-spot ( -- )
    edit-spot-men
    ['] .menu-edit-spot menu-display-xt !
    menu-done	['] noop	menu-key-default
    menu-done	['] noop	menu-default

    spot-vars@ spot-before spot-vars!
    do-menu-loop
    free-menus
    spot-vars@ spot-after spot-vars!
    spot-before  field-planes# cells  spot-after over  compare IF \ changed?
	?record-edit-spot
    THEN ;

: page-type ( addr count -- )	page type ;

variable (nuc-spot-known)	(nuc-spot-known) off \ also removed flag
defer ?record-remove-cell
: |remove-cell| ( -- )
    bell
    page cr red color-foreground
    cr ." Do you want this cell to be taken out of the play? "
    white color-foreground ." (k) y/n "
    green color-foreground
    cr cr
    key [char] y over = swap [char] k = or IF	\ 'y' and 'k' accepted as 'yes'
	log-mask @ IF
	    log-cat-id
	    s" : removed by user" log-death log-it
	THEN
	?record-remove-cell
	die
	1 (nuc-spot-known) !	\ magic flag for removed cell
	." cell removed."
    ELSE ." not removed" THEN
    cr 1500 ms
    default-color color-foreground ;

MENU: nuc-men
: .nuc-menu ( -- )
    cp@ IF					\ cp seems valid
	nuc-flags @ wake-me-trial and 0= IF
	    wake-me-xt @ setup-wake-me		\ so cost is always up to date
	THEN

	page

	title-colors ." Nucleus menu:" end-title

	cr
	nuc-flags @ wake-me-trial and IF	\ trial phase: genom string
	    s" wake-me: does a recently mutated gene which is on trial."
	    ['] see-genome	redisplay	menu-wait	menu-entry cr
	    s" gGwW" menu-same-key-entry
	ELSE
	    s" Choose genome of the cell.  (These genes can get mutated)."
	    title-choose-xt!
	    wake-me-actions	>stack		wake-me-xt	>stack-2
	    s" wake-me   does: "	['] choose-xt-menu	menu-entry
	    s" wW" menu-same-key-entry
	    1 4 screen-column
	    wake-me-xt @ xt>name  menu-wait redisplay ['] see-genome menu-entry
	    s" lLgG" menu-same-key-entry
	    2 4 screen-column
	    wake-me-xt @ gene-xt>internal-xt >stack
	    menu-wait	redisplay	['] .gene-info	name-menu-entry
	    cr
	THEN

	s" Choose the way the cells get rated and rewarded.  Eat action:"
	title-choose-xt!
	eat-actions		>stack		eat-xt		>stack-2
	s" eat       does: "	['] choose-xt-menu		menu-entry
	s" E" menu-same-key-entry
	1 4 screen-column
	eat-xt @	menu-wait	redisplay
	dup >stack	xt>name	['] page-see		menu-entry
	cr

	s" Choose how to reproduce:" title-choose-xt!
	reproduce-actions	>stack		reproduce-xt	>stack-2
	s" reproduce does: "	['] choose-xt-menu		menu-entry
	s" R" menu-same-key-entry
	1 4 screen-column
	reproduce-xt @	menu-wait	redisplay
	dup >stack	xt>name	['] page-see		menu-entry
	cr

	s" How does the cell show up in spot display?" title-choose-xt!
	show-me-actions		>stack		show-me-xt	>stack-2
	s" show-me   does: "	['] choose-xt-menu		menu-entry
	s" sS" menu-same-key-entry
	1 4 screen-column
	show-me-xt @
	menu-wait	redisplay
	dup >stack	xt>name	['] page-see		menu-entry cr

	cr
	s" Nuc id: "	ping	id		simple-menu-entry-value
	1 4 screen-column
	s" genom id  : " ping	genome-id	simple-menu-entry-value
	2 4 screen-column
	s" code cost: " code-cost @ ['] .sorry ping menu-entry-value
	3 4 screen-column
	s" nuc length: "  length @  ['] .sorry ping menu-entry-value cr

	s" age   : "		age		simple-menu-entry-value
	1 4 screen-column
	s" age limit : "	age-threshold	simple-menu-entry-value
	s" a" menu-same-key-entry
	2 4 screen-column
	s" generation: " ping	generation	simple-menu-entry-value
	3 4 screen-column
	s" genome gen: " ping  genom-generation	simple-menu-entry-value cr

	s" energy: "		energy		simple-menu-entry-value
	s" e" menu-same-key-entry
	1 4 screen-column
	appearance >stack   redisplay   s" appearance: "   appearance @
	['] char-to-addr menu-entry-value appearance @ .ascii up-to-here
	>last-xy
	2 4 screen-column
	s" reproduction energy: "
		reproduction-threshold		simple-menu-entry-value cr
	s" r" menu-same-key-entry

	cr
	at?		( 0 line )
	13 menu-scroll-lines !
	nuc-organs# 0 scrolled-range ?DO
	    from-here  nuc-organs i + >r r@ nuc-var-name type
	    s" : "  r> nuc-addr	simple-menu-entry-value cr
	    [char] A i + #key-same-entry
	LOOP
	at? nip >r				\ store y position

	at-xy
	nuc-parameters# 0 scrolled-range ?DO
	    1 3 screen-column
	    from-here  nuc-parameters i + >r r@ nuc-var-name type
	    s" : "  r> nuc-addr	simple-menu-entry-value cr
	    \ [char] a i + #key-same-entry
	LOOP
	0  at? nip r> max  at-xy		\ correct y position

	s" diversify"	['] diversification-menu	menu-entry
	s" d" menu-same-key-entry

	nuc-organs# IF
	    ."  organs: "
	    my-diversification-bitmask @
	    nuc-organs# 0 DO
		dup  1 i lshift  and IF  [char] A  ELSE  [char] a  THEN i +
		pad c!
		1 i lshift >stack	my-diversification-bitmask >stack-2
		pad 1		['] xor!	redisplay	menu-entry
		\ [char] A i + #key-same-entry
	    LOOP
	    drop
	THEN

	nuc-parameters# IF
	    ."  parameters: "
	    my-diversification-bitmask @
	    nuc-parameters# 0 DO
		dup  1 i nuc-organs# + lshift
		and IF  [char] A  ELSE  [char] a  THEN i + pad c!
		1 i nuc-organs# + lshift >stack
		my-diversification-bitmask >stack-2
		pad 1		['] xor!	redisplay	menu-entry
		\ [char] a i + #key-same-entry
	    LOOP
	    drop
	THEN
	at? nip 1+ last-line < IF cr THEN

	at? nip last-line <> IF
	    cr
	    s" Sow some of these"	false >stack	menu-done
	    ['] sow-some				menu-entry
	    1 3 screen-column
	    s" Sow diversified"		true >stack
	    ['] sow-some		menu-done	menu-entry
	    (nuc-spot-known) @ IF
		3 5 screen-column
		s" remove cell" menu-done ['] |remove-cell|	menu-entry
		s" kKrR" menu-same-key-entry
		4 5 screen-column
		s" Spot" redisplay	['] menu-edit-spot	menu-entry
	    THEN
	THEN

	at? nip last-line < IF
	    cr
	    s" (Diversification is done like specified in the diversification menu)"
	    ['] diversification-menu	menu-entry
	    s" d" menu-same-key-entry
	    at? nip last-line < IF cr THEN
	THEN
    ELSE					\ cp not valid
	." no living beeing here..." cr
    THEN

    common-menu-entries ;

: nuc-menu ( -- )
    nuc-men
    ['] .nuc-menu menu-display-xt !
    menu-done	['] noop	menu-key-default
    menu-done	['] noop	menu-default
    do-menu-loop
    free-menus ;

MENU: actual-pool-men
: .actual-pool-menu ( -- )
    page

    title-colors ." Gene pool: " actual-genepool-xt @ xt>name type
    10 spaces ." Set relative probabilities." clear-line-to-end
    reset-default-colors cr

    cr
    20 menu-scroll-lines !
    actual-genepool-xt @ execute
    dup nodes 0 scrolled-range ?DO
	i over n'th-node >r
	s" " r@ >probability simple-menu-entry-value 
	1 10 screen-column
	r> >xt @
	dup >stack  xt>name  ['] page-see menu-wait	redisplay menu-entry cr
    LOOP drop
    common-menu-entries ;

: actual-pool-menu ( -- )
    actual-pool-men
    ['] .actual-pool-menu menu-display-xt !

    menu-done	['] noop	menu-key-default
    menu-done	['] noop	menu-default
    do-menu-loop
    free-menus ;
' actual-pool-menu function-key-actions >list

MENU: gene-pool-men
: .gene-pool-menu ( -- )
    page

    title-colors ." Gene pools menu:" clear-line-to-end
    reset-default-colors cr

    cr
    s" Select actual gene pool:" title-choose-xt!
    gene-pools >stack	actual-genepool-xt >stack-2
    s" Actual gene pool: "	['] choose-xt-menu	menu-entry
    s" aA" menu-same-key-entry
    actual-genepool-xt @ xt>name ['] actual-pool-menu	menu-entry cr
    s" edit " ['] actual-pool-menu   menu-entry
    actual-genepool-xt @ xt>name type up-to-here cr
    s" eE" menu-same-key-entry

    common-menu-entries ;

: gene-pool-menu ( -- )
    gene-pool-men
    ['] .gene-pool-menu menu-display-xt !

    menu-done	['] noop	menu-key-default
    menu-done	['] noop	menu-default
    do-menu-loop
    free-menus ;
' gene-pool-menu function-key-actions >list

\ dummy individual to indicate missing selection
\ : (none) ;
' (none) individuals >list

MENU: individuals-men
variable selected-individual-xt		' (none) selected-individual-xt !
: .individuals-menu
    selected-individual-xt ?perform	\ makes the selected one the actual one
    page

    title-colors ." Individuals menu:" clear-line-to-end
    reset-default-colors cr

    cr
    s" Select an individual:" title-choose-xt!
    individuals >stack		selected-individual-xt >stack-2
    s" Selected Individual: "	['] choose-xt-menu		menu-entry
    s" sSiI" menu-same-key-entry

    .tab selected-individual-xt @ xt>name
    selected-individual-xt @ ['] (none) = IF type cr ELSE
	['] nuc-menu	redisplay		menu-entry cr

	s" See and edit it."	redisplay	['] nuc-menu	menu-entry cr

	cr
	s" Sow some clones"			false >stack	menu-done
	['] sow-some				menu-entry cr
	s" cC" menu-same-key-entry
	s" Sow some diversified copies"		true >stack	menu-done
	['] sow-some				menu-entry cr
	s" d" menu-same-key-entry

	s" (Diversification is done like specified in the diversification menu)"
	['] diversification-menu			menu-entry cr
	s" D" menu-same-key-entry
    THEN

    s" qQxX" menu-done ['] noop		menu-key-entry

    common-menu-entries ;

: individuals-menu
    individuals-men
    ['] .individuals-menu menu-display-xt !

    menu-done	['] noop	menu-key-default
    menu-done	['] noop	menu-default
    do-menu-loop
    free-menus
    ['] (none) selected-individual-xt @ <> IF
	selected-individual-xt @ EXECUTE
	cp@ (manually-selected-cell) !
    THEN ;
' individuals-menu function-key-actions >list

: change-world-do-direction ( -- )
    world-do-direction @ CASE
	1  OF -1 ENDOF
	-1 OF  0 ENDOF
	0  OF  2 ENDOF
	2  OF -2 ENDOF
	-2 OF  1 ENDOF
    ENDCASE world-do-direction ! ;

: display-on-off ( -- )
    display-switch dup @
    spot-display-on xor

    [ spot-display-on step-display-on or ] literal >r	\ both on?
    dup r@ and r> = IF
	spot-display-prefered if	\ switch to prefered on
	    step-display-on
	ELSE
	    spot-display-on
	THEN
	invert and
    THEN

    swap ! ;

: toggle-display-type ( -- )
    display-switch dup @
    [ spot-display-on step-display-on or ] literal >r
    r@ xor
    dup r@ and r> = IF			\ step AND spot display on
	spot-display-prefered IF
	    step-display-on		\ switch one off
	ELSE
	    spot-display-on
	THEN
	invert and
    THEN
    swap !
    ?reset-continuous-column
    ?step-display-sanity ;

variable (brew-show-or-go)	(brew-show-or-go) off
: brew-show-or-go ( -- )	(brew-show-or-go) on ;

: toggle-display-&-go ( -- )
    toggle-display-type menu-redisplay on ;
' toggle-display-&-go function-key-actions >list
' toggle-display-&-go F2-xt !

' do-FORTH function-key-actions >list
' do-FORTH F9-xt !

variable anything		anything off
: toggle-anything ( -- )	-1 anything xor! ;	\ switch something
' toggle-anything function-key-actions >list
' toggle-anything F11-xt !

' goodbye F12-xt !

MENU: function-key-men
: sfk ( -- )    s" Select action for this function key" title-choose-xt! ;
\ braindead cut&paste code ;-)
: .function-key-menu
    page

    title-colors ." Select actions of the function keys:" clear-line-to-end
    reset-default-colors cr

    cr
    sfk
    function-key-actions >stack		F1-xt	>stack-2
    s" F1   does: "	['] choose-xt-menu		menu-entry
    F1-xt @ xt>name type
    s" 1" menu-same-key-entry
    1 2 screen-column
    sfk
    function-key-actions >stack		shift-F1-xt	>stack-2
    s" shift-F1  does: "	['] choose-xt-menu		menu-entry
    shift-F1-xt @ xt>name type
    cr
    sfk
    function-key-actions >stack		F2-xt	>stack-2
    s" F2   does: "	['] choose-xt-menu		menu-entry
    F2-xt @ xt>name type
    s" 2" menu-same-key-entry
    1 2 screen-column
    sfk
    function-key-actions >stack		shift-F2-xt	>stack-2
    s" shift-F2  does: "	['] choose-xt-menu		menu-entry
    shift-F2-xt @ xt>name type
    cr
    sfk
    function-key-actions >stack		F3-xt	>stack-2
    s" F3   does: "	['] choose-xt-menu		menu-entry
    F3-xt @ xt>name type
    s" 3" menu-same-key-entry
    1 2 screen-column
    sfk
    function-key-actions >stack		shift-F3-xt	>stack-2
    s" shift-F3  does: "	['] choose-xt-menu		menu-entry
    shift-F3-xt @ xt>name type
    cr
    sfk
    function-key-actions >stack		F4-xt	>stack-2
    s" F4   does: "	['] choose-xt-menu		menu-entry
    F4-xt @ xt>name type
    s" 4" menu-same-key-entry
    1 2 screen-column
    sfk
    function-key-actions >stack		shift-F4-xt	>stack-2
    s" shift-F4  does: "	['] choose-xt-menu		menu-entry
    shift-F4-xt @ xt>name type
    cr
    sfk
    function-key-actions >stack		F5-xt	>stack-2
    s" F5   does: "	['] choose-xt-menu		menu-entry
    F5-xt @ xt>name type
    s" 5" menu-same-key-entry
    1 2 screen-column
    sfk
    function-key-actions >stack		shift-F5-xt	>stack-2
    s" shift-F5  does: "	['] choose-xt-menu		menu-entry
    shift-F5-xt @ xt>name type
    cr
    function-key-actions >stack		F6-xt	>stack-2
    s" F6   does: "	['] choose-xt-menu		menu-entry
    F6-xt @ xt>name type
    s" 6" menu-same-key-entry
    1 2 screen-column
    sfk
    function-key-actions >stack		shift-F6-xt	>stack-2
    s" shift-F6  does: "	['] choose-xt-menu		menu-entry
    shift-F6-xt @ xt>name type
    cr
    sfk
    function-key-actions >stack		F7-xt	>stack-2
    s" F7   does: "	['] choose-xt-menu		menu-entry
    F7-xt @ xt>name type
    s" 7" menu-same-key-entry
    1 2 screen-column
    sfk
    function-key-actions >stack		shift-F7-xt	>stack-2
    s" shift-F7  does: "	['] choose-xt-menu		menu-entry
    shift-F7-xt @ xt>name type
    cr
    sfk
    function-key-actions >stack		F8-xt	>stack-2
    s" F8   does: "	['] choose-xt-menu		menu-entry
    F8-xt @ xt>name type
    s" 8" menu-same-key-entry
    1 2 screen-column
    sfk
    function-key-actions >stack		shift-F8-xt	>stack-2
    s" shift-F8  does: "	['] choose-xt-menu		menu-entry
    shift-F8-xt @ xt>name type
    cr
    sfk
    function-key-actions >stack		F9-xt	>stack-2
    s" F9   does: "	['] choose-xt-menu		menu-entry
    F9-xt @ xt>name type
    s" 9" menu-same-key-entry
    cr
    sfk
    function-key-actions >stack		F10-xt	>stack-2
    s" F10  does: "	['] choose-xt-menu		menu-entry
    F10-xt @ xt>name type
    cr
    sfk
    function-key-actions >stack		F11-xt	>stack-2
    s" F11  does: "	['] choose-xt-menu		menu-entry
    F11-xt @ xt>name type
    cr
    sfk
    function-key-actions >stack		F12-xt	>stack-2
    s" F12  does: "	['] choose-xt-menu		menu-entry
    F12-xt @ xt>name type
    cr

    ['] redisplay default-function-keys

    common-menu-entries ;

: function-key-menu
    function-key-men
    ['] .function-key-menu menu-display-xt !
    menu-done	['] noop	menu-key-default
    menu-done	['] noop	menu-default
    do-menu-loop
    free-menus ;
' function-key-menu function-key-actions >list

: step-OR-spot
    display-switch @ >r			( -- r: actual-display-switch )
    spot-display-on step-display-on or	( display-type-mask    r: switch)
    r@ and				( display-type         r: switch)
    CASE \ type
	spot-display-on OF toggle-display-type ENDOF
	step-display-on OF toggle-display-type ENDOF
	\ else we put spot display on and go to step display setup:
	r@ spot-display-on or	step-display-on invert and   display-switch !
    ENDCASE rdrop

    display-switch @ step-display-on and IF	\ step display switched on
	menu-step-display			\ allow setup
    THEN ;

\ toggle-step-display ( -- )
:NONAME ( -- )
    display-switch >r
    r@ @ step-display-on xor r@ !
    r> @ step-display-on and IF			\ step display switched on
	menu-step-display			\ allow setup
    THEN ; IS toggle-step-display

MENU: display-men
: .display-menu
    page

    title-colors ." Display menu:" end-title

    cr
    [ decimal ] 16 stringbuf-open >r
    display-switch @ step-display-on and IF
	s" STEP OR "
    ELSE
	s" step OR "
    THEN
    r@ cat
    display-switch @ spot-display-on and IF
	s" SPOT"
    ELSE
	s" spot"
    THEN
    r@ cat
    s"  display. " r@ cat
    r@ buffer-as-string	redisplay	['] step-OR-spot	menu-entry
    r> stringbuf-close
    [ step-display-on spot-display-on or ] literal dup display-switch @ and =
    IF
	green color-foreground
	."  SETTING STEP *AND* SPOT DISPLAY IS NOT RECOMMENDED..."
	up-to-here default-color color-foreground
    THEN cr
    s" oO" menu-same-key-entry

    cr
    s" Step display "	redisplay	['] toggle-step-display	menu-entry
    s" eE" menu-same-key-entry
    display-switch @ step-display-on and dup .ON-off
    IF
	s" displaying " step-display-items simple-menu-entry-value
	." items." up-to-here cr

	s" Edit step display" redisplay ['] menu-step-display menu-entry
	s" eE" menu-same-key-entry
\  	s" Zoom steadiness, horizontal: "
\  	horizontal-zoom-scale	simple-menu-entry-scale
\  	s"   vertical: "	vertical-zoom-scale	simple-menu-entry-scale
    THEN cr

    cr
    s" Spot display "	spot-display-on	>stack		redisplay
    display-switch	>stack-2	['] xor!	menu-entry
    s" pP" menu-same-key-entry
    display-switch @ spot-display-on and dup .ON-off	cr
    IF
	s" What does 'look up' show?" title-choose-xt!
	show-me-actions	>stack		look-at-xt	>stack-2
	s" Look-at function: "	['] choose-xt-menu redisplay	menu-entry
	s" lL" menu-same-key-entry
	.tab	menu-wait	redisplay	look-at-xt @
	dup >stack	xt>name	['] page-see			menu-entry cr

	look-at-xt @ ['] <look-at> = IF		\ uups!
	    bell ." ==> No, don't set the function to itself, thats nonsense!"
	    cr
	    ['] show-generation look-at-xt !
	THEN

	s" Color menu "	['] color-menu	menu-done do-after	menu-entry
	s" cC" menu-same-key-entry
	cr
    THEN

    display-switch @
    [ step-display-on spot-display-on or ] literal and 0= IF
	cr
	display-switch @ no-snapshots and IF
	    s" Don't show snapshots "
	ELSE
	    s" Show snapshots "
	THEN
	no-snapshots >stack	display-switch >stack-2
	redisplay		['] xor!	menu-entry cr
    THEN

    display-slots @	dup >r		\ assure range for display-slots
    0 max	max-display-slots# min
    dup display-slots !
    r> <> IF bell cr ." Slot number was out of range, corrected." THEN

    cr
    s" The last line can display different infos.  slots: "
    display-slots	simple-menu-entry-value cr
    s" nNsSiI" menu-same-key-entry

    display-slots @ 6 > IF
	." (caution, too many slots might mess up your display)"
	up-to-here cr
\	display-slots >stack redisplay
\	['] change-value-at-addr menu-entry cr
    THEN

    display-slots @ 0 ?DO
	slot-display-words		>stack
	display-slots	i 1+ cells +	>stack-2
	s" Slot does: "	redisplay	['] choose-xt-menu	menu-entry
	s" What to display in this slot?" title-choose-xt!
	i [char] 1 + pad c!  pad 1 menu-same-key-entry
	.tab i 1+ cells display-slots + @
	menu-wait	redisplay
	dup >stack	xt>name		['] page-see		menu-entry cr
    LOOP

    common-menu-entries ;

: display-menu
    display-men
    ['] .display-menu menu-display-xt !
    menu-done	['] noop	menu-key-default
    menu-done	['] noop	menu-default
    do-menu-loop
    free-menus ;
' display-menu function-key-actions >list

: accept>stringbuf ( handle -- )
    ." new:            " .bs .bs .bs .bs .bs .bs .bs .bs .bs .bs .bs
    pad dup [ decimal ] 80  accept
    ?dup IF
	rot dup stringbuf-empty cat
    ELSE 2drop THEN ;

\ only use inside menus.
: .code-file-ON-off ( mask -- )   code-file-mask @ and .ON-off up-to-here ;

MENU: code-file-men
: .code-file-menu ( -- )
    page

    title-colors ." Code file menu:" end-title

    cr
    s" Writing a human readable code file to the hd "	redisplay
    write-code-file >stack	code-file-mask >stack-2	['] xor!  menu-entry
    write-code-file .code-file-ON-off cr
    s" cCwWhHoO" menu-same-key-entry
    write-code-file  code-file-mask @  AND IF
	cr
	s" Code file: "	(code-file-name) >stack
	redisplay		['] accept>stringbuf	menu-entry
	.tab (code-file-name) buffer-as-string type up-to-here
	.tab >last-xy cr
	s" cCfF" menu-same-key-entry

	cr
	s" file on trial SUCCESS "	file-end-trial	>stack	redisplay
	code-file-mask	>stack-2	['] xor! menu-entry
	.tab 		file-end-trial	.code-file-ON-off cr
	s" tTsS"	menu-same-key-entry
	s" file ALL MUTATIONs "	file-mutating	>stack		redisplay
	code-file-mask	>stack-2	['] xor! menu-entry
	.tab 		file-mutating	.code-file-ON-off cr
	s" aAmM"	menu-same-key-entry
	cr

	[ file-mutating file-end-trial or ] literal code-file-mask @ and IF
	    cr
	    ." What to put in the code file: "		redisplay
	    .tab s"  ALL  "	code-file-mask >stack	['] on	   menu-entry
	    s" aA"	menu-same-key-entry
	    .tab s"  nothing "
	    code-file-mask >stack	['] off	redisplay menu-entry
	    s" nN"	menu-same-key-entry
	    .tab s"  CODE & stack "	redisplay
	    [ write-code-file file-end-trial file-code file-stack or or or ]
	    literal >stack	code-file-mask >stack-2	  ['] !	 menu-entry cr
	    s" cC"	menu-same-key-entry
	THEN

	[ file-mutating file-end-trial or ] literal code-file-mask @ and IF
	    s" code: "	file-code	>stack		redisplay
	    code-file-mask	>stack-2	['] xor! menu-entry
	    .tab .tab	file-code	.code-file-ON-off cr
	    \ s" cC"	menu-same-key-entry
	    s" stack: "	file-stack	>stack		redisplay
	    code-file-mask	>stack-2	['] xor! menu-entry
	    .tab .tab	file-stack	.code-file-ON-off cr 
	    \ s" sS"		menu-same-key-entry
	THEN

	file-mutating code-file-mask @ and IF
	    s" frames: "	file-frames	>stack		redisplay
	    code-file-mask	>stack-2	['] xor! menu-entry
	    .tab		file-frames	.code-file-ON-off cr 
	    s" f"		menu-same-key-entry
	    s" depth&cost: " file-depth&cost	>stack		redisplay
	    code-file-mask	>stack-2	['] xor! menu-entry
	    .tab		file-depth&cost	.code-file-ON-off cr 
	    s" dD"		menu-same-key-entry
	    s" start&stop: " file-start&stop-entries	>stack	redisplay
	    code-file-mask	>stack-2	['] xor! menu-entry
	    .tab file-start&stop-entries	.code-file-ON-off cr 
	    \ s" tT"		menu-same-key-entry
	    s" step,spot,id: " file-step&spot&id	>stack	redisplay
	    code-file-mask	>stack-2	['] xor! menu-entry
	    .tab file-step&spot&id	.code-file-ON-off cr 
	    s" iI"		menu-same-key-entry
	    s" structures: " file-structure	>stack		redisplay
	    code-file-mask	>stack-2	['] xor! menu-entry
	    .tab file-structure	.code-file-ON-off cr 
	    \ s" S"		menu-same-key-entry
	    s" item-numbers: " file-item-number	>stack		redisplay
	    code-file-mask	>stack-2	['] xor! menu-entry
	    .tab file-item-number	.code-file-ON-off cr
	THEN
    THEN
    common-menu-entries ;

: code-file-menu
    code-file-men
    ['] .code-file-menu menu-display-xt !
    menu-done	['] noop	menu-key-default
    menu-done	['] noop	menu-default
    do-menu-loop
    free-menus ;
' code-file-menu function-key-actions >list

: .log-ON-off ( mask -- )  log-mask @ and .ON-off ;

: fix-log-flags ( -- )
    log-mask dup @	( log-mask-addr log-mask )

    [ log-mask @ 0> ] [IF] \ log extra information
	dup log-m-extra and IF
	    log-m-much or
	THEN
    [THEN]

    [ log-mask @ ] [IF]
	dup log-m-much and IF
	    log-m-more or
	THEN
	dup log-m-more and IF
	    log-m-some or
	THEN
    [THEN]

    dup log-m-some and IF
	log-m-type or
    THEN
    dup log-m-type and IF
	log-mutation or
    THEN

    swap ! ;

MENU: log-men
: .log-menu ( -- )
    page

    fix-log-flags
    title-colors ." Menu logging:" end-title

    ." Log files can grow huge, don't turn this on unless you really want to!"
    cr

    [ log-mask @ ] [IF]	\ compiled with fine grain mutation debugging
	[ log-mask @ 0> ] [IF] \ log extra information
	    19
	[ELSE]
	    18
	[THEN]
	menu-scroll-lines !
    [THEN]

    cr
    at? nip	( start-line )
    log-masks
    dup nodes 0 [ log-mask @ ] [IF] scrolled-range [THEN]  ?DO
	i over n'th-node @ >r	( list  r: xt )
	r@ xt>name 4 /string			\ title for menu-entry
	r@ execute >stack			\ mask >stack
	log-mask >stack-2
	redisplay	['] xor! menu-entry
	12 at-x r@ execute .log-ON-off up-to-here cr
	r> xt>name drop 4 + c@ #key-same-entry
    LOOP
    drop
    s" q" menu-done ['] noop	menu-key-entry

    cr
    s" Log file: "	(log-file-name) >stack
    redisplay		['] accept>stringbuf	menu-entry
    .tab (log-file-name) buffer-as-string type
    .tab >last-xy
    s" fFrR" menu-same-key-entry

    s" Writing code file: "	redisplay	['] code-file-menu menu-entry
    write-code-file .code-file-ON-off cr
    s" wWcC" menu-same-key-entry
    at? 2>r

    ( start-line ) 0 swap at-xy 1 2 screen-column
    s" All  "	log-mask >stack	['] on	redisplay   menu-entry
    s" aA"	menu-same-key-entry
    s"  Nothing "	log-mask >stack	['] off	redisplay menu-entry
    s" nN"	menu-same-key-entry
    s"  Some "	redisplay
    [ log-spot  log-birth or  log-death or  log-meal or  log-costs or
    log-trial or  log-emergency or  log-step or  log-user or  log-mutation or
    log-m-type or  log-m-some or ] literal >stack
    log-mask >stack-2 ['] ! menu-entry
    s" S"	menu-same-key-entry

    2r> at-xy	\ position cursor
    common-menu-entries ;

: log-menu
    log-men
    ['] .log-menu menu-display-xt !
    menu-done	['] noop	menu-key-default
    menu-done	['] noop	menu-default
    do-menu-loop
    free-menus ;
' log-menu function-key-actions >list

\ these are old demos
\ they are not guaranteed to work on this version
\ newer demos use the playback feature, these here don't
LIST: demos
: symbiotic-waves ( -- )	\ demo
    page
    ." Demo symbiotic behavior:	" cr
    cr 
    ." 2 species of cells, one increasing one decreasing the A-quality," cr
    ." whatever this might be." cr
    cr
    ." Each kind dislikes the change that it makes to 'A' himself" cr
    ." but does like the change the other one makes." cr
    ." Hmm, sounds familiar," cr
    ." somehow." cr
    cr
    ." Parameters are set, that both species just can survive and reproduce," cr
    ." if they start on a zero 'A' level." cr
    cr
    ." On a spot, another cell of the same species was living before," cr
    ." a cell will not be able to reproduce." cr
    ." On the other hand is living, where a cell of the other species happened to" cr
    ." be before, very easy.  Selection lets this effect increase over time."
    cr cr
    ." So the cell population will always move." cr
    ." The two species make waves and whirles following each other." cr
    ." If one species vanishes, the other one will vanish soon too." cr
    cr
    ." Press 'b' to let the background color show the enviromental change." cr
    cr
    ." Give any number as seed for the random generator: "
    seed-BRODIE dup . change-value-at-addr

    free-field
    \    1085634929 seed-BRODIE !
    \    1234567890 seed-BRODIE !
    \ 0 seed-BRODIE !
    \ 34567 seed-BRODIE !
    red-eater 3 sow drop
    blue-eater 5 sow drop
    time-step
    1		world-do-direction !
    0		world-food-supply !
    200		food-share/spot !
    0		individual-fixed-food-share !
    0		leave-energy-after-death !
    1		additive-stress !
    1 1		stress-rate 2!
    640		high-water-mark !
    spots	flood-mark !
    0		sos-mark !
    100		low-water-mark !
    1000	up-regulation-start !
    33		diversification-range !
    1 4		diversification-rate 2!
    2		diversification-closeness !
    diversification-mask dup @ 1 or swap !
    0 1		mutation-rate 2!
    ['] A-quality>color		background-color-xt !
    ['] generation>color	foreground-color-xt !
    192		A-quality>color-scale !
    4 display-slots !
    ['] .step 0 display-slot !
    ['] .cells 1 display-slot !
    ['] .burden 2 display-slot !
    ['] .living 3 display-slot !
    (scan-index) off
    14 (scan-detail-index) !
    ['] toggle-display-&-go F2-xt ! ;
' symbiotic-waves demos >list

: nomadic-bursts ( -- )		\ demo
    page cr
    ." This demo needs some moments to develop." cr
    ." After a few hundred steps islands of resident cell cultures emerge." cr cr
    ." As food is spread everywhere, food level increases in the empty spaces."
    cr
    ." This leads to sudden bursts of nomadic mutants, eating that food up." cr
    cr
    ." You can watch food level by pressing 'b'." cr
    ." Calibrate the food to color scale in the color menu 'C'." cr
    cr
    ." The waves of moving cells change some qualities of the places they pass by" cr
    ." in a mutation specific way." cr
    ." These changed qualities influence subsequent waves of nomadic cells." cr
    cr
    ." Change what background color shows to 'A','B', or 'C-quality' to see it." cr cr
    ." Sometimes some nomades settle down and form a new inhabited island." cr
    cr
    ." Try changing food policy in the food menu, reached by pressing 'F'."
    cr cr
    ." <SPACE> starts and stops the demo." cr
    cr
    ." You can give a number as seed for the random generator: " cr
    seed-BRODIE dup . change-value-at-addr

    free-field
    rainbow-eater
    ['] <look-at> show-me-xt !
    ['] show-ABC*. look-at-xt !
    3 sow drop
    ['] show-ABC-X| show-me-xt !
    4 sow drop
    time-step
    1		world-do-direction !
    0 		world-food-supply !
    10 		food-share/spot !
    0 		individual-fixed-food-share !
    0 		leave-energy-after-death !
    -1 		nuc-cost-can-be-help? !
    0 		leave-energy-after-death !
    1 		additive-stress !
    1 1		stress-rate 2!
    320 	high-water-mark !
    spots	flood-mark !
    0		sos-mark !
    100 	low-water-mark !
    1000 	up-regulation-start !
    -1 		nuc-cost-can-be-help? !
    50 		diversification-range !
    1 4		diversification-rate 2!
    2 		diversification-closeness !
    diversification-mask dup @ 7 or swap !
    0 1 mutation-rate 2!
    ['] food>color 	 background-color-xt !
    ['] generation>color 	 foreground-color-xt !
    \ ['] show-~|'.* 	 look-at-xt !
    600 		 food>color-scale !
    300 		 A-quality>color-scale !
    300 		 B-quality>color-scale !
    300 		 C-quality>color-scale !
    100 		 (A-scale) !
    100 		 (B-scale) !
    100 		 (C-scale) !
    4 display-slots !
    ['] .step 0 display-slot !
    ['] .cells 1 display-slot !
    ['] .burden 2 display-slot !
    ['] .living 3 display-slot ! ;
' nomadic-bursts demos >list

INDIVIDUAL: fertile
also genes
' g-noop setup-wake-me
previous

' cell-division reproduce-xt !
\ 120 reproduction-threshold !
\ 4   age-threshold !
60 reproduction-threshold !
2   age-threshold !
char * char 0 -   organ-A !
0 my-diversification-bitmask !
' eat-all eat-xt !
' <look-at>     show-me-xt !
' fertile individuals >list

: yeast1 ( -- )		\ demo
    page cr
    ." Demo population control." cr
    cr
    ." Fertile but quite short living cells showing an explosive population growth." cr
    cr
    ." Population is controlled by food sharing and increasing the life cost 'burden'," cr
    ." until there are no more individuals than the high water mark." cr
    cr
    ." Mutation does only change the look of these cells, not the behavior." cr
    wait

    free-field

    prototype
    [char] * [char] 0 - organ-A !
    ['] show-A look-at-xt !
    1 2-ascii-scale !
    1 food-share/spot !
    ['] food>color background-color-xt !
    12 food>color-scale !
    5 individual-fixed-food-share !
    2000 world-food-supply !
    1 additive-stress !
    1 1 stress-rate 2!
    420 high-water-mark !
    spots flood-mark !
    0 sos-mark !
    100 low-water-mark !
    100 up-regulation-start !
    diversification-mask dup @ 1 or swap !
    1 my-diversification-bitmask !
    0 leave-energy-after-death !
    -1 nuc-cost-can-be-help? !
    0 leave-energy-after-death !
    0 1 mutation-rate 2!

    ['] <look-at> show-me-xt !
    ['] show-A look-at-xt !

    ['] random-BRODIE random-xt !
    1334532601 seed-BRODIE !
    
    3 sow drop
    time-step
    1 world-do-direction !
    1 100 diversification-rate 2!
    2 diversification-range !
    1 diversification-closeness !
    1 2-ascii-scale !
    4 display-slots !
    ['] .step 0 display-slot !
    ['] .cells 1 display-slot !
    ['] .burden 2 display-slot !
    ['] .living 3 display-slot ! ;
' yeast1 demos >list

: yeast2 ( -- )		\ demo
    page cr
    ." Demo population control." cr
    cr
    ." Very fertile but short living cells showing an explosive population growth." cr
    cr
    ." Population is controlled only by the fact, that most food is divided equally" cr
    ." among all living cells. A small amount of food is spread everywhere," cr
    ." to let the cell cultures move." cr
    cr
    ." Mutation does only change the look of these cells, not the behavior." cr
    wait

    1 world-do-direction !
    2000 world-food-supply !
    1 food-share/spot !
\    7 individual-fixed-food-share !
    0 individual-fixed-food-share !
    0 nuc-do-cost !
    0 leave-energy-after-death !
    1 additive-stress !
    1 1 stress-rate 2!
    640 high-water-mark !
    spots flood-mark !
    0 sos-mark !
    100 low-water-mark !
    100 up-regulation-start !
    -1 nuc-cost-can-be-help? !
    diversification-mask dup @ 1 or swap !
    ['] food>color background-color-xt !
    ['] generation>color foreground-color-xt !
    ['] show-A look-at-xt !
    1 2-ascii-scale !
    9 food>color-scale !
    ['] random-BRODIE random-xt !
    1334532601 seed-BRODIE !
    0 cell-division-moves-both !
    0 cell-division-diversify-both !
    0 cell-division-mutate-both !
    0 1 mutation-rate 2!
    spot-display-on display-switch !
    3 display-slots !
    ['] .step 0 display-slot !
    ['] .cells 1 display-slot !
    ['] .living 2 display-slot !

    free-field
    fertile
    ['] <look-at> show-me-xt !
    ['] show-A look-at-xt !

    1 my-diversification-bitmask !
    5 diversification-range !
    1 diversification-closeness !
    3 sow-diversified drop
    time-step
    2 diversification-range !
    1 100 diversification-rate 2! ;
' yeast2 demos >list

: selecting-appetite ( -- )
    page
    ." This demo shows a very simple selective adaption." cr
    cr
    ." There is abundance of food here." cr
    ." But these cells eat only a limited amount, controlled by the value in organ-A." cr
    ." This appetite is diversified in mutations." cr
    ." The numbers shown in the beginning are appetite divided by 100." cr
    cr
    ." The high water mark is set to 100 individuals, so very soon the conditions" cr
    ." will automatically get harder.  Individuals that eat more survive better." cr
    ." So there is a constant positive selection of the hungriest cells, which" cr
    ." lets the value in organ-A increase constantly.  Press <F2> to see." cr
    cr
    ." You can see a summary of the values in all living cells by pressing 'n'." cr
    cr
    ." As food supply is not the problem here the cell cultures have no need to move." cr
    ." As the appetite grows the cells eat up all the food that summed up and the cell" cr
    ." cultures start to move more and more.  Finally they form a repeating pattern" cr
    ." (or eventually die out)." cr
    cr
    ." Things could develop similar in the real world if we get too greedy..." cr
    cr
    ." Here you can change it by increasing the food amount in the food menu reached" cr
    ." by pressing 'F'."
    wait

    free-field
    Yuppie
    5 sow drop
    time-step

    1 world-do-direction !
    0 world-food-supply !
    200 food-share/spot !
    0 individual-fixed-food-share !
    0 nuc-do-cost !
    0 leave-energy-after-death !
    1 additive-stress !
    1 1 stress-rate 2!
    100 high-water-mark !
    spots flood-mark !
    0 sos-mark !
    40 low-water-mark !
    1000 up-regulation-start !
    -1 nuc-cost-can-be-help? !
    50 diversification-range !
    1 4 diversification-rate 2!
    2 diversification-closeness !
    diversification-mask dup @ 1 or swap !
    ['] food>color background-color-xt !
    ['] generation>color foreground-color-xt !
    100 2-ascii-scale !
    0 cell-division-moves-both !
    0 cell-division-diversify-both !
    0 cell-division-mutate-both !
    0 1 mutation-rate 2!
    spot-display-on display-switch !
    4 display-slots !
    ['] .step 0 display-slot !
    ['] .cells 1 display-slot !
    ['] .burden 2 display-slot !
    ['] .living 3 display-slot !
    (scan-index) off
    14 (scan-detail-index) !
    ['] toggle-display-&-go F2-xt ! ;
' selecting-appetite demos >list

variable (demo)
: demo-menu ( -- )		\ technically speaking it's a pseudo menu
    s" Choose a demo:" title-choose-xt!
    (demo) off
    demos (demo) choose-xt-menu ( addr-of-list addr-to-store-xt -- )
    cursor-off
    (demo) @ IF start-help THEN
    (demo) ?perform ;


\ Words to save variables and the like to files
\ They are used to store the state of all relevant variables
\
\ Also used while recording to determine changes the user has made
\ by comparing two saved states.
\ In this case *all* variables are saved.
\ After comparing all lines that differ will be written to the record file.

decimal
256 STRINGBUF-HANDLE: (out-buffer)
: cat2out ( addr count -- )   (out-buffer) cat ;
: char2out ( c -- )   (out-buffer) char-cat ;

\ I want some variables to be set in play files anyway, even if te're equal.
\ the trick is to make them look different by adding a unique comment
variable (unique#)		-1 (unique#) !
variable (record-anyway)	(record-anyway) off
\
\ same, new style: with a stringbuffer
decimal 64 STRINGBUF-HANDLE: (unique)
: unique-start ( -- addr count )	s" 	\ unique: " ;
: ?cat-unique-comment ( -- )
    (unique) buffer-as-string dup IF
	unique-start cat2out
	cat2out
	(unique) stringbuf-empty
    ELSE 2drop THEN

    (record-anyway) @ IF		\ obsolete old style
	s" 	 \ gets recorded changed or not "	cat2out
	(unique#) dup @ num>string			cat2out
	1 swap +!
	(record-anyway) off
    THEN ;

: ?record-anyway ( flag -- )   IF (record-anyway) on THEN ;

variable (outfile-id)
: set-outfile ( id -- )   (outfile-id) ! ;

\ variable (out-type)	(out-type) off	\ out to string, .... buffer, screen
: out-line ( -- )	\ write string in (out-buffer) as line to (outfile-id)
    ?cat-unique-comment
    (out-buffer) buffer-as-string (outfile-id) @ write-line
    IF bell cr ." out-line: Error writing to file." 100 ms THEN
    [ flush-files ] [IF]
	(outfile-id) @ flush-file drop
    [THEN]
    (out-buffer) stringbuf-empty ;

: save-variable ( xt -- )	\ saves variable as FORTH code in (outfile-id)
    >r				\ 'dup' would not work with offsets
    r@ EXECUTE @ num>string	cat2out
    s"  "			cat2out
    r> xt>name			cat2out
    s"  !"			cat2out	out-line ;

\ save a variable given in the 'base >offset' form.
\ 'base' must be on stack when saving *and* when reading the file.
: save-offset-variable ( base-addr xt -- base-addr )
    >r dup r@ EXECUTE @		( base-addr value   r: xt )
    num>string			cat2out
    s"  over "			cat2out
    r> xt>name			cat2out
    s"  !"			cat2out	out-line ;

: save-2variable ( xt -- )	\ saves 2variable as FORTH code in (outfile-id)
    >r r@ EXECUTE 2@ swap	\ works for offsets too like this
    num>string			cat2out
    s"  "			cat2out
    num>string			cat2out
    s"  "			cat2out
    r> xt>name			cat2out
    s"  2!"			cat2out	out-line ;

: save-xt-variable ( xt -- )	\ saves xt-variable as FORTH in (outfile-id)
    \ like save-variable for variables containing xt's    

    \ dup EXECUTE @	( xt value=xt2|0 )		\ not ok for offsets
    >r r@ EXECUTE @ r> swap	( xt value=xt2|0 )	\ offsets also
    dup 0= IF drop save-variable EXIT THEN	\ empty xt-variable

    s" ' "			cat2out
    xt>name			cat2out
    bl				char2out
    xt>name			cat2out
    s"  !"			cat2out	out-line ;

\ save a xt variable given in the 'base >offset' form.
\ 'base' must be on stack when saving *and* when reading the file.
: save-offset-xt-variable ( base-addr xt -- base-addr )
    >r dup r@ EXECUTE @		( base-addr value-xt|0   r: xt )
    dup 0= IF drop r> save-offset-variable EXIT THEN

    s" ' "			cat2out
    xt>name			cat2out
    s"  over "			cat2out
    r> xt>name			cat2out
    s"  !"			cat2out	out-line ;

\ word to save bitmasks as readable text
: save-listed-mask ( xt-of-bitmask-variable list-addr -- )
    swap >r			( list   r: xt )
    r@ EXECUTE @ swap		( bitmask node   r: xt )
    s" 0 "		cat2out
    dup nodes 0 ?DO		( bitmask node )
	next-node >r		( bitmask  r: node )
	r@ @ EXECUTE		( bitmask actual-mask  r: node )
	2dup and IF		( bitmask actual-mask  r: node )
	    r@ @ xt>name cat2out
	    s"  OR "	cat2out
	    invert and
	ELSE drop THEN		( bitmask  r: node )
	r>			( bitmask node )
    LOOP			( bitmask node  r: xt )
    2drop
    s"  "		cat2out
    r> xt>name 		cat2out
    s"  !"		cat2out out-line ;

: save-display-slots ( flag-if-to-save-all -- )	\ display-slots to (outfile-id)
    ['] display-slots save-variable

    IF   max-display-slots#	\ while recording we save them all
    ELSE display-slots @ 	\ otherwise only the ones that are displayed
    THEN 0 DO
	s" ' "				cat2out
	i display-slot @ xt>name	cat2out
	s"  "				cat2out
	i num>string			cat2out
	s"  display-slot !"		cat2out	out-line
    LOOP ;

: which-random-seed ( -- xt u )		\  xt and cells
    random-xt @ CASE
	['] random-generalized OF
	    ['] (random-generalized)  2
	ENDOF
	['] random-BRODIE OF
	    ['] seed-BRODIE  1
	ENDOF
    ENDCASE ;

\ save kind and state of the actual random generator as line to (outfile-id)
\ or if flag is true save them all.
: save-random-generator ( flag-if-to-save-all -- )
    IF	\ while recording we save them all to catch differences
	['] random-xt			save-xt-variable
	\ do save the appropriate random seed anyway, changed or not:
	random-xt @ ['] random-generalized = IF true ?record-anyway THEN
	['] (random-generalized)	save-2variable
	random-xt @ ['] random-BRODIE	   = IF true ?record-anyway THEN
	['] seed-BRODIE			save-variable
    ELSE
	['] random-xt			save-xt-variable
	random-xt @ CASE
	    ['] random-generalized OF
		['] (random-generalized) save-2variable
	    ENDOF
	    ['] random-BRODIE OF
		['] seed-BRODIE		save-variable
	    ENDOF
	ENDCASE
    THEN ;

: save-nuc ( -- )	\ saves actual nuc to file (outfile-id)
    \ If the user has set a nuc from a selected individual 'nuc-is-word'
    \ would be set.  It *must* get cleared:
    nuc-flags @ >r
    nuc-flags dup @ [ nuc-is-word invert ] literal and swap !

    out-line						\ start with empty line
    s" \ define nuc and set it actual:"	cat2out	out-line \ comment

    length @ num>string		cat2out
    s"  new-nucleus DROP cp!"	cat2out	out-line
    nuc-genes-limit 0 DO
	s" gene' "		cat2out		\ save genes
	i nuc-addr @ xt>name	cat2out
	s"     	"		cat2out
	i nuc-var-name		cat2out
	s"  !"			cat2out	out-line
    LOOP

    nuc-xt-limit nuc-xt's DO			\ save xt's
	s" ' "			cat2out
	i nuc-addr @ xt>name	cat2out
	s"     	"		cat2out
	i nuc-var-name		cat2out
	s"  !"			cat2out	out-line
    LOOP

    nuc-scan-limit nuc-variables DO		\ save variables
	i nuc-addr @ num>string	cat2out
	s"  "			cat2out
	i nuc-var-name		cat2out
	s"  !"			cat2out	out-line
    LOOP

    r> nuc-flags ! ;

decimal
128 STRINGBUF-HANDLE: (record-file-name)
variable (record-file-id)

\ : ?record-free-field ( -- )
:noname ( -- )
    run-mode @ recording and 0= IF EXIT THEN	\ only if we do recording
    (record-file-id) @ set-outfile

    out-line							\ empty line
    s" \ restart with an unpopulated world:" cat2out  out-line	\ comment

    s" free-field"		cat2out	out-line ; IS ?record-free-field

\ : ?record-sow ( n -- )
:noname  ( n -- )
    run-mode @ recording and 0= IF drop EXIT THEN	\ only if we do record

    (record-file-id) @ set-outfile
    save-nuc					\ save actual nuc

    out-line					\ empty line
    s" \ sow above nuc:"      cat2out  out-line	\ comment

    false save-random-generator			\ save actual random generator
    ['] (sow-diversified)	save-variable	\ diversified?
    num>string			cat2out		\ number to sow
    s"  sow drop time-step"	cat2out	out-line ; IS ?record-sow


\ ?record-remove-cell
:noname ( -- )
    run-mode @ recording and 0= IF EXIT THEN	\ only if we do recording

    (record-file-id) @ set-outfile
    out-line						\ empty line
    s" \ cell removed by user:" cat2out out-line	\ comment
    spot @ num>string		cat2out
    s"  >spot! fcp @ cp! die "	cat2out out-line ; IS ?record-remove-cell

\ ?record-edit-spot
:noname ( -- )
    run-mode @ recording and 0= IF EXIT THEN	\ only if we do recording

    (record-file-id) @ set-outfile
    out-line						\ empty line
    s" \ spot edited by user:"  cat2out out-line	\ comment
    spot @ num>string		cat2out
    s"  >spot! "		cat2out out-line

    field-planes# 1 ?DO		\ we leave pointers out
	spot-before i cells + @
	spot-after  i cells + @ <> IF
	    spot-after  i cells + @ num>string	cat2out
	    s"  "				cat2out
	    i num>string			cat2out
	    s"  n'th-spot-variable ! "		cat2out out-line
	THEN
    LOOP
; IS ?record-edit-spot

\ assert results of a benchmark
: assert-variable-entry ( xt -- )
    dup xt>name			cat2out
    s"  @ "			cat2out
    EXECUTE @ num>string	cat2out
    s"  = AND"			cat2out out-line ;

\ this is for the creation of benchmarks
\ put 'assert-state-entry' into a record file at the appropriate place
\ to get the FORTH code to check state.  Insert it in the record file then.
: assert-state-entry ( -- )
    (outfile-id) @			( old-outfile-id )

    file-names-length# stringbuf-open >r
    tmp-dir		r@ cat
    s" ASSERTIONS"	r@ cat
    cr ." Saving to: "	r@ buffer-as-string type
    r@ buffer-as-string w/o create-file  r> stringbuf-close
    IF  bell cr ." assert-state-entry: Couldn't create file."
    ELSE
	set-outfile
	out-line
	s" \ assert expected results:"	cat2out out-line
	out-line

	which-random-seed >r		( old-id xt   r: cells )
	xt>name cat2out			( old-id      r: cells )
	r@ CASE
	    1 OF s"  @ "  ENDOF
	    2 OF s"  2@ " ENDOF
	    true ABORT" assert-state-entry: Error saving random seed."
	ENDCASE cat2out
	(out-buffer) buffer-as-string EVALUATE	( old-id n|d  r: cells)
	r> CASE
	    1 OF
		num>string	cat2out
		s"  ="	cat2out
	    ENDOF
	    2 OF
		num>string	cat2out
		bl		char2out
		num>string	cat2out
		s"  d="		cat2out
	    ENDOF
	ENDCASE		out-line

	['] cloned		assert-variable-entry
	['] living		assert-variable-entry
	['] nuc-do-cost		assert-variable-entry
	['] code-price		assert-variable-entry
	['] (mutated-max)	assert-variable-entry
	['] compiled-genes	assert-variable-entry

	s" cr"				cat2out out-line
	s" [IF]"			cat2out out-line
	s"     .( 	Result is valid. )"	cat2out out-line
	s" [ELSE]"			cat2out out-line
	s"     bell .( 	Unexpected result. Not comparable to other systems! )"
					cat2out out-line
	s" [THEN]"			cat2out out-line
	out-line

	(outfile-id) @ close-file drop
    THEN
    set-outfile ;

: save-probability-pool ( -- )	\ saves to (outfile-id)
    out-line						\ start with empty line
    s" \  set probabilities in actual pool:"	cat2out	out-line      \ comment
    ['] actual-genepool-xt			save-xt-variable

    s" actual-genepool-xt @ execute nul-all-probabilities" cat2out
    \ As last line must be included in diff's while recording when pool changes
    \ I add a comment containing the pools name:
    run-mode @ recording and IF
	s" 	\ "				cat2out
	actual-genepool-xt @ xt>name		cat2out		\ trick...
    THEN					out-line

    actual-genepool-xt @ execute
    dup nodes 0 ?DO
	next-node >r	( r: actual-node )		
	s" internal' "			cat2out
	r@ >xt @ xt>name		cat2out  s" 	" cat2out
	r@ >probability @ num>string	cat2out  bl char2out
	s" actual-genepool-xt @ execute set-xt-probability" cat2out out-line
	r>
    LOOP
    drop ;

: other-nodes-(none) ( node -- )	\ inserts ' (none) in all other nodes
    BEGIN
	next-node
    dup WHILE
	['] (none) over ( >cont-xt ) !
    REPEAT drop ;

: cont-unique ( node -- )	\ makes a string unique for the node values
    base @ >r hex

    (unique) >r
    dup >cont-xt @		num>string r@ cat
    dup >cont-item @		num>string r@ cat
    dup >cont-lower @		num>string r@ cat
    dup >cont-upper @		num>string r@ cat
    dup >cont-foreground-xt @	num>string r@ cat
        >cont-char @		num>string r@ cat
    rdrop

    r> base ! ;

: save-continuous-display ( -- )	\ we don't need the flag
    run-mode @ write-diff and	( flag )	\ as we can determine it here

    s" \ save-continuous-display:" cat2out out-line 
    ['] (continuous-column)	save-variable	\ not sure about that one
    ['] cont-zoom-up-scale	save-2variable

    continuous-display-list
    dup nodes 0 ?DO		( flag node )
	next-node
	over IF dup cont-unique THEN
	i num>string			cat2out
	s"  continuous-display-list n'th-node " cat2out out-line

	['] >cont-xt		save-offset-xt-variable
	dup >cont-xt @ ['] get-variable = IF
	    ['] >cont-item	save-offset-xt-variable
	ELSE    
	    ['] >cont-item	save-offset-variable
	THEN
	['] >cont-lower		save-offset-variable
	['] >cont-upper		save-offset-variable
	['] >cont-foreground-xt	save-offset-xt-variable
	['] >cont-char		save-offset-variable
    LOOP nip			( flag )

    s" continuous-display-list dup nodes swap n'th-node " cat2out out-line
    s" other-nodes-(none)"				  cat2out out-line

    \ we need same number of lines in the diff files:
    ( flag ) IF
	max-continuous-items#  continuous-display-list nodes - 0 ?DO
	    s" (none)	\ placeholders:"	cat2out out-line
	    7 0 DO
		s" (none)"			cat2out out-line
	    LOOP
	LOOP
    THEN ;

: step-unique ( -- )	\ builds a unique string from the parameters
    base @ hex
    (unique) >r

    (scan-xt) @			num>string r@ cat
    (scan-detail-index) @	num>string r@ cat
    (scan-lines) @		num>string r@ cat
    (vertical-display-range) @	num>string r@ cat
    scan-background-xt @	num>string r@ cat
    scan-foreground-xt @	num>string r@ cat
    (scan-flags) @		num>string r@ cat

    (scan-min-max) 2@		num>string r@ cat  num>string r@ cat
    (last-scan-min-max) 2@	num>string r@ cat  num>string r@ cat
    horizontal-zoom-scale 2@	num>string r@ cat  num>string r@ cat
    vertical-zoom-scale 2@	num>string r@ cat  num>string r@ cat

    rdrop
    base ! ;

: save-step-display-settings ( flag -- )	\ save all or only relevant?
    step-display-on display-switch @ and 0= over 0= and IF drop EXIT THEN

    s" \ save-step-display-settings:" cat2out out-line 
    ['] step-display-items	save-variable

    dup IF max-step-display-items# ELSE step-display-items @ THEN
    0 ?DO			( flag )
	i (scan-index) !

	run-mode @ write-diff and
	step-display-on display-switch @ and
	OR IF step-unique THEN

	['] (scan-index)		save-variable

	['] (scan-xt)			save-xt-variable
	(scan-xt) @ ['] continuous-display = IF drop TRUE THEN	\ set flag
	['] (scan-detail-index)		save-variable
	['] (scan-lines)		save-variable
	['] scan-background-xt		save-xt-variable
	['] scan-foreground-xt		save-xt-variable
	['] (scan-min-max)		save-2variable
	['] (last-scan-min-max)		save-2variable
	['] horizontal-zoom-scale	save-2variable
	['] vertical-zoom-scale		save-2variable
	['] (vertical-display-range)	save-variable
	['] (scan-flags)		save-variable
    LOOP

    IF save-continuous-display THEN ;

: save-brew-variables ( -- )	\ save all variables to (outfile-id)

    out-line						\ empty line
    s" \ set brew variables:"	cat2out	out-line	\ comment

    \ for the sake of simplicity we out *all* variables while recording
    \ (this asks for special treatment of the rare case, when we save
    \ variables while recording)
    \ I do the same with log files now.
    run-mode @ [ recording write-diff or ] literal and
    0<> >r	( r: flag-if-we-are-recording )

    ['] world-do-direction		save-variable
    ['] spot-do-xt			save-xt-variable
    ['] cell-do-before-xt		save-xt-variable
    ['] cell-do-after-xt		save-xt-variable
    ['] future-quality-change		save-variable
    ['] cell-division-moves-both	save-variable
    ['] cell-division-diversify-both	save-variable
    ['] cell-division-mutate-both	save-variable
    ['] world-food-supply		save-variable
    ['] food-share/spot			save-variable
    ['] individual-fixed-food-share	save-variable
    ['] nuc-do-cost			save-variable
    ['] code-price			save-variable
    ['] code-price-scale		save-2variable
    \ ['] clone-cost			save-variable
    ['] leave-energy-after-death	save-variable

    ['] additive-stress			save-variable
    ['] additive-release		save-variable
    ['] stress-rate			save-2variable
    ['] multiplicative-release		save-variable
    ['] code-additive-stress		save-variable
    ['] code-stress-rate		save-2variable
    ['] high-water-mark			save-variable
    ['] flood-mark			save-variable
    ['] flood-stress-rate		save-2variable
    ['] flood-kill-rate			save-2variable
    ['] flood-energy-rate		save-2variable
    ['] flood-food-rate			save-2variable
    ['] sos-mark			save-variable
    ['] sos-sow				save-variable
    ['] sos-release-rate		save-2variable
    ['] low-water-mark			save-variable
    ['] up-regulation-start		save-variable
    ['] nuc-cost-can-be-help?		save-variable
    ['] code-price-can-be-help?		save-variable
    ['] diversification-mask		save-variable
    ['] diversification-range		save-variable
    ['] diversification-rate		save-2variable
    ['] diversification-closeness	save-variable
    ['] sporadic-value-range		save-variable
    ['] sporadic-value-rate		save-2variable
    ['] background-color-xt		save-xt-variable
    ['] foreground-color-xt		save-xt-variable
    ['] look-at-xt			save-xt-variable

    ['] 2-ascii-scale			save-variable

    ['] food>color-scale		save-variable
    ['] age>color-scale			save-variable
    ['] A-quality>color-scale		save-variable
    ['] B-quality>color-scale		save-variable
    ['] C-quality>color-scale		save-variable

    ['] (A-scale)			save-variable
    ['] (B-scale)			save-variable
    ['] (C-scale)			save-variable

    r@ save-random-generator

    ['] cell-division-moves-both	save-variable
    ['] cell-division-diversify-both	save-variable
    ['] cell-division-mutate-both	save-variable

    ['] mutation-rate			save-2variable
    mutation-rate @ 0<> r@ or IF
	['] top-level-insert		save-2variable
	['] top-level-replace		save-2variable
	['] top-snipping		save-2variable
	top-snipping @ 0<> r@ or IF
	    ['] top-random-snipping	save-variable
	    ['] top-short-snipping	save-variable
	    ['] top-long-snipping	save-variable
	THEN
	['] top-address-replace		save-2variable
	['] top-token-replace		save-2variable
	['] top-from-scratch		save-2variable
	['] stack-turning-point		save-variable
	['] mutations-threshold		save-variable
	['] trial-phase			save-variable
	['] conditional-token-price	save-variable
	['] resolve-flags		save-variable
	['] max-if-items		save-variable
    THEN

    ['] display-switch display-switch-masks save-listed-mask
    r@ save-display-slots

    r@ save-step-display-settings

    ['] log-mask log-masks		save-listed-mask
    ['] code-file-mask code-file-masks	save-listed-mask

    s" \ save function keys:"		cat2out out-line 
    ['] F1-xt      			save-xt-variable
    ['] F2-xt      			save-xt-variable
    ['] F3-xt      			save-xt-variable
    ['] F4-xt      			save-xt-variable
    ['] F5-xt      			save-xt-variable
    ['] F6-xt      			save-xt-variable
    ['] F7-xt      			save-xt-variable
    ['] F8-xt      			save-xt-variable
    ['] F9-xt      			save-xt-variable
    ['] F10-xt     			save-xt-variable
    ['] F11-xt     			save-xt-variable
    ['] F12-xt     			save-xt-variable
    ['] shift-F1-xt			save-xt-variable
    ['] shift-F2-xt			save-xt-variable
    ['] shift-F3-xt			save-xt-variable
    ['] shift-F4-xt			save-xt-variable
    ['] shift-F5-xt			save-xt-variable
    ['] shift-F6-xt			save-xt-variable
    ['] shift-F7-xt			save-xt-variable
    ['] shift-F8-xt			save-xt-variable

[ use-pool-select ] [IF]
    save-probability-pool
[THEN]

    rdrop ;

: |save-brew-variables| ( -- )
    page
    ." Give a filename to save brew variables in "	pad 80 accept
    dup 0= IF cr ." not saved " bell 1200 ms  drop EXIT THEN
    pad swap
    file-names-length# stringbuf-open >r
    out-dir		r@ cat
    ( pad count )	r@ cat
    r@ buffer-as-string r/o open-file
    IF			\ test if file exists already
	drop
    ELSE
	cr ." File exists, not saved." 2000 ms
	close-file
	r> stringbuf-close
	drop EXIT
    THEN

    r@ buffer-as-string w/o CREATE-FILE
    IF
	drop
	bell cr ." |save-brew-variables|: Error creating "
	r@ buffer-as-string type 1000 ms
    ELSE
	set-outfile
	save-brew-variables
	(outfile-id) @ close-file
	IF ." |save-brew-variables|: Error closing file" THEN
    THEN
    r> stringbuf-close ;

variable (before)		\ file id
: save-before
    tmp-dir s" before" file-name-cat >r
    r@ buffer-as-string r/w CREATE-FILE
    r> stringbuf-close
    IF cr ." save-before: Error opening file." bell 600 ms drop  EXIT THEN
    dup (before) !		\ remember file
    set-outfile
    run-mode dup @ write-diff or swap !
    save-brew-variables
    run-mode dup @ write-diff invert and swap ! ;

variable (after)		\ file id
: save-after
    tmp-dir s" after" file-name-cat >r
    r@ buffer-as-string r/w CREATE-FILE
    r> stringbuf-close
    IF cr ." save-after: Error opening file." bell 600 ms drop  EXIT THEN
    dup (after) !		\ remember file
    set-outfile
    run-mode dup @ write-diff or swap !
    save-brew-variables
    run-mode dup @ write-diff invert and swap ! ;

: remove-unique ( addr count -- addr count' )
    2dup unique-start search IF
	nip -
    ELSE 2drop THEN ;

\ that's dangerous!
decimal 2048 constant record-line-max#	\ increase if needed.
: record-changes ( file-id -- )
    save-after			\ present state to file

    \ allocate two buffers to keep two lines to compare
    record-line-max# allocate
    ABORT" record-changes: Couldn't allocate."
    record-line-max# allocate
    ABORT" record-changes: Couldn't allocate."
    2>r		( r: addr-after addr-before )

    \ open the files
    tmp-dir s" before" file-name-cat >r
    r@ buffer-as-string  r/o open-file  r> stringbuf-close
    ABORT" record-changes: Error opening 'tmp/before'."		(before) !
    tmp-dir s" after" file-name-cat >r
    r@ buffer-as-string  r/o open-file  r> stringbuf-close
    ABORT" record-changes: Error opening 'tmp/after'."		(after) !

    ( file-id ) set-outfile

    (scratch) on	\ flag: add start comment
    BEGIN
	2r@		( addr-before addr-after   r: addr-before addr-after )
	record-line-max#
	(after) @	( addr-before addr-after length after-file-id  r: b a )
	read-line	( addr-before count-after flag wior   r: before after )
	ABORT" record-changes: Error reading 'after' line in."
    WHILE
	>r		( addr-before   r: addr-before addr-after count-after)
	dup record-line-max# (before) @ read-line
	ABORT" record-changes: Error reading 'before' line in."
	0= ABORT" record-changes: Error, maybe lines not balanced."
	2r@ compare IF	( --   r: addr-before addr-after count-after )
	    (scratch) @ IF	\ first time add start comment
		out-line					\ empty line
		s" \ changed parameters:" cat2out  out-line	\ comment
		(scratch) off	\ only first time
	    THEN
	    2r@ remove-unique cat2out	out-line
	THEN rdrop
    REPEAT 2drop

    r> free r> free or ABORT" record-changes: Couldn't free." ;

variable (nuc-before)		\ file id
: save-nuc-before
    tmp-dir s" nuc-before" file-name-cat >r
    r@ buffer-as-string r/w CREATE-FILE
    r> stringbuf-close
    IF cr ." save-nuc-before: Error opening file." bell 600 ms  drop EXIT THEN
    dup (nuc-before) !		\ remember file
    set-outfile
    save-nuc ;

variable (nuc-after)		\ file id
: save-nuc-after
    tmp-dir s" nuc-after" file-name-cat >r
    r@ buffer-as-string r/w CREATE-FILE
    r> stringbuf-close
    IF cr ." save-nuc-after: Error opening file." bell 600 ms  drop EXIT THEN
    dup (nuc-after) !		\ remember file
    set-outfile
    save-nuc ;

variable (nuc-changed?)		(nuc-changed?) off	\ only get change flag
\ if (nuc-changed?) is on, it only compares, without writing recording entries.
: record-nuc-changes ( -- )
    save-nuc-after			\ present state to file

    \ allocate two buffers to keep two lines to compare
    record-line-max# allocate
    ABORT" record-nuc-changes: Couldn't allocate."
    record-line-max# allocate
    ABORT" record-nuc-changes: Couldn't allocate."
    2>r		( r: addr-after addr-before )

    \ open the files
    tmp-dir s" nuc-before" file-name-cat >r
    r@ buffer-as-string  r/o open-file  r> stringbuf-close
    ABORT" record-nuc-changes: Error opening 'tmp/nuc-before'."	(nuc-before) !
    tmp-dir s" nuc-after" file-name-cat >r
    r@ buffer-as-string  r/o open-file  r> stringbuf-close
    ABORT" record-nuc-changes: Error opening 'tmp/nuc-after'."	(nuc-after) !

    (record-file-id) @ set-outfile
    (scratch) on	\ flag: add start comment
    BEGIN
	2r@		( addr-before addr-after   r: addr-before addr-after )
	record-line-max#
	(nuc-after) @	( addr-before addr-after length after-file-id  r: b a )
	read-line	( addr-before count-after flag wior   r: before after )
	ABORT" record-nuc-changes: Error reading 'nuc after' line in."
    WHILE
	>r		( addr-before   r: addr-before addr-after count-after)
	dup record-line-max# (nuc-before) @ read-line
	ABORT" record-nuc-changes: Error reading 'nuc before' line in."
	0= ABORT" record-nuc-changes: Error, maybe lines not balanced."
	2r@ compare IF	( --   r: addr-before addr-after count-after )
	    (nuc-changed?) @ IF	\ special case, only flag needed?
		rdrop
		r> free r> free or ABORT" record-nuc-changes: Couldn't free."
		true EXIT	\ only flag needed, TRUE
	    THEN

	    \ normal case, record changes:
	    (scratch) @ IF	\ first time add start comment
		out-line					\ empty line
		s" \ changed nuc parameters:" cat2out  out-line	\ comment
		(scratch) off	\ only first time
	    THEN
	    2r@ cat2out	out-line
	THEN rdrop
    REPEAT 2drop

    r> free r> free or ABORT" record-nuc-changes: Couldn't free."
    (nuc-changed?) @ IF	false EXIT THEN	\ special case, only flag needed?
;

: nuc-changed? ( -- flag )
    (nuc-changed?) on
    record-nuc-changes
    (nuc-changed?) off ;

: record-on/off ( -- )			\ toggles recording
    run-mode
    dup @ recording xor over !

    @ recording and IF			\ start recording
	page cr

	." Start from scratch:"

	cr
	cr ." Please give record file name: "			\ file name
	(record-file-name) >r
	r@ stringbuf-empty s" brew-recorded" r@ cat		\ default
	r@ buffer-as-string type  bl emit
	r@ accept>stringbuf

	r> buffer-as-string					\ rec-play-dir
	file-names-length# stringbuf-open >r
	rec-play-dir r@ cat r@ cat
	r@ buffer-as-string  r/w  CREATE-FILE			\ create file
	r> stringbuf-close
	IF bell  cr ." record-on/off: Couldn't create record file."
	    1000 ms  drop EXIT
	THEN
	dup set-outfile   (record-file-id) !

	out-line						\ empty line
	s" \ starting from scratch:"	cat2out	out-line	\ comment

	free-field						\ clear up

	\ start by recording all variables
	save-brew-variables					\ save all

	save-before						\ remember all
	save-after
\   ELSE				\ stop recording
    THEN ;

decimal
128 STRINGBUF-HANDLE: (playback-file-name)
variable (playback-file-id)

: play ( wfileid -- )
    dup (playback-file-id) !
    single-step off
    include-file ;

use-fileselect [IF]
    include fileselect.fs  		\ #################
[THEN]

: playback-on/off ( -- )
    run-mode dup @ playback xor swap !

    run-mode @ playback and IF				\ switch to playback

	run-mode @ recording and IF
	    run-mode dup @ recording xor swap !
	    cr ." Simultaneous recording and playback is not implemented."
	    cr ." Recording put off."
	THEN

[ use-fileselect ] [IF] \ use fileselect
	rec-play-dir open-directory 0= IF bell EXIT THEN
	s" Select playback file: " fileselect-menu
	close-current-dir
	( fileselect-menu-return-flag ) 0= IF bell EXIT THEN
	r/o OPEN-FILE
[ELSE] \ don't use fileselect
	cr
	cr ." Please give playback file name: "			\ file name
	(playback-file-name) >r
	r@ stringbuf-empty s" brew-recorded" r@ cat		\ default
	r@ buffer-as-string type  bl emit
	r@ accept>stringbuf

	r> buffer-as-string					\ rec-play-dir
	file-names-length# stringbuf-open >r
	rec-play-dir r@ cat r@ cat
	r@ buffer-as-string  r/w  OPEN-FILE			\ open file
	r> stringbuf-close
[THEN]

	IF bell  cr ." playback-on/off: Couldn't open playback file."
	    1000 ms  drop EXIT
	THEN
\	play	\ much better for debugging...
	['] play catch
	cursor-visible
	IF	\ let's say the user selected nonsence...
	    bell cr ." Something went wrong with playing. " 2000 ms
	THEN
    \ ELSE						\ playback off
    THEN ;

MENU: rec/play-men
: .rec/play-menu
    page

    title-colors ." Record and Playback menu:" end-title

    cr								\ recording
    s" Record "	    ['] record-on/off	redisplay	menu-entry
    run-mode @ recording and .ON-off cr
    s" rR" menu-same-key-entry

    run-mode @ recording and IF
	cr
	s" Record file:    "	(record-file-name) >stack
	redisplay		['] accept>stringbuf	menu-entry
	.tab (record-file-name) buffer-as-string type
	.tab >last-xy cr
	s" fF" menu-same-key-entry
    THEN

    cr								\ playback
    s" Playback "	['] playback-on/off	redisplay	menu-entry
    run-mode @ playback and .ON-off cr
    s" pP" menu-same-key-entry

    run-mode @ playback and IF
	cr
	s" Playback file: "	(playback-file-name) >stack
	redisplay		['] accept>stringbuf	menu-entry
	.tab (playback-file-name) buffer-as-string type
	.tab >last-xy cr
	s" fF" menu-same-key-entry
    THEN
    
    common-menu-entries ;

: rec/play-menu
    rec/play-men
    ['] .rec/play-menu menu-display-xt !
    menu-done	['] noop	menu-key-default
    menu-done	['] noop	menu-default
    do-menu-loop
    \ free-menus
;
' rec/play-menu function-key-actions >list

: intro ( -- )
    (playback-file-name) >r
    r@ stringbuf-empty s" OUTPUT/rec-play/intro.fs" r@ cat
    r@ buffer-as-string  r/w  OPEN-FILE
    r> stringbuf-close
    IF bell  cr ." Demo couldn't be loaded." 1000 ms  drop EXIT THEN
    run-mode @ >r
    playback run-mode !
    play
    r> run-mode !
    single-step on
    s" Now it's your turn. Press some keys and try out..." 1 >message ;
s" bigFORTH" environment? [IF] 2drop \ drop version number
[ELSE]
    ' intro demos >list
[THEN]

MENU: var-men
: .var-menu
    page

    title-colors ." Various menu:" end-title

    cr
    s" Choose random generator:" title-choose-xt!
    random-generators >stack		random-xt	>stack-2
    s" Random generator: "		['] choose-xt-menu	menu-entry
    s" rR" menu-same-key-entry
    .tab random-xt @ dup >stack	xt>name	['] page-see		menu-entry cr
    
    random-xt @ CASE
	['] random-BRODIE OF
	    .tab .tab .tab
	    s" Random seed: "	seed-BRODIE	simple-menu-entry-value cr
	ENDOF
	['] random-generalized OF
	    .tab .tab .tab
	    s" Random seed:      "
		(random-generalized)  simple-menu-entry-value cr
	    .tab .tab .tab
	    s" Random generator: "
		(random-generalized) cell+ simple-menu-entry-value cr
	ENDOF
    ENDCASE

    cr
    s" Can cells change qualities of the future? "
    future-quality-change >stack  ['] toggle-addr  redisplay	menu-entry
    future-quality-change @ .YES-NO cr

    cr
    s" Cell division moves cell mother too? "
    cell-division-moves-both >stack  ['] toggle-addr  redisplay	menu-entry
    cell-division-moves-both @ .YES-NO cr

    s" Cell division diversifies cell mother too? "
    cell-division-diversify-both >stack ['] toggle-addr  redisplay menu-entry
    cell-division-diversify-both @ .YES-NO cr
    
    s" Cell division mutates cell mother too? "
    cell-division-mutate-both >stack ['] toggle-addr  redisplay menu-entry
    cell-division-mutate-both @ .YES-NO cr

    cr
    s" Scale for value to ASCII transformation: " 2-ascii-scale
    simple-menu-entry-value cr
    s" S" menu-same-key-entry

[DEFINED] calibrate-cur-esc-wait [IF]
    cr
    s" Beeping or other problems moving the cursor with the arrow key's?"
    ['] calibrate-cur-esc-wait	redisplay	menu-entry cr
    s" bBcCpPmM" menu-same-key-entry
[ELSE]
    cr ." calibrate-cur-esc-wait undefined ############### " cr
[THEN]

    cr
    s" Save brew variables "
    run-mode @ recording and IF
	." is disabled during recording." cr
    ELSE    
	['] |save-brew-variables| 	redisplay		menu-entry cr
	s" s" menu-same-key-entry
    THEN

    cr
    s" World scan"
    ['] spot-scan-menu redisplay menu-entry cr
    s" wW" menu-same-key-entry
    1 6 screen-column  s" Nuc scan" ['] nuc-scan-menu redisplay	menu-entry cr
    s" nN" menu-same-key-entry

    cr
    s" Order to call the spots: "
    ['] change-world-do-direction	redisplay		menu-entry
    world-do-direction @ CASE
	0  OF ." randomly calling as many spots as there are. Slow."	ENDOF
	1  OF ." sequential calling spots forwards. Fastest."		ENDOF
	-1 OF ." sequential calling spots backwards. Slow."		ENDOF
	2  OF ." alternate direction, starting forwards. Quite slow."	ENDOF
	-2 OF ." alternate direction, starting backwards. Quite slow."	ENDOF
    ENDCASE cr

    cr
    s" Function keys" redisplay ['] function-key-menu	menu-entry cr
    s" fF" menu-same-key-entry

    common-menu-entries ;

: var-menu
    var-men
    ['] .var-menu menu-display-xt !
    menu-done	['] noop	menu-key-default
    menu-done	['] noop	menu-default
    do-menu-loop
    free-menus ;
' var-menu function-key-actions >list

MENU: menus-men
: .menus-menu
    page

    title-colors ." Menus menu:" end-title

    cr
    s" menu Individuals	"	['] individuals-menu		menu-entry cr
    s" iI"	menu-same-key-entry
    s" menu Food	"	['] food-menu			menu-entry cr
    s" fF"	menu-same-key-entry
    s" menu Diversification "	['] diversification-menu	menu-entry cr
    s" d"	menu-same-key-entry
    s" menu Population control"	['] menu-population-control	menu-entry cr
    s" aApP"	menu-same-key-entry
    s" menu Color	"	['] color-menu			menu-entry cr
    s" C"	menu-same-key-entry
    s" menu Display	"	['] display-menu		menu-entry cr
    s" D"	menu-same-key-entry
    s" menu Step Display	" ['] menu-step-display		menu-entry cr
\    s" sS"	menu-same-key-entry
    s" menu Continuous Display	" ['] menu-continuous-display	menu-entry cr
\    s" c"	menu-same-key-entry
    s" menu Various	"	['] var-menu			menu-entry cr
    s" vV"	menu-same-key-entry
    s" menu Mutation"		['] mutation-menu		menu-entry cr
    s" mM"	menu-same-key-entry
    s" menu Gene pools"		['] gene-pool-menu		menu-entry cr
    s" gG"	menu-same-key-entry
    s" menu Actual pool"	['] gene-pool-menu		menu-entry cr
    s" aA"	menu-same-key-entry
    s" menu Log/Log"		['] log-menu			menu-entry cr
    s" lLR"	menu-same-key-entry
    s" menu Code files"		['] code-file-menu		menu-entry cr
    s" c"	menu-same-key-entry
    s" menu Select nuc detail"	['] |menu-select-nuc-var|	menu-entry cr
    s" menu Scan nucs"		['] nuc-scan-menu		menu-entry cr
    s" s"	menu-same-key-entry
    s" menu Scan spots"		['] spot-scan-menu		menu-entry cr
    s" S"	menu-same-key-entry
    s" menu Function Keys"	['] function-key-menu		menu-entry cr
    s" kK"	menu-same-key-entry
    s" menu Record Playback"	['] rec/play-menu		menu-entry cr
    s" rR"	menu-same-key-entry

    cr
    s" menu Demos	"	['] demo-menu	menu-done	menu-entry cr

    common-menu-entries ;

: menus-menu
    menus-men
    ['] .menus-menu menu-display-xt !
    menu-done	['] noop	menu-key-default
    menu-done	['] noop	menu-default
    do-menu-loop
    \ free-menus
;
' menus-menu function-key-actions >list
' menus-menu F1-xt !


variable minimal-dictionary-space
10000 minimal-dictionary-space !
: sanity-checks ( -- )
    unused minimal-dictionary-space @ < IF
	s" Running out of dictionary space! (" 1 >message 
	unused num>string (message) cat
	s"  address units remaining)" (message) cat
	['] red message-fg-color-xt !
	single-step on
    THEN ;

\ brewing the gene soup: 'brew ( steps -- steps' )'
\ 'steps' counts down while brewing
\ starting with a positive value it does the given number of steps or less
\ starting with zero it counts the steps (downwards)
\ it stops,	when 'steps' reaches zero
\		when a key is pressed
\		or when the population dies out
\ 1 value brew-depth	\ depth control
: brew ( max-steps|0 -- steps-made )
    \ depth to brew-depth		\ easy stack control. insecure.
    dup >r

    ?step-display-sanity \ do this anyway, to set 'continuous-display-used'

    display-switch @
    [ spot-display-on step-display-on or ] literal and IF
	?step-display-sanity
	cursor-off
    THEN

    BEGIN ( steps )
	log-mask @ IF	\ checking logging (for speed reasons)
	    s" step " step @ log-step log-string-and-number
	THEN

	world-do
	sanity-checks

	display-switch @ 
	[ spot-display-on step-display-on no-snapshots or or ] literal
	and 0= IF 
	    step @ 127 and 0= IF
		.world
		cursor-off		\ quick cosmetic bug fix ;-)
	    THEN
	THEN

	\ display some infos
	.info-line

	population-control

	depth brew-depth <> IF
	    bell 0 l-s 4 - at-xy ." stack violation! " .s ."    "
	    do-FORTH
	THEN

	\ check if we're not done:
	1-  dup 0=				\ decrease counter, zero?
	single-step @ 0<> or			\ or single step?
	is-key? 0<> or				\ or key pressed?
	living @ 0= or				\ or no cells left!
    UNTIL					\ done? else repeat.

    (manually-selected-cell) off	\ deselect, could have died...

    r> swap - ;		( steps-made )

\ normally the display is shown while running the evolution
\ this word displays the world without brewing.
: brew-redisplay ( -- )

    display-switch @ spot-display-on AND IF
	.world
    THEN

    display-switch @ step-display-on AND IF
	run-mode @ continuous-display-used and IF
	    (continuous-column) dup @ >r r@ 1- 0 max swap !
	    step-display
	    r> (continuous-column) !
	ELSE
	    step-display
	THEN
    THEN

    .info-line

    \ set cursor if spot display (only) is on:
    display-switch @ dup spot-display-on and
    swap step-display-on and 0=
    and IF
	menu-2-at 2@ at-xy		\ cursor reset to where it was before
	c-l 2/ l-s 2/ at-xy		\ cursor in the middle of the screen
  \ ELSE
	\ don't know yet
    THEN
    cursor-visible ;

: edit-nuc-this-spot ( -- )
    run-mode @ recording and IF
	save-nuc-before
    THEN

    (nuc-spot-known) on
    nuc-menu

    run-mode @ recording and IF
	(nuc-spot-known) @ 1 <> IF		\ only if not removed by user
	    nuc-changed? IF
		(record-file-id) @ set-outfile

		out-line			\ empty line and comment
		s" \ changing nuc parameters of nuc at spot: " cat2out out-line

		spot @ num>string		cat2out
		s"   >spot!"			cat2out	out-line

		s" fcp @ cp!"			cat2out	out-line
		record-nuc-changes
	    THEN
	THEN
    THEN
    (nuc-spot-known) off ;

: clone-to-here ( cell-addr -- )
    cp!

    clone IF
	dup cp!			\ set actual				
	new-id id !		\ new id
	fcp !			\ put into field

	run-mode @ recording and IF
	    (record-file-id) @ set-outfile
	    out-line					\ empty line, comment
	    s" \ define a nuc and set to spot:" cat2out out-line
	    save-nuc

	    out-line					\ empty line, comment
	    s" \ now set defined nuc to spot:"	cat2out out-line
	    spot @ num>string			cat2out
	    s"  >spot!"				cat2out out-line
	    s" |cp@| fcp !"			cat2out out-line
	    s" brew-redisplay"			cat2out out-line
	THEN

	brew-redisplay		\ show it
    THEN ;

: mouse-in-field? ( x y k -- x y k false | spot k true )
   >r 2dup >r >r
   c-y / c-l *
   swap
   c-x / +
   dup spots <
   IF           ( spot )
     rdrop rdrop
     r> true    ( spot k true )
   ELSE drop
     r> r>  r>  false
 THEN ;

\ mouse interaction on a screen after (or while) brewing
\ (or from 'brew-redisplay')
\ cheques info? step? spot? 
\ interactions outside range are treated as noop's
: react-on-mouse ( -- ) \ mouse reaction info-line nuc-scan-display spot
    mousek@		( x y k )

    over c-y / last-line = IF	\ on info line?
	drop 2drop	( -- )
	display-menu EXIT	\ allow change of display characteristics
    THEN

    \ step display overwrites spot display, so we check for step first:
    display-switch @ step-display-on and IF	\ on step display?
	over c-y / l-s 2 - = IF			\ on scan info line?
	    |menu-select-nuc-var|
	ELSE
	    page cr ." Step display interaction not implemented yet." key drop
	THEN

	drop 2drop
	brew-redisplay EXIT
    THEN

    display-switch @ spot-display-on and IF	\ on spot display?
	mouse-in-field? IF swap  ( k spot )	\ in the field?
	    >spot!
	    drop		\ mousek
	    fcp @ dup IF   			\ occupied?
		dup cp!				\ set as actual nuc
		(manually-selected-cell) !
		edit-nuc-this-spot			\ does recording too
	    ELSE					\ spot was free
		drop
		(manually-selected-cell) @ IF		\ a cell preselected?
		    (manually-selected-cell) @ clone-to-here	\ clone cell
		ELSE
		    menu-edit-spot				\ look at spot
		THEN
	    THEN
	ELSE  ( x y k )		\ outside of the cell region (spot display)
	    drop 2drop
	THEN
    THEN ;

\ see genome at current spot, or spot if empty
: ?|see-genome-or-spot| ( -- )
    at? >r c-x * r> c-y * 0 mouse-in-field? nip IF
	dup someone-here? dup IF	( i addr|false )
	    nip
	    cp!
	    at? menu-2-at 2!		\ so cursor will be restored
	    see-genome
	    menu-redisplay on
	    key drop
	ELSE	( i false )
	    drop
	    >spot!
	    menu-edit-spot
	THEN
    ELSE 2drop THEN ;

: >look-at ( xt -- )	>stack look-at-xt >stack-2 ['] ! redisplay ;

: .keybindings ( -- )
    page

    title-colors
    ." Some keybindings in brews main screen:"
    ."    (Please see function keys too)."
    end-title

    cr
    ." <SPC>	 stop'n go"
    1 3 screen-column
    ." s        step"
    2 3 screen-column
    ." l  look at genome at spot"
    cr
    ." <arrow>	 move to spot"
    1 3 screen-column
    ." <RETURN> select"
    2 3 screen-column
    ." !  FORTH interpreter"
    cr
    ." r	 redisplay"
    2 3 screen-column
    ." w / n   scan world / nuc's"
    cr

    cr
    title-colors ." Menus:  ( B S m will port you to Brews Super Menu )" end-title

    cr
    ." M  Mutation menu"
    1 3 screen-column
    ." F  Food menu"
    2 3 screen-column
    ." P  Population control"
    cr
    ." D  Display menu"
    1 3 screen-column
    ." C  Color menu"
    2 3 screen-column
    ." I  Individuals menu"
    cr
    ." L  Logs Logs"
    1 3 screen-column
    ." V  Various menu"
    2 3 screen-column
    ." R  Record & Playback"
    cr
    ." G  Gene pool"
    1 3 screen-column
    ." A  Actual pool"
    2 3 screen-column
    ." K  Function Keys"
    cr
    cr
    title-colors ." Display:" end-title
    cr
    ." t  toggle step/spot"
    1 3 screen-column
    ." o  on/off (speedup)"
    cr
    ." f  foreground color"
    1 3 screen-column
    ." b  background color"
    2 3 screen-column
    ." c  colors back&forg"
    cr

    cr
    title-colors ." Keys to change what to look up:" end-title
    cr

    ." g  generation"
    1 4 screen-column
    ." 1  organ-A"
    2 3 screen-column
    ." p  sign-organ-A"
    cr
    ." a  age"
    1 4 screen-column
    ." *  ABC*."
    2 5 screen-column
    ." |  ABC-X|"
    cr
    ." e  energy"
    1 4 screen-column
    ." y  ABC"
    cr

    cr
    ." i  intro"
    1 4 screen-column
    ." d  demos"
    2 5 screen-column
    ." x  restart"
    2 3 screen-column
    ." QX leave menu to FORTH"

    ?reset-continuous-column ;

\ word called just after brewing to make record entry if appropriate
: ?record-brew ( steps -- )
    run-mode @ recording and 0= IF drop EXIT THEN	\ only if we do record

    (record-file-id) @ set-outfile

    out-line						\ empty line
    s" \ brew the gene soup:"	cat2out	out-line	\ comment

    num>string			cat2out
    s"  brew drop"		cat2out	out-line	\ brew
;

\  variable (brew-show-or-go)	(brew-show-or-go) off
\  : brew-show-or-go ( -- )	(brew-show-or-go) on ;
MENU: brew-men
: .brew ( -- )
    \ the way things are set up here 'redisplay' lets brewing continue
    \ while 'do-after' just shows the world and the info line again

    s"  "   single-step >stack	['] toggle-addr	redisplay
						do-after-2	menu-key-entry
    s" o"   ['] display-on-off			redisplay	menu-key-entry
    s" s"   single-step >stack	['] on		redisplay	menu-key-entry
    s" XQ"  menu-done		['] noop			menu-key-entry
    s" x"   redisplay		['] free-field			menu-key-entry
    s" khH"	wait-n'go   	['] .keybindings
				do-after-2	redisplay	menu-key-entry
    s" !"	['] do-FORTH			do-after	menu-key-entry
    s" r"	['] brew-redisplay				menu-key-entry
    s" l"	['] ?|see-genome-or-spot|	do-after-2	menu-key-entry
    s" w"	['] spot-scan-menu 		do-after	menu-key-entry
    s" n"	['] nuc-scan-menu 		do-after	menu-key-entry
    s" t"	['] toggle-display-&-go		do-after-2	menu-key-entry
    s" ?"	['] start-help			do-after	menu-key-entry

    s" L"	['] log-menu			do-after-2	menu-key-entry
    s" M"	['] mutation-menu		do-after-2	menu-key-entry
    s" BmS"	['] menus-menu			do-after-2	menu-key-entry
    s" F"	['] food-menu			do-after-2	menu-key-entry
    s" G"	['] gene-pool-menu		do-after-2	menu-key-entry
    s" A"	['] actual-pool-menu		do-after-2	menu-key-entry
    s" K"	['] function-key-menu		do-after-2	menu-key-entry
    s" P"	['] menu-population-control	do-after-2	menu-key-entry
    s" R"	['] rec/play-menu		do-after-2	menu-key-entry
    s" G"	['] gene-pool-menu		do-after-2	menu-key-entry
    s" D"	['] display-menu		do-after-2	menu-key-entry
    s" C"	['] color-menu			do-after-2	menu-key-entry
    s" V"	['] var-menu			do-after-2	menu-key-entry
    s" I"	['] individuals-menu		do-after-2	menu-key-entry

    s" c"   ['] toggle-colorizing do-after-2	redisplay	menu-key-entry
    s" b"   ['] toggle-background-colorizing
				do-after-2	redisplay	menu-key-entry
    s" f"   ['] toggle-foreground-colorizing
				do-after-2	redisplay	menu-key-entry
    s" d"			['] demo-menu	redisplay	menu-key-entry
    s" i"			['] intro	redisplay	menu-key-entry

    s" g"	['] show-generation  >look-at	do-after-2	menu-key-entry
    s" a"	['] show-age	>look-at	do-after-2	menu-key-entry
\   s" c"	['] show-ascii	>look-at	do-after-2	menu-key-entry
    s" e"	['] show-energy	>look-at	do-after-2	menu-key-entry
    s" 1"	['] show-A	>look-at	do-after-2	menu-key-entry
    s" p"	['] show-sign-A	>look-at	do-after-2	menu-key-entry
    s" *"	['] show-ABC*.	>look-at	do-after-2	menu-key-entry
    s" y"	['] show-ABC	>look-at	do-after-2	menu-key-entry
    s" |"	['] show-ABC-X|	>look-at	do-after-2	menu-key-entry
    s" .'"	['] show-~|'.*	>look-at	do-after-2	menu-key-entry

    ['] do-after-2 default-function-keys	\ work in progress ############

    do-not-cancel-keys on	\ otherwise the key will be cancelled in 'what'

    (brew-show-or-go) @ single-step @ and IF
	brew-redisplay		\ special: just show but don't make a live step
    ELSE			\ normal case: now we brew key? until
	log-mask @ log-user and IF
	    ?open-log-file
	    log-file-id @ record-changes
	THEN
	run-mode @ recording and IF
	    (record-file-id) @ record-changes
	THEN
	0 brew  ( steps )
	?record-brew
	save-before		\ save variables before user interaction
    THEN

    (brew-show-or-go) off
    c-l 2/ l-s 2/ at-xy			\ cursor in the middle of the screen
    cursor-visible ;

: brew-menu ( -- )
    single-step on
    start-help
    brew-men
    ['] .brew menu-display-xt !
    ['] noop  clear-screen-xt !
    ['] brew-redisplay to-do-after-xt !
    ['] brew-show-or-go to-do-after-2-xt !
    ['] react-on-mouse	do-after-2  menu-default
    ['] .keybindings  wait-n'go	redisplay do-after-2 menu-key-default
    do-menu-loop
    free-menus ;

' diversify?-some is <diversify>

log-mask off	\ was compile time switch, now run time switch

\ ok, we start now:
big-bang

include brew-init.fs

\ *********************************************************************
\ Run brew as a benchmark?  See 'brew-options.fs' for some info.

\ If your FORTH has an option to pass a string which is evaluate'd
\ during startup start benchmarks by defining a word with the
\ benchmarks name (without the '.fs' suffix).

\ So in Gforth you'd say:
\ time  gforth-fast --dictionary-size 2M  -e "create startup-bench"   brew.fs
\ time  gforth-fast --dictionary-size 2M  -e "create simple-bench"    brew.fs
\ time  gforth-fast --dictionary-size 2M  -e "create mutations-1"     brew.fs
\ time  gforth-fast --dictionary-size 2M  -e "create displayed-bench" brew.fs
\ time  gforth-fast --dictionary-size 2M  -e "create long-mutations"  brew.fs

\ In bigFORTH you say
\ time  bigforth -e "create simple-bench"  brew.fs

\ PFE also has the -e" option now.

[DEFINED] startup-bench   [IF] include benchmarks/startup-bench.fs 	[THEN]
[DEFINED] simple-bench    [IF] include benchmarks/simple-bench.fs 	[THEN]
[DEFINED] mutations-1     [IF] include benchmarks/mutations-1.fs 	[THEN]
[DEFINED] displayed-bench [IF] include benchmarks/displayed-bench.fs 	[THEN]
[DEFINED] long-mutations  [IF] include benchmarks/long-mutations.fs 	[THEN]

\ If your FORTH does not have such an option you uncomment the appropriate
\ 'include xxx' line in 'brew-options.fs'.

\  startup-bench
\	Startup only.

\  simple-bench
\      Testing speed of the very basic brew functions like calling the
\      'cells', reproduction, population control and the like.
\      Mutation is off.
\      This benchmark is the most portable one.
\      tested:    GForth 0.5.0
\                 time gforth-fast --dictionary-size 2M  -e "create simple-bench" brew.fs
\                 pfe Portable Forth Environment 0.30.38 (Dec 31 2000 09:38:15)
\                 ANS bigforth 386-Linux rev. 1.22 (only this bench works here)

\  mutations-1
\      This benchmark is far more complex.  Mutation is on.  It calls quite
\      a variety of additional functions.  Somehow unusual for a benchmark it
\      includes a fair amount of interpreting buffered strings and compilation.
\      Start with
\      time gforth-fast --dictionary-size 2M -e "create mutations-1" brew.fs
\      tested:	GForth 0.5.0
\               pfe Portable Forth Environment 0.30.38 (Dec 31 2000 09:38:15)
\                   Start with 'time pfe -k 2048 brew.fs'

\  displayed-bench
\      This benchmark has display switched on.
\      It does even use colors sometimes.
\      Use the others, if you don't want this.
\      tested:	GForth 0.5.0
\               pfe Portable Forth Environment 0.30.40 (Jan  1 2001 18:17:50)
\                    start with 'time pfe brew.fs'
\                    Don't use earlier pfe versions for this one.

\  long-mutations
\      This one is set up to stress interpretation of buffered strings.
\      Mutation builds very long gene sequences here.
\      In the second phase genes are kept longer than usual in trial phase,
\      which means that a lot of FORTH code is interpreted.
\      I deliberately switch to an unoptmized search order here to stress
\      dictionary searches.
\      tested:	GForth 0.5.0
\               
\               Portable Forth Environment 0.30.40 (Jan  1 2001 18:17:50)
\               Don't use earlier pfe versions for this one.

\ *********************************************************************

' include catch brew-options.fs
?dup [IF]
    dup -38 = [IF]
	drop .( brew-options.fs not present. )
    [ELSE]
	.( Error loading brew-options.fs ) . cr
	.( <press a key> ) cr key drop throw
    [THEN]
[THEN]


' include catch my-brew-options.fs
?dup [IF]
    -38 <> [IF] bell
	.( Error loading my-brew-options.fs ) . cr
	.( <press a key> ) cr key drop throw
    [THEN]
[THEN]

save-before	\ save variables before user interaction at startup

brew-menu
single-step off

page
cr .( you're in FORTH now.) cr
cr .( To go back to 'brew' you can type)
cr .( brew-menu) cr
cr
