\ genes-0.2.fs
\ 	$Id: genes-0.2.fs,v 1.22 2001/02/13 22:13:03 f Exp $	

\ Genes are ordinary Forth words.
\ They have an associated word though, which contains info needed mostly
\ at mutation time. It has the same name and is in gene-internals.
\ Gene names must be unique.

\ I dropped my plans to change the string evaluating during trial phase.
\ So genes are still evaluated as strings during trial.
\ During trial the stack data is in a comment at the end of the string.

\ Later on selected genes (strings) get compiled as Forth,
\ and the gene-internals word is created.

\ This version changed the internals to be used with the new
\ symbol stack version.
\ I sacrificed space to gain efficiency by including two double cells
\ as pointers to stack in and out data.
\ These strings are stored after the xt-chain.
decimal

false				\ true for testing, false otherwise
dup [if]
    cr .( Testing new genes version )
    include keybuf.fs
    include basics.fs
    include stringbuf-0.3.fs
    include lists.fs
    include random.fs
    VOCABULARY brew-words
    also brew-words
\ : xt>name ( xt -- addr length )   look IF name>string THEN ;  \ Bernd Paysan
[then]

\ ;gene is compiled into brew-words, to be always visible
: ;gene	( |u -- )			\ ends a gene definition
    state @ IF				\ or is a noop
	postpone ;
	previous			\ undo changes by GENE:
	current !			\ GENE: did put the value on stack
    THEN ;  immediate

vocabulary genes			\ holds genes
also genes

vocabulary gene-internals	\ holds words with the gene internals
current @			\ remember to restore
also gene-internals definitions
: gene-literal ;			\ just a token for the internals
previous
current !				\ restore current


\ GENE-INTERNALS: to create the internals data structure
\ do not change these without changing GENE-INTERNALS:
0
OFFSET: >gene-stack-in	( addr -- addr' )	\ count of input string
1+						\ address of input string
OFFSET: >gene-stack-out	( addr -- addr' )	\ count of output string
1+						\ address of output string
OFFSET: >gene-tokens#	( addr -- addr' )	\ number of xt's, literals...
OFFSET: >gene-cost	( addr -- addr' )	\ sum of costs of all sub genes
OFFSET: >gene-flags	( addr -- addr' )	\ flags
OFFSET: >gene-start	( addr -- addr' )	\ the xt array starts here
1- cells CONSTANT gene-info-bytes#		\ gene-start not counted

\ masks for gene-flags:
0 ( bit-position )
MASK: in-address		\ in tos is address
MASK: out-address		\ out tos is address
MASK: primitive			\ is it a gene primitive?
MASK: symbol-wildcard		\ are there symbol wildcards?
drop

VARIABLE tokens				\ make it easy counting tokens
VARIABLE (last-internal)		\ pointer to the body

\ don't want to rely on sp@ beeing here
[UNDEFINED] (scratch) [IF]	VARIABLE (scratch)	[THEN]

\ the mutation process uses special words in the evaluated strings
\ that do things like storing stack informations during internals compilation
\ or give the structure of the sequence of words,
\ which can be used for subgene compilation and such.
DEFER <internal-compiling-extensions> ( xt -- xt flag )
\ the flag says if the xt should be compiled into internals.
\ ( might have been changed by the deferred word called)
\ the word will be defined in the mutations section.
' TRUE IS <internal-compiling-extensions>

\ places a string as counted string at pad with a (uncounted) trailing bl
: counted>pad ( addr count -- pad )
    >r
    r@ pad !
    pad 1+ r@ move
    bl pad 1+ r> 1+ + c!
    pad ;

: [internal']   also gene-internals  postpone [']  previous ;  immediate  
: internal'     also gene-internals   '            previous ;

: [gene']   also genes	postpone [']	previous ;  immediate  
: gene'     also genes	 '		previous ;

100 CONSTANT (default-gene-cost#)	\ default gene cost of a primitive

\ variable and flag: don't try to get cost from the xt, but take this value
VARIABLE (cost-is-set)

VARIABLE conditional-token-price
(default-gene-cost#) 3 / conditional-token-price !	\ just a default

VARIABLE (compile-as-primitive)		(compile-as-primitive) OFF
: GENE-INTERNALS:
    \ misusing stringbufs to hold the xt's:
    [ decimal ] 2048 stringbuf-open >r	\ r: handle-of-xt-buffer
    current @				\ remember current wordlist
    also gene-internals definitions
    CREATE

    here (last-internal) !

    tokens off				\ token counter

    0					\ total cost
    BEGIN
	(cost-is-set) off

	bl word

	\ check for some special cases:
	dup count s" IF"   compare 0= IF
	    drop s" [IF]"   counted>pad
	    conditional-token-price @ (cost-is-set) !
	THEN

	dup count s" ELSE"   compare 0= IF
	    drop s" [ELSE]"   counted>pad
	    conditional-token-price @ (cost-is-set) !
	THEN

	dup count s" THEN"   compare 0= IF
	    drop s" [THEN]"   counted>pad
	    conditional-token-price @ (cost-is-set) !
	THEN

	(compile-as-primitive) @ IF
	    (default-gene-cost#) (cost-is-set) !
	THEN

	find			\ find the ingredients of the gene
				\ note: *internals* xt's are stored

	IF
	    \ extensions, see above
	    <internal-compiling-extensions> IF	\ into chain?
		1 tokens +!			\ count

		dup (scratch) !
		(scratch) 1 cells r@ cat	\ build a chain of xt's

		swap	( xt cost )	\ add cost:
		over ['] ;gene <> IF
		    (cost-is-set) @ ?dup IF	\ not a ordinary internals gene
			+
		    ELSE
			over >body >gene-cost @ + \ get cost from internals
		    THEN
		ELSE
		    (default-gene-cost#) +
		THEN
		swap ( cost xt )
	    THEN

	    ['] ;gene =				\ finished?
	ELSE					\ word not found

	    >r 0. r> count >number		\ let's try number conversion
	    IF ABORT" GENE-INTERNALS: Failed number conversion." THEN	\ I give up
	    IF					\ ok, was a number
		2 tokens +!			\ count literal and value

		d>s
		[internal'] gene-literal
		(scratch) !
		(scratch) 1 cells r@ cat	\ ' gene-literal into chain
		(scratch) !
		(scratch) 1 cells r@ cat	\ value too

		(default-gene-cost#) (cost-is-set) !
		false				\ we're not finished yet...
	    ELSE ABORT" GENE-INTERNALS: ';gene' expected."  \ that's an error!
	    THEN
	THEN
    UNTIL				\ until we reach ;gene

    here 4 cells dup allot erase	\ for input and output stack pointers
    
    tokens @ ,				\ compile number of tokens

    ,					\ >gene-cost

    0 ,					\ 1 cell for flags

    \ prepare arguments to 'move':
    r@	buffer-as-string >r	( buf-addr R:n )
    here r@ allot	( buf-a here )	 \ allot space for token chain
    r> move				 \ and move it into data space

    r> stringbuf-close

    previous
    current ! ;				\ restore current compilation wordlist

\ if the last compiled gene has symbol wildcards
\ this must be called after >internal
: symbol-wildcards-on ( -- )
    (last-internal) @  >gene-flags dup @ symbol-wildcard OR swap ! ;

: gene-body>stack-in-string ( body-addr -- addr count )   >gene-stack-in 2@ ;

: gene-body>stack-out-string ( body-addr -- addr count )   >gene-stack-out 2@ ;

: gene-body>n'th-xt-addr ( body-addr n -- addr )   cells  swap >gene-start + ;

: .gene-info ( internal-xt -- )	\ currently just for a quick test...
    page
    dup xt>name			cr type
    cr
    >body >r
    r@ >gene-stack-in 2@	cr ." stack in:  " type
    r@ >gene-stack-out 2@	cr ." stack out: " type
    r@ >gene-tokens# @		cr ." tokens:    " .
    r@ >gene-cost @		cr ." cost:      " .
    r@ >gene-flags @
    base dup @ 2>r  2 base !	cr ." flags      " .   r> r> !
    cr rdrop ;

: gene-stack-effect ( body-addr -- n )
    dup >r >gene-stack-out @  r> >gene-stack-in @  - ;

: string-tos-is-address? ( addr count -- flag )
    dup IF
	1- + c@ [char] a =
    ELSE nip THEN ;

\ set stack-in data string and flags in the last compiled internals gene
\ must be called *before* making any dictionary changes, because it allot's.
\ must be called *before* 'string>stack-out'
: string>stack-in ( addr count -- )
    (last-internal) @ >gene-stack-in >r	\ r: pointer to future string
    here over allot		\ remember here and allot space for the string
    swap			\ addr count where string will be stored
    2dup			\ remember for moving it later on
    r@ 2!			\ store addr count in the pointer
    move			\ and move the string

    r> 2@ string-tos-is-address? IF	\ set flag if tos was address
	(last-internal) @ >gene-flags dup @ in-address or swap !
    THEN ;

\ set stack-out data string and flags in the last compiled internals gene
\ must be called *after* 'string>stack-in'
\ must be called *before* making any other dictionary changes (it allot's).
: string>stack-out ( addr count -- )
    (last-internal) @ >gene-stack-out >r	\ r: pointer to future string
    here over allot	align	\ remember here and allot space for the string
    swap			\ addr count where string will be stored
    2dup			\ remember for moving it later on
    r@ 2!			\ store addr count in the pointer
    move			\ and move the string

    r> 2@ string-tos-is-address? IF	\ set flag if tos was address
	(last-internal) @ >gene-flags dup @ out-address or swap !
    THEN ;

\ word to set stackdata of the last compiled internals word from a string
\ must be called before making any dictionary changes, because it allot's.
\ the string has the form s" xxx-yyy" where 'xxx' is stack-in data, 'yyy' out
: string>stackdata! ( addr count -- )
    2dup		( addr count addr count )
    s" -" search 0= ABORT" string>stackdata! string does not contain '-'"
    2>r			( addr count r: addr2 count2 )
    r@ - string>stack-in
    r> 1- r> 1+ swap	\ skip '-' in the remaining string
    string>stack-out ;

VARIABLE actual-genepool-xt

use-pool-select [IF] \ new pool with relative probability.

    include gene-pool.fs
    ' gene-primitives actual-genepool-xt !

[ELSE] \ old lists with multiple entries

    LIST: gene-primitives	\ note that the pool holds the *internals* xt's
    ' gene-primitives actual-genepool-xt !

[THEN]



\ to use instack" and outstack" it's convenient to have a switch
\ between skipping the data or compiling them
variable (compiling-gene-internals)	(compiling-gene-internals) off

\ temporary buffers for name and gene code.
\ the buffers get automatically opened on first usage.
\ (these two buffers will get reused, but never closed).
decimal
4096 STRINGBUF-HANDLE: (gene-code-buf)	\ written by 'GENE:'
32   STRINGBUF-HANDLE: (name-buf)

: compile-internals ( addr u -- )
    [ decimal ] 4096 stringbuf-open >r	\ buffer for the string to evaluate
    s" GENE-INTERNALS: " r@ cat   ( addr u ) r@ cat

    (compiling-gene-internals) on
    r@ buffer-as-string EVALUATE	\ compile it
    (compiling-gene-internals) off

    r> stringbuf-close ;

\ builds the internal word and puts the stack data as specified in a string
\ remembers last-internal-xt
variable last-internal-xt
: >internal ( addr count -- )
    (gene-code-buf) buffer-as-string compile-internals
    string>stackdata!

    \ get xt from name and set 'last-internal-xt'
    64  stringbuf-open >r
    s" ' "			  r@ cat
    (name-buf) buffer-as-string r@ cat
    also gene-internals
    r@ buffer-as-string evaluate  last-internal-xt !
    previous
    r> stringbuf-close ;

\ makes the last compiled gene a gene primitive
\ must be called after >internal
: (gene-primitive) ( -- )
    (last-internal) @ >r
    r@ >gene-flags dup @ primitive OR swap !
    (default-gene-cost#) r> >gene-cost ! ;

: primitive>internal ( addr count -- )
    (compile-as-primitive) on   >internal	(compile-as-primitive) OFF
    (gene-primitive) ;

\ declare a normal FORTH word as a gene.  Builds the internals word.
\ It's an ugly hack, but string based mutation will probably be dropped soon.
: declare-primitive ( addr count xt -- )
    (name-buf) >r				\ hack: name to (name-buf)
    r@ stringbuf-empty  dup xt>name r> cat

    (gene-code-buf) >r			\ hack: set (gene-code-buf)
    r@ stringbuf-empty
    dup xt>name r@ cat bl r@ char-cat		\ name as name
    xt>name     r@ cat bl r@ char-cat		\ same as action
    s" ;gene " r> cat

    primitive>internal ;			\ now do it as usual

: GENE:					\ creates a named gene
    save-input				\ we need the name twice
    (name-buf) stringbuf-empty
    bl word count (name-buf) cat	\ put name into string buffer
    bl (name-buf) char-cat		\ append space

    restore-input    ABORT" Troubles restoring input source"

    save-input				\ we need it twice
    (gene-code-buf) stringbuf-empty
    BEGIN
	bl word count			\ get next sub gene
	2dup  (gene-code-buf) cat	\ append to string buffer
	bl (gene-code-buf) char-cat	\ append space
	dup 0= IF			\ input exhausted?
	    2drop
	    refill			\ try to refill
	ELSE
	    s" ;gene" compare
	THEN
    0= UNTIL				\ invert flag
    restore-input
    ABORT" Troubles restoring input source"

    \ first we compile the gene:
    current @				\ passeed to ';gene'
    also genes definitions		\ this will be undone by ';gene'
    : ;					\ compile gene with ':'

current @   also genes definitions    
GENE: g-noop			;gene s" -"	primitive>internal
previous current !


\ the way from the *gene's* xt to the xt of the *internals* word:
\ Use it *only* on uniquely named genes (genomes) *not* on unwrapped primitives
: gene-xt>internal-xt ( xt -- xt|0 )
    xt>name					\ name as string
    dup pad c!  pad 1+ swap move		\ build counted string at pad
    also gene-internals  pad find  previous	\ find same name in internals
    0= IF drop false THEN ;

\ the way from the *internals* xt to the xt of the *gene's* word:
: internal-xt>gene-xt ( xt -- xt|0 )
    xt>name					\ name as string
    dup pad c!  pad 1+ swap move		\ build counted string at pad
    also genes  pad find  previous		\ find same name in genes
    0= IF drop false THEN ;

\ get internal xt from the name
: name>internal-xt ( addr count -- xt )
    [ decimal ] 64 stringbuf-open >r
    s" internal'  " r@ cat   r@ cat
    r@ buffer-as-string EVALUATE
    r> stringbuf-close ;

decimal
1024 STRINGBUF-HANDLE: (internal-code)
: internal-xt>string ( internal-xt -- addr count )
    (internal-code) >r  r@ stringbuf-empty

    also gene-internals
    >body >gene-start			\ start address of xt chain on stack
    BEGIN
	dup @				\ get xt
	dup CASE
	    [internal'] gene-literal OF
		swap
		cell+
		dup @ num>string r@ cat
		swap
	    ENDOF

	    dup xt>name r@ cat
	ENDCASE
	bl r@ char-cat
	
	['] ;gene = 0= WHILE
	cell+
    REPEAT drop
    previous
    r> buffer-as-string ;

use-pool-select [IF] \ new pool with relative probability

: last-gene-into-pool ( relative-probability -- )
    last-internal-xt @ swap
    actual-genepool-xt @ execute	 ( xt probability pool )
    new-xt-to-pool ;

[ELSE]

\ put last gene relative-probability times in the gene-pool
: last-gene-into-pool ( relative-probability -- )
    [ decimal ] 64 stringbuf-open >r
    s" ' " r@ cat				\ we want the xt of the gene
    (name-buf) buffer-as-string r@ cat	\ append name of the gene
    bl r@ char-cat				\ space
    actual-genepool-xt @ xt>name r@ cat		\ append list name of genepool
    s"  >list" r@ cat			\ constructing string to evaluate
    also gene-internals			\ takes the xt of the *internals* word
    r>
    swap 0 DO
	dup buffer-as-string EVALUATE	\ puts the gene in the pool
    LOOP
    stringbuf-close
    previous ;

[THEN]

drop
previous
